"use strict";

var ScanOutLabel = require("./constants.cjs");
var express = require("express");
var app = express();
app.use(express.static(__dirname));
app.use(express.json())
app.use(express.urlencoded({ extended: true }));

const fs_node = require("fs");

var sslobj = {};
try {
	sslobj = JSON.parse(fs_node.readFileSync("./web_app.json", "utf8"));
}
catch (e) {
	console.log("Configuration file not found or not valid. Running as an http server.")
	sslobj = {protocol: "http", port: 3000}
}

var server;
var ishttp = sslobj.hasOwnProperty("protocol") && sslobj.protocol == "https";
if(ishttp){
	var options = {
		key: fs_node.readFileSync(sslobj.ssl.key),
		cert: fs_node.readFileSync(sslobj.ssl.cert)
	};	
	server = require("https").Server(options, app);
}
else{
	server = require("http").Server(app);
	sslobj.protocol = "http";
}

const io = require("socket.io")(server, {maxHttpBufferSize: 1e10});
const fs_path = require("path");
const fs_process = require("process");
const PORT = process.env.PORT || sslobj.port;

server.listen(PORT, function(){
	// process.setuid("admin");
    console.log("Node.js is listening to PORT:"+PORT);
});

function load_file(id, dirpath, fname)
{
	let relpath = fs_path.join(dirpath, fname);
	let ltxt = fs_node.readFileSync(relpath, "utf8");
	io.to(id).emit("loadtext", dirpath, ltxt);
}

function convert_filename2json(prmfile)
{
	let idx = prmfile.lastIndexOf(".prm");
	if(idx >= 0){
		return prmfile.substr(0, idx)+".json";
	}
	return prmfile+".json";
}

io.on("connection", (socket) => {
	console.log(socket.id);
	io.emit("message", "TEST");

	socket.on("preproc", function(prms, progid){
		console.log("Preprocessing...");
		let spawn = require("child_process").spawn;
		fs_node.writeFileSync("preproc.json", prms, "utf8");
		let childp = spawn("./spectra_solver_nompi", ["-f", "preproc.json"]);
		var precresult = "";
		childp.stdout.on("data", (data) => {
			precresult += data.toString();
		});
		childp.on("exit", (code) => {
			if(code == 0){
				io.to(socket.id).emit(progid, precresult);
			}
		});
	});

	socket.on("run", function(prms, dataname, progid, argstrs)
	{
		try {
			fs_process.chdir("./data");
		}
		catch (e) {
			io.to(socket.id).emit("error", "ERROR: \"data\" directory not found.");
			return;
		}
		let pexec = fs_path.join("..", "spectra_solver_nompi");
		console.log("Process:"+progid);
		let spawn = require("child_process").spawn;

		fs_node.writeFileSync(dataname, prms, "utf8");
		let args = argstrs.split("::");
		let childp = spawn(pexec, args);	

		fs_process.chdir("..");

		io.to(socket.id).emit(progid, "pid"+childp.pid.toString(), "", "");
		childp.stdout.on("data", (data) => {
			let status = data.toString();
			if(status.includes(ScanOutLabel)){
				try {
					let scanout = status.replace(ScanOutLabel, "").trim();
					let result = fs_node.readFileSync(fs_path.join("data", scanout), "utf8");
					io.to(socket.id).emit(progid, "scannext", result, scanout);		
				}
				catch (e) {
				}
			}
			else{
				io.to(socket.id).emit(progid, status);
			}
		});
		childp.on("exit", (code) => {
			if(code == 0){
				try {
					let result = fs_node.readFileSync(fs_path.join("data", dataname), "utf8");
					io.to(socket.id).emit(progid, "finish", result, dataname);		
				}
				catch (e) {
					io.to(socket.id).emit(progid, "finish", "", "");
				}
			}
			else{
				io.to(socket.id).emit(progid, "finish", "", "");	
			}
		});
	});
	
	socket.on("kill", function(msg){
		let pid = parseInt(msg);
		process.kill(pid);
		console.log("process "+msg+" killed.")
	});

	socket.on("getfiles", function(dirpath){
		fs_node.readdir(dirpath, {withFileTypes: true}, (err, dirents) => {
			let paths = [];
			if (err) {
				io.to(socket.id).emit("error", "ERROR: Cannot get the file list in the server: "+err.toString());
				return;
			}	
			for (const dirent of dirents) {
				if (dirent.isDirectory() == false) {
					paths.push(dirent.name);
				}
			}
			let retstr = JSON.stringify({files:paths});
			io.to(socket.id).emit("files", dirpath, retstr);
		});
	});

	socket.on("load", function(dirpath, fname){
		load_file(socket.id, dirpath, fname);
	});

	socket.on("save", function(dirpath, fname, ltxt, isconv){
		fs_node.writeFileSync(
			fs_path.join(dirpath, fname), ltxt, "utf8");
		if(isconv){
			try {
				fs_process.chdir("./prm");
			}
			catch (e) {
				io.to(socket.id).emit("error", "ERROR: \"prm\" directory not found.");
				return;
			}	
			let pexec = fs_path.join("..", "conv_spectra");
			let spawn = require("child_process").spawn;
			let jsonfile = convert_filename2json(fname);
			let childp = spawn(pexec, [fname, jsonfile]);	
			fs_process.chdir("..");

			childp.on("exit", (code) => {
				if(code == 0){
					load_file(socket.id, "prm", convert_filename2json(fname));
				}
				else{
					io.to(socket.id).emit("error", "ERROR: Conversion failed.");	
				}
			});
		}
	});

});

/*
let url = sslobj.protocol+"://localhost:"+port.toString();
var opn = require("opn");
opn(url);
*/