"use strict";

// constants
const COEF_GAMMA_EMEV = 0.510999;
const COEF_BPEAK = 1.80063;
const CC = 2.9979246e+8;
const COEF_BM_RADIUS = 3.33564;
const COEF_EC = 665.025;
const ONE_ANGSTROM_eV = 12398.4247;
const COEF_LINPWD = 4.2208e-3;
const COEF_K_VALUE = 93.3729;
const COEF_E1ST = 9.49634;
const COEF_TOTAL_POWER_ID = 3.6284427e-5;
const COEF_FLUX_UND =  1.7443e+14;
const LOGOFFSET = 1.1;
const ALL_UPDATE = "all";
const TypeLabel = "Type";
const OrgTypeLabel = "Original Type"
const OptionLabel =  "Options";
const AutomaticLabel = "Automatic";
const NoneLabel = "None";
const UpdateBeam = "EBeam";
const UpdateCurr = "ECurrent";
const UnselLabel = "Unselected";
const SeparatorLabel = "Separator";
const FileLabel = "FILE";
const FloatLabel = "FLOATNUMBER";
const FolderLabel = "FOLDER";
const CommentLabel = "COMMENT";
const PlotObjLabel = "PLOT";
const PrmTableLabel = "PRMTABLE";
const CalcStatusLabel = "Calculation Status: ";
const Fin1ScanLabel = "Scan Process: ";
const ScanOutLabel = "Output File: ";
const ErrorLabel = "Error: ";
const NumberLabel = "number";
const IntegerLabel = "integer";
const TilteLabel = "titles";
const DataLabel = "data";
const GridLabel = "grid";
const BoolLabel = "boolean";
const StringLabel = "string";
const SelectionLabel = "selection";
const ArrayLabel = "array";
const TimeLabel = "time (fs)"
const DepthLabel = "Depth (mm)"
const NormCurrLabel = "j (A/100%.E.S)";
const CurrLabel = "I (A)";
const EdevLabel = "DE/E";
const ZLabel = "z (m)";
const BxLabel = "Bx (T)";
const ByLabel = "By (T)";
const GapLabel = "Gap (mm)";
const EnergyLabel = "Energy (eV)";
const TransmLabel = "Transmission";
const ScanPrmItems = "ScanPrmItems";
const SeedWavelLabel = "Wavelength (nm)";
const SeedFluxLabel = "Intensity";
const SeedPhaseLabel = "Phase (deg.)";

const WindowHeightOffset = 50;
const JSONIndent = 2;

// labels
const MainTitle = "Main Parameters";
const PreProcTitle = "Pre-Processing";
const PostProcTitle = "Post-Processing";
const PrmSetLabel = "PrmSet";
const BLLabel = "Beamline";
const AccLabel = "Accelerator";
const SrcLabel = "Light Source";
const ConfigLabel =  "Configurations";
const CurrentLabel = "Current";
const FMaterialLabel = "Filter Materials"
const FALabel = "Filter/Absorber";
const ScanLabel = "Scan Parameter";
const BundleScanlabel = "Bundle the output data";
const ScanOptionlabel = "Scan Option";
const AccuracyLabel = "Accuracy Settings";
const UnitLabel = "Unit for Data Import";
const PCLabel = "Parallel Computing";
const PhaseErrLabel = "Phase Error Evaluation"
const PlotFilterLabel = "Transmission/Absorption Plot";
const POLabel = "Plot Options";
const PSNameLabel = "Parameter Set";
const OutFileLabel = "Output File";
const ReplaceLabel = "Replace Data";
const AccTypeCaption = "Accelerator Type";
const SrcTypeCaption = "Light Source Type";
const CalcTypeLabel = "Calculation Type";
const CalcProcessLabel ="Calculation Status/Processes";
const CalcResultLabel = "Calculation Result";
const SpatialDepLabel = "Spatial Dependence";
const InputLabel = "Input";
const OutputLabel = "Output";
const DataDimLabel = "dimension";
const DataTitlesLabel = "titles";
const UnitsLabel = "units";
const VariablesLabel = "variables";
const DetailsLabel = "details";
const FELSecIdxLabel = "Section";
const PErrThreshLabel = "Threshold for End Sections (%)";
const PErrZcoorLabel = "Longitudinal Coordinate";
const FiltPlotTypeLabel = "Plot Configuration";
const FiltPlotEmin = "Minimum Energy (eV)";
const FiltPlotEmax = "Maximum Energy (eV)";
const FiltPlotPoints = "Energy Points";
const FiltPlotEscale = "Energy Step";

const CancellAllLabel = "Cancel All";
const CancelLabel = "Cancel";
const RemoveLabel = "Remove";
const FixedPointLabel = "Fixed Point Calculation";
const VPDLabel = "Volume Power Density";
const CMDLabel = "Coherent Mode Decomposition"

const DimensionLabel = "Dimension";
const TargetsLabel = "TargetItems";
const TitlesLabel = "TargetTitles";
const OrdinateLabel = "OrdinateTitle";
const CalcIDLabel = "CalculationID";
const PreprocIDLabel = "PreprocessID";
const SimplifiedLabel = "Simplified";

const RelateDataLabel = "Related Data";

const CMDResultLabel = "CMD Result";
const CMDModalFluxLabel = "Modal Flux";
const CMDFieldLabel = "Modal Profile";
const CMDIntensityLabel = "Modal Intensity";
const CMDCompareIntLabel = "Flux Density Profile";
const CMDCompareXLabel = "Wigner Func. (X,X')";
const CMDCompareYLabel = "Wigner Func. (Y,Y')";
const CMDErrorLabel = "numerical validity";

const FELCurrProfile = "Current Profile";
const FELEtProfile = "E-t Profile";
const FELCurrProfileR56 = "Current Profile with Dispersion";
const FELEtProfileR56 = "E-t Profile with Dispersion";

const FELBunchFactor = "Bunch Factor";
const FELPulseEnergy = "Pulse Energy";
const FELEfield = "On-Axis Field";
const FELInstPower = "Inst. Power";
const FELSpectrum = "Spectrum";

const MaxOrderLabel = "maxorder";
const WavelengthLabel = "wavelength";
const SrcSizeLabel = "size";
const OrderLabel = "order";
const AmplitudeReLabel = "anm.re";
const AmplitudeImLabel = "anm.im";

const AmplitudeVReLabel = "anmv.re";
const AmplitudeVImLabel = "anmv.im";
const AmplitudeIndexReLabel = "anmidx.re";
const AmplitudeIndexImLabel = "anmidx.im";

const NormFactorLabel = "coefficient";
const FluxCMDLabel = "photon flux";
const MatrixErrLabel = "Matrix Error";
const FluxErrLabel = "Flux Consistency";
const WignerErrLabel = "Wigner Function Consistency";

const SuffixButton = "_btn";
const SuffixFile = "_file";
const SuffixPlotBody = "_plot";
const SuffixPlotDiv = "_div";
const IDSeparator = "::";

const GridColLabel = "ColLabel";
const GridTypeLabel = "ColType";
const AdditionalRows = 2;

// accelerator parameters
const AccPrmsLabel = {
    eGeV:"Energy (GeV)",
    imA:"Current (mA)",
    aimA:"Avg. Current (mA)",
    cirm:"Circumference (m)",
    bunches:"Bunches",
    pulsepps:"Pulses/sec",
    bunchlength:"&sigma;<sub>z</sub> (mm)",
    bunchcharge:"Bunch Charge (nC)",
    emitt:"Nat. Emittance (m.rad)",
    coupl:"Coupling Constant",
    espread:"Energy Spread",
    beta:"&beta;<sub>x,y</sub> (m)",
    alpha:"&alpha;<sub>x,y</sub>",
    eta:"&eta;<sub>x,y</sub> (m)",
    etap:"&eta;'<sub>x,y</sub>",
    peakcurr:"Peak Current (A)",
    epsilon:"&epsilon;<sub>x,y</sub> (m.rad)",
    sigma:"&sigma;<sub>x,y</sub> (mm)",
    sigmap:"&sigma;'<sub>x,y</sub> (mrad)",
    gaminv:"&gamma;<sup>-1</sup> (mrad)",
    R56add:"Additional R<sub>56</sub> (m)",
    minsize:"Beam Size Lower Limit (m)",
};

const AccLabelOrder = [
    "eGeV",
    "imA",
    "aimA",
    "cirm",
    "bunches",
    "pulsepps",
    "bunchlength",
    "bunchcharge",
    "emitt",
    "coupl",
    "espread",
    "beta",  
    "alpha",
    "eta",
    "etap",
    "peakcurr",
    "epsilon",
    "sigma",
    "sigmap",
    "gaminv",
    "R56add",
    "minsize"
];

var AccPrmsFormat = {};
var AccIsScan = {};
for(let i = 0; i < AccLabelOrder.length; i++){
    AccPrmsFormat[AccLabelOrder[i]] = NumberLabel;
    AccIsScan[AccLabelOrder[i]] = true;
}
AccPrmsFormat.beta = ArrayLabel;
AccPrmsFormat.alpha = ArrayLabel;
AccPrmsFormat.eta = ArrayLabel;
AccPrmsFormat.epsilon = ArrayLabel;
AccPrmsFormat.etap = ArrayLabel;
AccPrmsFormat.sigma = ArrayLabel;
AccPrmsFormat.sigmap = ArrayLabel;
AccIsScan.cirm = false;
AccIsScan.bunches = false;
AccIsScan.pulsepps = false;

// accelerator injection conditions
const InjectionPrmsLabel = {
    xy:"x,y (mm)",
    xyp:"x',y' (mrad)"
};

const InjectionLabelOrder = [
    "xy",
    "xyp"
];

var InjectionPrmsFormat = {
    xy:ArrayLabel,
    xyp:ArrayLabel
};

var InjectionIsScan =
{
    xy:true,
    xyp:true
}

// accelerator types
const AccTypeLabel = {
    RING:"Storage Ring",
    LINAC:"Linear Accelerator"
};

const AccTypeOrder = [
    "RING",
    "LINAC"
];

// accelerator options
const GaussianLabel = "Gaussian";
const ImportPaticleLabel = "Import Particles";
const CurrProfLabel = "Import Current Profile";
const EtProfLabel = "Import E-t Profile";
const CustomLabel = "Custom";
const EntranceLabel = "Align at Entrance";
const CenterLabel = "Align at Center";
const ExitLabel = "Align at Exit";

var AccOptionsLabel = {
    bunchtype:["Bunch Profile",
        [GaussianLabel, CurrProfLabel, EtProfLabel, ImportPaticleLabel]],
    bunchdata:["Particle Data", FileLabel],
    currdata:["Current Profile", PlotObjLabel],    
    Etdata:["E-t Profile", PlotObjLabel],    
    injectionebm:["Injection Condition",
        [AutomaticLabel, EntranceLabel, CenterLabel, ExitLabel, CustomLabel]],
    injecprm:["InjecPrm", PrmTableLabel],
    zeroemitt:["Zero Emittance", false],
    zerosprd:["Zero Energy Spread", false],
    singlee:["Single Electron", false]
};

const AccOptionsOrder = [
    "bunchtype",
    "bunchdata",
    "currdata",
    "Etdata",
    SeparatorLabel,
    "injectionebm",
    "injecprm",
    SeparatorLabel,
    "zeroemitt",
    "zerosprd",
    "singlee"
];

const AccOptionsPrmTbl = {
    injecprm:{index:InjectionLabelOrder, label:InjectionPrmsLabel, fmt:InjectionPrmsFormat}
};

// light source parameters
const SrcPrmsLabel = {
    gap:"Gap (mm)",
    bxy:"B<sub>x,y</sub> (T)",
    b:"B (T)",
    bmain:"Main Field (T)",
    subpoleb:"Sub Field (T)",
    lu:"&lambda;<sub>u</sub> (mm)",
    devlength:"Device Length (m)",
    reglength:"Reg. Magnet Length (m)",
    periods:"# of Reg. Periods",
    Kxy0:"K<sub>0x,0y</sub>",
    phase:"Phase Shift (mm)",
    Kxy:"K<sub>x,y</sub>",
    K:"K value",
    Kperp:"K<sub>&perp;</sub>",
    e1st:"&epsilon;<sub>1st</sub> (eV)",
    radius:"&rho; (m)",
    bendlength:"BM Length (m)",
    fringelen:"BM Fringe Length (m)",
    mplength:"Main Length (m)",
    subpolel:"Sub Length (m)",
    bminterv:"BM Interval (m)",
    csrorg:"Origin for CSR (m)",
    fvsz:"Field Profile",
    fvsz1per:"Field Profile (1 Period)",
    fmap:"Field Mapping Data",
    multiharm:"Harmonic Component",
    sigmar:"&sigma;<sub>r,r'</sub> (mm,mrad)",
    sigmarx:"&sigma;<sub>rx,rx'</sub> (mm,mrad)",
    sigmary:"&sigma;<sub>ry,ry'</sub> (mm,mrad)",
    Sigmax:"&Sigma;<sub>x,x'</sub> (mm,mrad)",
    Sigmay:"&Sigma;<sub>y,y'</sub> (mm,mrad)",
    lambda1:"&lambda;<sub>1st</sub> (nm)",
    fd:"Flux Density",
    flux:"Flux<sub>1st</sub>",
    brill:"Brilliance<sub>1st</sub>",
    pkbrill:"Peak Brilliance",
    degener:"Bose Degeneracy",
    ec:"&epsilon;<sub>c</sub> (eV)",
    lc:"&lambda;<sub>c</sub> (nm)",
    tpower:"Total Power (kW)",
	  tpowerrev:"Total Power/Rev. (kW)",
    linpower:"Lin. Pow. Density (kW/mrad)"
};

const SrcLabelOrder = [
    "fvsz",
    "fvsz1per",
    "gap",
    "bxy",
    "b",
    "bmain",
    "subpoleb",
    "lu",
    "devlength",
    "reglength",
    "periods",
    "Kxy0",
    "phase",
    "Kxy",
    "K",
    "Kperp",
    "e1st",
    "multiharm",
    "radius",
    "bendlength",
    "fringelen",
    "mplength",
    "subpolel",
    "bminterv",
    "csrorg",
    "fmap",
    //------ no input paramters below
    "sigmar",
    "sigmarx",
    "sigmary",
    "Sigmax",
    "Sigmay",
    "lambda1",
    "fd",
    "flux",
    "brill",
    "pkbrill",
    "degener",
    "ec",
    "lc",
    "tpower",
    "tpowerrev",
    "linpower" 
];

var SrcPrmsFormat = {};
var SrcIsScan = {};
for(let i = 0; i < SrcLabelOrder.length; i++){
    SrcPrmsFormat[SrcLabelOrder[i]] = NumberLabel;
    SrcIsScan[SrcLabelOrder[i]] = true;
}

SrcPrmsFormat.bxy = ArrayLabel;
SrcPrmsFormat.Kxy0 = ArrayLabel;
SrcPrmsFormat.Kxy = ArrayLabel;
SrcPrmsFormat.sigmar = ArrayLabel;
SrcPrmsFormat.sigmarx = ArrayLabel;
SrcPrmsFormat.sigmary = ArrayLabel;
SrcPrmsFormat.Sigmax = ArrayLabel;
SrcPrmsFormat.Sigmay = ArrayLabel;
SrcPrmsFormat.fvsz = DataLabel;
SrcPrmsFormat.fvsz1per = DataLabel;
SrcPrmsFormat.fmap = FileLabel;
SrcPrmsFormat.multiharm = GridLabel;

// light source field error parameters
const FerrPrmsLabel = {
    boffset:"Offset x,y (T)",
    ltaper:"Lin. Taper x,y (/m)",
    qtaper:"Quad. Taper x,y (/m<sup>2</sup>)"
};

const FerrLabelOrder = [
    "boffset",
    "ltaper",
    "qtaper"
];

var FerrPrmsFormat = {};
var FerrIsScan = {};
for(let i = 0; i < FerrLabelOrder.length; i++){
    FerrPrmsFormat[FerrLabelOrder[i]] = ArrayLabel;
    FerrIsScan[FerrLabelOrder[i]] = true;
}

// light source field error parameters
const PerrPrmsLabel = {
    seed:"Random Number Seed",
    fsigma:"&sigma;<sub>B</sub> (%)",
    psigma:"&sigma;<sub>&phi;</sub> (deg.)",
    xysigma:"&sigma;<sub>x,y</sub> (mm);"
};

const PerrLabelOrder = [
    "seed",
    "fsigma",
    "psigma",
    "xysigma"
];

var PerrPrmsFormat = {};
var PerrIsScan = {};
for(let i = 0; i < PerrLabelOrder.length; i++){
    PerrPrmsFormat[PerrLabelOrder[i]] = NumberLabel;
    PerrIsScan[PerrLabelOrder[i]] = true;
}
PerrPrmsFormat.xysigma = ArrayLabel;
PerrIsScan.seed = 1;

// light source segment parameters
const SegPrmsLabel = {
    segments:"Number of Segments",
    hsegments:"Half Number of Segments",
    interval:"Segment Interval (m)",
    pslip:"Number of Phase Slip@&lambda;<sub>1</sub>",
    phi0:"&Delta;&phi; (&pi;)",
    phi12:"&Delta;&phi;<sub>1,2</sub> (&pi;)",
    mdist:"Matching Distance (m)"
};

const SegLabelOrder = [
    "segments",
    "hsegments",
    "interval",
    "pslip",
    "phi0",
    "phi12",
    "mdist"
];

var SegPrmsFormat = {};
var SegIsScan = {};
for(let i = 0; i < SegLabelOrder.length; i++){
    SegPrmsFormat[SegLabelOrder[i]] = NumberLabel;
    SegIsScan[SegLabelOrder[i]] = true;
}
SegIsScan.segments = 1;
SegIsScan.hsegments = 1;
SegPrmsFormat.phi12 = ArrayLabel;

// light source magnet conditions
const MagnetPrmsLabel = {
    br:"B<sub>r</sub> (T)",
    geofactor:"Geometrical Factor (x,y)",
};

const MagnetLabelOrder = [
    "br",
    "geofactor",
];

var MagnetPrmsFormat = 
{
    br: NumberLabel,
    geofactor: ArrayLabel
};

var MagnetIsScan = 
{
    br: false,
    geofactor: false
};

// light source types
const SrcTypeLabel = {
    LIN_UND:"Linear Undulator",
    VERTICAL_UND:"Vertical Undulator",
    HELICAL_UND:"Helical Undulator",
    ELLIPTIC_UND:"Elliptic Undulator",
    FIGURE8_UND:"Figure-8 Undulator",
    VFIGURE8_UND:"Vertical Figure-8 Undulator",
    MULTI_HARM_UND:"Multi-Harmonic Undulator",
    BM:"Bending Magnet",
    WIGGLER:"Wiggler",
    EMPW:"EMPW",
    WLEN_SHIFTER:"Wavelength Shifter",
    FIELDMAP3D:"Field Mapping",
    CUSTOM_PERIODIC:"Periodic: User Defined",
    CUSTOM:"User Defined"
}
const SrcTypeOrder = [
    "LIN_UND",
    "VERTICAL_UND",
    "HELICAL_UND",
    "ELLIPTIC_UND",
    "FIGURE8_UND",
    "VFIGURE8_UND",
    "MULTI_HARM_UND",
    "WIGGLER",
    "EMPW",
    "WLEN_SHIFTER",
    "BM",
    "FIELDMAP3D",
    "CUSTOM_PERIODIC",
    "CUSTOM"    
];
var SrcTypeObj = [];

// light source options
const BxOnlyLabel = "Bx";
const ByOnlyLabel = "By";
const BothLabel = "Both";
const IdenticalLabel = "Identical";
const ImpGapTableLabel = "Import Table"
const SwapBxyLabel = "2nd: Swap Bx,y";
const FlipBxLabel = "2nd: Flip Bx";
const FlipByLabel = "2nd: Flip By";
const AntiSymmLabel = "Antisymmetric";
const SymmLabel = "Symmetric";

const SrcOptionsLabel = {
    gaplink:["Gap-Field Relation",
        [NoneLabel, AutomaticLabel, ImpGapTableLabel]],
    gaptbl:["Gap vs. Field", PlotObjLabel],

    apple:["APPLE Configuration", false],

    field_str:["Field Structure", [AntiSymmLabel, SymmLabel]],
    endmag:["End Correction Magnet", true],

    natfocus:["Natural Focusing",
        [NoneLabel, BxOnlyLabel, ByOnlyLabel, BothLabel]],
    fielderr:["Field Offset & Taper", false],
    ferrprm:["FerrPrms", PrmTableLabel],

    phaseerr:["Add Phase Error", false],
    perrprm:["PerrPrms", PrmTableLabel],

    bmtandem:["Tandem Arrangement", false],

    segment_type:["Segmentation",
        [NoneLabel, IdenticalLabel, SwapBxyLabel, FlipBxLabel, FlipByLabel]],
    segprm:["SegPrms", PrmTableLabel],
    perlattice:["Periodic &beta; Function", false],
    magconf:["Magnet Configurations", PrmTableLabel]
};

const SrcOptionsOrder = [
    "gaplink",
    "gaptbl",
    "apple",
    "field_str",
    "endmag",
    SeparatorLabel,
    "natfocus",
    "fielderr",
    "ferrprm",
    "phaseerr",
    "perrprm",
    "bmtandem",
    SeparatorLabel,
    "segment_type",
    "segprm",
    "perlattice",
    "magconf"
];

const SrcOptionsPrmTbl = {
    ferrprm:{index:FerrLabelOrder, label:FerrPrmsLabel, fmt:FerrPrmsFormat},
    perrprm:{index:PerrLabelOrder, label:PerrPrmsLabel, fmt:PerrPrmsFormat},
    segprm:{index:SegLabelOrder, label:SegPrmsLabel, fmt:SegPrmsFormat},
    magconf:{index:MagnetLabelOrder, label:MagnetPrmsLabel, fmt:MagnetPrmsFormat}
};

// configurations
const ConfigPrmsLabel = {
    // z positions
    slit_dist:"Distance from the Source (m)",
    zrange:"z range (m)",
    zmesh:"Points (z)",

    // energy 
    erange:"Energy Range (eV)",
    de:"Energy Pitch (eV)",
    epitch:"Energy Pitch for Integration (eV)",
    emesh:"Points (Energy)",
    efix:"Target Energy (eV)",
    nefix:"Normalized Energy",
    detune:"Detuning",

    // obs. points
    xyfix:"Position x,y (mm)",
    qxyfix:"Angle &theta;<sub>x,y</sub> (mrad)",
    spdxfix:"Surface Pos. x (mm)",
    spdyfix:"Surface Pos. y (mm)",
    spdrfix:"Surface Radius (mm)",
    Qnorm:"&Theta; (deg.)",
    Phinorm:"&Phi; (deg.)",
    Qgl:"Glancing Angle (deg.)",
    Phiinc:"Azimuth of Incidence (deg.)",

    // slit
    slitpos:"Slit Pos.: x,y (mm)",
    qslitpos:"Slit Pos.: &theta;<sub>x,y</sub> (mrad)",
    nslitapt:"&Delta;/&Sigma;<sub>s</sub>: x,y",
    slitapt:"&Delta;x,&Delta;y (mm)",
    qslitapt:"&Delta;&theta;<sub>x,y</sub> (mrad)",
    slitr:"Slit r<sub>1,2</sub> (mm)",
    slitq:"Slit &theta;<sub>1,2</sub> (mrad)",
    pplimit:"Power Upper Limit (kW)",
    illumarea:"Illuminated x,y (mm)",

    // obs. range 
    xrange:"x Range (mm)",
    qxrange:"&theta;<sub>x</sub> Range (mrad)",
    xmesh:"Points (x)",
    yrange:"y Range (mm)",
    qyrange:"&theta;<sub>y</sub> Range (mrad)",
    ymesh:"Points (y)",
    rrange:"r Range (mm)",
    qrange:"&theta; Range (mrad)",
    rphimesh:"Points (r)",
    qphimesh:"Points (&theta;)",
    phirange:"&phi; Range (deg.)",
    phimesh:"Points (&phi;)",
    drange:"Depth Range (mm)",
    dmesh:"Points (Depth)",

    // beam size (flux, power)
    fsize:"&Sigma;<sub>x,y</sub>@&epsilon;<sub>1st</sub> (mm)",
    psize:"&Sigma;<sub>px,py</sub> (mm)",
    fdiv:"&Sigma;<sub>x',y'</sub>@&epsilon;<sub>1st</sub> (mrad)",
    pdiv:"&Sigma;<sub>px',py'</sub> (mrad)",

    // K
    krange:"K Range",
    ckrange:"K<sub>&perp;</sub> Range",
    kmesh:"Points (K)",
    e1strange:"&epsilon;<sub>1st</sub> Range (eV)",

    // harmonic
    hrange:"Harmonic Range",
    hfix:"Target Harmonic",
    hmax:"Maximum Harmonic",

    // temporal
    trange:"Temporal Range (fs)",
    tmesh:"Points (Temporal)",

    // source point
    gtacc:"&gamma;&Delta;&theta;<sub>x,y</sub>",
    horizacc:"X' Acceptance (mrad)",
    Xfix:"Slice X (mm)",
    Yfix:"Slice Y (mm)",
    Xpfix:"Slice X' (mrad)",
    Ypfix:"Slice Y' (mrad)",
    Xrange:"X Range (mm)",
    Xmesh:"Points (X)",
    Xprange:"X' Range (mrad)",
    Xpmesh:"Points (X')",
    Yrange:"Y Range (mm)",
    Ymesh:"Points (Y)",
    Yprange:"Y' Range (mrad)",
    Ypmesh:"Points (Y')",
};

const ConfigLabelOrder = [
    // observation point
    "slit_dist",

    // harmonic
    "hrange",
    "hfix",
    "hmax",

    // energy 
    "detune",
    "erange",
    "de",
    "epitch",
    "emesh",
    "nefix",
    "efix",

    // obs. points
    "xyfix",
    "qxyfix",
    "spdxfix",
    "spdyfix",
    "spdrfix",
    "Qnorm",
    "Phinorm",
    "Qgl",
    "Phiinc",

    // slit
    "slitpos",
    "qslitpos",
    "nslitapt",
    "slitapt",
    "qslitapt",
    "slitr",
    "slitq",
    "pplimit",
    "illumarea",

    // obs. range 
    "xrange",
    "qxrange",
    "xmesh",
    "yrange",
    "qyrange",
    "ymesh",
    "rrange",
    "qrange",
    "rphimesh",
    "qphimesh",
    "phirange",
    "phimesh",
    "drange",
    "dmesh",

    // z range
    "zrange",
    "zmesh",

    // beam size (flux, power)
    "fsize",
    "psize",
    "fdiv",
    "pdiv",

    // K
    "krange",
    "ckrange",
    "e1strange",
    "kmesh",

    // temporal
    "trange",
    "tmesh",

    // Wigner
    "gtacc",
    "horizacc",
    "Xfix",
    "Yfix",
    "Xpfix",
    "Ypfix",
    "Xrange",
    "Xmesh",
    "Xprange",
    "Xpmesh",
    "Yrange",
    "Ymesh",
    "Yprange",
    "Ypmesh"
];

var ConfigPrmsFormat = {};
var ConfigPrmsIsScan = {};
for(let i = 0; i < ConfigLabelOrder.length; i++){
    ConfigPrmsFormat[ConfigLabelOrder[i]] = ArrayLabel;
    ConfigPrmsIsScan[ConfigLabelOrder[i]] = true;
}
ConfigPrmsIsScan.zrange = false;
ConfigPrmsIsScan.zmesh = false;
ConfigPrmsIsScan.hrange = false;
ConfigPrmsIsScan.hfix = 1;
ConfigPrmsIsScan.hmax = false;
ConfigPrmsIsScan.erange = false;
ConfigPrmsIsScan.de = false;
ConfigPrmsIsScan.epitch = false;
ConfigPrmsIsScan.emesh = false;
ConfigPrmsIsScan.xrange = false;
ConfigPrmsIsScan.qxrange = false;
ConfigPrmsIsScan.xmesh = false;
ConfigPrmsIsScan.yrange = false;
ConfigPrmsIsScan.qyrange = false;
ConfigPrmsIsScan.ymesh = false;
ConfigPrmsIsScan.rrange = false;
ConfigPrmsIsScan.qrange = false;
ConfigPrmsIsScan.rphimesh = false;
ConfigPrmsIsScan.qphimesh = false;
ConfigPrmsIsScan.phirange = false;
ConfigPrmsIsScan.phimesh = false;
ConfigPrmsIsScan.drange = false;
ConfigPrmsIsScan.dmesh = false;
ConfigPrmsIsScan.krange = false;
ConfigPrmsIsScan.ckrange = false;
ConfigPrmsIsScan.kmesh = false;
ConfigPrmsIsScan.trange = false;
ConfigPrmsIsScan.tmesh = false;
ConfigPrmsIsScan.Xrange = false;
ConfigPrmsIsScan.Xmesh = false;
ConfigPrmsIsScan.Xprange = false;
ConfigPrmsIsScan.Xpmesh = false;
ConfigPrmsIsScan.Yrange = false;
ConfigPrmsIsScan.Ymesh = false;
ConfigPrmsIsScan.Yprange = false;
ConfigPrmsIsScan.Ypmesh = false;

ConfigPrmsFormat.slit_dist = NumberLabel;
ConfigPrmsFormat.zmesh = NumberLabel;
ConfigPrmsFormat.de = NumberLabel;
ConfigPrmsFormat.epitch = NumberLabel;
ConfigPrmsFormat.emesh = NumberLabel;
ConfigPrmsFormat.efix = NumberLabel;
ConfigPrmsFormat.nefix = NumberLabel;
ConfigPrmsFormat.detune = NumberLabel;
ConfigPrmsFormat.spdxfix = NumberLabel;
ConfigPrmsFormat.spdyfix = NumberLabel;
ConfigPrmsFormat.spdrfix = NumberLabel;
ConfigPrmsFormat.Qnorm = NumberLabel;
ConfigPrmsFormat.Phinorm = NumberLabel;
ConfigPrmsFormat.Qgl = NumberLabel;
ConfigPrmsFormat.Phiinc = NumberLabel;
ConfigPrmsFormat.xmesh = NumberLabel;
ConfigPrmsFormat.ymesh = NumberLabel;
ConfigPrmsFormat.rphimesh = NumberLabel;
ConfigPrmsFormat.qphimesh = NumberLabel;
ConfigPrmsFormat.phimesh = NumberLabel;
ConfigPrmsFormat.dmesh = NumberLabel;
ConfigPrmsFormat.kmesh = NumberLabel;
ConfigPrmsFormat.hfix = NumberLabel;
ConfigPrmsFormat.hmax = NumberLabel;
ConfigPrmsFormat.tmesh = NumberLabel;
ConfigPrmsFormat.pplimit = NumberLabel;
ConfigPrmsFormat.horizacc = NumberLabel;
ConfigPrmsFormat.Xfix = NumberLabel;
ConfigPrmsFormat.Yfix = NumberLabel;
ConfigPrmsFormat.Xpfix = NumberLabel;
ConfigPrmsFormat.Ypfix = NumberLabel;
ConfigPrmsFormat.Xmesh = NumberLabel;
ConfigPrmsFormat.Xpmesh = NumberLabel;
ConfigPrmsFormat.Ymesh = NumberLabel;
ConfigPrmsFormat.Ypmesh = NumberLabel;

// configuration of BPF conditions
const BPFPrmsLabel = {
    bpfcenter:"Central Energy (eV)",
    bpfwidth:"Width (eV)",
    bpfsigma:"Width (&sigma;, eV)",
    bpfmaxeff:"Max. Trans. Rate"
};

const BPFLabelOrder = [
    "bpfcenter",
    "bpfwidth",
    "bpfsigma",
    "bpfmaxeff"
];

var BPFPrmsFormat = {};
var BPFIsScan = {};
for(let i = 0; i < BPFLabelOrder.length; i++){
    BPFPrmsFormat[BPFLabelOrder[i]] = NumberLabel;
    BPFIsScan[BPFLabelOrder[i]] = true;
}

// configuration of CMD conditions
const CMDPrmsLabel = {
    HGorderxy:"HG Order Limit (X,Y)",
    HGorderx:"HG Order Limit (X)",
    HGordery:"HG Order Limit (Y)",
    maxHGorderxy:"Max. HG Order (X,Y)",
    maxHGorderx:"Max. HG Order (X)",
    maxHGordery:"Max. HG Order (Y)",
    maxmode:"Maximum CMD Order",
    fcutoff:"Flux Cutoff",
    cutoff:"Amplitude Cutoff",
    fieldgridxy:"Export Step: X,Y (mm)",
    fieldgridx:"Export Step: X (mm)",
    fieldgridy:"Export Step: Y (mm)"
};

const CMDLabelOrder = [
    "HGorderxy",
    "HGorderx",
    "HGordery",
    "maxHGorderxy",
    "maxHGorderx",
    "maxHGordery",
    "maxmode",
    "fcutoff",
    "cutoff",
    "fieldgridxy",
    "fieldgridx",
    "fieldgridy"
];

var CMDPrmsFormat = {};
var CMDIsScan = {};
for(let i = 0; i < CMDLabelOrder.length; i++){
    CMDPrmsFormat[CMDLabelOrder[i]] = NumberLabel;
    CMDIsScan[CMDLabelOrder[i]] = false;
}
CMDPrmsFormat.HGorderxy = ArrayLabel;
CMDPrmsFormat.maxHGorderxy = ArrayLabel;
CMDPrmsFormat.fieldgridxy = ArrayLabel;

// configuration of FEL-related parameters
const FELPrmsLabel = {
  pulseE:"Pulse Energy (mJ)",
  wavelen:"Wavelength (nm)",
  pulselen:"Pulse Length (FWHM, fs)",
  tlpulselen:"TL. Pulse Length (FWHM, fs)",
  srcsize:"Source Size (FWHM, mm)",
  waistpos:"Waist Position (m)",
  timing: "Timing (fs)",
  gdd: "GDD (fs<sup>2</sup>)",
  tod: "TOD (fs<sup>3</sup>)",

  pulseE_d:"Pulse Energy: 1,2 (mJ)",
  wavelen_d:"Wavelength: 1,2 (nm)",
  tlpulselen_d:"TL. Pulse Length: 1,2 (FWHM, fs)",
  srcsize_d:"Source Size: 1,2 (FWHM, mm)",
  waistpos_d:"Waist Position: 1,2 (m)",
  timing_d: "Timing: 1,2 (fs)",
  gdd_d: "GDD: 1,2 (fs<sup>2</sup>)",
  tod_d: "TOD: 1,2 (fs<sup>3</sup>)",

  svstep: "Step: Initial, Interval (m)",
  radstep: "Substeps for Radiation",
  eproi: "Photon Energy ROI (eV)",
  particles: "Number of Particles",
  edevstep: "e- Energy Interval",
  R56: "R<sub>56</sub> (m)"
};

const FELLabelOrder = [
  "pulseE",
  "wavelen",
  "pulselen",
  "tlpulselen",
  "srcsize",
  "waistpos",
  "timing",
  "gdd",
  "tod",

  "pulseE_d",
  "wavelen_d",
  "tlpulselen_d",
  "srcsize_d",
  "waistpos_d",
  "timing_d",
  "gdd_d",
  "tod_d",

  "svstep",
  "radstep",
  "eproi",
  "particles",
  "edevstep",
  "R56"
];

var FELPrmsFormat = {};
var FELIsScan = {};
for(let i = 0; i < FELLabelOrder.length; i++){
  FELPrmsFormat[FELLabelOrder[i]] = NumberLabel;
  FELIsScan[FELLabelOrder[i]] = true;
}
FELPrmsFormat.svstep = ArrayLabel;
FELPrmsFormat.eproi = ArrayLabel;
FELIsScan.edevstep = false;

FELPrmsFormat.pulseE_d = ArrayLabel;
FELPrmsFormat.wavelen_d = ArrayLabel;
FELPrmsFormat.tlpulselen_d = ArrayLabel;
FELPrmsFormat.srcsize_d = ArrayLabel;
FELPrmsFormat.waistpos_d = ArrayLabel;
FELPrmsFormat.timing_d = ArrayLabel;
FELPrmsFormat.gdd_d = ArrayLabel;
FELPrmsFormat.tod_d = ArrayLabel;

// configuration options
const FixedSlitLabel = "Fixed";
const NormSlitLabel = "Normalized";
const GenFilterLabel = "Generic Filter";
const BPFGaussianLabel = "BPF: Gaussian";
const BPFBoxCarLabel = "BPF: Boxcar";
const LinearLabel = "Linear";
const LogLabel = "Logarithmic";
const ArbPositionsLabel = "Export at Arbitrary Positions";
const ObsPointDist = "Position";
const ObsPointAngle = "Angle";
const CMDParameterLabel = "CMD Parameter";
const FELConfigLabel = "FEL Configurations"
const FELPrebunchedLabel = "Prebunched FEL";
const FELSeedLabel = "Seeded FEL";
const FELSeedCustomLabel = "Seeded width Custom Pulse";
const FELCPSeedLabel = "Seeded with Chirped Pulse";
const FELDblSeedLabel = "Seeded with Double Pulse";
const FELReuseLabel = "Reuse Bunch Factor";
const PhaseErrZPole = "Magnetic Pole Number";
const PhaseErrZPos = "z (m)";

const JSONOnly = "JSON";
const ASCIIOnly = "ASCII";
const BinaryOnly = "BINARY";
const BothFormat = "Both";
const XOnly = "X";
const YOnly = "Y";
const OutFormat = "Format";
const LineLabel = "Line";
const LineSymbolLabel = "Line & Symbol";
const SymbolLabel = "Symbol";
const SurfaceLabel = "Surface";
const ContourLabel = "Contour";
const DefaultLabel = "Default";
const RainbowLabel = "Rainbow";
const BlackBodyLabel = "Blackbody";
const EarthLabel = "Earth";
const GreysLabel = "Greys";
const ByMaxLabel = "By Maximum";
const ForEachLabel = "For Each";

const ConfigOptionsLabel = {
    filter:["Filtering",
        [NoneLabel, GenFilterLabel, BPFGaussianLabel, BPFBoxCarLabel, CustomLabel]],
    fmateri:["Filters", GridLabel],
    amateri:["Absorbers", GridLabel],
    fcustom:["Custom Filter", PlotObjLabel],
    bpfprms:["BPF Parameter", PrmTableLabel],
    estep:["Energy Step", [LinearLabel, LogLabel]],
    aperture:["Slit Aperture Size", [FixedSlitLabel, NormSlitLabel]],
    dstep:["Depth Step", 
        [LinearLabel, LogLabel, ArbPositionsLabel]],
    depthdata:["Depth-Position Data", PlotObjLabel],
	  defobs:["Define Obs. Point in", [ObsPointDist, ObsPointAngle]],
	  normenergy:["Normalize Photon Energy", false],
    powlimit:["Set Upper Limit on Power", false],
    optDx:["Optimize &Delta;X' for Computation", true],
    xsmooth:["Level of Smoothing Along X", 1],
    fouriep:["Observation in the Fourier Plane", false],
    wiggapprox:["Wiggler Approximation", false],
    esmooth:["Spectral Smoothing", false],
    smoothwin:["Smoothing Window (%)", 1, 0, 0.1],
    acclevel:["Accuracy Level", 1],
    accuracy:["Accuracy", [DefaultLabel, CustomLabel]],
    CMD:["Perform CMD?", false],
    GSModel:["Apply GS Model", false],
    GSModelXY:["GS Model X/Y", [NoneLabel, XOnly, YOnly, BothFormat]],
    CMDfld:["Export Field Profile", [NoneLabel, JSONOnly, BinaryOnly, BothFormat]],
    CMDint:["Export Intensity Profile", false],
    CMDcmp:["Compare Wigner Function", false],
    CMDcmpint:["Compare Intensity Profile", false],
    CMDprms:[CMDParameterLabel, PrmTableLabel],
    fel:["FEL Mode", [NoneLabel, FELPrebunchedLabel, FELSeedLabel, 
      FELCPSeedLabel, FELDblSeedLabel, FELSeedCustomLabel, FELReuseLabel]],
    seedspec:["Seed Spectrum", PlotObjLabel],
    FELprms:[FELConfigLabel, PrmTableLabel],
    exportInt:["Export Intermediate Data", true], 
    R56Bunch:["Bunch with Dispersion", false],
    exportEt:["E-t Data", false]
};

const ConfigOptionsOrder = [
    "filter",
    "fmateri",
    "fcustom",
    "bpfprms",
    SeparatorLabel,
    "amateri",
    "dstep",
    "depthdata",
    SeparatorLabel,
	  "defobs",
    "normenergy",
    "estep",
    "aperture",
    "powlimit",
    "optDx",
	  "xsmooth",
    "fouriep",
    "wiggapprox",
    "esmooth",
    "smoothwin",
    "acclevel",
    "accuracy",
    SeparatorLabel,
    "CMD",
    "GSModel",
    "GSModelXY",
    "CMDfld",
    "CMDint",
    "CMDprms",
    "CMDcmp",
    "CMDcmpint",
    SeparatorLabel,
    "fel",
    "seedspec",
    "FELprms",
    "exportInt",
    "R56Bunch",
    "exportEt"
];

const ConfigOptionsPrmTbl = {
    bpfprms:{index:BPFLabelOrder, label:BPFPrmsLabel, fmt:BPFPrmsFormat},
    CMDprms:{index:CMDLabelOrder, label:CMDPrmsLabel, fmt:CMDPrmsFormat},
    FELprms:{index:FELLabelOrder, label:FELPrmsLabel, fmt:FELPrmsFormat}
};

// built-in filters
const FilterMaterial = {
    Air:{
        dens:0.001184,
        comp:[[7,0.755],[8,0.232],[18,0.013]]
    },
    He:{
        dens:0.000179,
        comp:[[2,1]]
    },
    Ar:{
        dens:0.001784,
        comp:[[18,1]]
    },
    Be:{
        dens:1.84,
        comp:[[4,1]]
    },
    C:{
        dens:2.25,
        comp:[[6,1]]
    },
    Diamond:{
        dens:3.52,
        comp:[[6,1]]
    },
    N2:{
        dens:0.00125,
        comp:[[7,1]]
    },
    Al:{
        dens:2.69,
        comp:[[13,1]]
    },
    Si:{
        dens:2.34,
        comp:[[14,1]]
    },
    Ti:{
        dens:4.506,
        comp:[[22,1]]
    },
    Cu:{
        dens:8.93,
        comp:[[29,1]]
    },
    Mo:{
        dens:10.28,
        comp:[[42,1]]
    },
    Rh:{
        dens:12.41,
        comp:[[45,1]]
    },
    Pt:{
        dens:21.45,
        comp:[[78,1]]
    },
    Au:{
        dens:19.32,
        comp:[[79,1]]
    },
    Pb:{
        dens:11.35,
        comp:[[82,1]]
    },
    Kapton:{
        dens:1.42,
        comp:[[1,0.026],[6,0.69],[7,0.073],[8,0.21]]
    }
}

// parameter scan configurations
const ScanPrmsLabel = {
    initial:"Initial Value",
    final:"Final Value",
    points:"Number of Points",
    interv:"Interval",
    iniser:"Initial Serial Number",
    initial2:"Initial (1st, 2nd)",
    final2:"Final  (1st, 2nd)",
    points2:"Points (1st, 2nd)",
    interv2:"Interval (1st, 2nd)",
    iniser2:"Initial Serial Number (1st, 2nd)"
};

const ScanLabelOrder = [
    "initial",
    "final",
    "points",
    "interv",
    "iniser",
    "initial2",
    "final2",
    "points2",
    "interv2",
    "iniser2"
];

var ScanPrmsFormat = {};
ScanPrmsFormat.initial = NumberLabel;
ScanPrmsFormat.final = NumberLabel;
ScanPrmsFormat.points = NumberLabel;
ScanPrmsFormat.interv = NumberLabel;
ScanPrmsFormat.iniser = NumberLabel;
ScanPrmsFormat.initial2 = ArrayLabel;
ScanPrmsFormat.final2 = ArrayLabel;
ScanPrmsFormat.points2 = ArrayLabel;
ScanPrmsFormat.interv2 = ArrayLabel;
ScanPrmsFormat.iniser2 = ArrayLabel;

const ScanOptionsLabel = {
  bundle:[BundleScanlabel, false]
};

const ScanOptionsOrder = [
  "bundle"
];

// units
const LengthUnitLabel = ["mm", "cm", "m"];
const LengthUnitFactor = [1.0e-3, 0.01, 1.0];
const GapUnitFactor = [1.0, 10.0, 1000.0];
const DepthUnitLabel = ["mm", "cm", "m"];
const DepthUnitFactor = [1.0, 10.0, 1.0e+3];
const UnitOptionsLabel = {
    gap:["Gap", LengthUnitLabel, GapUnitFactor], 
    zpos:["Longitudinal Position (z)", LengthUnitLabel, LengthUnitFactor],
    magf:["Magnetic Field (B<sub>x,y</sub>)", ["Tesla", "Gauss"], [1.0, 1.0e-4]],
    depth:["Depth for "+VPDLabel, DepthUnitLabel, DepthUnitFactor],
    time:["Time for Bunch Profile", 
        ["mm", "m", "fs", "ps", "s"], [1.0e+12/CC, 1.0e+15/CC, 1.0, 1.0e+3, 1.0e+15]],
    seedwavel:["Wavelength", ["nm"], [1.0]],
    seedflux:["Intensity", ["-"], [1.0]],
    seedphase:["Phase", ["degree"], [1.0]],
      
};

const UnitOptionsOrder = [
    "gap", 
    "zpos",
    "magf",
    "depth",
    "time"
];

// parallel computing
const PCOptionsLabel = {
    enable:["Enable Parallel Computing", false],
    processes:["Number of Processes", 4]
};

const PCOptionsOrder = [
    "enable", 
    "processes"
];

// phase error configuration
const PhaseErrOptionsLabel = {
  thresh:[PErrThreshLabel, 90, 0],
  zcoord:[PErrZcoorLabel, [PhaseErrZPole, PhaseErrZPos]]
};

const PhaseErrOptionsOrder = [
  "thresh",
  "zcoord"
];

// filter configuration
const FilterPlotOptionsLabel = {
  filtauto:[FiltPlotTypeLabel, [AutomaticLabel, CustomLabel]],
  filtemin:[FiltPlotEmin, 1000, 1],
  filtemax:[FiltPlotEmax, 100000, 1],
  filtpoints:[FiltPlotPoints, 1000, 2],
  filtscale:[FiltPlotEscale, [LinearLabel, LogLabel]]
};

const FilterPlotOptionsOrder = [
  "filtauto",
  "filtemin",
  "filtemax",
  "filtpoints",
  "filtscale"
];

// custom accuracy
const AccuracyOptionsLabel = {
  accdisctra:["Trajectory Step", 1],
  accinpE:  ["Photon Energy Integral Convergence", 1],
  acclimpE: ["Photon Energy Integral Range", 1],
  accdiscpE:["Photon Energy Discretization", 1],
  accineE:  ["Electron Energy Integral Convergence", 1],
  acclimeE:["Electron Energy Integral Range", 1],
  accinobs: ["Spatial Integral Convergence", 1],
  acclimobs:["Spatial Integral Range", 1],
  accdiscobs:["Spatial Discretization", 1],
  accintime:["Temporal Integral Convergence", 1],
  acclimtime:["Temporal Integral Range", 1],
  accinsrc: ["Source Integral Convergence", 1],
  acclimsrc:["Source Integral Range", 1],
  accdiscsrc:["Source Discretization", 1],
  accconvharm:["Harmonic Convergence", 1],
  accothers:["Others", 1],
  accEcorr:["Energy Consistency", false],
  accconvMC:["Monte Carlo Integral Tolerance", FloatLabel],
  accconvMCcoh:["Coherent Radiation Integral Tolerance", FloatLabel],
  acclimMCpart:["Limit Macroparticles", false],
  accMCpart:["Maximum Macroparticles", FloatLabel]
};

const AccuracyOptionsOrder = [
  "accdisctra",
  SeparatorLabel,
  "accinpE", 
  "acclimpE",
  "accdiscpE",
  SeparatorLabel,
  "accineE",
  "acclimeE",
  SeparatorLabel,
  "accinobs",
  "acclimobs",
  "accdiscobs",
  SeparatorLabel,
  "accintime",
  "acclimtime",
  SeparatorLabel,
  "accinsrc",
  "acclimsrc",
  "accdiscsrc",
  SeparatorLabel,
  "accconvharm",
  SeparatorLabel,
  "accothers",
  SeparatorLabel,
  "accEcorr",
  SeparatorLabel,
  "accconvMC",
  "accconvMCcoh",
  "acclimMCpart",
  "accMCpart"
];

// parameter set new (re-) name
const PSNameOptionsLabel = {
  newname:["Input a New Name", CommentLabel]
};

const PSNameOptionsOrder = [
  "newname"
];

// Output file
const OutputOptionsLabel = {
    comment:["Comment", CommentLabel],
    format:[OutFormat, [JSONOnly, ASCIIOnly, BothFormat]],
    folder:["Folder", FolderLabel],
    prefix:["Prefix", CommentLabel],
    serial:["Serial Number", -1]
};

const OutputOptionsOrder = [
    "format", 
    "folder", 
    "prefix",
    "comment",
    "serial"
];

// Plot Config
const PlotOptionsLabel = {
  xscale:["X-axis Scale", [LinearLabel, LogLabel]],
  yscale:["Y-axis Scale", [LinearLabel, LogLabel]],
  type:["Plot Type", [LineLabel, LineSymbolLabel, SymbolLabel]],
  type2d:["2D Plot Type", [ContourLabel, SurfaceLabel]],
  colorscale:["Color Scale", [DefaultLabel, RainbowLabel, BlackBodyLabel, EarthLabel, GreysLabel]],
  normalize:["Normalization", [ByMaxLabel, ForEachLabel]]
};

const PlotOptionsOrder = [
  "xscale", 
  "yscale", 
  "type",
  "type2d",
  "colorscale",
  "normalize"
];

var PlotlyScatterType = "scattergl";

// preprocessing menus
const PPBetaLabel = "betatron Functions";
const PPFDlabel = "Field Distribution";
const PP1stIntLabel = "1st Integral";
const PP2ndIntLabel = "2nd Integral";
const PPPhaseErrLabel = "Phase Error";
const PPRedFlux = "Harmonic Intensity";
const PPTransLabel = "Transmission Rate";
const PPAbsLabel = "Absorption Rate";
const PPDataImpLabel = "Imported Data";
var PreProcessLabel = [];
var DefaultPPItems;
var preprocess_result = "";

// Menu Items
const FileNew = "Create a New Parameter File";
const FileOpen = "Open a Parameter File";
const FileOpenList = "Open a Parameter File in the Server";
const FileOpenUp = "Upload a Local Parameter File";
const FileImport = "Append Parameter Sets";
const FileConvert = "Convert & Open a Parameter File";
const FileConvertUp = "Convert & Upload a Parameter File";
const FileSave = "Save";
const FileSaveAs = "Save As";
const FileDownload = "Download This Parameter File";
const FileExit = "Exit";

const PrmSetDuplicate = "Duplicate";
const PrmSetRename = "Rename";
const PrmSetDelete = "Delete";
const Scan1D = "Scan This Parameter";
const Scan2D1D = "Scan Parameter (1D)";
const Scan2D2D = "Scan Parameter (2D)";

const EditMaterial = "Filter/Absorber Material";
const EditMagnet = "Magnet Configurations";
const EditAccuracy = "Numerical Accuracy";
const EditUnits = "Units for Data Import";
const EditMPIConfig = "Parallel Computing";
const EditPhaseErrConfig = "Phase Error Evaluation";
const EditFiltPlotConfig = "Transmission/Absorption Rate Plot";
const HelpReference  = "Open Reference Manual";
const HelpAbout = "About SPECTRA";

const FileMenus = 
[
  {
    label:FileNew
  },
  {
    label:FileOpen,
  },
  {
    label:FileOpenList,
  },
  {
    label:FileOpenUp,
  },
  {
    label:FileImport,
  },
  {
    type:"separator"
  },
  {
    label:FileConvert
  },
  {
    label:FileConvertUp
  },
  {
    type:"separator"
  },
  {
    label:FileSave
  },
  {
    label:FileSaveAs
  },
  {
    label:FileDownload
  },
  {
    label:FileExit
  }
];

const CreateProcessLabel = "Create New Process";
const ExportSettingLabel = "Export Calculation Settings";
const SettingDownloadLabel = "Download Calculation Settings";
const StartCalcLabel = "Start Calculation";

const RunMenus = 
[
  {
    label:CreateProcessLabel
  },
  {
    label:ExportSettingLabel
  },
  {
    label:SettingDownloadLabel
  },
  {
    label:StartCalcLabel
  }
];

const EditViewMenus = [
  {
    label:EditMaterial
  },
  {
    label:EditFiltPlotConfig
  },
  {
    type:"separator"
  },
  {
    label:EditMagnet
  },
  {
    label:EditPhaseErrConfig
  },
  {
    label:EditUnits
  },
  {
    type:"separator"
  },
  {
    label:EditAccuracy
  },
  {
    label:EditMPIConfig
  }
];

const HelpMenus = [
  {
    label:HelpReference
  },
  {
    label:HelpAbout
  }
];

const MenuLabels = {
  far:"Far Field & Ideal Condition",
  near:"Near Field",
  cohrad:"Coherent Radiation",
  srcpoint: "Characterization at the Source Point",
  CMD: CMDLabel,
  fixed: FixedPointLabel,

  energy:"Energy Dependence",
  spatial: SpatialDepLabel,
  Kvalue:"K Dependence",
  temporal:"Time Dependence",
  wigner:"Wigner Function",

  fdensa:"Angular Flux Density",
  pflux:"Partial Flux",
  tflux:"Total Flux",

  slitrect:"Rectangular Slit",
  slitcirc:"Circular Slit",

  pdensa:"Angular Power Density",
  pdensr:"Resolved Power Density",
  vpdens:VPDLabel,

  along:"Along Axis",
  meshxy:"Mesh: x-y",
  meshrphi:"Mesh: r-&phi;",

  CMD2d:"CMD with the Wigner Function",
  CMDPP:"Post-processing",
  CMDcheck:"Check Validity",

  simpcalc:"Simplified Calculation",
  fluxfix:"Flux at a Fixed Energy",
  fluxpeak:"Peak Flux Curve",
  powercv:"Power",

  tgtharm:"Target Harmonics",
  allharm:"All Harmonics",

  ppower:"Partial Power",

  fdenss:"Spatial Flux Density",
  pdenss:"Spatial Power Density",
  spdens:"Surface Power Density",

  xzplane:"Planar Surface: x-z",
  yzplane:"Planar Surface: y-z",
  pipe:"Cylindrical Surface",

  efield:"Electric Field",
  camp:"Complex Amplitude",
  phasespace:"Phase-Space Distribution",

  XXpslice:"X-X' (Sliced)",
  XXpprj:"X-X' (Projected)",
  YYpslice:"Y-Y' (Sliced)",
  YYpprj:"Y-Y' (Projected)",
  XXpYYp:"X-X'-Y-Y'",
  sprof:"Spatial Profile",

  Wslice:"Sliced",
  WprjX:"Projected on X-X'",
  WprjY:"Projected on Y-Y'",
  Wrel:"Related Characteristics"
}

const CalcMenus = 
[
  {
    label:MenuLabels.far,
    submenu:[
      {
        label:MenuLabels.energy,
        submenu:[
          {
            label:MenuLabels.fdensa
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            label:MenuLabels.tflux
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.simpcalc,
          }
        ]
      },      
      {
        label:MenuLabels.spatial,
        submenu:[
          {
            label:MenuLabels.fdensa,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.pdensa,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.pdensr,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.vpdens,
          }
        ]
      },
      {
        label:MenuLabels.Kvalue,
        submenu:[
          {
            label:MenuLabels.simpcalc,
            submenu:[
              {
                label:MenuLabels.tgtharm,
              },
              {
                label:MenuLabels.allharm,
              }
            ]
          },
          {
            label:MenuLabels.fluxfix,
            submenu:[
              {
                label:MenuLabels.fdensa
              },
              {
                label:MenuLabels.pflux,
                submenu:[
                  {
                    label:MenuLabels.slitrect
                  },
                  {
                    label:MenuLabels.slitcirc
                  },
                ]          
              }
            ]
          },
          {
            label:MenuLabels.fluxpeak,
            submenu:[
              {
                label:MenuLabels.fdensa,
                submenu:[
                  {
                    label:MenuLabels.tgtharm,
                  },
                  {
                    label:MenuLabels.allharm,
                  }
                ]
              },
              {
                label:MenuLabels.pflux,
                submenu:[
                  {
                    label:MenuLabels.slitrect,
                    submenu:[
                      {
                        label:MenuLabels.tgtharm,
                      },
                      {
                        label:MenuLabels.allharm,
                      }
                    ]
                  },
                  {
                    label:MenuLabels.slitcirc,
                    submenu:[
                      {
                        label:MenuLabels.tgtharm,
                      },
                      {
                        label:MenuLabels.allharm,
                      }
                    ]
                  },
                ]          
              }
            ]
          },
          {
            label:MenuLabels.powercv,
            submenu:[
              {
                label:MenuLabels.pdensa
              },
              {
                label:MenuLabels.ppower,
                submenu:[
                  {
                    label:MenuLabels.slitrect
                  },
                  {
                    label:MenuLabels.slitcirc
                  },
                ]          
              }
            ]
          }
        ]
      },
    ]
  },
  {
    type:"separator"
  },
  {
    label:MenuLabels.near,
    submenu:[
      {
        label:MenuLabels.energy,
        submenu:[
          {
            label:MenuLabels.fdenss
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          }
        ]
      },
      {
        label:MenuLabels.spatial,
        submenu:[
          {
            label:MenuLabels.fdenss,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.pdenss,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.spdens,
            submenu:[
              {
                label:MenuLabels.xzplane,
              },
              {
                label:MenuLabels.yzplane,
              },
              {
                label:MenuLabels.pipe,
              }
            ]
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.vpdens,
          }
        ]
      },
    ]
  },
  {
    label:MenuLabels.cohrad,
    submenu:[
      {
        label:MenuLabels.energy,
        submenu:[
          {
            label:MenuLabels.fdenss
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          }
        ]
      },
      {
        label:MenuLabels.spatial,
        submenu:[
          {
            label:MenuLabels.camp,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.fdenss,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          },
          {
            label:MenuLabels.pdenss,
            submenu:[
              {
                label:MenuLabels.along,
              },
              {
                label:MenuLabels.meshxy,
              },
              {
                label:MenuLabels.meshrphi,
              }
            ]
          }
        ]
      },
      {
        label:MenuLabels.temporal,
        submenu:[
          {
            label:MenuLabels.efield
          },
          {
            label:MenuLabels.pdenss
          },
          {
            label:MenuLabels.ppower,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]
          },
        ]
      }
    ]
  },
  {
    label:MenuLabels.srcpoint,
    submenu:[
      {
        label:MenuLabels.wigner,
        submenu:[
          {
            label:MenuLabels.energy
          },
          {
            label:MenuLabels.Kvalue,
            submenu:[
              {
                label:MenuLabels.tgtharm,
              },
              {
                label:MenuLabels.allharm,
              }
            ]
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.phasespace,
            submenu:[
              {
                label:MenuLabels.XXpslice
              },
              {
                label:MenuLabels.XXpprj
              },
              {
                label:MenuLabels.YYpslice
              },
              {
                label:MenuLabels.YYpprj
              },
              {
                label:MenuLabels.XXpYYp
              },
            ]
          },
        ]
      },
      {
        type:"separator"
      },
      {
        label:MenuLabels.sprof
      }
    ]
  },
  {
    type:"separator"
  },
  {
    label:MenuLabels.fixed,
    submenu:[
      {
        label:MenuLabels.far,
        submenu:[
          {
            label:MenuLabels.fdensa
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            label:MenuLabels.tflux
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.pdensa
          },
          {
            label:MenuLabels.ppower,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.simpcalc
          }
        ]
      },
      {
        type:"separator"
      },
      {
        label:MenuLabels.near,
        submenu:[
          {
            label:MenuLabels.fdenss
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.pdenss
          },
          {
            label:MenuLabels.ppower,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            label:MenuLabels.spdens
          }
        ]
      },
      {
        label:MenuLabels.cohrad,
        submenu:[
          {
            label:MenuLabels.fdenss
          },
          {
            label:MenuLabels.pflux,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          },
          {
            type:"separator"
          },
          {
            label:MenuLabels.pdenss
          },
          {
            label:MenuLabels.ppower,
            submenu:[
              {
                label:MenuLabels.slitrect
              },
              {
                label:MenuLabels.slitcirc
              },
            ]          
          }
        ]
      },
      {
        label:MenuLabels.wigner,
        submenu:[
          {
            label:MenuLabels.Wslice
          },
          {
            label:MenuLabels.WprjX
          },
          {
            label:MenuLabels.WprjY
          },
          {
            label:MenuLabels.Wrel
          }
        ]
      }        
    ]
  },
  {
    type:"separator"
  },
  {
    label:MenuLabels.CMD,
    submenu:[
      {
        label:MenuLabels.CMD2d
      },
      {
        label:MenuLabels.CMDPP
      }
      /*,{
        label:MenuLabels.CMDcheck
      }*/
    ]
  }
]

const SelectPrmSetMenus = 
[
  {
    label:BLLabel,
    submenu:[
    ]
  },
  {
    label:AccLabel,
    submenu:[
    ]
  },
  {
    label:SrcLabel,
    submenu:[
    ]
  },
  {
    label:ConfigLabel,
    submenu:[
    ]
  }
];

const PrmEditMenus = [
  {
    label:PrmSetDuplicate
  },
  {
    label:PrmSetRename
  },
  {
    label:PrmSetDelete
  }
];

const ScanMenus = [
  {
    label:Scan1D
  }
];

const Scan2DMenus = [
  {
    label:Scan2D1D
  },
  {
    label:Scan2D2D
  }
];
