"use strict";

// global variables
var VersionNumber = "11.2";
var FullVersionNumber = "11.2.1";

var SPGUI = {};
SPGUI.filereader = {};
SPGUI.sprdsheetcol = {};
SPGUI.calcprocess = null;
SPGUI.socketio = null;
SPGUI.isimported = false;
SPGUI.disabledcalcid = [];
SPGUI.targetid = {};
SPGUI.importdata = {};
SPGUI.setting = {defpaths:{}};
SPGUI.orginobj = "";
SPGUI.idxpair4d = [
    [[0, 1], [2, 3]], /* x, y */
    [[2, 3], [0, 1]], /* x', y' */
    [[0, 2], [1, 3]], /* x, x' */
    [[1, 3], [0, 2]]  /* y, y' */
];
SPGUI.idxpair3d = [
    [[0], [1, 2]],
    [[1], [0, 2]],
    [[2], [0, 1]],
    [[0, 1], [2]], 
    [[0, 2], [1]], 
    [[1, 2], [0]], 
];
SPGUI.idxpair2d = [
    [[0], [1]], /* x */
    [[1], [0]], /* y */
    [[0, 1], []], /* x, y */
];
SPGUI.currCalcType = "";
SPGUI.plobj = null;
SPGUI.subwindows = 0;

var PlotlyEditIcon = {
    'width': 1792,
    'height': 1792,
    'path': "M491 1536l91-91-235-235-91 91v107h128v128h107zm523-928q0-22-22-22-10 0-17 7l-542 542q-7 7-7 17 0 22 22 22 10 0 17-7l542-542q7-7 7-17zm-54-192l416 416-832 832h-416v-416zm683 96q0 53-37 90l-166 166-416-416 166-165q36-38 90-38 53 0 91 38l235 234q37 39 37 91z",
    'transform': 'matrix(1 0 0 1 0 1)'
  };

var PlotlyPrms = {};
PlotlyPrms.config = {
    displaylogo: false, 
    responsive: true,
    scrollZoom: true,
    editable: true,
    edits: {
        axisTitleText: false,
        titleText: false
    },

    modeBarButtonsToAdd:[
        {
          name: "Edit",
          icon: PlotlyEditIcon,
          click: function(e) {
            let eventup = new CustomEvent("editplotly");
            e.dispatchEvent(eventup);    
          }
        }
      ],
  
    modeBarButtonsToRemove:["toggleSpikelines"]
};
PlotlyPrms.colorbar = {
    thickness: 10,
    thicknessmode: "pixels",
    len: 0.8,
    lenmode: "fraction",
    outlinewidth: 0,
    tickformat: ".2g",
};
PlotlyPrms.heatmap =  {
    coloring: "heatmap",
    tickformat: ".2g"
};
PlotlyPrms.clscale = [
    [0, "rgb(0,0,255)"], 
    [0.5, "rgb(220,220,220)"], 
    [1, "rgb(255,0,0)"]
];
PlotlyPrms.margin1d = {l:70,r:20,t:20,b:40};
PlotlyPrms.margin2d = {l:20,r:20,t:20,b:20};


// default parameters and options
var def_injecprm = 
{
    [InjectionPrmsLabel.xy]:[0,0],
    [InjectionPrmsLabel.xyp]:[0,0]
};

var def_acc = 
{
    [AccPrmsLabel.eGeV]:8,
    [AccPrmsLabel.imA]:100,
    [AccPrmsLabel.cirm]:1435.95,
    [AccPrmsLabel.bunches]:203,
    [AccPrmsLabel.pulsepps]:60,
    [AccPrmsLabel.bunchlength]:3.9,
    [AccPrmsLabel.bunchcharge]:0.1,
    [AccPrmsLabel.emitt]:2.4e-9,
    [AccPrmsLabel.coupl]:0.002,
    [AccPrmsLabel.espread]:0.0011,
    [AccPrmsLabel.beta]:[31.2, 5],
    [AccPrmsLabel.alpha]:[0, 0],
    [AccPrmsLabel.eta]:[0.146, 0],
    [AccPrmsLabel.etap]:[0, 0],
    [AccPrmsLabel.R56add]: 0,
    [AccPrmsLabel.minsize]:1.0e-12,
    [OptionLabel]:{
        [AccOptionsLabel.bunchtype[0]]:AccOptionsLabel.bunchtype[1][0],
        [AccOptionsLabel.bunchdata[0]]:UnselLabel,
        [AccOptionsLabel.injectionebm[0]]:AccOptionsLabel.injectionebm[1][0],
        [AccOptionsLabel.injecprm[0]]:def_injecprm,
        [AccOptionsLabel.singlee[0]]:false,
        [AccOptionsLabel.zeroemitt[0]]:false,
        [AccOptionsLabel.zerosprd[0]]:false
    },
    [TypeLabel]:AccTypeLabel.RING,
}

var def_magconf = 
{
    [MagnetPrmsLabel.br]:1.2,
    [MagnetPrmsLabel.geofactor]:[0.6,0.9]                
};

var def_ferrprm =
{
    [FerrPrmsLabel.boffset]:[0,0],
    [FerrPrmsLabel.ltaper]:[0,0],
    [FerrPrmsLabel.qtaper]:[0,0]
};

var def_perrprm =
{
    [PerrPrmsLabel.seed]:1,
    [PerrPrmsLabel.fsigma]:0.5,
    [PerrPrmsLabel.psigma]:5,
    [PerrPrmsLabel.xysigma]:[1e-3,1e-3]
};

var def_segprm =
{
    [SegPrmsLabel.segments]:2,
    [SegPrmsLabel.hsegments]:1,
    [SegPrmsLabel.interval]:5,
    [SegPrmsLabel.phi0]:1,
    [SegPrmsLabel.phi12]:[0.5,1.5],
    [SegPrmsLabel.mdist]:5
};

var def_src =
{
    [SrcPrmsLabel.gap]:20,
    [SrcPrmsLabel.bxy]:[0.5, 0.5],
    [SrcPrmsLabel.b]:0.26,
    [SrcPrmsLabel.bmain]:1.0,
    [SrcPrmsLabel.subpoleb]:1.0,
    [SrcPrmsLabel.lu]:32,
    [SrcPrmsLabel.devlength]:4.564,
    [SrcPrmsLabel.Kxy0]:[0.6,1.0],
    [SrcPrmsLabel.phase]:0,
    [SrcPrmsLabel.Kxy]:[1.0, 1.0],
    [SrcPrmsLabel.K]:1,
    [SrcPrmsLabel.e1st]:12000,
    [SrcPrmsLabel.radius]:34.2,
    [SrcPrmsLabel.bendlength]:2.8,
    [SrcPrmsLabel.fringelen]:0.05,
    [SrcPrmsLabel.mplength]:0.1,
    [SrcPrmsLabel.subpolel]:0.05,
    [SrcPrmsLabel.bminterv]:5.0,
    [SrcPrmsLabel.csrorg]:0,
    [SrcPrmsLabel.fmap]:"",
    [SrcPrmsLabel.multiharm]:[],
    [OptionLabel]:{
        [SrcOptionsLabel.gaplink[0]]:SrcOptionsLabel.gaplink[1][0],
        [SrcOptionsLabel.apple[0]]:false,
        [SrcOptionsLabel.field_str[0]]:SrcOptionsLabel.field_str[1][0],
        [SrcOptionsLabel.endmag[0]]:true,
        [SrcOptionsLabel.natfocus[0]]:SrcOptionsLabel.natfocus[1][0],
        [SrcOptionsLabel.fielderr[0]]:false,
        [SrcOptionsLabel.ferrprm[0]]:def_ferrprm,
        [SrcOptionsLabel.phaseerr[0]]:false,
        [SrcOptionsLabel.perrprm[0]]:def_perrprm,
        [SrcOptionsLabel.bmtandem[0]]:false,
        [SrcOptionsLabel.segment_type[0]]:SrcOptionsLabel.segment_type[1][0],
        [SrcOptionsLabel.perlattice[0]]:false,
        [SrcOptionsLabel.segprm[0]]:def_segprm,
        [SrcOptionsLabel.magconf[0]]:def_magconf
    },
    [TypeLabel]:SrcTypeLabel.LIN_UND,
};

var def_bpfprm =
{
    [BPFPrmsLabel.bpfcenter]:10000,
    [BPFPrmsLabel.bpfwidth]:100,
    [BPFPrmsLabel.bpfsigma]:100,
    [BPFPrmsLabel.bpfmaxeff]:1
};

var def_CMDprm = 
{
    [CMDPrmsLabel.HGorderxy]:[50,50],
    [CMDPrmsLabel.HGorderx]:50,
    [CMDPrmsLabel.HGordery]:50,
    [CMDPrmsLabel.maxHGorderxy]:[50,50],
    [CMDPrmsLabel.maxHGorderx]:50,
    [CMDPrmsLabel.maxHGordery]:50,
    [CMDPrmsLabel.maxmode]:100,
    [CMDPrmsLabel.fcutoff]:1e-3,
    [CMDPrmsLabel.cutoff]:1e-4,
    [CMDPrmsLabel.fieldgridxy]:[0.005,0.002],
    [CMDPrmsLabel.fieldgridx]:0.005,
    [CMDPrmsLabel.fieldgridy]:0.002
};

var def_FELprm = 
{
    [FELPrmsLabel.pulseE]: 0.1,
    [FELPrmsLabel.wavelen]: 267,
    [FELPrmsLabel.pulselen]: 40,
    [FELPrmsLabel.tlpulselen]: 40,
    [FELPrmsLabel.srcsize]: 1,
    [FELPrmsLabel.waistpos]: 0,
    [FELPrmsLabel.timing]: 0,
    [FELPrmsLabel.gdd]: 0,
    [FELPrmsLabel.tod]: 0,

    [FELPrmsLabel.pulseE_d]: [0.1, 0.1],
    [FELPrmsLabel.wavelen_d]: [267, 280],
    [FELPrmsLabel.tlpulselen_d]: [40, 40],
    [FELPrmsLabel.srcsize_d]: [1, 1],
    [FELPrmsLabel.waistpos_d]: [0, 0],
    [FELPrmsLabel.timing_d]: [0, 0],
    [FELPrmsLabel.gdd_d]: [0, 0],
    [FELPrmsLabel.tod_d]: [0, 0],

    [FELPrmsLabel.svstep]: [-1,0.1],
    [FELPrmsLabel.radstep]: 5,    
    [FELPrmsLabel.eproi]: [1,120],
    [FELPrmsLabel.particles]: 1000000,
    [FELPrmsLabel.edevstep]: 1e-4,
    [FELPrmsLabel.R56]: 0
};

var def_conf =
{
    // z positions
    [ConfigPrmsLabel.slit_dist]:30,
    [ConfigPrmsLabel.zrange]:[5,6],
    [ConfigPrmsLabel.zmesh]:21,

    // energy 
    [ConfigPrmsLabel.erange]:[5000,50000],
    [ConfigPrmsLabel.de]:2,
    [ConfigPrmsLabel.epitch]:2,
    [ConfigPrmsLabel.emesh]:101,
    [ConfigPrmsLabel.efix]:12661,
    [ConfigPrmsLabel.nefix]:1,
    [ConfigPrmsLabel.detune]:0,

    // obs. points
    [ConfigPrmsLabel.xyfix]:[0,0],
    [ConfigPrmsLabel.qxyfix]:[0,0],
    [ConfigPrmsLabel.spdxfix]:10,
    [ConfigPrmsLabel.spdyfix]:3,
    [ConfigPrmsLabel.spdrfix]:3,
    [ConfigPrmsLabel.Qnorm]:90,
    [ConfigPrmsLabel.Phinorm]:90,
    [ConfigPrmsLabel.Qgl]:90,
    [ConfigPrmsLabel.Phiinc]:90,

    // slit
    [ConfigPrmsLabel.slitpos]:[0,0],
    [ConfigPrmsLabel.qslitpos]:[0,0],
    [ConfigPrmsLabel.slitapt]:[2,1],
    [ConfigPrmsLabel.qslitapt]:[0.06,0.03],
    [ConfigPrmsLabel.nslitapt]:[4,4],
    [ConfigPrmsLabel.slitr]:[0,1],
    [ConfigPrmsLabel.slitq]:[0,0.03],
    [ConfigPrmsLabel.pplimit]:0.3,  

    // obs. range 
    [ConfigPrmsLabel.xrange]:[-1,1],
    [ConfigPrmsLabel.qxrange]:[-0.03,0.03],
    [ConfigPrmsLabel.xmesh]:41,
    [ConfigPrmsLabel.yrange]:[-0.5,0.5],
    [ConfigPrmsLabel.qyrange]:[-0.02,0.02],
    [ConfigPrmsLabel.ymesh]:41,
    [ConfigPrmsLabel.rrange]:[0,1],
    [ConfigPrmsLabel.qrange]:[0,0.03],
    [ConfigPrmsLabel.rphimesh]:41,
    [ConfigPrmsLabel.qphimesh]:41,
    [ConfigPrmsLabel.phirange]:[0,90],
    [ConfigPrmsLabel.phimesh]:41,
    [ConfigPrmsLabel.drange]:[0,1],
    [ConfigPrmsLabel.dmesh]:41,

    // K
    [ConfigPrmsLabel.krange]:[0,2.3],
    [ConfigPrmsLabel.ckrange]:[0,5.0],
    [ConfigPrmsLabel.kmesh]:51,

    // harmonic
    [ConfigPrmsLabel.hrange]:[1,5],
    [ConfigPrmsLabel.hfix]:1,
    [ConfigPrmsLabel.hmax]:15,

    // temporal
    [ConfigPrmsLabel.trange]:[-0.05,0.05],
    [ConfigPrmsLabel.tmesh]:10001,

    // source point
    [ConfigPrmsLabel.gtacc]:[1,1],
    [ConfigPrmsLabel.horizacc]:0.1,
    [ConfigPrmsLabel.Xfix]:0,
    [ConfigPrmsLabel.Yfix]:0,
    [ConfigPrmsLabel.Xpfix]:0,
    [ConfigPrmsLabel.Ypfix]:0,
    [ConfigPrmsLabel.Xrange]:[-1,1],
    [ConfigPrmsLabel.Xmesh]:51,
    [ConfigPrmsLabel.Xprange]:[-0.05,0.05],
    [ConfigPrmsLabel.Xpmesh]:41,
    [ConfigPrmsLabel.Yrange]:[-0.04,0.04],
    [ConfigPrmsLabel.Ymesh]:51,
    [ConfigPrmsLabel.Yprange]:[-0.02,0.02],
    [ConfigPrmsLabel.Ypmesh]:41,

    // options
    [OptionLabel]:{
        [ConfigOptionsLabel.filter[0]]:ConfigOptionsLabel.filter[1][0],
        [ConfigOptionsLabel.fmateri[0]]:[],
        [ConfigOptionsLabel.amateri[0]]:[],
        [ConfigOptionsLabel.bpfprms[0]]:def_bpfprm,
        [ConfigOptionsLabel.estep[0]]:ConfigOptionsLabel.estep[1][0],
        [ConfigOptionsLabel.aperture[0]]:ConfigOptionsLabel.aperture[1][0],
        [ConfigOptionsLabel.dstep[0]]:ConfigOptionsLabel.dstep[1][0],
        [ConfigOptionsLabel.defobs[0]]:ConfigOptionsLabel.defobs[1][0],
        [ConfigOptionsLabel.normenergy[0]]:false,
        [ConfigOptionsLabel.powlimit[0]]:false,
        [ConfigOptionsLabel.optDx[0]]:true,
        [ConfigOptionsLabel.xsmooth[0]]:1,
        [ConfigOptionsLabel.fouriep[0]]:false,
        [ConfigOptionsLabel.wiggapprox[0]]:false,
        [ConfigOptionsLabel.esmooth[0]]:false,
        [ConfigOptionsLabel.smoothwin[0]]:1,
        [ConfigOptionsLabel.accuracy[0]]:DefaultLabel,
        [ConfigOptionsLabel.CMD[0]]:false,
        [ConfigOptionsLabel.GSModel[0]]:false,
        [ConfigOptionsLabel.GSModelXY[0]]:NoneLabel,
        [ConfigOptionsLabel.CMDfld[0]]:JSONOnly,
        [ConfigOptionsLabel.CMDint[0]]:false,
        [ConfigOptionsLabel.CMDcmp[0]]:false,
        [ConfigOptionsLabel.CMDcmpint[0]]:false,
        [ConfigOptionsLabel.CMDprms[0]]:def_CMDprm,
        [ConfigOptionsLabel.fel[0]]:ConfigOptionsLabel.fel[1][0],
        [ConfigOptionsLabel.exportInt[0]]:true,
        [ConfigOptionsLabel.exportEt[0]]:false,
        [ConfigOptionsLabel.R56Bunch[0]]:false,
        [ConfigOptionsLabel.FELprms[0]]:def_FELprm
    },
    [TypeLabel]:""
};

var def_spjson =
{
    [CurrentLabel]:"Sample",
    [BLLabel]:{
        "Sample":
        {
            [AccLabel]:"Accelerator Sample",
            [SrcLabel]:"Light Source Sample",
            [ConfigLabel]:"Configurations Sample"
        }
    },
    [AccLabel]:{
        "Accelerator Sample":def_acc
    },
    [SrcLabel]:{
        "Light Source Sample":def_src,
    },
    [ConfigLabel]:{
        "Configurations Sample":def_conf        
    },
    [FMaterialLabel]:{}
};

let def_scan = 
{
    [ScanPrmsLabel.initial]:0.5,
    [ScanPrmsLabel.final]:1.5,
    [ScanPrmsLabel.points]:11,
    [ScanPrmsLabel.interv]:1,
    [ScanPrmsLabel.iniser]:0,
    [ScanPrmsLabel.initial2]:[0.5,0.5],
    [ScanPrmsLabel.final2]:[1.5,1.5],
    [ScanPrmsLabel.points2]:[11,11],
    [ScanPrmsLabel.iniser2]:[0,0]
};

let def_scan_option =
{
    [ScanOptionsLabel.bundle[0]]:true
};

let def_accuracy = 
{
    [AccuracyOptionsLabel.accdisctra[0]]:1,
    [AccuracyOptionsLabel.accinpE[0]]:1,
    [AccuracyOptionsLabel.acclimpE[0]]:1,
    [AccuracyOptionsLabel.accdiscpE[0]]:1,
    [AccuracyOptionsLabel.accineE[0]]:1,
    [AccuracyOptionsLabel.acclimeE[0]]:1,
    [AccuracyOptionsLabel.accinobs[0]]:1,
    [AccuracyOptionsLabel.acclimobs[0]]:1,
    [AccuracyOptionsLabel.accdiscobs[0]]:1,
    [AccuracyOptionsLabel.accintime[0]]:1,
    [AccuracyOptionsLabel.acclimtime[0]]:1,
    [AccuracyOptionsLabel.accinsrc[0]]:1,
    [AccuracyOptionsLabel.acclimsrc[0]]:1,
    [AccuracyOptionsLabel.accdiscsrc[0]]:1,
    [AccuracyOptionsLabel.accconvharm[0]]:1,
    [AccuracyOptionsLabel.accconvMC[0]]:0.05,
    [AccuracyOptionsLabel.accconvMCcoh[0]]:0.5,
    [AccuracyOptionsLabel.accMCpart[0]]:1000000,
    [AccuracyOptionsLabel.acclimMCpart[0]]:false,
    [AccuracyOptionsLabel.accEcorr[0]]:false,
    [AccuracyOptionsLabel.accothers[0]]:1
};

let def_units = 
{
    [UnitOptionsLabel.gap[0]]:UnitOptionsLabel.gap[1][0],
    [UnitOptionsLabel.zpos[0]]:UnitOptionsLabel.zpos[1][0],
    [UnitOptionsLabel.magf[0]]:UnitOptionsLabel.magf[1][0],
    [UnitOptionsLabel.depth[0]]:UnitOptionsLabel.depth[1][0],
    [UnitOptionsLabel.time[0]]:UnitOptionsLabel.time[1][0]
};

let def_pc_configs =
{
    [PCOptionsLabel.enable[0]]:false,
    [PCOptionsLabel.processes[0]]:4
};

let def_perr_configs =
{
    [PhaseErrOptionsLabel.thresh[0]]:90,
    [PhaseErrOptionsLabel.zcoord[0]]:PhaseErrZPole
};

let def_filtplot_configs =
{
    [FilterPlotOptionsLabel.filtauto[0]]:AutomaticLabel,
    [FilterPlotOptionsLabel.filtemin[0]]: FilterPlotOptionsLabel.filtemin[1],
    [FilterPlotOptionsLabel.filtemax[0]]: FilterPlotOptionsLabel.filtemax[1],
    [FilterPlotOptionsLabel.filtpoints[0]]: FilterPlotOptionsLabel.filtpoints[1],
    [FilterPlotOptionsLabel.filtscale[0]]:LinearLabel
};

let def_psname_configs =
{
    [PSNameOptionsLabel.newname[0]]:""
};

let def_output_configs =
{
    [OutputOptionsLabel.format[0]]:OutputOptionsLabel.format[1][0],
    [OutputOptionsLabel.folder[0]]:"",
    [OutputOptionsLabel.prefix[0]]:"",
    [OutputOptionsLabel.serial[0]]:1,
    [OutputOptionsLabel.comment[0]]:""
};

let def_plot_configs = 
{
    [PlotOptionsLabel.xscale[0]]:PlotOptionsLabel.xscale[1][0],
    [PlotOptionsLabel.yscale[0]]:PlotOptionsLabel.yscale[1][0],
    [PlotOptionsLabel.type[0]]:PlotOptionsLabel.type[1][0],
    [PlotOptionsLabel.type2d[0]]:PlotOptionsLabel.type2d[1][0],
    [PlotOptionsLabel.colorscale[0]]:PlotOptionsLabel.colorscale[1][0],
    [PlotOptionsLabel.normalize[0]]:PlotOptionsLabel.normalize[1][0]
};
