"use strict";

//----- classes -----

// Parameter Table
class PrmTable {
    constructor(categ, prmlabels, format, defprm, isscan){
        this.m_table = document.createElement("table");
        this.m_prmlabels = prmlabels;
        this.m_categ = categ;
        this.m_format = format;
        this.m_defprm = defprm;
        this.m_isscan = isscan;
        this.m_skipupdate = false;
        this.m_fdlist = {};
        this.m_grids = {};
    }

    get JSONObj(){
        return this.m_jsonobj;
    }

    set JSONObj(jsonobj){
        this.m_jsonobj = jsonobj;
    }

    GetTable(){
        return this.m_table;
    }

    DisableTable(disable){
        for(let i = 0; i < this.m_prmlabels.length; i++){
            let prmtype = this.m_prmtypes == null ? 0 : this.m_prmtypes[this.m_prmlabels[i]];
            if((typeof prmtype) == "number"){
                if(prmtype < 0){ // skip
                    continue;
                }
            }
            if(prmtype == GridLabel){
                this.m_grids[this.m_prmlabels[i]].DisableGrid(disable);
            }
            else if(prmtype == FileLabel || prmtype == FolderLabel){
                let btnid = AttachSuffix(this.m_prmlabels[i], SuffixButton);
                this.DisableInput(btnid, disable);
                this.DisableInput(this.m_prmlabels[i], disable);
            }
            else{
                let fmt = this.m_format[this.m_prmlabels[i]];
                if(fmt == "array"){
                    this.DisableInput(this.m_prmlabels[i], [disable, disable]);
                }
                else{
                    this.DisableInput(this.m_prmlabels[i], disable);
                }    
            }
        }
    }

    GetReferenceTable(types, isopt, simplified, showprms, hide, skiptitle, subtitle){
        let categ = this.m_categ;
        if(isopt){
            let items = categ.split(IDSeparator);
            items.splice(1, 0, OptionLabel);
            categ = items.join(".");
        }
        let table = document.createElement("table");
        /*
        let caption = document.createElement("caption");
        caption.innerHTML = "Parameters in \""+categ+"\" object.";
        table.caption = caption;
        */
        let rows = [];
        let cell;

        if(!skiptitle){
            ArrangeObjectTblTitle(table, rows);
        }
        if(subtitle != ""){
            rows.push(table.insertRow(-1)); 
            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = subtitle;
            cell.className += "subtitle";
            cell.setAttribute("colspan", "5");
        }

        if(types.length > 0){
            rows.push(table.insertRow(-1)); 
            for(let j = 0; j < types.length; j++){
                cell = rows[rows.length-1].insertCell(-1);
                cell.innerHTML = types[j];
                if(j == 1 || j == 2){
                    cell.className += " prm";
                }
            }
        }

        for(let i = 0; i < this.m_prmlabels.length; i++){
            let tgtlabel = this.m_prmlabels[i];
            let fmt = this.m_format[tgtlabel];
            if(hide.indexOf(tgtlabel) >= 0){
                continue;
            }
            if(showprms.indexOf(tgtlabel) >= 0){
                continue;
            }

            rows.push(table.insertRow(-1)); 

            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = tgtlabel;

            cell = rows[rows.length-1].insertCell(-1);
            let label = 
                tgtlabel.replace(/&/g, "&amp;").
                replace(/</g, "&lt;").replace(/>/g, "&gt;");
            cell.innerHTML = label;
            cell.className += " prm";

            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = simplified[i];
            cell.className += " prm";

            cell = rows[rows.length-1].insertCell(-1);
            if(fmt == FileLabel){
                cell.innerHTML = "file name";
            }
            else{
                cell.innerHTML = fmt;
            }

            let val = this.m_jsonobj[tgtlabel];
            let valstr = "";
            cell = rows[rows.length-1].insertCell(-1);
            if(fmt == ArrayLabel){
                try {
                    valstr = ToPrmString(val[0], 3)+", "+ToPrmString(val[1], 3);
                } catch(e) {
                    valstr = "";
                }
            }
            if(fmt == NumberLabel){
                try {
                    valstr = ToPrmString(val, 5);
                } catch(e) {
                    valstr = "";
                }
            }
            cell.innerHTML = valstr;
        }
        let retstr = table.outerHTML;   
        return retstr;
    }
    
    SetPrmTypes(prmtypes){
        if(prmtypes != undefined){
            this.m_prmtypes = prmtypes;
        }
        this.m_table.innerHTML = "";
        this.m_fdialog = {};
        let rows = [];
        for(let i = 0; i < this.m_prmlabels.length; i++){
            let prmtype = this.m_prmtypes == null ? 0 : this.m_prmtypes[this.m_prmlabels[i]];
            if(prmtype != PlotObjLabel && 
                this.m_jsonobj.hasOwnProperty(this.m_prmlabels[i]) == false){
                if(this.m_defprm.hasOwnProperty(this.m_prmlabels[i])){
                    this.m_jsonobj[this.m_prmlabels[i]] 
                    = this.m_defprm[this.m_prmlabels[i]];    
                }
                else{
                    if(this.m_format[this.m_prmlabels[i]] == "array"){
                        this.m_jsonobj[this.m_prmlabels[i]] = [0, 0];
                    }
                    else{
                        this.m_jsonobj[this.m_prmlabels[i]] = 0;
                    }
                }
            }

            if((typeof prmtype) == "number"){
                if(prmtype < 0){ // skip
                    continue;
                }
            }
            rows.push(this.m_table.insertRow(-1)); 

            let cell = rows[rows.length-1].insertCell(-1);
            let val = this.m_jsonobj[this.m_prmlabels[i]];

            let inputid = 
                GetIDFromItem(this.m_categ, this.m_prmlabels[i], -1);
            if(prmtype == PlotObjLabel){
                cell.setAttribute("colspan","3");
                let cellcont = document.createElement("div");
                cellcont.className = "flexcont flexright";
                let item = document.createElement("button");
                item.innerHTML = "Import/View Data";
                item.addEventListener("click", (e)=>
                {
                    ShowDataImport(e.currentTarget.id);
                    let items = GetItemFromID(e.currentTarget.id)
                    if(this.m_jsonobj.hasOwnProperty(items.item)){
                        if(CheckDataObj(this.m_jsonobj[items.item])){
                            return;
                        }
                    }
                });
                item.id = inputid;
                cellcont.appendChild(item);               
                cell.appendChild(cellcont);               
                continue;
            }
            if(prmtype == FileLabel || prmtype == FolderLabel){
                this.m_fdlist[this.m_prmlabels[i]] = prmtype;
                cell.setAttribute("colspan","3");
                let cellcont = document.createElement("div");
                let filesel = GetFileInputText(inputid, val, cellcont, this.m_prmlabels[i]);
                this.m_fdialog[inputid] = filesel.dialog;
                this.m_fdialog[inputid].button.addEventListener("click",
                    (e) => {
                        let iditem = DetachSuffix(e.currentTarget.id, SuffixButton);
                        let item = GetItemFromID(iditem);
                        let inputid = e.currentTarget.id;
                        let isfile = this.m_fdlist[item.item] == FileLabel;
                        let title = "Select a directory to save the output file."
                        if(isfile){
                            title = "Select a data file."
                        }
                        let options  = {
                            title: title,
                            directory: !isfile,
                        };
                        if(SPGUI.setting.defpaths.hasOwnProperty(inputid)){
                            options.defaultPath = SPGUI.setting.defpaths[inputid];
                        }
                        window.__TAURI__.dialog.open(options).then((retstr) => {
                            if(retstr != null){
                                this.m_jsonobj[item.item] = retstr;
                                SPGUI.setting.defpaths[inputid] = retstr;
                                document.getElementById(iditem).value = retstr;
                            }
                        });                        
                    });
                let item = filesel.item;
                item.id = inputid;
                cellcont.appendChild(item);
                cell.appendChild(cellcont);
                item.addEventListener("change", (e) => {this.Change(e);} );
                continue;
            }
            if(prmtype == GridLabel){
                cell.setAttribute("colspan","3");
                let cellcont = document.createElement("div");
                cellcont.className = "sheetarea";

                let grid = new Grid(inputid, SPGUI.sprdsheetcol[inputid]);
                this.m_grids[this.m_prmlabels[i]] = grid;
                let data = this.m_jsonobj[this.m_prmlabels[i]];
                if(Array.isArray(data) == false){
                    data = [];
                }
                grid.SetData(data);
                cellcont.appendChild(grid.GetTable());
                grid.GetTable().style.width = "100px";

                let celltitle = document.createElement("div");
                celltitle.innerHTML = this.m_prmlabels[i];

                cell.appendChild(celltitle);
                cell.appendChild(cellcont);
                grid.GetTable().addEventListener("gridchange", (e) => {
                    this.Change(e);
                });
                continue;
            }

            let fmt = this.m_format[this.m_prmlabels[i]];
            if(fmt == "array"){
                cell.setAttribute("colspan","3");

                let cellcont = document.createElement("div");
                cellcont.className = "flexcont flexspace";
                cell.appendChild(cellcont);

                let celltitle = document.createElement("div");
                celltitle.innerHTML = this.m_prmlabels[i];

                cellcont.appendChild(celltitle);

                let cellvalue = document.createElement("div");
                cellvalue.className = "flexcont";
                for(let j = 0; j < 2; j++){
                    let spdiv = document.createElement("span")
                    if(j == 0){
                        spdiv.className = "xycomma";
                        spdiv.innerHTML = "";
        	            cellvalue.appendChild(spdiv);
                    }
                    else{
                        if(this.m_prmlabels[i].indexOf(",") >= 0){
                            spdiv.className = "xycomma";
                            spdiv.innerHTML = ",";
                        }
                        else{
                            spdiv.className = "range";
                            spdiv.innerHTML = "~";
                        }
        	            cellvalue.appendChild(spdiv);
                    }
                    inputid 
                       = GetIDFromItem(this.m_categ, this.m_prmlabels[i], j);
                    
                    let valstr;
                    try {
                        valstr = ToPrmString(val[j], 3);
                    } catch(e) {
                        valstr = "";
                    }
                    
		            if(prmtype == 0){ // not editable
		            	let div = document.createElement("div");
	                    div.innerHTML = valstr;
	                    div.id = inputid;
                        div.className = "statext";
        	            cellvalue.appendChild(div);
		            }
		            else{
	                    let input = document.createElement("input");
    	                input.className = "half";
        	            input.setAttribute("type", "text");
            	        input.setAttribute("value", valstr);
	                    input.id = inputid;
    	                input.addEventListener("change", (e) => {this.Change(e);} );
        	            cellvalue.appendChild(input);
                        if(this.m_isscan[this.m_prmlabels[i]]){
                            input.addEventListener("contextmenu", (e) => {
                                let item = GetItemFromID(e.currentTarget.id);
                                let target = item.item;
                                if(item.subitem != ""){
                                    target = item.subitem;
                                }
                                if(this.m_prmtypes[target] == 1){
                                    SPGUI.targetid = this.GetTargetPrmId(e.currentTarget.id);
                                    PopContextMenu(e, "scan2d", "");    
                                }
                                return false;
                            });    
                        }
                    }
                }
                cellcont.appendChild(cellvalue);
                continue;
            }
    
            cell.innerHTML = this.m_prmlabels[i];

            cell = rows[rows.length-1].insertCell(-1);
            cell.style.width = "5px";

            cell = rows[rows.length-1].insertCell(-1);
            cell.className = "statext";

            let valstr;
            try {
                valstr = ToPrmString(val, 5);
            } catch(e) {
                valstr = "";
            }

            if(prmtype == 0){ // not editable
                cell.innerHTML = valstr;
                cell.id = inputid;
            }
            else{
                let input = document.createElement("input");
                input.setAttribute("type", "text");
                input.setAttribute("value", valstr);
                input.id = inputid;
                input.addEventListener("change", (e) => {this.Change(e);} );
                cell.appendChild(input);
                if(this.m_isscan[this.m_prmlabels[i]]){
                    input.addEventListener("contextmenu", (e) => {
                        let item = GetItemFromID(e.currentTarget.id);
                        let target = item.item;
                        if(item.subitem != ""){
                            target = item.subitem;
                        }
                        if(this.m_prmtypes[target] == 1 
                                || this.m_prmtypes[target] == 3){
                            SPGUI.targetid = this.GetTargetPrmId(e.currentTarget.id);
                            PopContextMenu(e, "scan1d", "");
                        }
                        return false;
                    });
                }
            }
        }
    }

    DisableInput(prmlabel, disables){    
        if(Array.isArray(disables)){
            for(let j = 0; j < 2; j++){
                let inputid = GetIDFromItem(this.m_categ, prmlabel, j);
                let el = document.getElementById(inputid);
                if(el == null){
                    continue;
                }            
                if(disables[j]){
                    el.setAttribute("disabled", true);
                }
                else{
                    el.removeAttribute("disabled");                        
                }
            }
        }
        else{
            let inputid = GetIDFromItem(this.m_categ, prmlabel, -1);
            let el = document.getElementById(inputid);
            if(el == null){
                return;
            }            
            if(disables){
                el.setAttribute("disabled", true);
            }
            else{
                el.removeAttribute("disabled");                        
            }
        }
    }

    GetTargetPrmId(id){
        let item = GetItemFromID(id);
        let isscan;
        if(item.subitem != ""){
            isscan = this.m_isscan[item.subitem];
        }
        else{
            isscan = this.m_isscan[item.item];
        }
        return {id:id, isscan:isscan};
    }

    Change(event){
        if(event.type == "gridchange"){
            let id = event.detail.id;
            if(!this.m_skipupdate){
                UpdatePrms(id);
            }
            return;
        }
        let tgt = event.currentTarget; 
        let idc = GetItemFromID(tgt.id);
        if(idc.categ != this.m_categ){
            if(idc.subitem == ""){
                return;
            }
            idc.item = idc.subitem;
        }

        if(idc.jxy < 0){
            this.m_jsonobj[idc.item] = parseFloat(tgt.value);
        }
        else{
            this.m_jsonobj[idc.item][idc.jxy] = parseFloat(tgt.value);
        }
        if(!this.m_skipupdate){
            UpdatePrms(tgt.id);
        }
    }

    ExportInputPrms(){
        let obj = {};
        for(let i = 0; i < this.m_prmlabels.length; i++){
            let prmtype = this.m_prmtypes == null ? 0 : this.m_prmtypes[this.m_prmlabels[i]];
            if((typeof prmtype) == "number"){
                if(prmtype == 1){ // only input files
                    obj[this.m_prmlabels[i]] = 
                        CopyJSON(this.m_jsonobj[this.m_prmlabels[i]]);
                }
            }
            else{ // PlotObjLabel, FileLabel
                obj[this.m_prmlabels[i]] = 
                    CopyJSON(this.m_jsonobj[this.m_prmlabels[i]]);
            }
        }
        return obj;
    }
}

// Option List Table
class OptionList {
    constructor(categ, optionlabels, defopt){
        this.m_table = document.createElement("div");
        this.m_table.className = "flexgrow flexcont flexcolumn";
        this.m_optionlabels = optionlabels;
        this.m_categ = categ;
        this.m_defopt = defopt;
        this.m_grids = {};
    };

    get JSONObj(){
        return this.m_jsonobj;
    };

    set JSONObj(jsonobj){
        this.m_jsonobj = jsonobj;
    };

    GetTable(){
        return this.m_table;
    };

    DisableInput(prmlabel, isdisable){
        let inputid = GetIDFromItem(this.m_categ, prmlabel, -1);
        let el = document.getElementById(inputid);
        if(el == null){
            return;
        }
        if(isdisable){
            el.setAttribute("disabled", true);
        }  
        else{
            el.removeAttribute("disabled");                        
        }
    }

    DisableOptions(disable)
    {
        for(let i = 0; i < this.m_optionlabels.length; i++){
            if(this.m_optionlabels[i] == SeparatorLabel){
                continue;
            }
            if(this.m_validlist[this.m_optionlabels[i][0]] == false){
                continue;
            }
            if(this.m_optionlabels[i][1] == GridLabel){
                this.m_grids[this.m_optionlabels[i][0]].DisableGrid(disable);
            }
            else if(this.m_optionlabels[i][1] == FileLabel 
                || this.m_optionlabels[i][1] == FolderLabel){
                let btnid = AttachSuffix(this.m_optionlabels[i][0], SuffixButton);
                this.DisableInput(btnid, disable);
                this.DisableInput(this.m_optionlabels[i][0], disable);
            }
            else if(this.m_optionlabels[i][1] == PrmTableLabel){
                this.m_tbllist[this.m_optionlabels[i][0]].DisableTable(disable);
            }
            else{
                this.DisableInput(this.m_optionlabels[i][0], disable)
            }
        }
    }

    DisableSelection(label, item, isdisable)
    {
        let id = GetIDFromItem(this.m_categ, label, -1);
        let el = document.getElementById(id);
        if(el == undefined){
            return;
        }
        let children = el.childNodes;
        for(let i = 0; i < children.length; i++){
            if(children[i].text == item){
                if(isdisable){
                    children[i].setAttribute("disabled", true);
                }
                else{
                    children[i].removeAttribute("disabled");
                }
                break;
            }
        }

    }

    GetReferenceList(simplified)
    {
        let tbl = document.createElement("table");

        /*
        let caption = document.createElement("caption");
        caption.innerHTML = "Parameters in \""+this.m_categ+"."+OptionLabel+"\" object.";
        tbl.caption = caption;
        */

        let rows = [];
        let cell;

        ArrangeObjectTblTitle(tbl, rows);

        for(let i = 0; i < this.m_optionlabels.length; i++){
            if(this.m_optionlabels[i] == SeparatorLabel){
                continue;
            }
            let tgtlabel = this.m_optionlabels[i][0];

            rows.push(tbl.insertRow(-1)); 

            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = tgtlabel;

            if(rows.length == 1){
                cell = rows[rows.length-1].insertCell(-1);
                cell.innerHTML = this.m_categ+"."+OptionLabel;
                cell.setAttribute("rowspan", idx.length.toString());
            }          

            cell = rows[rows.length-1].insertCell(-1);           
            let label = tgtlabel
                .replace(/&/g, "&amp;")
                .replace(/</g, "&lt;").replace(/>/g, "&gt;");
            cell.innerHTML = label;
            cell.className += " prm";

            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = simplified[i];
            cell.className += " prm";

            let fmt;
            if((typeof this.m_optionlabels[i][1]) == "boolean"){
                fmt = "boolean";
            }
            else if((typeof this.m_optionlabels[i][1]) == "number"){
                fmt = "number";
            }
            else if(this.m_optionlabels[i][1] == PrmTableLabel){
                fmt = "object";
            }
            else if(this.m_optionlabels[i][1] == PlotObjLabel){
                fmt = "data";
            }
            else if(this.m_optionlabels[i][1] == FileLabel){
                fmt = "file name";
            }
            else if(this.m_optionlabels[i][1] == FolderLabel){
                fmt = "directory name";
            }
            else if(this.m_optionlabels[i][1] == CommentLabel){
                fmt = "string";
            }
            else{
                fmt = this.m_optionlabels[i][1];                
                if(Array.isArray(fmt)){
                    fmt = fmt.join(", ");
                    fmt = "Select from: "+fmt;
                }
            }
            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = fmt;          

            let val = "";
            if(this.m_optionlabels[i][1] != PrmTableLabel){
                val = this.m_jsonobj[OptionLabel][tgtlabel];
            }
            if(val == undefined){
                val = "";
            }
            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = val;
        }
        let retstr = tbl.outerHTML;   
        return retstr;
    }

    IsShown(label)
    {
        return this.m_validlist[label];
    }

    SetValidity(validlist, tbllist, prmtypes){
        this.m_validlist = validlist;
        this.m_table.innerHTML = "";
        this.m_tbllist = tbllist;
        this.m_fdialog = {};
        this.m_fdlist = {};
        this.m_types = {};
        this.m_skipupdate = false;
        let isitem = false;

        for(let i = 0; i < this.m_optionlabels.length; i++){
            if(this.m_optionlabels[i] == SeparatorLabel){
                if(isitem){
                    let isshow = false;
                    for(let j = i+1; j < this.m_optionlabels.length; j++){
                        if(this.m_optionlabels[j] == SeparatorLabel){
                            continue;
                        }
                        if(this.m_validlist[this.m_optionlabels[j][0]]){
                            isshow = true;
                            break;
                        }
                    }
                    if(isshow){
                        let cell = document.createElement("hr");
                        cell.style.width = "100%";
                        this.m_table.appendChild(cell);
                        isitem = false;
                    }
                }
                continue;
            }
            if(this.m_jsonobj[OptionLabel].hasOwnProperty(this.m_optionlabels[i][0]) == false){
                if(this.m_optionlabels[i][1] == PlotObjLabel){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]] = {};
                }
                else if(this.m_optionlabels[i][1] == FileLabel || 
                    this.m_optionlabels[i][1] == FolderLabel){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]] = "";
                }
                else if(this.m_optionlabels[i][1] == GridLabel){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]] = [];
                }
                else if(this.m_defopt.hasOwnProperty(this.m_optionlabels[i][0])){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]]
                        = this.m_defopt[this.m_optionlabels[i][0]];
                }
            }

            if(this.m_validlist[this.m_optionlabels[i][0]] == false){
                continue;
            }
            isitem = true;
            let cell = document.createElement("div");
            cell.className = "flexcont flexspace flexcenter";
            this.m_table.appendChild(cell);

            this.m_types[this.m_optionlabels[i][0]] = this.m_optionlabels[i][1];

            let changable = this.m_optionlabels[i][1] != PlotObjLabel
                && this.m_optionlabels[i][1] != FileLabel
                && this.m_optionlabels[i][1] != FolderLabel
                && this.m_optionlabels[i][1] != GridLabel
                && this.m_optionlabels[i][1] != PrmTableLabel;

            let val = this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]];
            let inputid = 
                GetIDFromItem(this.m_categ, this.m_optionlabels[i][0], -1);

            if((typeof this.m_optionlabels[i][1]) == "boolean"){
                let chkdiv = document.createElement("div");
                chkdiv.className = "flexcont flexcenter";
                let chkbox = document.createElement("input");
                chkbox.setAttribute("type", "checkbox");
                if(val == true){
                    chkbox.setAttribute("checked", "checked");
                }
                chkbox.id = inputid;
                chkbox.addEventListener("change", (e) => {this.Change(e);} );
                let label = document.createElement("label");
                label.setAttribute("for", inputid);
                label.innerHTML = this.m_optionlabels[i][0];
                chkdiv.appendChild(chkbox);
                chkdiv.appendChild(label);
                cell.appendChild(chkdiv);
                continue;                
            }

            if(changable){
                let celltitle = document.createElement("div");
                celltitle.innerHTML = this.m_optionlabels[i][0];
                celltitle.className = "rpadding";
                cell.appendChild(celltitle);
            }

            let item;
            if((typeof this.m_optionlabels[i][1]) == "object"){
                item = document.createElement("select");
                item.className = "option";
                let isexist = false;
                for(let j = 0; j < this.m_optionlabels[i][1].length; j++){
                    let option = document.createElement("option");
                    option.text = this.m_optionlabels[i][1][j];
                    if(val == this.m_optionlabels[i][1][j]){
                        option.selected = true;
                        isexist = true;
                    }
                    item.appendChild(option);
                }
                if(!isexist){
                    item.selectedIndex = -1;
                }
            }
            else if(this.m_optionlabels[i][1] == GridLabel){
                cell.className = "flexcont flexcolumn";
                item = document.createElement("div");
                item.className = "sheetarea";
                let grid = new Grid(inputid, SPGUI.sprdsheetcol[inputid]);
                this.m_grids[this.m_optionlabels[i][0]] = grid;
                if(Array.isArray(
                        this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]]) == false){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]] = [];
                }
                grid.SetData(
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]]);
                item.appendChild(grid.GetTable());
                grid.GetTable().addEventListener("gridchange", (e) => {
                    this.Change(e);
                });
                let celltitle = document.createElement("div");
                celltitle.innerHTML = this.m_optionlabels[i][0];
                cell.appendChild(celltitle);
            }
            else if(this.m_optionlabels[i][1] == PrmTableLabel)
            {
                item = document.createElement("div");
                item.className = "flexgrow flexright flexcont";
                let tbl = this.m_tbllist[this.m_optionlabels[i][0]];
                let prmtype = prmtypes[this.m_optionlabels[i][0]];
                if(this.m_jsonobj[OptionLabel].hasOwnProperty(this.m_optionlabels[i][0]) == false){
                    this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]]
                        = CopyJSON(this.m_defopt[this.m_optionlabels[i][0]]);
                }
                tbl.JSONObj = this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]];
                tbl.SetPrmTypes(prmtype);
                let tblht = tbl.GetTable();
                item.appendChild(tblht);
            }
            else if(this.m_optionlabels[i][1] == FileLabel 
                    || this.m_optionlabels[i][1] == FolderLabel)
            {
                this.m_fdlist[this.m_optionlabels[i][0]] = this.m_optionlabels[i][1];
                let filesel = GetFileInputText(inputid, val, cell, this.m_optionlabels[i][0]);
                item = filesel.item;
                this.m_fdialog[inputid] = filesel.dialog;
                this.m_fdialog[inputid].button.addEventListener("click",
                    (e) => {
                        let iditem = DetachSuffix(e.currentTarget.id, SuffixButton);
                        let item = GetItemFromID(iditem);
                        let inputid = e.currentTarget.id;
                        let isfile = this.m_fdlist[item.item] == FileLabel;
                        let title = "Select a directory to save the output file."
                        if(isfile){
                            title = "Select a data file."
                        }
                        let options  = {
                            title: title,
                            directory: !isfile,
                        };
                        if(SPGUI.setting.defpaths.hasOwnProperty(inputid)){
                            options.defaultPath = SPGUI.setting.defpaths[inputid];
                        }
                        window.__TAURI__.dialog.open(options).then((retstr) => {
                            if(retstr != null){
                                this.m_jsonobj[OptionLabel][item.item] = retstr;
                                SPGUI.setting.defpaths[inputid] = retstr;
                                document.getElementById(iditem).value = retstr;
                            }
                        });                        


                    });
                item.addEventListener("change", (e) => {this.Change(e);} );
            }
            else if(this.m_optionlabels[i][1] == PlotObjLabel)
            {
                cell.classList.remove("flexspace");
                cell.className += " flexright";
                item = document.createElement("button");
                item.innerHTML = "Import/View Data";
                item.addEventListener("click", (e)=>
                {
                    ShowDataImport(e.currentTarget.id);
                    let items = GetItemFromID(e.currentTarget.id)
                    if(this.m_jsonobj[OptionLabel].hasOwnProperty(items.item)){
                        if(CheckDataObj(this.m_jsonobj[OptionLabel][items.item])){
                            return;
                        }
                    }
                });
            }
            else if(this.m_optionlabels[i][1] == CommentLabel ||
                    this.m_optionlabels[i][1] == StringLabel){
                item = document.createElement("textarea");
                item.setAttribute("rows", "1");
                item.className = "prmtbl";
                item.value = val;
            }
            else if(this.m_optionlabels[i][1] == FloatLabel){                
                item = document.createElement("input");
                item.setAttribute("type", "text");
                let valstr = ToPrmString(val, 5);
                item.setAttribute("value", valstr);
            }
            else{
                item = document.createElement("input");
                item.setAttribute("type", "number");
                if(this.m_optionlabels[i].length > 3){
                    item.setAttribute("step", 
                        this.m_optionlabels[i][3].toString());
                }
                if(this.m_optionlabels[i].length > 2){
                    item.setAttribute("min", 
                        this.m_optionlabels[i][2].toString());
                }
                else{
                    if(this.m_optionlabels[i][1] == 0){
                        item.setAttribute("min", "0");
                    }
                    else if(this.m_optionlabels[i][1] > 0){
                        item.setAttribute("min", "1");
                    }
                    else{
                        item.setAttribute("min", "-1");
                    }    
                }
                item.setAttribute("value", val);
            }
            item.id = inputid;
            if(changable){
                item.addEventListener("change", (e) => {this.Change(e);} );
            }
            cell.appendChild(item);
        }
    }

    Change(event){
        if(event.type == "gridchange"){
            let id = event.detail.id;
            if(!this.m_skipupdate){
                UpdatePrms(id);
            }
            return;
        }
        let tgt = event.currentTarget; 
        let idc = GetItemFromID(tgt.id);
        if(idc.categ != this.m_categ){
            return;
        }
        if(this.m_types[idc.item] == FloatLabel){
            this.m_jsonobj[OptionLabel][idc.item] = parseFloat(tgt.value);
        }
        else if(tgt.type == "text" || 
                tgt.type == "textarea" || tgt.type == "select-one"){
            this.m_jsonobj[OptionLabel][idc.item] = tgt.value;
        }
        else if(tgt.type == "checkbox"){
            this.m_jsonobj[OptionLabel][idc.item] = tgt.checked;
        }
        else if(tgt.type == "number"){
            this.m_jsonobj[OptionLabel][idc.item] = parseFloat(tgt.value);
        }
        UpdateOptions(tgt.id);
    };

    ExportInputOptions(){
        let obj = {};
        for(let i = 0; i < this.m_optionlabels.length; i++){
            if(this.m_optionlabels[i] == SeparatorLabel){
                continue;
            }
            if(this.m_validlist[this.m_optionlabels[i][0]] == false){
                continue;
            }
            if(this.m_optionlabels[i][1] == PrmTableLabel){
                let prmobj = this.m_tbllist[this.m_optionlabels[i][0]].ExportInputPrms();
                obj[this.m_optionlabels[i][0]] = CopyJSON(prmobj);
            }
            else{
                obj[this.m_optionlabels[i][0]] = 
                    CopyJSON(this.m_jsonobj[OptionLabel][this.m_optionlabels[i][0]]);
            }
        }
        return obj;
    }
}

// Notebook Panel
class NoteBook
{
    constructor(parent, titles, issub){
        this.m_titles = [];
        this.m_panels = [];
        this.m_parent = parent;
        if(issub){
            this.m_parent.className += " sub";
        }
        this.m_label = document.createElement("div");
        this.m_label.className = "note_label";
        this.m_parent.appendChild(this.m_label);
        for(let j = 0; j < titles.length; j++){
            this.AddPanel(titles[j], j == 0, issub);        
        }
        this.m_currpage = 0;
    }

    AddPanel(title, isactive, issub){
        let ltitle = document.createElement("div");
        ltitle.className = "note_title";
        ltitle.innerHTML = title;
        let panel = document.createElement("div");
        panel.className = "note_panel";
        if(issub){
            ltitle.className += " sub";
            panel.className += " sub";
        }
        if(isactive){
            ltitle.className += " active";
            panel.className += " active";
        }
        this.m_panels.push(panel);
        this.m_titles.push(ltitle);
        this.m_label.appendChild(ltitle);
        this.m_parent.appendChild(panel);
        ltitle.addEventListener("click",
            (e) => {
                let idx = Array.prototype.indexOf.call(this.m_titles, e.currentTarget);
                this.ShowPage(idx);
            }
        );
    }

    ShowPage(index){
        for(let j = 0; j < this.m_titles.length; j++){
            this.m_titles[j].classList.remove("active");
        }
        for(let j = 0; j < this.m_panels.length; j++){
            this.m_panels[j].classList.remove("active");
        }
        this.m_titles[index].classList.add("active");
        this.m_panels[index].classList.add("active");
        this.m_currpage = index;

        let eventup = new CustomEvent("pagechange", { detail:{ page:index} });
        this.m_parent.dispatchEvent(eventup);
    }

    GetPanel(index){
        return this.m_panels[index];
    }

    GetCurrentPage(){
        return this.m_currpage;
    }
}

// Grid Controls (spread sheed)
class Grid {
    constructor(id, gridconf){
        this.m_id = id;
        this.m_coltypes = gridconf.coltypes;
        this.m_table = document.createElement("table");
        this.m_table.className = "grid";
        this.m_table.style.tableLayout = "auto";
        this.m_withnum = gridconf.withnum;
        this.m_readonly = false;
        if(gridconf.hasOwnProperty("readonly")){
            this.m_readonly = gridconf.readonly;
        }
    }

    DisableGrid(isdisable){
        for(let i = 1; i < this.m_table.rows.length; i++){
            for(let j = 0; j < this.m_coltypes.length; j++){
                let id = GetIdFromCell(this.m_id, i-1, j);
                let el = document.getElementById(id);
                if(el == null){
                    continue;
                }            
                if(isdisable){
                    el.setAttribute("disabled", true);
                }
                else{
                    el.removeAttribute("disabled");
                }
            }
        }
    }

    SetData(data){
        this.m_data = data;
        this.m_table.innerHTML = "";
        let titlerow = this.m_table.insertRow(-1);
        let cell;

        if(this.m_withnum){
            cell = titlerow.insertCell(-1);
            cell.innerHTML = "";    
        }
        for(let j = 0; j < this.m_coltypes.length; j++){
            cell = titlerow.insertCell(-1);
            cell.innerHTML = this.m_coltypes[j][GridColLabel];
            cell.className = "title";
        }
        let rowdata;
        let nrows = data.length;
        if(!this.m_readonly){
            nrows += AdditionalRows;
        } 
        for(let i = 0; i < nrows; i++){
            rowdata = i >= data.length ? "" : data[i];
            this.AppendRow(rowdata);
        }
    }

    AppendRow(rowdata){
        let cell, item, val;
        let row = this.m_table.insertRow(-1);

        if(this.m_withnum){
            cell = row.insertCell(-1);
            cell.innerHTML = this.m_table.rows.length-1;
        }
        for(let j = 0; j < this.m_coltypes.length; j++){
            cell = row.insertCell(-1);
            val = rowdata == "" ? "" : rowdata[j];
            if((typeof this.m_coltypes[j][GridTypeLabel]) == "object"){
                item = document.createElement("select");
                let isselect = false;

                for(let k = 0; k < this.m_coltypes[j][GridTypeLabel].length; k++){
                    let option = document.createElement("option");
                    option.text = this.m_coltypes[j][GridTypeLabel][k];
                    if(val == this.m_coltypes[j][GridTypeLabel][k]){
                        option.selected = true;
                        isselect = true;
                    }
                    item.appendChild(option);
                }
                if(!isselect){
                    item.selectedIndex = -1;
                }
                item.addEventListener("keydown", (e) => {
                    if(e.key == "Delete"){
                        e.currentTarget.selectedIndex = -1;
                        let cell = GetCellFromId(this.m_id, e.currentTarget.id);
                        this.m_data[cell[0]][cell[1]] = "";
                    }
                });
            }
            else if(this.m_readonly){
                item = document.createElement("div");
                if(typeof val == "number"){
                    item.innerHTML = ToPrmString(val, 4);
                }
                else{
                    item.innerHTML = val;
                }
            }
            else{
                item = document.createElement("input");
                item.setAttribute("type", "text");
                if(typeof val == "number"){
                    item.setAttribute("value", ToPrmString(val, 4));
                }
                else{
                    item.setAttribute("value", val);
                }
                if(this.m_readonly){
                    item.setAttribute("readonly", "readonly");
                }
            }
            item.addEventListener("change", (e) => {this.Change(e);} );
            item.id = GetIdFromCell(this.m_id, this.m_table.rows.length-2, j);
            cell.appendChild(item);
        }        
    }

    GetTable(){
        return this.m_table;
    }

    Change(event){
        let cell = GetCellFromId(this.m_id, event.currentTarget.id);
        let nadd = cell[0]+AdditionalRows-(this.m_table.rows.length-2);
        for(let n = 0; n < nadd; n++){
            this.AppendRow("");
        }
        nadd = cell[0]-(this.m_data.length-1);
        for(let n = 0; n < nadd; n++){
            this.m_data.push([]);
            for(let j = 0; j < this.m_coltypes.length; j++){
                this.m_data[this.m_data.length-1].push("");
            }
        }
        this.m_data[cell[0]][cell[1]] = event.currentTarget.value;
        let eventup = new CustomEvent("gridchange", { detail: {id: this.m_id} });
        this.m_table.dispatchEvent(eventup);
    }
}

// GUI to edit filter material
class EditFilterMaterial {
    constructor(id, builtin, custom){      
        this.m_id = id;
        let title = document.createElement("caption");
        title.innerHTML = "Filter/Absorber Materials Available";
        this.m_table = document.createElement("table");
        this.m_table.appendChild(title);
        this.m_table.className = "grid fmaterial";
        this.m_builtinobjs = builtin;
        this.m_materials = custom;

        this.m_elements = 0;
        for(let el in builtin){
            this.m_elements = Math.max(this.m_elements, builtin[el].comp.length);
        }
        for(let el in custom){
            this.m_elements = Math.max(this.m_elements, custom[el].comp.length);
        }

        let cell;
        for(let k = 0; k < 2; k++){
            let titlerow = this.m_table.insertRow(-1);
            if(k == 0){
                cell = titlerow.insertCell(-1);
                cell.setAttribute("rowspan","2");
                cell.innerHTML = "Name";
                cell.className = "title";
                cell = titlerow.insertCell(-1);
                cell.setAttribute("rowspan","2");
                cell.innerHTML = "Density (g/cm<sup>3</sup>)";
                cell.className = "title";    
            }
            for(let i = 0; i < this.m_elements; i++){
                cell = titlerow.insertCell(-1);
                cell.className = "title";
                if(k == 0){
                    cell.setAttribute("colspan","2");
                    cell.innerHTML = "Element"+(i+1).toString();
                }
                else{
                    cell.innerHTML = "Z";
                    cell = titlerow.insertCell(-1);
                    cell.innerHTML = "Ratio";
                    cell.className = "title";
                }
            }    
        }

        let filters = [builtin, custom];
        for(let k = 0; k < filters.length; k++){
            let keys = Object.keys(filters[k]);
            if(k == 0){
                this.m_builtins = keys.length;
            }
            for(let n = 0; n < keys.length; n++){
                this.AddMaterial(keys[n], filters[k][keys[n]], k == 0);
            }
        }
        for(let n = 0; n < AdditionalRows; n++){
            this.AddMaterial("", "", false);
        }
        this.AddElement();
    }

    GetTable(){
        return this.m_table;
    }

    AddMaterial(matname, matobj, isreadonly){
        let cell, item, val;
        let row = this.m_table.insertRow(-1);
        for(let j = -2; j < this.m_elements*2; j++){
            cell = row.insertCell(-1);
            if(matname == "" || j >= matobj.comp.length*2){
                val = "";
            }
            else if(j == -2){
                val = matname;
            }
            else if(j == -1){
                val = matobj.dens;
            }
            else if(j%2 ==0){
                val = matobj.comp[j/2][0];
            }
            else{
                val = matobj.comp[(j-1)/2][1];
            }
            if(isreadonly){
                cell.innerHTML = val;
                cell.className = "rdonly";    
            }
            else{
                item = document.createElement("input");
                item.setAttribute("type", "text");
                item.setAttribute("value", val);
                item.addEventListener("change", (e) => {this.Modify(e);} );
                item.id = GetIdFromCell(this.m_id, this.m_table.rows.length-3, j);
                cell.appendChild(item);    
            }
        }        

    }

    AddElement(){
        this.m_elements++;
        let rows = this.m_table.rows;
        let cell, item;

        cell = rows[0].insertCell(-1);
        cell.setAttribute("colspan","2");
        cell.innerHTML = "Element"+this.m_elements.toString(); cell.className = "title";

        cell = rows[1].insertCell(-1);
        cell.innerHTML = "Z"; cell.className = "title";
        cell = rows[1].insertCell(-1);
        cell.innerHTML = "Ratio"; cell.className = "title";

        for(let n = 2; n < rows.length; n++){
            for(let j = 0; j < 2; j++){
                cell = rows[n].insertCell(-1);
                if(n < this.m_builtins+2){
                    cell.innerHTML = "";
                    cell.className = "rdonly";        
                }
                else{
                    item = document.createElement("input");
                    item.setAttribute("type", "text");
                    item.setAttribute("value", "");
                    item.addEventListener("change", (e) => {this.Modify(e);} );
                    let cellidx = 2*(this.m_elements-1)+j;
                    item.id = GetIdFromCell(this.m_id, n-2, cellidx);
                    cell.appendChild(item);    
                }
            }
        }
    }

    Modify(event){
        let cell = GetCellFromId(this.m_id, event.currentTarget.id);
        let naddrow = cell[0]+AdditionalRows-(this.m_table.rows.length-3);
        for(let n = 0; n < naddrow; n++){
            this.AddMaterial("", "");
        }
        if(cell[1] >= 2*(this.m_elements-1)){
            this.AddElement();
        }
        let name = document.getElementById(GetIdFromCell(this.m_id, cell[0], -2));
        let dens = document.getElementById(GetIdFromCell(this.m_id, cell[0], -1));
        let conts = [];
        for(let n = 0; n < this.m_elements; n++){
            let znum = parseInt(document.getElementById(GetIdFromCell(this.m_id, cell[0], 2*n)).value);
            let ratio = parseFloat(document.getElementById(GetIdFromCell(this.m_id, cell[0], 2*n+1)).value);
            if(!isNaN(znum) && !isNaN(ratio)){
                conts.push([znum, ratio]);
            }
        }
        if(cell[1] == -2){
            let oldname = event.currentTarget.defaultValue;
            if(this.m_materials.hasOwnProperty(oldname)){
                delete this.m_materials[oldname];
            }
            name.setAttribute("value", name.value);
        }
        let ndens = parseFloat(dens.value);
        if(!isNaN(ndens) && conts.length > 0){
            this.m_materials[name.value] = {dens:ndens, comp:conts};
        }
    }

    GetMaterialNames(){
        let bnames = Object.keys(this.m_builtinobjs);
        let mnames = Object.keys(this.m_materials);
        return bnames.concat(mnames);
    }
}

// GUI to operate the calculation process
class CalcProcess {
    constructor(issingle, spobj, dataname, id){
        this.m_id = id;
        this.m_running = false;
        this.m_issingle = issingle;
        this.m_spobj = [];
        this.m_spobj.push(spobj);
        this.m_datanames = [];
        this.m_datanames.push(dataname);
        this.m_div = document.createElement("div");
        this.m_div.className = "flexcolumn flexstretch flexgrow calcproc";
        this.m_div.id = id;

        let progcnt = document.createElement("div");
        progcnt.className = "flexcont flexspace flexcenter";
        this.m_div.appendChild(progcnt);

        if(SPGUI.runEnvironment == "tauri" && SPGUI.runOS == "Darwin"){
            this.m_progress = document.createElement("div");
            this.m_progress.className = "flexgrow"
            this.m_progress.id = GetIDFromItem(id, "progress", -1);    
        }
        else{
            this.m_progress = document.createElement("progress");
            this.m_progress.setAttribute("max", "100");
            this.m_progress.setAttribute("value", "0");
            this.m_progress.className = "flexgrow"
            this.m_progress.id = GetIDFromItem(id, "progress", -1);    
        }
        this.m_progress.style.visibility = "hidden";

        progcnt.appendChild(this.m_progress);

        let btncnt = document.createElement("div");
        btncnt.className = "flexcont flexcenter";
        if(issingle){
            let title = document.createElement("div");
            title.style.width = "220px";
            title.style.whiteSpace = "normal";
            title.style.wordBreak = "break-all";
            title.innerHTML = dataname;
            btncnt.appendChild(title);
        }
        else{
            let btn = document.createElement("input");
            btn.setAttribute("type", "button");
            btn.setAttribute("value", CancellAllLabel);
            btn.addEventListener("click", (e) => {
                this.CancelAll();
            });
            btncnt.appendChild(btn);

            this.m_list = document.createElement("select");
            this.m_list.setAttribute("size", "5");
            this.m_list.style.minWidth = "100%";
            this.m_list.setAttribute("multiple", "multiple");
            let item = document.createElement("option");
            item.innerHTML = dataname;
            this.m_list.appendChild(item);

            let divlist = document.createElement("div");
            divlist.style.maxWidth = "280px";
            divlist.style.overflow = "auto";
            divlist.appendChild(this.m_list);
            this.m_div.appendChild(divlist);
        }

        this.m_cancelbtn = document.createElement("input");
        this.m_cancelbtn.setAttribute("type", "button");
        this.m_cancelbtn.setAttribute("value", RemoveLabel);
        this.m_cancelbtn.addEventListener("click", (e) => {
            this.Cancel();
        });
        btncnt.appendChild(this.m_cancelbtn);
        if(issingle){
            this.m_div.appendChild(btncnt);
        }
        else{
            progcnt.appendChild(btncnt);
        }
    }

    AppendProcess(spobj, dataname){
        this.m_spobj.push(spobj);
        this.m_datanames.push(dataname);
        let item = document.createElement("option");
        item.innerHTML = dataname;
        this.m_list.appendChild(item);
        if(this.m_datanames.length > 5 && this.m_datanames.length <= 10){
            this.m_list.setAttribute("size", this.m_datanames.length.toString());
        }
    }

    GetList(){
        return this.m_div;
    }

    Start(){
        this.m_currindex = 0;
        this.m_running = true;
        this.m_cancelbtn.setAttribute("value", CancelLabel);
        this.m_progress.style.visibility = "visible";
        this.StartSingle();
    }   

    async StartSingle(){
        let dataname = this.m_datanames[this.m_currindex];
        let currspobj = this.m_spobj[this.m_currindex];
        this.m_resultstr = "";

        let args;
        if(dataname == ""){// fixed point
            let prefix = SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]];
			SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]] = Date.now().toString();
            SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]] = prefix;
            dataname = GetCurrentDataName(true);
            if(dataname == ""){                
                this.m_currindex++;
                this.FinishSingle();
                return false;
            }
            args = ["-f", dataname, "-clear"];
        }
        else{
            args = ["-f", dataname];
        }
        let comtype = "solver_nompi";
        if(SPGUI.setting.hasOwnProperty(PCLabel) && SPGUI.setting[PCLabel][PCOptionsLabel.enable[0]]){
            comtype = "solver";
            args.unshift("./spectra_solver");
            args.unshift(SPGUI.setting[PCLabel][PCOptionsLabel.processes[0]].toString());
            args.unshift("-n");
        }

        if(SPGUI.runEnvironment == "node"){
            let argstrs = args.join("::");
            let prms = FormatArray(JSON.stringify(currspobj, null, JSONIndent));
            SPGUI.socketio.on(this.m_progress.id, (msg, result, dataname) => {
                if(msg == "finish" || msg == "scannext"){
                    if(this.m_resultstr != ""){
                        this.SetCalcResult();
                    }
                    else if(result != ""){
                        let files = [{name:dataname}];
                        SPGUI.postprocessor.SetImportFiles(files);
                        SPGUI.postprocessor.LoadOutputFile(result); 
                    }
                    if(msg == "finish"){
                        this.FinishSingle();
                    }
                }
                else if(msg.indexOf("pid") >= 0){
                    this.m_pid = parseInt(msg.replace("pid", ""));                    
                }
                else{
                    this.HandleStatus(msg);
                }
            });
            SPGUI.socketio.emit("run", prms, dataname, this.m_progress.id, argstrs);
        }
        else{
            let prms = FormatArray(JSON.stringify(currspobj, null, JSONIndent));
            let result = await window.__TAURI__.tauri.invoke(
                    "write_file", { path: dataname, data: prms});
            if(result != ""){
                Alert(result);
                this.m_currindex++;
                this.FinishSingle();
                return;
            }

            const command = new window.__TAURI__.shell.Command(comtype, args);
            command.on("close", (data) => {
                let msg = "Done";
                if(this.m_errmsgs.length > 0){
                    Alert(this.m_errmsgs.join(" "));
                    msg = "Terminated incorrectly.";
                }
                else if(this.m_process == null){
                    msg = "Canceled";
                }
                else if(data.code != 0){
                    msg = "Solver terminated incorrectly; exit code "+data.code.toString();
                }
                else if(this.m_resultstr != ""){
                    this.SetCalcResult();
                }
                else{
                    let files = [{name:this.m_datanames[this.m_currindex-1]}];
                    if(this.m_datanames[this.m_currindex-1] != ""){
                        window.__TAURI__.tauri.invoke("read_file", { path: this.m_datanames[this.m_currindex-1]})
                        .then((data) => {
                            SPGUI.postprocessor.SetImportFiles(files);
                            SPGUI.postprocessor.LoadOutputFile(data);
                        });
                    }
                }
                this.FinishSingle(msg);
            });
            command.stdout.on("data", (data) => {
                this.HandleStatus(data);
            });
            command.stderr.on("data", (data) => {
                this.m_errmsgs.push(data);
                this.m_process = null;
            });
    
            this.m_errmsgs = [];
            try{
                this.m_process = await command.spawn();
            } catch(e) {
                Alert("Cannot launch the solver: "+e+".");
                this.m_currindex++;
                this.FinishSingle();
                return;
            }
    
        }

        if(!this.m_issingle){
            this.m_list.childNodes[this.m_currindex].innerHTML 
            = this.m_datanames[this.m_currindex]+": in progress";
        }

        this.m_currindex++;
    }

    SetCalcResult(){
        try {
            let resobj = JSON.parse(this.m_resultstr);
            let items = resobj["titles"];
            let units = resobj["units"];
            let data = resobj["data"];
            let grdata = [];
            for(let i = 0; i < items.length; i++){
                if(units[i] != ""){
                    grdata.push([items[i]+" ("+units[i]+")", data[i]]);
                }
                else{
                    grdata.push([items[i], data[i]]);
                }
            }
            SPGUI.caclresultgrid.SetData(grdata);    
        } catch (err) {
            Alert("Error: Cannot get the output result.\n"+err.message);
        }
    }

    HandleStatus(status){
        if(status.indexOf(CalcStatusLabel) >= 0){
            let pct = parseInt(status.replace(CalcStatusLabel, ""));
            if(typeof pct == "number" && pct <= 100){
                if(SPGUI.runEnvironment == "tauri" && SPGUI.runOS == "Darwin"){
                    pct = parseFloat(status.replace(CalcStatusLabel, ""));
                    this.m_progress.innerHTML = "Running: "+pct+"% done";
                }
                else{
                    this.m_progress.value = pct;
                }
            }
        }
        else if(status.indexOf(Fin1ScanLabel) >= 0){
            let procs = status.replace(Fin1ScanLabel, "").split("/");
            if(procs.length >= 2){
                let total = parseInt(procs[1]);
                let curr = Math.min(total, parseInt(procs[0])+1);
                if(typeof curr == "number" && typeof total == "number"){
                    this.m_list.childNodes[this.m_currindex-1].innerHTML 
                    = this.m_datanames[this.m_currindex-1]
                    +": "+curr.toString()+"/"+total.toString()+" in Progress";
                }    
            }
        }
        else if(status.includes(ScanOutLabel)){
            let outfile = status.replace(ScanOutLabel, "").trim();
            let files = [{name:outfile}];
            if(SPGUI.runEnvironment == "tauri"){
                if(outfile != ""){
                    window.__TAURI__.tauri.invoke("read_file", {path: outfile})
                    .then((data) => {
                        SPGUI.postprocessor.SetImportFiles(files);
                        SPGUI.postprocessor.LoadOutputFile(data);
                    });
                }
            }
            else{
                let result = SPGUI.fs_node.readFileSync(outfile, "utf8");
                SPGUI.postprocessor.SetImportFiles(files);
                SPGUI.postprocessor.LoadOutputFile(result);    
            }
        }
        else if(status.indexOf(ErrorLabel) >= 0){
            Alert(status);
        }
        else if(this.m_resultstr != ""){
            this.m_resultstr += status;
        }

        let outstr = "\"Output\": "
        let idx = status.indexOf(outstr);
        if(idx >= 0){
            this.m_resultstr = status.slice(idx+outstr.length);
        }    
    }

    ReleaseProcess(id)
    {
        let cprocdiv = document.getElementById("calcproc");
        let item = document.getElementById(id);
        cprocdiv.removeChild(item);
        if(cprocdiv.childElementCount == 0){
            document.getElementById("calcprocfld").classList.replace("flexcont", "none");
        }
        SPGUI.calcprocess = null;
    }
    
    FinishSingle(msg = "Done"){
        if(this.m_currindex < this.m_spobj.length){
            this.m_list.childNodes[this.m_currindex-1].disabled = true;
            this.m_list.childNodes[this.m_currindex-1].innerHTML 
                = this.m_datanames[this.m_currindex-1]+": "+msg;
            this.StartSingle();
        }
        else{
            this.m_div.innerHTML = "";
            this.m_div.style.display = "none";
            this.ReleaseProcess(this.m_id);
        }
    }

    async CancelAll(){
        if(this.m_running){
            if(SPGUI.runEnvironment == "node"){
                SPGUI.socketio.emit("kill", this.m_pid.toString());
            }
            else{
                await this.m_process.kill();
                this.m_process = null;
            }
        }
        this.m_div.innerHTML = "";
        this.m_div.style.display = "none";
        this.ReleaseProcess(this.m_id);
    }

    Cancel(){
        if(this.m_issingle){
            this.KillCurrent();
            this.FinishSingle();
            return;
        }
        let selects = [];
        for(let n = 0; n < this.m_list.options.length; n++){
            if(this.m_list.options[n].selected){
                selects.push(n);
            }
        }        
        for(let n = selects.length-1; n >= 0; n--){
            if(this.m_running){
                if(selects[n] >= this.m_currindex){
                    this.RemoveProcess(selects[n]);
                }
                else{
                    this.KillCurrent();
                }
            }
            else{
                this.RemoveProcess(selects[n]);
            }        
        }        

    }

    RemoveProcess(index){
        let item = this.m_list.childNodes[index];
        this.m_list.removeChild(item);
        this.m_spobj.splice(index, 1);
        this.m_datanames.splice(index, 1)
    }

    async KillCurrent(){
        if(SPGUI.runEnvironment == "node"){
            SPGUI.socketio.emit("kill", this.m_pid.toString());
        }
        else{
            await this.m_process.kill();
            this.m_process = null;
        }
    }

    ExportProcesses(){
        let prms;
        if(this.m_spobj.length == 1){            
            prms = FormatArray(JSON.stringify(this.m_spobj[0], null, JSONIndent));
        }
        else{
            let calcobjs = [];
            for(let n = 0; n < this.m_datanames.length; n++){
                let calcobj = {};
                calcobj[this.m_datanames[n]] = this.m_spobj[n];
                calcobjs.push(calcobj);
            }
            prms = FormatArray(JSON.stringify(calcobjs, null, JSONIndent));
        }
        return prms;
    }
}

// Class to hold the ascii data (for pre-processing)
class AsciiData {
    constructor(dim, ntargets, titles, ordinate, extitle = ""){
        this.m_dim = dim; 
            // dimension of argument 1 (2D) or 2 (3D)
        this.m_ntargets = ntargets; 
            // number of target items as a function of x (and y) 
        this.m_nitems = this.m_ntargets+this.m_dim;
            // number of items to load
        this.m_labels = new Array(this.m_nitems);
        this.m_values = new Array(this.m_nitems);
        this.m_ordinate = ordinate;
        this.m_titles = [];
        for(let m = 0; m < this.m_nitems; m++){
            this.m_values[m] = [];
            if(m < titles.length){
                this.m_titles.push(titles[m]);
            }
            else{
                this.m_titles.push("col"+m);
            }
        }
        this.m_extitle = extitle;
        this.m_exdata = [];
    }

    GetXYMeshNumber(x, y)
    {
        let dx, dy, isx1st;
        let ndata = Math.min(x.length, y.length);

        if(ndata < 2){
            return {meshx:1, meshy:1, isx1st:true};
        }

        dy = Math.abs(y[ndata-1]-y[0])/(ndata-1)*1.0e-8;
        dx = Math.abs(x[ndata-1]-x[0])/(ndata-1)*1.0e-8;

        let n;
        for(n = 1; n < ndata; n++){
            if(n == 1){
                isx1st = Math.abs(x[n]-x[n-1]) > dx;
            }
            if(Math.abs(y[n]-y[n-1]) > dy 
                    && Math.abs(x[n]-x[n-1]) > dx){
                break;
            }
        }

        let meshx, meshy;
        if(isx1st){
            meshx = n;
            meshy = ndata/meshx;
        }
        else{
            meshy = n;
            meshx = ndata/meshy;
        }
        return {meshx:meshx, meshy:meshy, isx1st:isx1st};
    }

    SetData(data, defunits){
        let lines;
        if(this.m_dim == 0){
            lines = data.split(/\n|,/);
        }
        else{
            lines = data.split(/\n/);
        }
        let items, n;
        let tmp = new Array(this.m_nitems);

        let units = [];
        for(let m = 0; m < this.m_nitems; m++){
            this.m_values[m].length = 0;
            if(Array.isArray(defunits)){
                if(m < defunits.length){
                    units.push(defunits[m]);
                }
                else{
                    units.push(1.0);    
                }    
            }
            else{
                units.push(1.0);
            }
        }

        this.m_exdata = [];
        let exdata = [], extitle;
        for(n = 0; n < lines.length; n++){
            items = lines[n].trim().split(/\s*,\s*|\s+/);
            if(items.length < this.m_nitems){
                continue;
            }
            let isheader = false;
            for(let m = 0; m < this.m_nitems; m++){
                tmp[m] = parseFloat(items[m]);
                if(isNaN(tmp[m])){
                    isheader = true;
                    break;
                }
            }
            if(isheader){
                for(let m = 0; m < this.m_nitems; m++){
                    this.m_labels[m] = items[m];
                }
                if(items.length > this.m_nitems){
                    extitle = items[this.m_nitems];
                }
            }
            else{
                for(let m = 0; m < this.m_nitems; m++){
                    this.m_values[m].push(tmp[m]*units[m]);
                }
                if(items.length > this.m_nitems){
                    exdata.push(parseFloat(items[this.m_nitems]));
                }
            }
        }
        if(exdata.length == this.m_values[0].length){
            this.m_exdata = exdata;
        }

        if(this.m_dim < 2){
            return;
        }

        this.m_z = new Array(this.m_ntargets);

        let index;
        let xyspec = this.GetXYMeshNumber(this.m_values[0], this.m_values[1]);
        this.m_x = new Array(xyspec.meshx);
        this.m_y = new Array(xyspec.meshy);
        for(let i = 0; i < this.m_ntargets; i++){
            this.m_z[i] = new Array(xyspec.meshy);            
        }
        for(let m = 0; m < xyspec.meshy; m++){
            for(let i = 0; i < this.m_ntargets; i++){
                this.m_z[i][m] = new Array(xyspec.meshx);
            }
            index = xyspec.isx1st?(m*xyspec.meshx):m;
            this.m_y[m] = this.m_values[1][index];
        }
        for(let n = 0; n < xyspec.meshx; n++){
            index = xyspec.isx1st?n:(n*xyspec.meshy);
            this.m_x[n] = this.m_values[0][index];
        }
        for(let m = 0; m < xyspec.meshy; m++){
            for(let n = 0; n < xyspec.meshx; n++){
                if(xyspec.isx1st){
                    index = m*xyspec.meshx+n;
                }
                else{
                    index = n*xyspec.meshy+m;
                }
                for(let i = 0; i < this.m_ntargets; i++){
                    this.m_z[i][m][n] = this.m_values[i+2][index];                    
                }
            }
        }
    }

    GetTitle(j){
        return this.m_titles[j];
    }

    GetTitles(){
        return this.m_titles;
    }

    GetOrdinate(){
        return this.m_ordinate;
    }

    GetDimension(){
        return this.m_dim;
    }

    GetItems(){
        return this.m_nitems;
    }

    GetObj(){
        let titles = [];
        let data = [];
        if(this.m_dim < 2){
            for(let m = 0; m < this.m_nitems; m++){
                titles.push(this.m_titles[m]);
                data.push(this.m_values[m]);
            }
        }
        else{
            titles.push(this.m_titles[0]);
            titles.push(this.m_titles[1]);
            data.push(this.m_x);
            data.push(this.m_y);
            for(let m = 0; m < this.m_ntargets; m++){
                titles.push(this.m_titles[m+2]);
                data.push(this.m_z[m]);
            }
        }
        let obj = {};
        obj[TilteLabel] = titles;
        obj[DataLabel] = data;
        if(this.m_extitle != "" && this.m_exdata.length > 0){
            obj[this.m_extitle] = this.m_exdata;
        }
        return obj;
    }
}

// Class to plot the data held by AsciiData
class PlotArea {
    constructor(id, asciifmt, extitle = ""){
        this.m_currentId = id;
        let items =  GetItemFromID(id);
        this.m_label = items.item;
        this.m_unitids = asciifmt[UnitLabel];
        this.m_newdata = true;
        this.m_asciifmt = 
            new AsciiData(
                asciifmt[DimensionLabel],
                asciifmt[TargetsLabel],
                asciifmt[TitlesLabel],
                asciifmt[OrdinateLabel],
                extitle);
        this.m_plotwindow = null;
        this.m_extitle = extitle;

        this.m_filereader = new FileReader();
        this.m_filereader.addEventListener("load", () => {
            let convfactors = [];
            for(let j = 0; j < this.m_unitids.length; j++){
                if(this.m_unitids[j] == ""){
                    convfactors.push(1.0);
                    continue;
                }
                let unititem = this.m_unitids[j][0];
                let selstr = SPGUI.setting[UnitLabel][unititem];
                let unitval = this.m_unitids[j][2][0];
                for(let i = 0; i < this.m_unitids[j][1].length; i++){
                    if(selstr == this.m_unitids[j][1][i]){
                        unitval = this.m_unitids[j][2][i];
                        break;
                    }
                }
                convfactors.push(unitval);
            }
            this.m_asciifmt.SetData(this.m_filereader.result, convfactors);
            this.m_jsonobj[this.m_label] = CopyJSON(this.m_asciifmt.GetObj());
            this.UpdatePlot(true);
            document.getElementById("importbutton").innerHTML = ReplaceLabel;
            UpdatePrms(this.m_currentId);
        });
    };

    Refresh(){
        if(this.m_plotwindow != null){
            this.m_plotwindow.RefreshPlotObject();
        }
    }

    GetPlotWindow(){
        return this.m_plotwindow;
    }

    SetJSONObj(jsonobj){
        this.m_jsonobj = jsonobj;
        this.m_newdata = true;
    };

    GetContentString(){
        let titles = this.m_jsonobj[this.m_label].titles;
        let data = this.m_jsonobj[this.m_label].data;
        if(titles.length != data.length){
            return "Data import failed.";
        }
        let obj = {};
        for(let i = 0; i < titles.length; i++){
            obj[titles[i]] = data[i];
        }
        if(this.m_extitle != "" && this.m_jsonobj[this.m_label].hasOwnProperty(this.m_extitle)){
            obj[this.m_extitle] = this.m_jsonobj[this.m_label][this.m_extitle];
        }
        let contstr = JSON.stringify(obj, null, JSONIndent);
        return contstr;
    }

    GetPlotArea(label)
    {
        let body = document.createElement("div");
        body.className = "flexcont flexcolumn flexgrow";

        let fdialog = CreateFileDialogElement(label, "Import", false, false);
        fdialog.file.id = AttachSuffix(this.m_currentId, SuffixFile);
        fdialog.button.id = AttachSuffix(this.m_currentId, SuffixButton);
        this.m_file = fdialog.file;
        
        fdialog.button.addEventListener("click", 
            () => {
                let idfile = AttachSuffix(this.m_currentId, SuffixFile);
                document.getElementById(idfile).click();
            });
        fdialog.file.addEventListener("change", (e) => {
            this.m_filereader.readAsText(e.target.files[0]);
            this.m_file.value = "";
            });

        fdialog.body.classList.add("none");
        body.appendChild(fdialog.body);
        body.id = this.m_currentId;

        let item = document.createElement("div");
        item.id = AttachSuffix(this.m_currentId, SuffixPlotDiv);
        item.className = "flexcont flexgrow";

        this.m_showinfdiv = document.createElement("div");
        this.m_showinfdiv.innerHTML = "No data available. Click to import.";
        this.m_showinfdiv.style.textAlign = "center";
        this.m_showinfdiv.style.fontSize = "x-large";
        this.m_showinfdiv.style.cursor = "pointer";
        this.m_showinfdiv.addEventListener("click", (e)=>{
            let btnid = AttachSuffix(this.m_currentId, SuffixButton);
            document.getElementById(btnid).click();
        });
        this.m_showinfdiv.className = "none";
        body.appendChild(this.m_showinfdiv);
        body.appendChild(item);
        return body;
    };

    UpdatePlot(force){
        if(!force && !this.m_newdata){
            return;
        }
        ShowImportCtrls();

        let plotdiv = document.getElementById(AttachSuffix(this.m_currentId, SuffixPlotDiv));
        if(CheckDataObj(this.m_jsonobj[this.m_label]) == false){
            this.m_showinfdiv.classList.replace("none", "flexgrow");
            plotdiv.classList.replace("flexgrow", "none");
            return;
        }
        else{
            this.m_showinfdiv.classList.replace("flexgrow", "none");
            plotdiv.classList.replace("none", "flexgrow");
        }

        let obj = this.GetContentString();
        document.getElementById("datacont").innerHTML = obj;
        
        let dim = this.m_asciifmt.GetDimension();
        let titles = this.m_asciifmt.GetTitles();
        let ordinate = this.m_asciifmt.GetOrdinate();
        let axtitles = [];
        if(dim == 0){
            axtitles.push("Index");
        }
        else if(dim == 1){
            axtitles.push(titles[0]);
            axtitles.push(ordinate);
        }
        else{
            for(let j = 0; j < 3; j++){
                axtitles.push(titles[j]);
            }
        }

        let plotid = AttachSuffix(this.m_currentId, SuffixPlotBody);
        let plobj = {
            data: [this.m_jsonobj[this.m_label]],
            dimension: dim,
            titles: titles,
            axtitles: axtitles,
            legprefix: [],
            isdata2d: true
        }
        plotdiv.innerHTML = "";
        this.m_plotwindow = new PlotWindow(plotdiv, 
            plotid, plobj, SPGUI.setting[POLabel][this.m_currentId], this.m_currentId);
        this.m_newdata = false;
    };
}

//----- utility functions -----
function ExportAsciiData(data)
{
    if(SPGUI.runEnvironment == "tauri"){
        let options = {
            title: "Select the file name to export the data."
        };
        if(SPGUI.setting.defpaths.hasOwnProperty("exportAscii")){
            options.defaultPath = SPGUI.setting.defpaths.exportAscii;
        }
        window.__TAURI__.dialog.save(options).then((datapath) => {
            if(datapath != null){
                SPGUI.setting.defpaths.exportAscii = datapath;
                window.__TAURI__.tauri.invoke("write_file", { path: datapath, data: data});
            }
        });
    }
    else{
        let blob = new Blob([data], {type:"application/text"});
        let link = document.createElement("a");
        link.href = window.URL.createObjectURL(blob);
        link.download = SPGUI.setting.prmfile;
        link.click();
        link.remove();       
    }    
}

function ArrangeObjectTblTitle(tbl, rows)
{
    let cell;
    let titles = [["GUI Notation", ""], 
        ["Key", ["Full", "Simplified"]], ["Format", ""], ["Default", ""]];

    rows.push(tbl.insertRow(-1));
    for(let j = 0; j < titles.length; j++){
        cell = rows[rows.length-1].insertCell(-1);
        if(typeof titles[j][0] == "string"){
            cell.innerHTML = titles[j][0];
        }
        cell.className += " title";
        if(titles[j][0] == "Format"){
            cell.setAttribute("width", "200px");
        }
        if(typeof titles[j][1] == "string" && titles[j][1] == ""){
            cell.setAttribute("rowspan", "2");
        }
        else if(Array.isArray(titles[j][1])){
            cell.setAttribute("colspan", titles[j][1].length.toString());
        }
    }
    rows.push(tbl.insertRow(-1));
    for(let j = 0; j < titles.length; j++){
        if(Array.isArray(titles[j][1])){
            for(let i = 0; i < titles[j][1].length; i++){
                cell = rows[rows.length-1].insertCell(-1);
                cell.innerHTML = titles[j][1][i];
                cell.className += " title";
            }
        }
        else if(titles[j][1] != ""){
            cell = rows[rows.length-1].insertCell(-1);
            cell.innerHTML = titles[j][1];
            cell.className += " title";
        }
    }
}

function CheckDataObj(obj)
{
    if(obj.hasOwnProperty(TilteLabel) == false){
        return false;
    }
    if(Array.isArray(obj[TilteLabel]) == false){
        return false;
    }
    if(obj.hasOwnProperty(DataLabel) == false){
        return false;
    }
    if(Array.isArray(obj[DataLabel]) == false){
        return false;
    }
    return true;
}

function ToPrmString(numvalr, prec) {
    if((typeof numvalr) == "undefined"){
        return "";
    }
    if(numvalr == 0){
        return numvalr;
    }

    let numexp = numvalr.toExponential().toLowerCase();
    let numchar = numexp.split('e');
    let signif = Math.floor(numchar[0]*1.0e+5+0.5)/1.0e+5;
    let numval = parseFloat(signif.toFixed(5)+'e'+numchar[1]);
    let absval = Math.abs(numval);

    let dec = Math.floor(Math.log10(absval));
    numexp = (numvalr < 0 ? -absval : absval).toExponential().toLowerCase();
    if(numexp.length < prec+6){
        if(Math.abs(dec) > 5){
            return numexp;
        }
        return numval;
    }

    let lval;
    if(dec <= -3 || dec > 3){
        lval = numval.toExponential(Math.max(prec, 3));
    }
    else{        
        lval = numval.toFixed(Math.max(prec, 3)-dec);
    }
    return lval;
}

function BundleFieldSet(label, comps, ids){
    let fldset = document.createElement("div");
    let legend = document.createElement("div");

    fldset.className = "flexcont flexcolumn";
    fldset.style.margin = "2px";
    fldset.style.border = "solid 1px";

    legend.style.backgroundColor = "#444";
    legend.style.color = "#eee";
    legend.style.padding = "2px";
    fldset.appendChild(legend);

    if(ids.hasOwnProperty("body")){
        fldset.id = ids.body;
    }
    if(ids.hasOwnProperty("title")){
        // append DOM later, so create div for text
        legend.id = ids.title;
        let legtext = document.createElement("div");     
        legtext.innerHTML = label;
        legtext.className = "flexgrow";
        legend.appendChild(legtext);
    }
    else{
        legend.innerHTML = label;
    }

    if(comps.length > 0){
        let fpanel = document.createElement("div");
        fpanel.className = "flexgrow flexcolumn";
        fpanel.style.margin = "5px";
        fldset.appendChild(fpanel);
    
        for(let i = 0; i < comps.length; i++){
            let item = document.createElement(comps[i].type);
            if(typeof comps[i].class != "undefined" && comps[i].class != ""){
                item.className = comps[i].class;
            }
            if(comps[i].id != ""){
                item.id = comps[i].id;
            }
            fpanel.appendChild(item);
        }    
    }
    return fldset;
}

function CreateFileDialogElement(title, btnlabel, ismulti, isjson)
{
    let cont = document.createElement("div");
    cont.className = "flexcont flexspace flexcenter";

    if(title.length > 0){
        let tlabel = document.createElement("div");
        tlabel.innerHTML = title;
        cont.appendChild(tlabel);    
    }

    let div = document.createElement("div");
    div.className = "flexcont";

    let fileel =  document.createElement("input");
    fileel.setAttribute("type", "file");
    if(ismulti){
        fileel.setAttribute("multiple", "multiple");
    }
    if(isjson){
        fileel.setAttribute("accept", "application/json");
    }
    fileel.className = "none";

    let btn = document.createElement("input");
    btn.setAttribute("type", "button");
    btn.setAttribute("value", btnlabel);

    div.appendChild(fileel);
    div.appendChild(btn);
    cont.appendChild(div);

    return {file:fileel, button:btn, body:cont};
}

function CreateSelectBox(title, lines, ismulti, iscolumn)
// used in main.js and postprocess.js
{
    let div = document.createElement("div");
    if(iscolumn){
        div.className = "flexcont flexcolumn";
    }
    else{
        div.className = "flexcont selcont";
    }

    let label = document.createElement("div");        
    label.innerHTML = title;
    if(!iscolumn){
        label.className = "seltitle";
    }

    let select = document.createElement("select");
    select.setAttribute("size", lines.toString());
    select.multiple = ismulti;

    div.appendChild(label);
    div.appendChild(select);

    return {select:select, body:div};
}

function GetSelections(select)
{
    let indices = [];
    let texts = [];
    for(let i = 0; i < select.length; i++){
        if(select.options[i].selected){
            indices.push(i);
            texts.push(select.options[i].text);
        }
    }
    return {index:indices, text:texts};
}

function SetSelection(select, label, isvalue)
{
    let curridx = -1;
    for(let i = 0; i < select.length; i++){
        if(select.options[i].selected){
            curridx = i;
        }
        select.options[i].selected = false; 
    }
    select.selectedIndex = -1;
    for(let i = 0; i < select.length; i++){
        if(label == (isvalue ? 
                select.options[i].value : select.options[i].text)){
            select.options[i].selected = true;
            select.selectedIndex = i;
            break;
        }
    }
    if(select.selectedIndex < 0 && curridx >= 0){
        select.options[curridx].selected = true;
        select.selectedIndex = curridx;
    }
}

function GetFileInputText(inputid, val, cellcont, label)
{
    cellcont.className = "flexcont flexcolumn";               
    let fdialog = CreateFileDialogElement(label, "Browse", false, false);
    fdialog.button.id = AttachSuffix(inputid, SuffixButton);
    fdialog.file.id = AttachSuffix(inputid, SuffixFile);
    cellcont.appendChild(fdialog.body);

    let item = document.createElement("textarea");
    item.className = "filearea";   
    item.value = val;
    item.id = inputid;
    return {dialog:fdialog, item:item};
}

function CopyJSON(obj)
{
    if(typeof obj == "undefined"){
        return "";
    }
    return JSON.parse(JSON.stringify(obj));
}

function FormatArray(strobj)
{
    let ist, iend, bef, mid, newmid, aft, inext;
    
    iend = 0;
    do{
        ist = strobj.indexOf("[", iend);
        inext = strobj.indexOf("[", ist+1);
        iend = strobj.indexOf("]", ist);

        if(inext >= 0){
            if(inext < iend){
                iend = inext;
                continue;
            }
        }

        if(ist == -1 || iend == -1){
            break;
        }   
        bef = strobj.slice(0, ist+1);
        mid = strobj.slice(ist+1, iend);
        aft = strobj.slice(iend);    
        newmid = mid.replace(/\n\s*/g, "");
        iend -= mid.length-newmid.length;
        strobj = bef+newmid+aft;
    } while(ist != -1 && iend != -1);
    return strobj;
}

function CreateModalInfDialog(dlgid, title, conttext, footers)
{
    document.getElementById(dlgid).innerHTML = "";

    let maindiv = document.createElement("div");
    maindiv.className = "flexcont flexgrow flexcolumn";
    maindiv.style.whiteSpace = "normal";

    document.getElementById(dlgid).appendChild(maindiv);

    let divs = [];

    divs.push(document.createElement("h2"));
    divs[0].style.padding = "0px";
    divs[0].style.margin = "5px";
    divs[0].innerHTML = title;

    divs.push(document.createElement("p"));
    divs[1].innerHTML = conttext;

    for(let j = 0; j < footers.length; j++){
        if(footers[j] == "hr"){
            divs.push(document.createElement("hr"));
        }
        else{
            divs.push(document.createElement("div"));
            divs[j+2].innerHTML = footers[j];    
        }
    }

    for(let j = 0; j < divs.length; j++){
        maindiv.appendChild(divs[j]);
    }
  
    let okbtn = document.createElement("button");
    okbtn.innerHTML = "OK";
    let btndiv = document.createElement("div");
    btndiv.className = "flexcont flexright";
    btndiv.appendChild(okbtn);
    maindiv.appendChild(btndiv);
    okbtn.addEventListener("click", (e) => {
        document.getElementById(dlgid).close();
        e.stopPropagation();
    });
}

function CreateModalDialog(prmtbls, dlgid, divid, caption, titleid, func, iscancel)
// used in main.js
{
    let magdiv = document.createElement("div");
    document.getElementById(dlgid).appendChild(magdiv);
    magdiv.className = "flexcont flexcolumn";
    magdiv.style.alignItems = "flex-end";
    
    let comps = [{type:"div", id:divid, class:"flexcont flexcolumn"}];
    let fldset = BundleFieldSet(caption, comps, {});
    magdiv.appendChild(fldset);
    
    let okbtn = document.createElement("button");
    okbtn.innerHTML = "OK";
    let btndiv;
    if(iscancel){
        btndiv = document.createElement("div");
        let cancelbtn = document.createElement("button");
        cancelbtn.innerHTML = CancelLabel;
        btndiv.appendChild(cancelbtn);
        cancelbtn.addEventListener("click", (e) => {
            document.getElementById(dlgid).close();
            e.stopPropagation();
        });    
        btndiv.appendChild(okbtn);
        magdiv.appendChild(btndiv);
    }
    else{
        magdiv.appendChild(okbtn);
    }
    okbtn.addEventListener("click", (e) => {
        document.getElementById(dlgid).close();
        if(func != null){
            func();
        }
        e.stopPropagation();
    });

    if(titleid != ""){
        let titlediv = document.createElement("div");    
        titlediv.id = titleid;
        titlediv.style.margin = "5px 0px";
        document.getElementById(divid).appendChild(titlediv);    
    }
    for(let j = 0; j < prmtbls.length; j++){
        document.getElementById(divid).appendChild(prmtbls[j].GetTable());
    }
}

function SetSelectMenus(select, menuobj, disables)
// used in main.js
{
    let ncounts = 0;
    for(let j = 0; j < menuobj.length; j++){
        let optgroup = document.createElement("optgroup");
        let grpname = Object.keys(menuobj[j]);
        optgroup.label = grpname[0];
        for(let i = 0; i < menuobj[j][grpname[0]].length; i++){
            let option = document.createElement("option");
            option.innerHTML = menuobj[j][grpname[0]][i];
            optgroup.appendChild(option);
            ncounts++;
            for(let k = 0; k < disables.length; k++){
                if(disables[k] == menuobj[j][grpname[0]][i]){
                    option.setAttribute("disabled", true);
                }
            }
        }
        select.appendChild(optgroup);
    }
    select.selectedIndex = -1;
    return ncounts;
}

function AttachSuffix(id, suffix)
{
    return id+suffix;
}

function DetachSuffix(id, suffix)
{
    let idx = id.lastIndexOf(suffix);
    let ids = id.slice(0, idx);
    return ids;
}

function GetIDFromItem(categ, item, jxy)
{
    let id;
    if(jxy >= 0){
        id = categ+IDSeparator+item+IDSeparator+jxy;
    }
    else{
        id = categ+IDSeparator+item;
    }
    return id;
}

function GetItemFromID(id)
{
    let categ = "", item = "", subitem = "", jxy = -1;
    if((typeof id) == "string"){
        let labels = id.split(IDSeparator);
        if(labels.length > 2){
            let xy = parseInt(labels[labels.length-1]);
            if(isNaN(xy)== false){
                jxy = xy;
            }
        }
        if(labels.length >= 2){           
            categ = labels[0];
            item = labels[1];
            if(labels.length > 3 || 
                (labels.length == 3 && jxy < 0)){
                subitem = labels[2];
            }
        }
    }
    return {categ:categ, item:item, subitem:subitem, jxy:jxy};
}

function GetIdFromCell(parent, i, j)
{
    return parent+"/"+i+"/"+j;
}

function GetCellFromId(parent, id)
{
    let cell = id.split("/");
    return [parseInt(cell[1]), parseInt(cell[2])];
}

function ShowImportCtrls()
{
    let seldata = document.getElementById("dataselect");
    let datacont = document.getElementById("datacont");
    let selections = GetSelections(seldata).index;
    let index = -1;
    if(selections.length > 0){
        index = selections[0];
    }

    if(index >= DefaultPPItems){
        datacont.classList.remove("none");
    }
    else{
        if(!datacont.classList.contains("none")){
            datacont.classList.add("none");
        }
    }
}

function ShowDataImport(id)
{
    SPGUI.mainnotebook.ShowPage(1);
    let select = document.getElementById("dataselect");
    let plotid = AttachSuffix(id, SuffixPlotBody);
    SetSelection(select, plotid, true);
    ShowImportPlot(id);
}

function ShowImportPlot(id)
{
    ClearDataPlot();
    SPGUI.importplotarea.appendChild(SPGUI.importdata[id].GetPlotArea(""));
    SPGUI.importdata[id].UpdatePlot(true);
}

function ClearDataPlot()
{
    while(SPGUI.importplotarea.childNodes.length > 0){
        SPGUI.importplotarea.removeChild(SPGUI.importplotarea.firstChild);
    }
    SPGUI.preproc = null;
    document.getElementById("datacont").innerHTML = "";
}

function Alert(msg)
{
    // message modal dialog
    CreateModalInfDialog("msgDialog", "WARNING", msg, []);
    document.getElementById("msgDialog").showModal();
}

function PopContextMenu(event, id, type)
{
    let ppmenu = document.getElementById(id);
    ppmenu.style.zIndex = 2;
    ppmenu.style.top = event.pageY+"px";
    ppmenu.style.left = event.pageX+"px";
    ppmenu.style.visibility = "visible";
}

//-------

var AnimationInterval = 100;

class PlotWindow {
    constructor(parent, id, plobj, setting, plotname, frameindices = [], tmval = -1){
        this.m_parent = parent;
        this.m_isnew = true;
        this.m_ploptions = setting;
        this.m_plotname = plotname;
        this.m_layout = {};

        if(tmval >= 0){
            AnimationInterval = tmval;
        }

        let plotdiv = document.createElement("div");
        this.m_parent.appendChild(plotdiv);
        plotdiv.id = id;
        plotdiv.className = "flexgrow";
        plotdiv.style.padding = "10px";
        plotdiv.addEventListener("editplotly", (e)=>{    
            this.EditPlot();
        });      

        this.m_data = plobj.data;
        this.m_dimension = plobj.dimension;
        this.m_id = id;
        this.m_isanim = frameindices.length > 0;
        this.m_axtitles = plobj.axtitles;
        let config = this.GetConfig();

        if(this.m_isanim){
            this.m_detvars = plobj.detvars;
            this.m_animlegend = plobj.animlegend;
            this.m_compdata = plobj.compdata;
            this.m_vardata = plobj.vardata;
            this.m_ftitles = plobj.ftitles;
            this.m_fdata = plobj.fdata;

            // slider setting
            this.m_sliderdiv = document.createElement("div");
            this.m_parent.appendChild(this.m_sliderdiv);
            this.m_sliderdiv.className = "gridrow";

            this.m_animdiv = document.createElement("div");
            this.m_animdiv.className = "flexcont flexright";
            let intrvtitle = document.createElement("div");
            intrvtitle.innerHTML = "Interval (ms)";
            intrvtitle.style.margin = "0px 5px";
            this.m_animintrv = document.createElement("input");
            this.m_animintrv.setAttribute("type", "number");
            this.m_animintrv.min = 100;
            this.m_animintrv.max = 2000;
            this.m_animintrv.step = 100;
            this.m_animintrv.style.width = "50px";
            this.m_animintrv.value = AnimationInterval;
            this.m_animintrv.addEventListener("change", (e) => 
                {
                    AnimationInterval = e.currentTarget.value;
                }
            )
   
            this.m_animbtn = document.createElement("button");
            this.m_animbtn.innerHTML = "&#9654;"
            this.m_animbtn.className = "animbtn";
            this.m_animbtn.addEventListener("click", (e)=>{
                if(this.onanimation){
                    this.StopAnimation();
                }
                else{
                    let slides = this.GetSlideNumber();
                    if(slides[0] == slides[1]){
                        // return to 0-th slide
                        this.SetSlideNumber(0);
                    }
                    this.StartAnimation();
                }
                this.SwitchAnimationButton(this.onanimation);
            });
            this.m_animdiv.appendChild(intrvtitle);
            this.m_animdiv.appendChild(this.m_animintrv);
            this.m_animdiv.appendChild(this.m_animbtn);
            this.m_sliderdiv.appendChild(this.m_animdiv);

            this.m_slider = [];
            this.m_sliderlabel = [];
            for(let i = 0; i < 2; i++){
                let divcont = document.createElement("div");
                divcont.className = "flexcont flexgrow";

                this.m_slider.push(document.createElement("input"));
                this.m_slider[i].type = "range";
                this.m_slider[i].max = this.m_fdata[i].length-1;
                this.m_slider[i].value = frameindices[i];
                this.m_slider[i].addEventListener("change",
                    (e) => {
                        this.ShowSlide();                
                    });
                divcont.appendChild(this.m_slider[i]);

                this.m_sliderlabel.push(document.createElement("div"));
                this.m_sliderlabel[i].innerHTML = "";
                divcont.appendChild(this.m_sliderlabel[i]);

                if(this.m_ftitles[i] == ""){
                    divcont.className += " none";
                }
                this.m_sliderdiv.appendChild(divcont);
            }
            if(this.m_ftitles[1] != ""){
                this.m_animdiv.classList.replace("flexcont", "none");
            }
            else{
                this.m_slider[1].value = 0;
            }
            this.SetRangesAnim();            
            if(this.m_dimension == 1){
                let range = this.GetRanges(config);
                this.m_layout = GetLayout1D(this.m_axtitles, range.linear, range.log, config.xyscale);
            }
            else{
                this.m_ztitle = plobj.ztitle;
                this.m_layout = {
                    "surface": GetLayout2D(1, false, this.m_axtitles, [this.m_axtitles[2]], null, this.m_ranges[2]),
                    "heatmap": GetLayout2D(1, true, this.m_axtitles, [this.m_axtitles[2]], null, this.m_ranges[2]),
                    "ztitle" : this.m_ztitle
                };
            }
            this.ShowSlide(true);
        }
        else{
            this.m_titles = plobj.titles;
            this.m_legprefix = plobj.legprefix;
            this.m_isdata2d = plobj.isdata2d;
            this.PlotObject(id, this.m_data, this.m_dimension, 
                this.m_titles, this.m_axtitles, this.m_legprefix, this.m_isdata2d, config.xyscale);
            this.SetupRelayout(id);
        }
    }

    GetClientSize(){
        return [this.m_parent.clientWidth, this.m_parent.clientHeight];
    }

    GetData(){
        let plobj = {};
        plobj.data = this.m_data;
        plobj.dimension = this.m_dimension;
        plobj.axtitles = this.m_axtitles;
        plobj.plotname  = this.m_plotname;
        plobj.size = this.GetClientSize()

        plobj.frameindices = [];
        plobj.tinv = 100;
        plobj.setting = CopyJSON(this.m_ploptions);
        if(this.m_isanim){
            plobj.detvars = this.m_detvars;
            plobj.animlegend = this.m_animlegend;
            plobj.compdata = this.m_compdata;
            plobj.vardata = this.m_vardata;
            plobj.ftitles = this.m_ftitles;
            plobj.fdata = this.m_fdata;
            plobj.frameindices = [this.m_slider[0].value, this.m_slider[1].value];
            plobj.tinv  = this.m_animintrv.value;
            if(this.m_dimension == 2){
                plobj.ztitle = this.m_ztitle;
            }
        }
        else{
            plobj.titles = this.m_titles;
            plobj.legprefix = this.m_legprefix;
            plobj.isdata2d = this.m_isdata2d;
        }
        return CopyJSON(plobj);
    }

    GetRanges(config){
        let ranges = null;
        let logranges = this.m_logranges;
        if(this.m_isanim){
            if(config.normeach){
                let index = [];
                for(let j = 0; j < 2; j++){
                    index.push(parseInt(this.m_slider[j].value));
                }
                logranges = this.m_lograneach[index[1]][index[0]];
            }
            else{
                ranges = this.m_ranges;
            }
        }
        return {linear:ranges, log:logranges};
    }

    PlotObject(id, obj, dim, titles, axtitles, legprefix, isdata2d, xyscale)
    // isdata2d: input data is 2-dimensional (true: data[x][y]) or 1-dim. (false: data[x+Ny])
    {
        let plotobj, layout;
        let config = this.GetConfig();
        if(dim == 0 || dim == 1){
            if(dim == 1){
                plotobj = [];
                for(let i = 0; i < obj.length; i++){
                    let legends = [];
                    if(i < legprefix.length){
                        if(titles.length > 2){
                            for(let j = 1; j < titles.length; j++){
                                legends.push(titles[j]+":"+legprefix[i]);
                            }
                        }
                        else{
                            legends.push(legprefix[i]);
                        }    
                    }
                    let plotobjr = GetPlotConfig1D(obj[i], titles[0], titles.slice(1), legends, config.mode);
                    for(let j = 0; j < plotobjr.length; j++){
                        plotobj.push(plotobjr[j]);
                    }
                }
            }
            else{
                plotobj = GetPlotConfig0D(obj[0], titles, config.mode);
                axtitles.push(titles[0]);
            }
            this.SetRanges(plotobj);
            this.m_layout = GetLayout1D(axtitles, null, this.m_logranges, xyscale);
            layout = this.m_layout;
        }
        else{
            let pobj = GetPlotConfig2D(obj, titles, axtitles, config, isdata2d); 
            plotobj = pobj.data;
            this.m_layout = pobj.layout;
            layout = this.m_layout[config.type2d];
            this.SetRanges(plotobj);
        }
        Plotly.newPlot(id, plotobj, layout, PlotlyPrms.config);    
    }

    RefreshPlotObject(changes = [])
    {
        let data = document.getElementById(this.m_id).data;
        let layoutnew, layout;
        let config = this.GetConfig();
        if(changes.length > 0){
            if(this.m_dimension == 1){
                let range = this.GetRanges(config);
                layoutnew = GetLayout1D(this.m_axtitles, range.linear, range.log, config.xyscale);
                for(let n = 0; n < data.length; n++){
                    data[n].mode = config.mode;
                }
            }
            else{
                for(let n = 0; n < data.length; n++){
                    data[n].type = config.type2d;
                    data[n].colorscale = config.colorscale;
                }
                layoutnew = GetLayout2D(data.length, config.type2d == "heatmap", 
                    this.m_axtitles, [this.m_axtitles[2]], null, this.m_ranges[2]);
            }            
        }
        for(let j = 0; j < changes.length; j++){
            if(changes[j] == PlotOptionsLabel.xscale[0]){
                this.m_layout.xaxis = CopyJSON(layoutnew.xaxis);
            }
            if(changes[j] == PlotOptionsLabel.yscale[0]){
                this.m_layout.yaxis = CopyJSON(layoutnew.yaxis);
            }
            if(changes[j] == PlotOptionsLabel.normalize[0]){
                if(this.m_dimension == 1){
                    this.m_layout.yaxis = CopyJSON(layoutnew.yaxis);
                }
                else{
                    if(config.type2d == "surface"){
                        this.m_layout["surface"].scene.zaxis = CopyJSON(layoutnew.scene.zaxis);
                    }
                    this.m_layout["surface"].zaxis = CopyJSON(layoutnew.zaxis);
                    this.m_layout["heatmap"].zaxis = CopyJSON(layoutnew.zaxis);
                }
            }
        }
        if(this.m_dimension == 1){
            layout = this.m_layout;
        }
        else{
            layout = this.m_layout[config.type2d];
        }
        Plotly.purge(this.m_id);
        Plotly.newPlot(this.m_id, data, layout, PlotlyPrms.config);
        this.SetupRelayout(this.m_id);
        if(this.m_isanim && this.m_dimension == 2){
            this.ShowSlide();
        }
    }
    
    Plot1DAmination(id, layout, nplots, vararr, sliceobj, legend, compdata, frameno)
    {
        let config = this.GetConfig();
        let data = [];
        if(nplots == 1){
            data.push(
                {
                    x: vararr,
                    y: sliceobj[frameno[1]][frameno[0]][0],
                    type: PlotlyScatterType,
                    name: legend,
                    mode: config.mode
                }
            );    
        }
        else{
            for(let j = 0; j < nplots; j++){
                data.push(
                    {
                        x: vararr[j],
                        y: sliceobj[frameno[1]][frameno[0]][j],
                        type: PlotlyScatterType,
                        name: legend[j],
                        mode: config.mode
                    }
                );        
            }
        }
        for(let i = 0; i < compdata.length; i++){
            compdata[i].mode = config.mode;
            data.push(compdata[i]);
        }
    
        if(this.m_isnew){
            Plotly.newPlot(id, data, layout, PlotlyPrms.config);
            this.m_isnew = false;
        }
        else{
            Plotly.react(id, data, layout);
        }
    }        

    Plot2DAnimation(id, layout, vararr, sliceobj, frameno, config)
    {
        let type = config.type2d;
        let normeach = config.normeach;
        let data = [];
        let clbar = CopyJSON(PlotlyPrms.colorbar);
        clbar.title = layout.ztitle;
        data.push(
            {
                x: vararr[0],
                y: vararr[1],
                z: sliceobj[frameno[1]][frameno[0]],
                type: type,
                zsmooth: "best",
                scene:"scene",
                showscale: true,
                zmin: this.m_ranges[2][0],
                zmax: this.m_ranges[2][1],
                colorbar: clbar,
                contours: CopyJSON(PlotlyPrms.heatmap),      
                colorscale: config.colorscale
            }
        );

        let layoutloc = layout[type];
        if(normeach){
            delete data[0].zmax;
            delete data[0].zmin;
            if(type == "surface"){
                delete layoutloc.scene.zaxis.range;
            }
        }

        if(this.m_isnew){
            Plotly.newPlot(id, data, layoutloc, PlotlyPrms.config);
            this.m_isnew = false;
        }
        else{
            if(type == "surface"){
                let plobj = document.getElementById(id);
                let camera = plobj._fullLayout.scene._scene.getCamera();
                let pllayout = {margin:layoutloc.margin, scene:layoutloc.scene};
                pllayout.scene.camera = camera;
                Plotly.react(id, data, pllayout);
            }
            else{
                Plotly.react(id, data, layoutloc);    
            }
        }
    }   
    
    SetRangesAnim(){
        this.m_ranges = [[], [], []];
        this.m_logranges = [[], [], []];
        let mesh = [
            this.m_data.length,
            this.m_data[0].length,
            this.m_data[0][0].length,
            this.m_data[0][0][0].length
        ];
        this.m_lograneach = new Array(mesh[0]);
        let ranges;
        for(let i = 0; i < mesh[0]; i++){
            this.m_lograneach[i] = new Array(mesh[1]);
            for(let j = 0; j < mesh[1]; j++){               
                let rangeeach;
                for(let k = 0; k < mesh[2]; k++){
                    for(let l = 0; l < mesh[3]; l++){
                        let f = this.m_data[i][j][k][l];
                        if(i+j+k+l == 0){
                            ranges = [f, f];
                        }
                        else{
                            ranges[0] = Math.min(ranges[0], f);
                            ranges[1] = Math.max(ranges[1], f);
                        }
                        if(k+l == 0){
                            rangeeach = [f, f];
                        }    
                        else{
                            rangeeach[0] = Math.min(rangeeach[0], f);
                            rangeeach[1] = Math.max(rangeeach[1], f);
                        }
                    }                        
                }
                this.m_lograneach[i][j] = rangeeach;
            }
        }

        if(this.m_detvars.length > 0){
            this.m_ranges[0] = [Math.min(...this.m_detvars[0]), Math.max(...this.m_detvars[0])];
            for(let n = 1; n < this.m_detvars.length; n++){
                this.m_ranges[0][0] = Math.min(this.m_ranges[0][0], Math.min(...this.m_detvars[n]));
                this.m_ranges[0][1] = Math.max(this.m_ranges[0][1], Math.max(...this.m_detvars[n]));
            }
        }
        else{
            for(let j = 0; j < this.m_dimension; j++){
                this.m_ranges[j] = [Math.min(...this.m_vardata[j]), Math.max(...this.m_vardata[j])];
            }
        }
        this.m_ranges[this.m_dimension] = ranges;
        for(let j = 0; j <= this.m_dimension; j++){
            this.m_logranges[j] = GetLogRange(this.m_ranges[j][0], this.m_ranges[j][1], LOGOFFSET);
        }
        let fullranges = ranges[1]-ranges[0];
        this.m_ranges[this.m_dimension][0] -= fullranges*(LOGOFFSET-1)/2;
        this.m_ranges[this.m_dimension][1] += fullranges*(LOGOFFSET-1)/2;
    }

    SetRanges(plotobj){
        this.m_ranges = [];
        this.m_logranges = [];
        for(let j = 0; j <= this.m_dimension; j++){
            let range;
            for(let n = 0; n < plotobj.length; n++){
                let data;
                if(j == 0){
                    data = plotobj[n].x;
                }
                else if(j == 1){
                    data = plotobj[n].y;
                }
                else{
                    data = plotobj[n].z;
                }
                if(n == 0){
                    range = [data.reduce((a, b)=>Math.min(a, b)), data.reduce((a, b)=>Math.max(a, b))];
                }
                else{
                    range[0] = Math.min(range[0], data.reduce((a, b)=>Math.min(a, b)));
                    range[1] = Math.max(range[1], data.reduce((a, b)=>Math.max(a, b)));
                }
            }
            this.m_ranges.push(range);
            this.m_logranges.push(GetLogRange(this.m_ranges[j][0], this.m_ranges[j][1], LOGOFFSET));
        }
    }
    
    EditPlot(){
        let pltypes = {};
        PlotOptionsOrder.forEach(function(element){
            pltypes[PlotOptionsLabel[element][0]] = true;
        });
        pltypes[PlotOptionsLabel.type2d[0]] = this.m_dimension == 2;
        pltypes[PlotOptionsLabel.colorscale[0]] = this.m_dimension == 2;
        pltypes[PlotOptionsLabel.xscale[0]] = this.m_dimension == 1;
        pltypes[PlotOptionsLabel.yscale[0]] = this.m_dimension == 1;
        pltypes[PlotOptionsLabel.type[0]] = this.m_dimension == 1;
        pltypes[PlotOptionsLabel.normalize[0]] = this.m_isanim;

        let ploption = new PlotOption();
        this.m_ploptionsold =  CopyJSON(this.m_ploptions);
        let plobj = {};
        plobj[OptionLabel] = this.m_ploptionsold;
        ploption.JSONObj = plobj;
        ploption.SetValidity(pltypes, {}, {});
    
        document.getElementById("plconfDialog").innerHTML = "";
        CreateModalDialog([ploption], "plconfDialog", "plotconfig", 
            "Edit Plot Configurations", "", this.ChangeOption.bind(this), true);
        document.getElementById("plconfDialog").showModal();
    }

    ChangeOption(){
        let changes = [];
        PlotOptionsOrder.forEach((e) => {
            if(this.m_ploptionsold[PlotOptionsLabel[e][0]] != 
                this.m_ploptions[PlotOptionsLabel[e][0]]){
                    changes.push(PlotOptionsLabel[e][0]);
            }            
        });
        if(changes.length == 0){
            return;
        }
        Object.keys(this.m_ploptionsold).forEach(el=>{
            this.m_ploptions[el] = this.m_ploptionsold[el];
        });
        this.RefreshPlotObject(changes);
    }
    
    GetSlideNumber(){
        return [parseInt(this.m_slider[0].value), parseInt(this.m_slider[0].max)];
    }

    SetSlideNumber(slide){
        this.m_slider[0].value = slide.toString();
    }   

    AdvanceSlide()
    {   
        let slides = this.GetSlideNumber();
        slides[0]++;
        if(slides[0] > slides[1]){
            this.StopAnimation();
            this.SwitchAnimationButton(false);
            return;
        }
        this.SetSlideNumber(slides[0]);
        this.ShowSlide();
    }
    
    StopAnimation()
    {
        clearInterval(this.animtimer);
        this.onanimation = false;
    }
   
    StartAnimation()
    {
        this.animtimer = setInterval(this.AdvanceSlide.bind(this), this.m_animintrv.value);
        this.onanimation = true;
    }

    GetSliceValue(){
        let slices = [];        
        for(let i = 0; i < 2; i++){
            slices.push(parseInt(this.m_slider[i].value));
            this.m_sliderlabel[i].innerHTML = this.m_ftitles[i]+this.m_fdata[i][slices[i]];    
        }
        return slices;
    }

    GetConfig(){
        let config = {};
        config.xyscale = [
            this.m_ploptions[PlotOptionsLabel.xscale[0]] == LinearLabel ? "linear" : "log",
            this.m_ploptions[PlotOptionsLabel.yscale[0]] == LinearLabel ? "linear" : "log"
        ];

        switch(this.m_ploptions[PlotOptionsLabel.type[0]]){
            case LineLabel:
                config.mode = "lines";
                break;
            case LineSymbolLabel:
                config.mode = "lines+markers";
                break;
            case SymbolLabel:
                config.mode = "markers";
                break;                    
        }
        config.type2d = 
            this.m_ploptions[PlotOptionsLabel.type2d[0]] == SurfaceLabel ? "surface" : "heatmap";
        config.normeach = this.m_ploptions[PlotOptionsLabel.normalize[0]] == ForEachLabel;
        if(this.m_ploptions[PlotOptionsLabel.colorscale[0]] == DefaultLabel){
            config.colorscale = PlotlyPrms.clscale;
        }
        else{
            config.colorscale = 
                this.m_ploptions[PlotOptionsLabel.colorscale[0]];
        }
        return config;        
    }
    
    ShowSlide(init = false){
        let slices = this.GetSliceValue();
        let config = this.GetConfig();        
        if(this.m_dimension == 1){
            let range = this.GetRanges(config);
            if(init){
                this.m_layout = GetLayout1D(this.m_axtitles, range.linear, range.log, config.xyscale);
            }
            if(this.m_detvars.length > 0){
                this.Plot1DAmination(this.m_id, this.m_layout, this.m_detvars.length,
                    this.m_detvars, this.m_data, this.m_animlegend, this.m_compdata, slices);
            }
            else{
                this.Plot1DAmination(this.m_id, this.m_layout, 1, 
                    this.m_vardata[0], this.m_data, this.m_animlegend, this.m_compdata, slices);
            }        
        }
        else{
            this.Plot2DAnimation(this.m_id, 
                this.m_layout, this.m_vardata, this.m_data, slices, config);
        }
    }

    SwitchAnimationButton(ison){
        this.m_animbtn.innerHTML = ison ? "&#9646;" : "&#9654;";
        if(ison){
            this.m_animintrv.setAttribute("disabled", "disabled");
        }
        else{
            this.m_animintrv.removeAttribute("disabled");
        }
    }

    SetupRelayout(id){
        let plyobj = document.getElementById(id);
        plyobj.on("plotly_relayout", (e) => {
            if(this.m_onrelayout){
                this.m_onrelayout = false;
                return;
            }
            let plobj = document.getElementById(id);
            let nplots = plobj.data.length;
            if(nplots == 1){
                return;
            }
            if(this.m_dimension == 2){
                let keys = Object.keys(e);
                let layout = {};
                if(keys[0].indexOf("dragmode") >= 0){
                    let dmode = e["scene.dragmode"];
                    for(let j = 0; j < keys.length; j++){
                        layout[keys[j]] = dmode;
                    }
                }
                else if(keys[0].indexOf("range[0]") >= 0){
                    let vranges = ["", "", "", ""];
                    let xylabels = ["xaxis", "xaxis", "yaxis", "yaxis"];
                    let rlabels = [".range[0]", ".range[1]", ".range[0]", ".range[1]"];
                    let ixy, iif;
                    for(let j = 0; j < keys.length; j++){
                        if(keys[j].indexOf("xaxis") >= 0){
                            ixy = 0;
                        }
                        else if(keys[j].indexOf("yaxis") >= 0){
                            ixy = 1;
                        }
                        else{
                            continue;
                        }
                        if(keys[j].indexOf("[0]") >= 0){
                            iif = 0;
                        }
                        else if(keys[j].indexOf("[1]") >= 0){
                            iif = 1;
                        }
                        else{
                            continue;
                        }
                        vranges[2*ixy+iif] = e[keys[j]];
                    }
                    for(let j = 0; j < nplots; j++){
                        let numstr = j == 0 ? "" : (j+1).toString();
                        for(let i = 0; i < 4; i++){
                            if(typeof vranges[i] != "number"){
                                continue;
                            }
                            layout[xylabels[i]+numstr+rlabels[i]] = vranges[i];
                        }
                    }
                }
                else if(keys[0].indexOf("camera") >= 0){
                    let camobj = e[keys[0]];
                    camobj = {camera: CopyJSON(camobj)};

                    let axtitles = ["xaxis", "yaxis", "zaxis"];
                    for(let n = 0; n < nplots; n++){
                        let label = "scene";                        
                        if(n > 0){
                            label += (n+1).toString()
                        }
                        let xyzobjs = [];
                        for(let j = 0; j < axtitles.length; j++){
                            xyzobjs.push(plobj.layout[label][axtitles[j]]);
                       }
                        layout[label] = CopyJSON(camobj);
                        for(let i = 0; i < xyzobjs.length; i++){
                            layout[label][axtitles[i]] = xyzobjs[i];
                        }
                    }
                }
                if(Object.keys(layout).length == 0){
                    return;
                }
                let config = this.GetConfig();
                this.m_layout[config.type2d] = CopyJSON(layout);
                this.m_onrelayout = true;
                Plotly.relayout(plobj, this.m_layout[config.type2d]);
            }
        });
    }

    ExportAsASCII(){
        let plobj = document.getElementById(this.m_id);
        let ztitles;
        if(this.m_dimension == 2){
            ztitles = this.m_axtitles.slice(2);
        }
        let data = GetAsciiData(this.m_dimension, plobj, ztitles);
        ExportAsciiData(data);
    }
}