"use strict";

function CheckCalcID()
{
    let calcid = GetCurrentCalcID();
    if(calcid != ""){
        let accobj = GetObj(AccLabel);
        let srcobj = GetObj(SrcLabel);
        SetDisableCalcIDs(accobj, srcobj);
        for(let j = 0; j < SPGUI.disabledcalcid.length; j++){
            if(calcid.indexOf(SPGUI.disabledcalcid[j]) >= 0){
                GetObj(ConfigLabel)[TypeLabel] = "";
                CloseCalculation();
                break;
            } 
        }
    }
}

async function PreProcess(menuitem)
{
    switch(menuitem){
        case PPBetaLabel:
        case PPFDlabel:
        case PP1stIntLabel:
        case PP2ndIntLabel:
        case PPPhaseErrLabel:
        case PPRedFlux:
        case PPTransLabel:
        case PPAbsLabel:
            break;
        default:
            return Promise.resolve(false);
    }

    let currspobj = GetCalcObj(null);
    let id = GetIDFromItem(PreprocIDLabel, SPGUI.totalpreprocs);
    SPGUI.totalpreprocs++;       

    currspobj.runid = menuitem;
    if(menuitem == PPPhaseErrLabel || menuitem == PPRedFlux){
        currspobj[PPPhaseErrLabel] = SPGUI.setting[PhaseErrLabel];
    }
    else if(menuitem == PPTransLabel || menuitem == PPAbsLabel){
        currspobj[PlotFilterLabel] = SPGUI.setting[PlotFilterLabel];
    }
    let prms = FormatArray(JSON.stringify(currspobj, null, JSONIndent));

    delete currspobj.runid;
    preprocess_result = "";
    
    if(SPGUI.runEnvironment == "node"){
        SPGUI.socketio.on(id, (msg) => {
           PlotPreprocess(msg, menuitem);
        });
        SPGUI.socketio.emit("preproc", prms, id);
    }
    else if(SPGUI.runEnvironment == "browser"){
        let blob = new Blob([prms], {type:"application/json"});
        let link = document.createElement("a");
        link.href = window.URL.createObjectURL(blob);
        link.download = SPGUI.setting.prmfile;
        link.click();
        link.remove();
    }
    else{
        let tmpname = await GetTemporaryFileName();
        let result = await window.__TAURI__.tauri.invoke("write_file", { path: tmpname, data: prms});
        if (result != "") {
            Alert(result);
            return Promise.resolve(true);
        }
        const command = new window.__TAURI__.shell.Command("solver_nompi", ["-f", tmpname]);
        command.on("close", (data) => {
            PlotPreprocess(preprocess_result, menuitem);
        });
        command.stdout.on("data", (data) => {
            preprocess_result += data.toString();
        });   
        command.spawn();
    }
    return Promise.resolve(true);
}

function PlotPreprocess(result, target)
{
    if(result.indexOf(ErrorLabel) >= 0){
        Alert(result);
        return;
    }

    ClearDataPlot();
    let obj;
    try {
        obj = JSON.parse(result);
    }
    catch (err) {
        Alert("Error: Cannot get the pre-processing result.\n"+err.message);
        return;
    }
    let titles = obj.titles;
    let units = obj.units;
    let dimension = obj.dimension;
    if(titles.length != units.length){
        Alert("Preprocessed Data Invalid Format.");
        return;
    }
    let axtitles = [];
    for(let j = 0; j < dimension; j++){
        if(units[j] == ""){
            axtitles.push(titles[j]);
        }
        else{
            axtitles.push(titles[j]+" ("+units[j]+")");
        }
    }
    if(obj.hasOwnProperty("ordinate")){
        axtitles.push(obj.ordinate);
    }
    else{
        let ordinateloc = "";
        for(let j = dimension; j < titles.length; j++){
            ordinateloc += titles[j];
        }
        axtitles.push(ordinateloc);
    }

    let legprefix = [];
    if(titles[1] == PPPhaseErrLabel){
        let rms = 0, avg = 0;
        let ndata = obj.data[1].length;
        if(ndata > 0){
            for(let n = 0; n < ndata; n++){
                avg += obj.data[1][n];
            }
            avg /= ndata;
            for(let n = 0; n < ndata; n++){
                rms += (avg-obj.data[1][n])**2;
            }
        }
        rms = Math.sqrt(rms/ndata);
        let strrms = "RMS(deg.) = "+ToPrmString(rms, 2);
        legprefix.push(strrms);
    }

    let plobj = {
        data: [obj],
        dimension: 1,
        titles: titles,
        axtitles: axtitles,
        legprefix: legprefix,
        isdatad2d: false
    }
    SPGUI.preproc = new PlotWindow(SPGUI.importplotarea, 
            "plotpreprop", plobj, SPGUI.setting[POLabel]["plotpreprop"], target);
}

function AppendDefault(spjson)
{
    let objlabels = [AccLabel, SrcLabel, ConfigLabel];
    let defobjs = [def_acc, def_src, def_conf];

    for(let j = 0; j < objlabels.length; j++){
        let okeys = Object.keys(spjson[objlabels[j]]);
        for(let i = 0; i < okeys.length; i++){
            let defobt = CopyJSON(defobjs[j][OptionLabel]);
            Object.assign(defobt, spjson[objlabels[j]][okeys[i]][OptionLabel]);
            spjson[objlabels[j]][okeys[i]][OptionLabel] = CopyJSON(defobt);        
            let defobj = CopyJSON(defobjs[j]);
            Object.assign(defobj, spjson[objlabels[j]][okeys[i]]);
            spjson[objlabels[j]][okeys[i]] = CopyJSON(defobj);        
        }
    }
}

function SetOutputTable()
{
    if(!SPGUI.setting.hasOwnProperty(OutFileLabel)){
        SPGUI.setting[OutFileLabel] = CopyJSON(def_output_configs);
    }
    let outtypes = {};
    OutputOptionsOrder.forEach(function(element){
        outtypes[OutputOptionsLabel[element][0]] = true;
    });
    if(SPGUI.runEnvironment == "browser" || SPGUI.runEnvironment == "node"){
        outtypes[OutputOptionsLabel.folder[0]] = false;
    }
    let outobj = {};
    outobj[OptionLabel] = SPGUI.setting[OutFileLabel];
    SPGUI.outputopt.JSONObj = outobj;
    SPGUI.outputopt.SetValidity(outtypes, {}, {});
    document.getElementById("output_option").appendChild(SPGUI.outputopt.GetTable());    
}

function HandleFileSelect(evt) 
{
    SwitchSpinner(true);
    SPGUI.filereader.readAsText(evt.target.files[0]);
    SPGUI.isimported = evt.currentTarget.id == "import";
    if(SPGUI.isimported){
        document.getElementById("import").value = "";
    }
    else{
        SPGUI.setting.prmfile = evt.target.files[0].name;
        document.getElementById("open").value = "";
        document.title = "SPECTRA "+FullVersionNumber+" - "+SPGUI.setting.prmfile;
    }
}

function LoadPrmFileFromList()
{
    let select = document.getElementById("prmfilesSelect");
    let prmfile = GetSelections(select).text[0];
    document.title = "SPECTRA "+FullVersionNumber+" - "+prmfile;
    SPGUI.socketio.emit("load", "prm", prmfile)
}

function LoadParameters(fresults)
{
    let spjson;

    if(SPGUI.runEnvironment == "node"){
        let isconv = SPGUI.upload_menu == FileConvertUp;
        SPGUI.socketio.emit("save", "prm", SPGUI.setting.prmfile, fresults, isconv);
        if(isconv){
            SPGUI.upload_menu = "";
            return;
        }
    }

    try {
        spjson = JSON.parse(fresults);
    }
    catch (err) {
        Alert("Error: Cannot load the parameter file.\n"+err.message);
        return false;
    }
    if(spjson.hasOwnProperty(SimplifiedLabel) && spjson[SimplifiedLabel]){
        let tmp = CopyJSON(spjson);
        spjson = GetDetailObj(tmp);
    }

    SPGUI.orginobj = "";
    if(IsFELBunchFactor(spjson)){
        SPGUI.orginobj = CopyJSON(spjson);
    }
    if(spjson.hasOwnProperty(InputLabel)){
        let cmdcont = Output4CMD(spjson);
        if(cmdcont.wigner){
            SPGUI.orginobj = CopyJSON(spjson);
        }
        spjson = CopyJSON(spjson[InputLabel]);
    }
    if(spjson.hasOwnProperty(OutFileLabel)){
        SPGUI.setting[OutFileLabel] = CopyJSON(spjson[OutFileLabel]);
        SetOutputTable();
    }
    spjson = ArrangeSpJSON(spjson);
    AppendDefault(spjson);

    if(SPGUI.isimported){
        let categs = [BLLabel, AccLabel, SrcLabel, ConfigLabel];
        for(let j = 0; j < categs.length; j++){
            let keys = Object.keys(spjson[categs[j]]);
            for(let n = 0; n < keys.length; n++){
                if(SPGUI.curr_spjson[categs[j]].hasOwnProperty(keys[n])){
                    let rep = 1, newkey;
                    do{
                        newkey = keys[n]+"_"+rep.toString();
                        rep++;
                    } while(SPGUI.curr_spjson[categs[j]].hasOwnProperty(newkey));
                    SPGUI.curr_spjson[categs[j]][newkey] = CopyJSON(spjson[categs[j]][keys[n]]);
                }
                else{
                    SPGUI.curr_spjson[categs[j]][keys[n]] = CopyJSON(spjson[categs[j]][keys[n]]);
                }
            }
        }
        if(spjson.hasOwnProperty(FMaterialLabel)){
            Object.assign(SPGUI.curr_spjson[FMaterialLabel], spjson[FMaterialLabel]);
        }
        SPGUI.isimported = false;
    }
    else{
        SPGUI.curr_spjson = CopyJSON(spjson);
    }
    
    CreatePrmSetMenu();
    ArrangeDataImport(false);
    LoadJSON();

    return true;
}

function GetDetailSingle(simplobj, labels, isoption, sublabels)
{
    let dobj = {};
    let keys = Object.keys(simplobj);
    let item;
    for(let i = 0; i < keys.length; i++){
        if(keys[i] == OptionLabel){
            item = GetDetailSingle(simplobj[OptionLabel], sublabels[OptionLabel], true, sublabels)
        }
        else if(typeof simplobj[keys[i]] == "object"
                && simplobj[keys[i]].hasOwnProperty("data") == false
                && Array.isArray(simplobj[keys[i]]) == false){
            item = GetDetailSingle(simplobj[keys[i]], sublabels[keys[i]], false, sublabels)
        }
        else{
            item = simplobj[keys[i]];
        }
        if(labels == undefined){
            continue;
        }
        if(labels.hasOwnProperty(keys[i])){
            let key = isoption ? labels[keys[i]][0] : labels[keys[i]];
            dobj[key] = item;
        }
        else{
            dobj[keys[i]] = item;
        }
    }
    return dobj;
}

function GetDetailObj(simplobj)
{
    let categs = [AccLabel, SrcLabel, ConfigLabel];
    let labels = [AccPrmsLabel, SrcPrmsLabel, ConfigPrmsLabel];
    let sublabels = [
        {[OptionLabel]:AccOptionsLabel, injecprm:InjectionPrmsLabel},
        {[OptionLabel]:SrcOptionsLabel, ferrprm:FerrPrmsLabel, 
            perrprm:PerrPrmsLabel, segprm:SegPrmsLabel},
        {[OptionLabel]:ConfigOptionsLabel, bpfprms:BPFPrmsLabel}
    ];

    let body = {};
    for(let j = 0; j < labels.length; j++){
        let obj = simplobj[categs[j]];
        let keys = Object.keys(obj);
        let dobjects = {};
        for(let i = 0; i < keys.length; i++){
            let dobj = GetDetailSingle(obj[keys[i]], labels[j], false, sublabels[j]);
            dobjects[keys[i]] = dobj;
        }
        body[categs[j]] = dobjects;
    }
    body[BLLabel] = simplobj[BLLabel];
    body[FMaterialLabel] = simplobj[FMaterialLabel];
    body[CurrentLabel] = simplobj[CurrentLabel];

    return body;
}

async function GetTemporaryFileName()
{
    if(SPGUI.setting[OutFileLabel][OutputOptionsLabel.folder[0]] == ""){
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.folder[0]] = SPGUI.wdname;
        let idoutdir = GetIDFromItem(OutFileLabel, OutputOptionsLabel.folder[0], -1);
        document.getElementById(idoutdir).value = SPGUI.wdname;
    }

    let base = await window.__TAURI__.path.join(SPGUI.setting[OutFileLabel][OutputOptionsLabel.folder[0]], "spectra_temp")
    let fname, sn = 0;
    do{
        fname = base+sn.toString()+".json";
        sn++;
    } while (await window.__TAURI__.tauri.invoke("exists", { path: fname}));
    return Promise.resolve(fname);
}

function CreateScanProcess()
{
    let item = GetItemFromID(SPGUI.targetid.id);
    let prmitem = item.item;
    let obj;
    if(item.subitem == ""){
        obj = SPGUI.setting[ScanLabel][item.categ];
    }
    else{
        obj = SPGUI.setting[ScanLabel][item.categ][OptionLabel][item.item];
        prmitem = item.subitem;
    }
    let prms = SPGUI.scanconftbl.ExportInputPrms();
    if(SPGUI.targetid.menuid == Scan1D){
        obj[prmitem] = [prms[ScanPrmsLabel.initial], prms[ScanPrmsLabel.final]];
        if(SPGUI.targetid.isinteger){
            obj[prmitem].push(prms[ScanPrmsLabel.interv]);
        }
        else{
            obj[prmitem].push(prms[ScanPrmsLabel.points]);
        }
        obj[prmitem].push(prms[ScanPrmsLabel.iniser]);
    }
    else{
        obj[prmitem] = [prms[ScanPrmsLabel.initial2], prms[ScanPrmsLabel.final2]];
        if(SPGUI.targetid.isinteger){
            obj[prmitem].push(prms[ScanPrmsLabel.interv2]);
        }
        else{
            obj[prmitem].push(prms[ScanPrmsLabel.points2]);
        }
        obj[prmitem].push(prms[ScanPrmsLabel.iniser2]);
    }
    let values = CopyJSON(obj[prmitem]);
    if(SPGUI.targetid.menuid == Scan2D1D){
        values[2][1-item.jxy] = 0;
    }
    if(SPGUI.targetid.isinteger){
        values.push(IntegerLabel);
    }

    SPGUI.targetid.scanobj = {};
    SPGUI.targetid.scanobj[item.categ] = {};
    if(item.subitem == ""){
        SPGUI.targetid.scanobj[item.categ][item.item] = values;
    }
    else{
        SPGUI.targetid.scanobj[item.categ][OptionLabel] = {}
        SPGUI.targetid.scanobj[item.categ][OptionLabel][item.item] = {};
        SPGUI.targetid.scanobj[item.categ][OptionLabel][item.item][item.subitem] = values;
    }
  
    SPGUI.setting[ScanOptionlabel] = SPGUI.scanconfopt.ExportInputOptions();

    let datapath = GetCurrentDataName();
    if(datapath == ""){
        return;
    }
       
    SPGUI.process_event_origin_id = 
        SPGUI.calcprocess == null ? "" : CreateProcessLabel;
    CreateProcess(datapath);
    IncrementSerialNumber();
}

function SetPrmSetName()
{
    let items = SPGUI.curritem;
    let newname0 = SPGUI.setting[PSNameLabel][PSNameOptionsLabel.newname[0]];
    let obj = SPGUI.curr_spjson[items.categ];
    let currkeys = Object.keys(obj);

    let isuf = 0;
    let newname = newname0;
    for(let i = 0; i < currkeys.length; i++){
        if(newname == currkeys[i]){
            newname = newname0+"_"+isuf.toString();
            isuf++;
            i = 0;
        }
    }

    let addobj = CopyJSON(obj[items.item]);
    if(SPGUI.eventid == PrmSetRename){
        delete obj[items.item];
    }
    obj[newname] = addobj;

    ArrangeCurrentPrmSet(SPGUI.eventid, items, newname);
}

function SetCustomAccuracy()
{
    GetObj(ConfigLabel)[AccuracyLabel] = CopyJSON(SPGUI.setting[AccuracyLabel]);
}

function GetAsciiFormat()
{
    let body = {};
    let labels = [];
    let simpid = {};
    let asciifmt = {}, fid;

    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 1;
    asciifmt[TitlesLabel] = [TimeLabel, CurrLabel];
    asciifmt[OrdinateLabel] = "Current";
    asciifmt[UnitLabel] = [UnitOptionsLabel.time, ""];
    fid = GetIDFromItem(AccLabel, AccOptionsLabel.currdata[0]);
    body[fid] = asciifmt;
    simpid[fid] = "currdata";
    labels.push({id:fid, isopt:true});

    asciifmt = {};
    asciifmt[DimensionLabel] = 2;
    asciifmt[TargetsLabel] = 1;
    asciifmt[TitlesLabel] = [TimeLabel, EdevLabel, NormCurrLabel];
    asciifmt[UnitLabel] = [UnitOptionsLabel.time, "", ""];
    fid = GetIDFromItem(AccLabel, AccOptionsLabel.Etdata[0]);
    body[fid] = asciifmt;
    simpid[fid] = "Etdata";
    labels.push({id:fid, isopt:true});

    asciifmt = {};
    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 2;
    asciifmt[TitlesLabel] = [ZLabel, BxLabel, ByLabel];
    asciifmt[OrdinateLabel] = "Magnetic Field (T)";
    asciifmt[UnitLabel] = [UnitOptionsLabel.zpos, UnitOptionsLabel.magf, UnitOptionsLabel.magf];
    fid = GetIDFromItem(SrcLabel, SrcPrmsLabel.fvsz);  
    body[fid] = asciifmt;
    simpid[fid] = "fvsz";
    labels.push({id:fid, isopt:false});

    asciifmt = {};
    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 2;
    asciifmt[TitlesLabel] = [ZLabel, BxLabel, ByLabel];
    asciifmt[OrdinateLabel] = "Magnetic Field (T)";
    asciifmt[UnitLabel] = [UnitOptionsLabel.zpos, UnitOptionsLabel.magf, UnitOptionsLabel.magf];
    fid = GetIDFromItem(SrcLabel, SrcPrmsLabel.fvsz1per);
    body[fid] = asciifmt;
    simpid[fid] = "fvsz1per";
    labels.push({id:fid, isopt:false});
    
    asciifmt = {};
    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 2;
    asciifmt[TitlesLabel] = [GapLabel, BxLabel, ByLabel];
    asciifmt[OrdinateLabel] = "Magnetic Field (T)";
    asciifmt[UnitLabel] = [UnitOptionsLabel.gap, UnitOptionsLabel.magf, UnitOptionsLabel.magf];
    fid = GetIDFromItem(SrcLabel, SrcOptionsLabel.gaptbl[0]);
    body[fid] = asciifmt;
    simpid[fid] = "gaptbl";
    labels.push({id:fid, isopt:true});
    
    asciifmt = {};
    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 1;
    asciifmt[TitlesLabel] = [EnergyLabel, TransmLabel];
    asciifmt[OrdinateLabel] = "Transmission Rate";
    asciifmt[UnitLabel] = ["", ""];
    fid = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.fcustom[0]);
    body[fid] = asciifmt;
    simpid[fid] = "fcustom";
    labels.push({id:fid, isopt:true});

    asciifmt = {};
    asciifmt[DimensionLabel] = 0;
    asciifmt[TargetsLabel] = 1;
    asciifmt[TitlesLabel] = [DepthLabel];
    asciifmt[OrdinateLabel] = "Depth (mm)";
    asciifmt[UnitLabel] = [UnitOptionsLabel.depth];
    fid = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.depthdata[0]);
    body[fid] = asciifmt;
    simpid[fid] = "depthdata";
    labels.push({id:fid, isopt:true});

    asciifmt = {};
    asciifmt[DimensionLabel] = 1;
    asciifmt[TargetsLabel] = 2;
    asciifmt[TitlesLabel] = [SeedWavelLabel, SeedFluxLabel, SeedPhaseLabel];
    asciifmt[OrdinateLabel] = "Intensity/Phase(deg.)";
    asciifmt[UnitLabel] = [UnitOptionsLabel.seedwavel, UnitOptionsLabel.seedflux, UnitOptionsLabel.seedphase];
    fid = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.seedspec[0]);
    body[fid] = asciifmt;
    simpid[fid] = "seedspec";
    labels.push({id:fid, isopt:true});

    return {fmt:body, label:labels, simpid:simpid};
}


function SwitchSpinner(enable)
{
    if(enable){
        document.getElementById("spinner").classList.remove("none");
    }
    else{
        document.getElementById("spinner").classList.add("none");
    }
}

var TWindow = {};

async function BeforeExit()
{
    const factor = await TWindow.appWindow.scaleFactor();
    let pos = await TWindow.appWindow.outerPosition();
    let size = await TWindow.appWindow.innerSize();
    pos = pos.toLogical(factor);
    size = size.toLogical(factor);
    SPGUI.setting.window = {...pos, ...size};
    SPGUI.setting.animinterv = AnimationInterval;
    let data = FormatArray(JSON.stringify(SPGUI.setting, null, JSONIndent));
    let conffile = await window.__TAURI__.path.join(SPGUI.wdname, "spectra_config.json");
    await window.__TAURI__.tauri.invoke("write_file", { path: conffile, data: data});
    return Promise.resolve(0);
}

async function OpenReference()
{
    try {
        let refpath;
        if(SPGUI.runOS == "Linux"){
            refpath = "/usr/share/spectra/help/reference.html";
            if(await window.__TAURI__.invoke("exists", {path: refpath}) == false){
                refpath = "help"+window.__TAURI__.path.sep+"reference.html";
            }
        }
        else{
            refpath = "help"+window.__TAURI__.path.sep+"reference.html";
        }
        await window.__TAURI__.shell.open(refpath);
    } catch (e) {
        Alert(e);
    }
}

async function GetConfiguration()
{
    try {
        let conffile = await window.__TAURI__.path.join(SPGUI.wdname, "spectra_config.json");
        let data = await window.__TAURI__.tauri.invoke("read_file", { path: conffile});
        if(data != ""){
            SPGUI.setting =  JSON.parse(data);
            if(SPGUI.setting.hasOwnProperty("animinterv")){
                AnimationInterval = SPGUI.setting.animinterv;
            };
            if(!SPGUI.setting.hasOwnProperty("defpaths")){
                SPGUI.setting.defpaths = {};
            }
            // backward compatibility (<= 11.1)
            if(SPGUI.setting.hasOwnProperty("main")){
                delete SPGUI.setting.main;
            }
            if(SPGUI.setting.hasOwnProperty("sub")){
                let subobj = CopyJSON(SPGUI.setting.sub);
                Object.keys(subobj).forEach((el) =>{
                    SPGUI.setting[el] = CopyJSON(subobj[el]);
                })
                delete SPGUI.setting.sub;
            }
        }
    } catch(e) {
        Alert(e);
    }
}

function DuplicatePlotObject()
{
    let plobj;
    if(SPGUI.plotobject.object != null){
        plobj = SPGUI.plotobject.object;
    }
    else{
        plobj = {data: SPGUI.plotobject.window[0].GetData(), size:SPGUI.plotobject.size};
    }
    return plobj;
}

function CreateNewplot()
{
    let size = SPGUI.plotobject.size;
    if(SPGUI.runEnvironment == "tauri"){       
        let newplot = new TWindow.WebviewWindow("subwindow"+SPGUI.subwindows.toString(), {
            url:"newplot.html",
            width: size[0],
            height: size[1],
        });
        newplot.listen("ready", e => {
            newplot.emit("message", DuplicatePlotObject());
        });
        SPGUI.subwindows++;
    }
    else{
        SPGUI.currnewwin = window.open("newplot.html", "", 
        "innerWidth="+size[0]+",innerHeight="+size[1]);
    }
}

// create the main GUI window
window.onload = async function()
{
    if(window.hasOwnProperty("__TAURI__")){
        SPGUI.runEnvironment = "tauri";
        SPGUI.runOS = await window.__TAURI__.os.type();
        TWindow = window.__TAURI__.window;
        TWindow.appWindow.setTitle("SPECTRA "+FullVersionNumber);
        if(SPGUI.runOS == "Linux"){
            PlotlyScatterType = "scatter";
        }
        TWindow.appWindow.listen('tauri://close-requested', (ev) => {
            BeforeExit().then((e) => {
                window.__TAURI__.process.exit(0);
            });
        });

        window.__TAURI__.tauri.invoke("get_exepath").then(async function(exepath){
            let exedir = await window.__TAURI__.path.dirname(exepath);
            if(SPGUI.runOS == "Darwin"){
                exedir = exedir.substring(0, exedir.lastIndexOf("Contents"));
            }
            if(await window.__TAURI__.tauri.invoke("change_directory", { path: exedir})){
            }
            else{
                Alert("Failed to move to the application directory.");
            }    
        });

        SPGUI.wdname = await window.__TAURI__.path.documentDir();

        await GetConfiguration();
        let winconf = {width:800, height: 600, x:100, y: 100};
        if(SPGUI.setting.hasOwnProperty("window")){
            winconf = SPGUI.setting.window;
            if(!winconf.hasOwnProperty("width")){
                winconf.width = 800;
            } 
            if(!winconf.hasOwnProperty("height")){
                winconf.height = 800;
            } 
            if(!winconf.hasOwnProperty("x")){
                winconf.x = 100;
            } 
            if(!winconf.hasOwnProperty("y")){
                winconf.y = 100;
            } 
        }
        try {
            await TWindow.appWindow.setSize(new TWindow.LogicalSize(winconf.width, winconf.height));
            await TWindow.appWindow.setPosition(new TWindow.LogicalPosition(winconf.x, winconf.y));
        } catch(e) {
            Alert(e);
        }
    }
    else{
        try {
            SPGUI.socketio = io();
            SPGUI.runEnvironment = "node";
        } catch(e) {
        	console.log("Local mode");
            SPGUI.runEnvironment = "browser";
        }
    }

    if(SPGUI.setting.hasOwnProperty("prmfile") == false){
        SPGUI.setting.prmfile = "untitled";
    }
    else if(SPGUI.setting.prmfile == ""){
        SPGUI.setting.prmfile = "untitled";
    }
    SPGUI.preproc = null;
    SPGUI.currnewwin = null;

    let tmpobj = {};
    tmpobj[AccLabel] = [PPBetaLabel];
    PreProcessLabel.push(CopyJSON(tmpobj)); tmpobj = {};
    tmpobj[SrcLabel] = [PPFDlabel, PP1stIntLabel, PP2ndIntLabel, PPPhaseErrLabel, PPRedFlux];
    PreProcessLabel.push(CopyJSON(tmpobj)); tmpobj = {};
    tmpobj[FALabel] = [PPTransLabel, PPAbsLabel]; 
    PreProcessLabel.push(CopyJSON(tmpobj)); tmpobj = {};

    tmpobj["Undulators"] = [
        SrcTypeLabel.LIN_UND,
        SrcTypeLabel.VERTICAL_UND,
        SrcTypeLabel.HELICAL_UND,
        SrcTypeLabel.ELLIPTIC_UND,
        SrcTypeLabel.FIGURE8_UND,
        SrcTypeLabel.VFIGURE8_UND,
        SrcTypeLabel.MULTI_HARM_UND
    ];
    SrcTypeObj.push(CopyJSON(tmpobj)); tmpobj = {};
    tmpobj["BMs and Wigglers"] = [
        SrcTypeLabel.BM,
        SrcTypeLabel.WIGGLER,
        SrcTypeLabel.EMPW,
        SrcTypeLabel.WLEN_SHIFTER
    ];
    SrcTypeObj.push(CopyJSON(tmpobj)); tmpobj = {};
    tmpobj["Custom"] = [
        SrcTypeLabel.CUSTOM_PERIODIC,
        SrcTypeLabel.CUSTOM
    ];
    if(SPGUI.runEnvironment == "tauri"){
        tmpobj["Custom"].push(SrcTypeLabel.FIELDMAP3D);
    }
    else{
        let bunchsel = AccOptionsLabel.bunchtype[1].filter(val => val != ImportPaticleLabel);
        AccOptionsLabel.bunchtype[1] = bunchsel;
    }
    SrcTypeObj.push(CopyJSON(tmpobj));
  
    SPGUI.acctbl = new Accelerator();
    SPGUI.injtbl = new InjectionCondition();
    SPGUI.accopt = new AccOption();

    SPGUI.srctbl = new LightSource();
    SPGUI.srcferrtbl = new FieldError();
    SPGUI.srcperrtbl = new PhaseError();
    SPGUI.srcsegtbl = new Segmentation();
    SPGUI.srcopt = new SrcOption();

    SPGUI.conftbl = new Configuration();
    SPGUI.configbpftbl = new BPFPrms();
    SPGUI.configCMDtbl = new CMDPrms();
    SPGUI.configFELtbl = new FELPrms();
    SPGUI.confopt = new ConfigOption();
    SPGUI.outputopt = new OutputOption();
    SPGUI.gapfieldtable = new GapFieldTable();
    SPGUI.postprocessor = new PostProcessor();

    SPGUI.curr_spjson = CopyJSON(def_spjson);

    SPGUI.totalcalcs = 0;
    SPGUI.totalpreprocs = 0;

    SPGUI.filereader = new FileReader();    
    SPGUI.filereader.addEventListener("load", ()=>{
        LoadParameters(SPGUI.filereader.result);
        SwitchSpinner(false);
    });
    document.getElementById("open").addEventListener("change", HandleFileSelect);
    document.getElementById("import").addEventListener("change", HandleFileSelect);
    
    // create panels
    let titles = [MainTitle, PreProcTitle, PostProcTitle];
    SPGUI.mainnotebook = new NoteBook(document.getElementById("notemain"), titles, false);

    document.getElementById("notemain").addEventListener("pagechange", (e)=>{
        let page = e.detail.page;
        switch(page){
            case 1:
                let select = document.getElementById("dataselect");
                let index = GetSelections(select).index;
                if(index.length < 1){
                    return;
                }
                if(index[0] >= DefaultPPItems){
                    let id = DetachSuffix(select.value, SuffixPlotBody);
                    SPGUI.importdata[id].Refresh();
                }
                else{
                    if(SPGUI.preproc != null){
                        SPGUI.preproc.RefreshPlotObject();
                    }
                }
                break;
            case 2:
                SPGUI.postprocessor.Refresh();
                break;
        }
    });

    // window size configuration
    let winh = window.innerHeight-WindowHeightOffset;
    document.getElementById("notemain").style.height = winh.toString()+"px";    

    window.addEventListener("resize", () => {
        let winh = window.innerHeight-WindowHeightOffset;
        document.getElementById("notemain").style.height = winh.toString()+"px";
    });

    // main parameters
    // create accelerator section
    let comps = [{type:"div", class:"flexcont", id:"acctypediv"}, {type:"div", id:"acc"}, 
                    {type:"hr"}, {type:"div", id:"acc_option"}];
    let fldset = BundleFieldSet("Accelerator", comps, {body:"accfld"});
    SPGUI.mainnotebook.GetPanel(0).appendChild(fldset);

    // create light source section
    comps = [{type:"div", class:"flexcont", id:"srctypediv"}, {type:"div", id:"src"},
                    {type:"hr"}, {type:"div", id:"src_option"}];
    fldset = BundleFieldSet("Light Source", comps, {body:"srcfld"});

    SPGUI.mainnotebook.GetPanel(0).appendChild(fldset);

    // create config. section
    let divconf = document.createElement("div");
    divconf.id = "";
    divconf.className = "flexcont flexcolumn flexstretch"

    comps = [{type:"div", class:"flexcont flexcolumn", id:"config"}, {type:"hr"}, 
                {type:"div", id:"config_option"}];
    fldset = BundleFieldSet("Configurations", comps, {body:"configfld", title:"closeconfig"});
    fldset.className += " none";
    divconf.appendChild(fldset);   
    SPGUI.mainnotebook.GetPanel(0).appendChild(divconf);

    let caclclosebtn = document.createElement("div");
    caclclosebtn.addEventListener("click", (e) => {
        GetObj(ConfigLabel)[TypeLabel] = "";
        CloseCalculation();
        LoadAccOption();
        LoadSrcOption();
        SwitchFELRelated();
        FreezeGUI(false, true, true, false);
    });
    caclclosebtn.innerHTML = "&times;";
//    caclclosebtn.innerHTML = "&#x2716;";
 //   caclclosebtn.innerHTML = "&#x2612;";
    caclclosebtn.className =  "closediv";
    document.getElementById("closeconfig").appendChild(caclclosebtn);
    document.getElementById("closeconfig").className = "flexcont flexcenter";

    // output file option
    comps = [{type:"div", id:"output_option"}];
    fldset = BundleFieldSet("Output File", comps, {body:"outputfld"});
    fldset.className += " flexgrow none";
    divconf.appendChild(fldset);   

    // create calculation process
    comps = [{type:"div", id:"calcproc"}];
    fldset = BundleFieldSet(CalcProcessLabel, comps, {body:"calcprocfld"});
    fldset.className += " flexgrow none";
    divconf.appendChild(fldset);
    document.getElementById("calcproc").className = "flexgrow flexcolumn";

    // create calculation result spread sheet
    comps = [{type:"div", id:"calcresult"}];
    fldset = BundleFieldSet(CalcResultLabel, comps, {body:"calcresultfld"});
    fldset.className += " flexgrow none sheetarea";
    divconf.appendChild(fldset);
    let btn = document.createElement("button");
    btn.innerHTML = StartCalcLabel;
    btn.addEventListener("click", (e) => {
        SPGUI.process_event_origin_id = StartCalcLabel;
        CreateProcess("");
    });
    let calcresdiv = document.getElementById("calcresult");
    calcresdiv.style.overflow = "auto";
    calcresdiv.appendChild(btn);
    let gridconf = {
        coltypes:[
            {ColLabel:"Item", ColType:"string"},
            {ColLabel:"Value", ColType:"number"}
        ],
        readonly:true,
        withnum:false
    };
    SPGUI.caclresultgrid = new Grid("fixpointresult", gridconf);
    calcresdiv.appendChild(SPGUI.caclresultgrid.GetTable());

    // create import data plot: select box
    SPGUI.importformats = GetAsciiFormat();
    let dimports = SPGUI.importformats.label;
    let iptitles = [];
    let poconfig = {};
    for(let n = 0; n < dimports.length; n++){
        let item = GetItemFromID(dimports[n].id);
        iptitles.push(item.item);
        poconfig[dimports[n].id] = CopyJSON(def_plot_configs);
    }
    poconfig["plotcurrent"] = CopyJSON(def_plot_configs);
    poconfig["plotpreprop"] = CopyJSON(def_plot_configs);    
    if(!SPGUI.setting.hasOwnProperty(POLabel)){
        SPGUI.setting[POLabel] = poconfig;
    }

    // create imported data view
    comps = [{type:"div", id:"preprocdiv"}, 
        {type:"div", id:"dataimpdiv", class:"flexcont flexcolumn flexstretch"}, 
        {type:"hr"}, 
        {type:"div", id:"datacontdiv", class:"flexcont flexgrow flexstretch"}];
    fldset = BundleFieldSet("Target Item", comps, {});
    SPGUI.mainnotebook.GetPanel(1).appendChild(fldset);
    fldset.style.width = "200px";

    // create import button
    let impbtn = document.createElement("button");
    impbtn.id = "importbutton";
    impbtn.innerHTML = "Update";
    impbtn.addEventListener("click", (e)=>
    {
        let tgtsel = document.getElementById("dataselect");
        let tgtitem = GetSelections(tgtsel);
        if(tgtitem.length == 0){
            return;
        }
        PreProcess(tgtitem.text[0]).then((done) => {
                if(!done){
                    let select = document.getElementById("dataselect");
                    let currtgt  = GetSelections(select);
                    if(currtgt.index.length == 0){
                        return;
                    }
                    let id = select.options[currtgt.index[0]].value;
                    id = DetachSuffix(id, SuffixPlotBody);
                    let btnid = AttachSuffix(id, SuffixButton);
                    let btn = document.getElementById(btnid);
                    btn.click();            
                }
            }
        );

    });
    document.getElementById("dataimpdiv").appendChild(impbtn);

    // select box for data type
    let selcont = CreateSelectBox("", 14, false, true);
    document.getElementById("dataimpdiv").appendChild(selcont.body);
    selcont.select.id = "dataselect";
    selcont.select.addEventListener("change", (e) => {
        ShowImportCtrls();
        let impbtn = document.getElementById("importbutton");
        PreProcess(e.currentTarget.value).then((done) => {
            if(done){
                impbtn.innerHTML = "Update";
            }
            else{
                impbtn.innerHTML = ReplaceLabel;
                let id = DetachSuffix(e.currentTarget.value, SuffixPlotBody);
                ShowImportPlot(id);        
            }
        });
    });   
    DefaultPPItems = SetSelectMenus(selcont.select, PreProcessLabel, 
            [PPPhaseErrLabel, PPRedFlux, PPTransLabel, PPAbsLabel]);

    function GetPPPlotWindow() {
        let select = document.getElementById("dataselect");
        let index = GetSelections(select).index;
        if(index.length < 1){
            return null;
        }
        if(index[0] >= DefaultPPItems){
            let id = DetachSuffix(select.value, SuffixPlotBody);
            return SPGUI.importdata[id].GetPlotWindow();
        }
        return SPGUI.preproc;
    }

    // export button
    let exportbtn = document.createElement("button");
    exportbtn.id = "dataexport";
    exportbtn.innerHTML = "Export as ASCII";
    exportbtn.addEventListener("click", (e)=>{
        let plotwin = GetPPPlotWindow();
        if(plotwin != null){
            plotwin.ExportAsASCII();   
        }
    });
    document.getElementById("dataimpdiv").appendChild(exportbtn);

    let newwinbtn = document.createElement("button");
    newwinbtn.innerHTML = "Open New Window";
    newwinbtn.addEventListener("click", (e)=>{
        let plot = GetPPPlotWindow();
        let size = plot.GetClientSize();
        SPGUI.plotobject = 
                {size:size, window:[plot], object:null, ispostp:false}; 
        CreateNewplot();
    });
    document.getElementById("dataimpdiv").appendChild(newwinbtn);

    // text area to show data content
    let datacontent = document.createElement("textarea");
    datacontent.id = "datacont";
    datacontent.readOnly = true;
    datacontent.setAttribute("rows", "20");
    datacontent.style.width = "100%";
    document.getElementById("datacontdiv").appendChild(datacontent);

    // create plot arer for imported data
    SPGUI.importplotarea = document.createElement("div");
    SPGUI.importplotarea.className = "flexcont flexgrow flexstretch";
    SPGUI.mainnotebook.GetPanel(1).appendChild(SPGUI.importplotarea);
    ArrangeDataImport(true);

    // create postprocessor
    comps = [];
    fldset = BundleFieldSet("Output Data", comps, {});
    SPGUI.mainnotebook.GetPanel(2).appendChild(fldset);
    fldset.appendChild(SPGUI.postprocessor.GetPanel());

    let plotdiv = document.createElement("div");
    plotdiv.id = "pp_plot";
    plotdiv.className = "flexcont flexcolumn flexgrow flexstretch";
    SPGUI.mainnotebook.GetPanel(2).appendChild(plotdiv);

    // menus
    let filemenucurr, runmenucurr;
    if(SPGUI.runEnvironment == "browser" || SPGUI.runEnvironment == "node"){
        filemenucurr = FileMenus.filter(obj => 
            (typeof obj.label) == "undefined" || 
                (obj.label != FileOpen && obj.label != FileImport && obj.label != FileSave 
                    && obj.label != FileSaveAs && obj.label != FileConvert)
        );
        runmenucurr = RunMenus.filter(obj => 
            (typeof obj.label) == "undefined" || obj.label != ExportSettingLabel
        );
    }
    else{
        filemenucurr = FileMenus.filter(obj => 
            (typeof obj.label) == "undefined" || 
                (obj.label != FileOpenList && obj.label != FileOpenUp 
                    && obj.label != FileDownload && obj.label != FileConvertUp)
        );
        runmenucurr = RunMenus.filter(obj => 
            (typeof obj.label) == "undefined" || obj.label != SettingDownloadLabel
        );
    }
    let filemenu = SetMenuItems(filemenucurr, "");
    let runmenu = SetMenuItems(runmenucurr, "");
    if(SPGUI.runEnvironment != "tauri"){
        let mpiidx = EditViewMenus.indexOf({label:EditMPIConfig});
        EditViewMenus.splice(mpiidx, 1);        
    }
    let vieweditmenu = SetMenuItems(EditViewMenus, "");
    let helpmenu = SetMenuItems(HelpMenus, "");

    CreateMenuList(filemenu, document.getElementById("file"));

    let calcmenu = SetMenuItems(CalcMenus, "");
    SPGUI.calcids = CreateMenuList(calcmenu, document.getElementById("calc"));

    CreatePrmSetMenu();
    CreateMenuList(runmenu, document.getElementById("run"));
    CreateMenuList(vieweditmenu, document.getElementById("view"));
    EnableEditAccuracy(false);

    CreateMenuList(helpmenu, document.getElementById("help"));

    let editmenu = SetMenuItems(PrmEditMenus, "");
    CreateMenuList(editmenu, document.getElementById("prmedit"))
    document.body.addEventListener("click", () => {
        document.getElementById("prmedit").style.visibility = "hidden";
    });

    let scanmenu = SetMenuItems(ScanMenus, "scan1d");
    CreateMenuList(scanmenu, document.getElementById("scan1d"))
    document.body.addEventListener("click", () => {
        document.getElementById("scan1d").style.visibility = "hidden";
    });      
    scanmenu = SetMenuItems(Scan2DMenus, "scan2d");
    CreateMenuList(scanmenu, document.getElementById("scan2d"))
    document.body.addEventListener("click", () => {
        document.getElementById("scan2d").style.visibility = "hidden";
    });      

    // accelerator type
    let selitem = document.createElement("select");
    for(let j = 0; j < AccTypeOrder.length; j++){
        let option = document.createElement("option");
        option.text = AccTypeLabel[AccTypeOrder[j]];
        selitem.appendChild(option);
    }
    selitem.className = "title";
    selitem.id = "acctype";
    document.getElementById("acctypediv").appendChild(selitem);
    selitem.addEventListener("change",
        (e) => {
            GetObj(AccLabel)[TypeLabel] = 
                e.currentTarget.options[e.currentTarget.selectedIndex].text;
            let ulink = LoadAcc(true);
            let indices = ulink.update;
            if(ulink.issrc){
                let srcindices = LoadSrc(false);
                Array.prototype.push.apply(indices, srcindices);
            }
            if(ulink.isconf){
                confindices = LoadConfig(false);               
                Array.prototype.push.apply(indices, confindices);
            }
            LoadAccOption();
            CheckCalcID();
            RefreshGUIPrms(indices);
        });

    // light source type
    selitem = document.createElement("select");
    SetSelectMenus(selitem, SrcTypeObj, []);

    selitem.className = "title";
    selitem.id = "srctype";
    document.getElementById("srctypediv").appendChild(selitem);
    selitem.addEventListener("change",
        (e) => {
            GetObj(SrcLabel)[TypeLabel] = 
                e.currentTarget.options[e.currentTarget.selectedIndex].text;
            SPGUI.gapfieldtable.SetSrcObj(GetObj(SrcLabel));
            LoadAccOption();
            let indices = LoadSrc(true);
            LoadSrcOption();
            CheckCalcID();
            let confindices = LoadConfig(true);
            Array.prototype.push.apply(indices, confindices);            
            RefreshGUIPrms(indices);
        });

    let id = GetIDFromItem(SrcLabel, SrcPrmsLabel.multiharm);
    SPGUI.sprdsheetcol[id] = {
        coltypes:[
            {ColLabel:"K<sub>x</sub> Ratio", ColType:"number"},
            {ColLabel:"K<sub>x</sub> Phase", ColType:"number"},
            {ColLabel:"K<sub>y</sub> Ratio", ColType:"number"},
            {ColLabel:"K<sub>y</sub> Phase", ColType:"number"}
        ],
        withnum:true    
    }

    // numerical accuracy modal dialog
    let accutypes = {};
    AccuracyOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){            
            accutypes[AccuracyOptionsLabel[element][0]] = true;
        }
    });
    SPGUI.accuracy = new AccuracyOption();
    if(!SPGUI.setting.hasOwnProperty(AccuracyLabel)){
        SPGUI.setting[AccuracyLabel] = CopyJSON(def_accuracy);
    }
    let accuobj = {};
    accuobj[OptionLabel] = SPGUI.setting[AccuracyLabel];
    SPGUI.accuracy.JSONObj = accuobj;
    SPGUI.accuracy.SetValidity(accutypes, {}, {});
    CreateModalDialog([SPGUI.accuracy], "accuDialog", "accuconfig", 
        EditAccuracy, "", SetCustomAccuracy, true);

    // magnet configuration modal dialog
    SPGUI.srcmagconftbl = new MagnetConfiguration();
    let magprmtypes = {};
    MagnetLabelOrder.forEach(function(element){
        magprmtypes[MagnetPrmsLabel[element]] = 1;
    });
    SPGUI.srcmagconftbl.JSONObj = 
        GetObj(SrcLabel)[OptionLabel][SrcOptionsLabel.magconf[0]];
    SPGUI.srcmagconftbl.SetPrmTypes(magprmtypes);
    CreateModalDialog([SPGUI.srcmagconftbl], "magDialog", "magconfig", 
        "Magnet Configurations", "", null, false);

    // scan configuration modal dialog
    SPGUI.scanconftbl = new ScanConfiguration();
    SPGUI.scanconfopt = new ScanOption();
    CreateModalDialog([SPGUI.scanconftbl, SPGUI.scanconfopt], "scanDialog", "scanconfig", 
        "Parameter Scan Configurations", "scanconftitle", CreateScanProcess, true);

    // import data unit modal dialog
    let unittypes = {};
    UnitOptionsOrder.forEach(function(element){
        unittypes[UnitOptionsLabel[element][0]] = true;
    });
    SPGUI.unitopt = new UnitOption();
    if(!SPGUI.setting.hasOwnProperty(UnitLabel)){
        SPGUI.setting[UnitLabel] = CopyJSON(def_units);
    }
    let unitobj = {};
    unitobj[OptionLabel] = SPGUI.setting[UnitLabel];
    SPGUI.unitopt.JSONObj = unitobj;
    SPGUI.unitopt.SetValidity(unittypes, {}, {});
    CreateModalDialog([SPGUI.unitopt], "unitDialog", "unitconfig", 
        EditUnits, "", null, false);

    // parallel computing configuration modal dialog
    SPGUI.pcopt = new PCOption();
    if(!SPGUI.setting.hasOwnProperty(PCLabel)){
        SPGUI.setting[PCLabel] = CopyJSON(def_pc_configs);
    }
    let pcobj = {};
    pcobj[OptionLabel] = SPGUI.setting[PCLabel];
    SPGUI.pcopt.JSONObj = pcobj;
    LoadParallelComp();
    CreateModalDialog([SPGUI.pcopt], "pcDialog", "pcconfig", 
        "Configurations - Memorgy/Parallel Computing", "", null, false);

    // phase error configuration modal dialog
    let perrtypes = {};
    PhaseErrOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){
            perrtypes[PhaseErrOptionsLabel[element][0]] = true;
        }
    });
    SPGUI.perropt = new PhaseErrOption();
    if(!SPGUI.setting.hasOwnProperty(PhaseErrLabel)){
        SPGUI.setting[PhaseErrLabel] = CopyJSON(def_perr_configs);
    }
    let perrobj = {};
    perrobj[OptionLabel] = SPGUI.setting[PhaseErrLabel];
    SPGUI.perropt.JSONObj = perrobj;
    SPGUI.perropt.SetValidity(perrtypes, {}, {});
    CreateModalDialog([SPGUI.perropt], "perrDialog", "perrconfig", 
        "Configurations - Phase Error Evaluation", "", null, false);

    // trans./abs. rate plot configuration modal dialog
    SPGUI.filtplot = new FilterPlotOption();
    if(!SPGUI.setting.hasOwnProperty(PlotFilterLabel)){
        SPGUI.setting[PlotFilterLabel] = CopyJSON(def_filtplot_configs);
    }
    let filtplotobj = {};
    filtplotobj[OptionLabel] = SPGUI.setting[PlotFilterLabel];
    SPGUI.filtplot.JSONObj = filtplotobj;
    LoadFiltPlot();
    CreateModalDialog([SPGUI.filtplot], "filtplotDialog", "filtplotconfig", 
        "Configurations - Transmission/Absorption Rate Plot", "", null, false);           

    // about modal dialog
    let conttext = 
        "SPECTRA is a computer software to numerically evaluate the characteristics of synchrotron radiation. "
        +"SPECTRA is free for use, however, the author retains the copyright to SPECTRA."
        +"If you are submitting articles to scientific journals with the results obtained by using SPECTRA, "
        +"cite the reference below.";
    let footers = [
        "J. Synchrotron Radiation 28, 1267 (2021)",
        "hr",
        "E-mail: admin@spectrax.org",
        "URL: http://spectrax.org/spectra/index.html"
    ];
    CreateModalInfDialog("aboutDialog", "SPECTRA (ver. "+FullVersionNumber+")", conttext, footers)
    document.getElementById("aboutDialog").style.width = "400px";

    // parameter set new name modal dialog
    let pstypes = {[PSNameOptionsLabel.newname[0]]: true};
    SPGUI.psnameopt = new PSNameOption();
    if(!SPGUI.setting.hasOwnProperty(PSNameLabel)){
        SPGUI.setting[PSNameLabel] = CopyJSON(def_psname_configs);
    }
    let psnameobj = {};
    psnameobj[OptionLabel] = SPGUI.setting[PSNameLabel];
    SPGUI.psnameopt.JSONObj = psnameobj;
    SPGUI.psnameopt.SetValidity(pstypes, {}, {});
    CreateModalDialog([SPGUI.psnameopt], "psDialog", "psnameconfig", 
        "Rename/Duplicate Parameter Set", "", SetPrmSetName, false);

    // parameter files list modal dialog (for node version)
    CreateModalDialog([], "prmfilesDialog", "prmfilesdlg", 
        "Parameter Files Available", "", LoadPrmFileFromList, true);
    let sel = document.createElement("select");
    sel.id = "prmfilesSelect";
    document.getElementById("prmfilesdlg").appendChild(sel);

    if(SPGUI.runEnvironment == "node"){
        SPGUI.socketio.on("files", (dirname, fileobj) => {
            if(dirname == "prm"){
                let files = JSON.parse(fileobj).files;
                let sel = document.getElementById("prmfilesSelect");
                sel.innerHTML = "";
                for(let j = 0; j < files.length; j++){
                    let option = document.createElement("option");
                    let filower = files[j].toLowerCase();
                    if(filower.indexOf(".json") >= 0 
                            && filower.lastIndexOf(".json") == filower.length-5){
                        option.text = files[j];
                        sel.appendChild(option);    
                    }
                }
                sel.size = Math.min(files.length, 10);
                document.getElementById("prmfilesDialog").showModal();
            }
        });
        SPGUI.socketio.on("error", (msg) => {
            Alert(msg);
        });
        SPGUI.socketio.on("loadtext", (dirname, ltxt) => {
            if(dirname == "prm"){
                LoadParameters(ltxt);
                SwitchSpinner(false);
            }
            else{
                let blob = new Blob([ltxt], {type:"application/text"});
                let link = document.createElement("a");
                link.href = window.URL.createObjectURL(blob);
                link.download = SPGUI.setting.prmfile;
                link.click();
                link.remove();       
            }
        });
    }
    
    // output file options
    SetOutputTable();

    // display default parameters;
    LoadJSON(); 

    // load parameter file
    document.title = "SPECTRA "+FullVersionNumber;
    if(SPGUI.setting.hasOwnProperty("prmfile") 
            && SPGUI.setting.prmfile != "untitled"){
        OpenParameterFile(SPGUI.setting.prmfile);
    }
};

window.addEventListener("unload", () => {
    if(SPGUI.runEnvironment == "node"){
        SPGUI.socketio.emit("exit");
    }
    else if(SPGUI.runEnvironment == "tauri"){
        BeforeExit().then((e) => {
            window.__TAURI__.process.exit(0);
        });
    }
});

window.addEventListener("message", (e) => {
    let plobj = DuplicatePlotObject();
    SPGUI.currnewwin.postMessage(plobj, "*");
    SPGUI.currnewwin = null;
});

// key binding for supplmental functions (generating C++ source code etc.)
window.addEventListener("keydown", (e) => {
    try {
        if(SPGUI.runEnvironment == "tauri"){
            let prmname = SplitPath(SPGUI.setting.prmfile).name;
            if(!ReferenceInputFiles.includes(prmname)){
                return;
            }
            let isX = false;
            if(e.ctrlKey && (e.key == "a" || e.key == "h")){
                if(e.altKey){ // only current parameter set
                    SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]] 
                        = SPGUI.curr_spjson[CurrentLabel];
                    if(isX){
                        CreateCalcProcesses4X(e.key == "h");
                    }
                    else{
                        CreateAllCalcProcesses(e.key == "h")
                    }
                }
                else{ // all parameter sets
                    CreateAllProcesses(e.key == "h", isX, false);
                }
            }
            if(e.ctrlKey && e.key == "s"){
                CreateAllProcesses(false, false, true);
            }
            return;
        }
        if(e.key == "e" && e.ctrlKey){
            GenerateHeaderFile();
        }
        if(e.key == "m" && e.ctrlKey){
            CreateMenuItems();
        }
        if(e.key == "q" && e.ctrlKey){
            ExportHelpFile();
        }
        if(e.key == "1" && e.ctrlKey){ //  helical retarder
            GenerateField4SC3(true, true);
        }
        if(e.key == "2" && e.ctrlKey){
            GenerateField4SC3(false, false); // horizontal retarder w/o flip
        }
        if(e.key == "3" && e.ctrlKey){ // horizontal retarder
            GenerateField4SC3(true, false);
        }
        if(e.key == "4" && e.ctrlKey){ // empty retarder
            GenerateField4SC3(true, false, true);
        }
        if(e.key == "5" && e.ctrlKey){ // double seed pulse
            let config = window.prompt(
                "Input 5 values in CSV:\nthreshold to search peak (peak/average),\nnormalized slippage,\nposition for slippage tuning &\nchicane length\ntemporal range to ignore", 
                "3, 1, 40, 0.3, 10");
            let prms = config.split(",");
            if(prms.length < 4){
                alert("CSV with two values need to be specified");
                return;
            }
            let thresh = parseFloat(prms[0]);
            let norsl = parseFloat(prms[1]);
            let slpos = parseInt(prms[2]);
            let Lchic = parseFloat(prms[3]);
            let skiptr = parseFloat(prms[4]);
            GenerateField4SC3(true, false, false, false, thresh, norsl, slpos, Lchic, skiptr);
        }
        if(e.key == "6" && e.ctrlKey){ // double pulse generation
            let td = parseFloat(window.prompt("Input time delay (fs)", "5"));
            GenerateField4SC3(true, false, false, true, 1.5, td);
        }
    }
    catch (e) {
        alert(e);
    }
}, false);


//-------------------------
// create processes to check the compatibility with former versions
//-------------------------

var ReferenceInputFiles = ["std.json", "misc.json", "csr.json"];

function CreateAllProcesses(iscoh, isX, issingle)
{
    let blnames = Object.keys(SPGUI.curr_spjson[BLLabel]);

    blnames.forEach(function(el) {
        SPGUI.curr_spjson[CurrentLabel] = el;
        CreatePrmSetMenu();
        ArrangeDataImport(false);
        LoadJSON();
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]] = el;
        if(issingle){
            SPGUI.process_event_origin_id = CreateProcessLabel;
            SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = -1;
            let datapath = GetCurrentDataName();
            if(datapath != ""){
                CreateProcess(datapath);
            }
        }
        else{
            SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = 0;
            SetOutputTable();
            if(isX){
                CreateCalcProcesses4X(iscoh);
            }
            else{
                CreateAllCalcProcesses(iscoh)
            }    
        }
    });
}

function AddOption(calcid, options, iscoh)
{
    let optobj = {}, option = {}, categ = {};

    optobj[ConfigOptionsLabel.aperture[0]] = [
        "Far Field & Ideal Condition::K Dependence::Peak Flux Curve::Partial Flux::Rectangular Slit::Target Harmonics",
        "Far Field & Ideal Condition::K Dependence::Peak Flux Curve::Partial Flux::Rectangular Slit::All Harmonics",
        "Far Field & Ideal Condition::K Dependence::Power::Partial Power::Rectangular Slit"
    ];
    option[ConfigOptionsLabel.aperture[0]] = NormSlitLabel;
    categ[ConfigOptionsLabel.aperture[0]] = iscoh ? "" : ConfigLabel;

    optobj[AccOptionsLabel.singlee[0]] = [
        "Coherent Radiation::Time Dependence::Electric Field",
        "Coherent Radiation::Spatial Dependence::Complex Amplitude::Mesh: x-y"
    ];
    option[AccOptionsLabel.singlee[0]] = true;
    categ[AccOptionsLabel.singlee[0]] = iscoh ? "" : AccLabel;

    optobj[ConfigOptionsLabel.fouriep[0]] = [
        "Coherent Radiation::Time Dependence::Electric Field",
        "Coherent Radiation::Time Dependence::Spatial Power Density",
        "Coherent Radiation::Time Dependence::Partial Power::Rectangular Slit",
        "Coherent Radiation::Time Dependence::Partial Power::Circular Slit"
    ];
    option[ConfigOptionsLabel.fouriep[0]] = true;
    categ[ConfigOptionsLabel.fouriep[0]] = iscoh ? ConfigLabel : "";

    Object.keys(optobj).forEach(el => {
        let cids = optobj[el];
        for(let n = 0; n < cids.length; n++){
            if(cids[n] == calcid){
                options.push({
                    id: calcid,
                    label: el,
                    categ: categ[el],
                    value: option[el]
                });
            }
        }
    });
}

function CreateAllCalcProcesses(iscoh)
{
    let options = [];
    SPGUI.process_event_origin_id = CreateProcessLabel;
    let serorg = SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]];
    SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = -1;
    for(let n = 0; n < SPGUI.calcids.length; n++){
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]]++;
        AddOption(SPGUI.calcids[n], options, iscoh);
        if(SPGUI.calcids[n].indexOf(MenuLabels.fixed) >= 0){
            continue;
        }
        else if(iscoh && SPGUI.calcids[n].indexOf(MenuLabels.cohrad) < 0){
            continue;
        }
        else if(!iscoh && SPGUI.calcids[n].indexOf(MenuLabels.cohrad) >= 0){
            continue;
        }
        let skip = false;
        for(let j = 0; j < SPGUI.disabledcalcid.length; j++){
            if(SPGUI.calcids[n].indexOf(SPGUI.disabledcalcid[j]) >= 0){
                skip = true;
                break;
            }
        }
        if(skip){
            continue;
        }
        let erange = null, Kvalue = null, accu = null;
        let de, acculim, accnp;
        let wige = MenuLabels.srcpoint+"::"+MenuLabels.wigner+"::"+MenuLabels.energy;
        let srccont = GetSrcContents(GetObj(SrcLabel));
        let srcobj = GetObj(SrcLabel)[OptionLabel];
        if(SPGUI.calcids[n] == wige){
            if(GetObj(SrcLabel)[TypeLabel] == SrcTypeLabel.FIELDMAP3D){
                continue;
            }
            if(srccont.isund >= 0){
                erange = GetObj(ConfigLabel)[ConfigPrmsLabel.erange];
                de = GetObj(ConfigLabel)[ConfigPrmsLabel.de];
                let e1st = GetObj(SrcLabel)[SrcPrmsLabel.e1st];

                let nemin = 0.9;
                let nemax = 1.05;
                let nepoints = 400;
                if(srcobj[SrcOptionsLabel.segment_type[0]] != NoneLabel 
                        && srcobj[SrcOptionsLabel.perlattice[0]])
                {
                    nemin = 0.975;
                    nemax = 1.015;
                }
                GetObj(ConfigLabel)[ConfigPrmsLabel.erange] = [e1st*nemin, e1st*nemax];
                GetObj(ConfigLabel)[ConfigPrmsLabel.de] = e1st/nepoints;
            }
        }
        else if(SPGUI.calcids[n].includes(MenuLabels.XXpYYp)){
            if(srccont.isund >= 0 
                && srcobj[SrcOptionsLabel.segment_type[0]] != NoneLabel 
                && srcobj[SrcOptionsLabel.perlattice[0]])
            {
                accu = GetObj(ConfigLabel)[OptionLabel][ConfigOptionsLabel.accuracy[0]];
                acculim = SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.acclimMCpart[0]];
                accnp = SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.accMCpart[0]];
                GetObj(ConfigLabel)[OptionLabel][ConfigOptionsLabel.accuracy[0]] = CustomLabel;
                SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.acclimMCpart[0]] = true;
                SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.accMCpart[0]] = 10000;
            }
        }
        else if(SPGUI.calcids[n].includes(VPDLabel)){
            Kvalue = GetObj(SrcLabel)[SrcPrmsLabel.Kxy];
            GetObj(SrcLabel)[SrcPrmsLabel.Kxy] = [1, 2];
        }

        SPGUI.currCalcType = SPGUI.calcids[n];
        GetObj(ConfigLabel)[TypeLabel] = SPGUI.calcids[n];
        LoadAccOption();
        LoadSrcOption();
        LoadConfig(true);
    
        let datapath = GetCurrentDataName();
        if(datapath != ""){
            CreateProcess(datapath);
            SPGUI.currCalcType = ""; // reset current calculationtype
        }
        if(erange != null){
            GetObj(ConfigLabel)[ConfigPrmsLabel.erange] = erange;
            GetObj(ConfigLabel)[ConfigPrmsLabel.de] = de;
        }
        if(accu != null){
            GetObj(ConfigLabel)[OptionLabel][ConfigOptionsLabel.accuracy[0]] = accu;
            SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.acclimMCpart[0]] = acculim;
            SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.accMCpart[0]] = accnp;
        }
        if(Kvalue != null){
            GetObj(SrcLabel)[SrcPrmsLabel.Kxy] = Kvalue;
        }
    }
    for(let n = 0; n < options.length; n++){
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]]++;
        let skip = false;
        for(let j = 0; j < SPGUI.disabledcalcid.length; j++){
            if(options[n].id.indexOf(SPGUI.disabledcalcid[j]) >= 0){
                skip = true;
                break;
            }
        }
        if(skip){
            continue;
        }
        if(options[n].categ == ""){
            continue;
        }
        SPGUI.currCalcType = options[n].id;
        GetObj(ConfigLabel)[TypeLabel] = options[n].id;
        let org = GetObj(options[n].categ)[OptionLabel][options[n].label];
        GetObj(options[n].categ)[OptionLabel][options[n].label] = options[n].value;

        LoadAccOption();
        LoadSrcOption();
        LoadConfig(true);   
        let datapath = GetCurrentDataName();
        if(datapath != ""){
            CreateProcess(datapath);
        }
        GetObj(options[n].categ)[OptionLabel][options[n].label] = org;
    }

    SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = serorg;
    document.getElementById("run").classList.remove("disable");
}

// for comparison with ver. 10
function CreateCalcProcesses4X(iscoh)
{
    let ids4x = 77; // 77 calculations in version X
    SPGUI.process_event_origin_id = CreateProcessLabel;
    let serorg = SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]];
    SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = -1;
    for(let n = 0; n <= ids4x; n++){
        let obj11 = ConvertCalcId(n);
        if(obj11.index == -1){
            continue;
        }
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = n;

        let singlee = obj11.option == AccOptionsLabel.singlee[0];
        let nn = Math.abs(obj11.index);
        if(singlee){
        //-------
        //    continue;
        }
        else if(iscoh && SPGUI.calcids[nn].indexOf(MenuLabels.cohrad) < 0){
            continue;
        }
        else if(!iscoh && SPGUI.calcids[nn].indexOf(MenuLabels.cohrad) >= 0){
            continue;
        }

        let skip = false;
        for(let j = 0; j < SPGUI.disabledcalcid.length; j++){
            if(SPGUI.calcids[nn].indexOf(SPGUI.disabledcalcid[j]) >= 0){
                skip = true;
                break;
            }
        }
        if(skip){
            continue;
        }
        GetObj(ConfigLabel)[TypeLabel] = SPGUI.calcids[nn];

        let orgvalue;
        if(obj11.option != ""){
            orgvalue = GetObj(obj11.categ)[OptionLabel][obj11.option];
            GetObj(obj11.categ)[OptionLabel][obj11.option] = obj11.selection;
        }
        LoadAccOption();
        LoadSrcOption();
        LoadConfig(true);
    
        let datapath = GetCurrentDataName();
        if(datapath != ""){
            CreateProcess(datapath);
        }

        if(obj11.option != ""){
            GetObj(obj11.categ)[OptionLabel][obj11.option] = orgvalue;
        }
    }
    SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] = serorg;
    document.getElementById("run").classList.remove("disable");
}

function ConvertCalcId(idx10)
{
    let obj11 = {};
    obj11.option = "";
    obj11.categ = ConfigLabel;

    if(idx10 == 14){
        // not consistent, skip
        obj11.index = -1;       
    }
    else if(idx10 <= 15){
        obj11.index = idx10;
    }
    else if(idx10 == 54){
        obj11.index = -1;       
    }
    else if(idx10 == 42 || idx10 == 45){
        // too much computation time, skip
        obj11.index = -1;       
    }
    else if(idx10 >= 16 && idx10 <= 21){
        obj11.index = idx10-1;
    }
    else if(idx10 >= 29 && idx10 <= 31){
        obj11.index = idx10-3;
    }
    else if(idx10 >= 33 && idx10 <= 41){
        obj11.index = idx10-4;
    }
    else if(idx10 >= 46 && idx10 <= 53){
        obj11.index = idx10+13;
    }
    else if(idx10 >= 55 && idx10 <= 57){
        obj11.index = idx10-13;
    }
    else if(idx10 >= 58 && idx10 <= 67){
        obj11.index = idx10-10;
    }
    else if(idx10 >= 68 && idx10 <= 70){
        obj11.index = idx10-23;
    }
    else if(idx10 >= 71 && idx10 <= 74){
        obj11.index = -(idx10-17);
        obj11.option = ConfigOptionsLabel.fouriep[0];
        obj11.selection = true;
    }
    else if(idx10 >= 75 && idx10 <= 77){
        obj11.index = idx10-37;
    }
    else if(idx10 == 22){
        obj11.index = 22;
    }
    else if(idx10 == 23){
        obj11.index = 24;
    }
    else if(idx10 == 24){
        obj11.index = -22;
        obj11.option = ConfigOptionsLabel.aperture[0];
        obj11.selection = NormSlitLabel;
    }
    else if(idx10 == 25){
        obj11.index = 21;
    }
    else if(idx10 == 26){
        obj11.index = 23;
    }
    else if(idx10 == 27){
        obj11.index = 25;
    }
    else if(idx10 == 28){
        obj11.index = -23;
        obj11.option = ConfigOptionsLabel.aperture[0];
        obj11.selection = NormSlitLabel;
    }
    else if(idx10 == 32){
        obj11.index = -27;
        obj11.option = ConfigOptionsLabel.aperture[0];
        obj11.selection = NormSlitLabel;
    }
    else if(idx10 == 42){
        obj11.index = 41;
    }
    else if(idx10 == 43){
        obj11.index = -54;
        obj11.option = AccOptionsLabel.singlee[0];
        obj11.selection = true;
        obj11.categ = AccLabel;
    }
    else if(idx10 == 44){
        obj11.index = -46;
        obj11.option = AccOptionsLabel.singlee[0];
        obj11.selection = true;
        obj11.categ = AccLabel;
    }
    else if(idx10 == 45){
        obj11.index = 58;
    }

    return obj11;
}

