"use strict";

function OpenParameterFile(datapath)
{
    SwitchSpinner(true); 
    window.__TAURI__.tauri.invoke("read_file", { path: datapath})
    .then((data) => {
        let isload = LoadParameters(data);
        if(isload){
            SPGUI.setting.prmfile = datapath;
            TWindow.appWindow.setTitle("SPECTRA "+FullVersionNumber+" - "+datapath);
        }
        SwitchSpinner(false);
    })
    .catch((err) => {
        SwitchSpinner(false);
        Alert(err);
    })

}

function GetLogRange(vmin, vmax, coef){
    if(vmax < 0){
        return[-1, 1];
    }
    let rmax = Math.log10(vmax*coef), rmin;
    if(vmin < 0){
        rmin = Math.log10(vmax*1.0e-6);
    }
    else{
        rmin = Math.max(Math.log10(vmin/coef), rmax-10);
    }
    return [rmin, rmax];
}

function SetMenuItems(list, id)
{
    let tmp = CopyJSON(list);
    tmp.forEach(element => {
        if(id.length > 0){
            element.id = GetIDFromItem(id, element.label, -1);
        }
        else{
            element.id = element.label;
        }
        if(element.hasOwnProperty("submenu")){
            element.submenu = SetMenuItems(element.submenu, element.id);
        }
    });
    return tmp;
}

function CreateMenuList(list, parent)
{
    if(list.length < 1){
        return;
    }
    let layer = list[0].id.split(IDSeparator);
    let menucont = document.createElement("ul");
    let ids = [];
    menucont.style.zIndex = layer.length;
    parent.appendChild(menucont);
    let menutype = [];
    list.forEach(element => {
        let menuitem =  document.createElement("li");
        let skip = false;
        if(element.type == "separator"){
            if(menutype.length > 0){
                if(menutype[menutype.length-1] == "separator"){
                    skip = true;
                }
            }            
            menuitem.className = "separator";
        }
        else{
            menuitem.className = "flexcont flexspace";
            let labelitem = document.createElement("div");
            labelitem.innerHTML = element.label;
            menuitem.appendChild(labelitem);
            if(element.hasOwnProperty("spinner")){
                let spinner = document.createElement("div");
                spinner.id = element.id+"::spinner";
                spinner.className = "spinner none";
                menuitem.appendChild(spinner);
            }
            menuitem.id = element.id;
    
            if(element.hasOwnProperty("submenu")){
                menuitem.setAttribute("tabindex", "-1");
                let subids = CreateMenuList(element.submenu, menuitem);
                ids = ids.concat(subids);
                labelitem = document.createElement("div");
                labelitem.innerHTML = "&#x25B6;";
                labelitem.style.fontSize = "x-small";
                menuitem.appendChild(labelitem);
            }
            else{
                menuitem.addEventListener("click", (e) => {MenuCommand(e);} );
                ids.push(element.id);
            }
        }
        if(!skip){
            menucont.appendChild(menuitem);
            menutype.push(element.type);    
        }
    });
    return ids;
}

function GetObjNames(label)
{
    let names = Object.keys(SPGUI.curr_spjson[label]);
    return names;
}

function GetBL()
{
    let currbl = SPGUI.curr_spjson[BLLabel][SPGUI.curr_spjson[CurrentLabel]];
    return currbl;
}

function GetObj(label)
{
    return SPGUI.curr_spjson[label][GetBL()[label]];
}

function GetCalcObj(scanobj)
{
    let calcid = GetCurrentCalcID();
    if(calcid.indexOf(MenuLabels.CMD) >= 0){
        if(SPGUI.orginobj[InputLabel][ConfigLabel].hasOwnProperty(OrgTypeLabel) == false){
            SPGUI.orginobj[InputLabel][ConfigLabel][OrgTypeLabel] 
                = SPGUI.orginobj[InputLabel][ConfigLabel][TypeLabel]
        }
        let calcobj = CopyJSON(SPGUI.orginobj);
        calcobj[InputLabel][ConfigLabel][OptionLabel] = CopyJSON(SPGUI.confopt.ExportInputOptions());
        calcobj[InputLabel][ConfigLabel][TypeLabel] = calcid;
        if(calcobj[InputLabel][ConfigLabel][OptionLabel][ConfigOptionsLabel.accuracy[0]] == CustomLabel){
            let accuobj = {};
            accuobj[AccuracyOptionsLabel.accothers[0]] = SPGUI.setting[AccuracyLabel][AccuracyOptionsLabel.accothers[0]];
            calcobj[InputLabel][ConfigLabel][AccuracyLabel] = accuobj;
        }
        calcobj[InputLabel][OutFileLabel][OutputOptionsLabel.format[0]] = JSONOnly;
        return calcobj;
    }

    let accobj = SPGUI.acctbl.ExportInputPrms();
    let srcobj = SPGUI.srctbl.ExportInputPrms();
    let confobj = SPGUI.conftbl.ExportInputPrms();

    accobj[TypeLabel] = GetObj(AccLabel)[TypeLabel];
    srcobj[TypeLabel] = GetObj(SrcLabel)[TypeLabel];
    confobj[TypeLabel] = GetCurrentCalcID();
    if(typeof confobj[TypeLabel] == "undefined"){
        confobj[TypeLabel] = "";
    }

    accobj[OptionLabel] = SPGUI.accopt.ExportInputOptions();
    srcobj[OptionLabel] = SPGUI.srcopt.ExportInputOptions();  
    confobj[OptionLabel] = SPGUI.confopt.ExportInputOptions();

    if(confobj[OptionLabel][ConfigOptionsLabel.accuracy[0]] == CustomLabel){
        let accutypes = CustomAccuracyOpt(accobj, srcobj, confobj, SPGUI.setting[AccuracyLabel]);
        let accuobj = {};
        Object.keys(accutypes).forEach(el => {
            if(accutypes[el]){
                accuobj[el] = SPGUI.setting[AccuracyLabel][el];
            }
        });
        confobj[AccuracyLabel] = accuobj;
    }

    let isgap = confobj[TypeLabel].indexOf(MenuLabels.Kvalue) >= 0;

    if(scanobj != null){
        if(scanobj.hasOwnProperty(SrcLabel)){
            if(scanobj[SrcLabel].hasOwnProperty(SrcPrmsLabel.gap) || 
                scanobj[SrcLabel].hasOwnProperty(SrcPrmsLabel.e1st)){
                    isgap = true;
            }    
        }
    }
    
    if(isgap && srcobj[OptionLabel].hasOwnProperty(SrcOptionsLabel.gaplink[0])){
        if(srcobj[OptionLabel][SrcOptionsLabel.gaplink[0]] == AutomaticLabel){
            srcobj[OptionLabel][SrcOptionsLabel.magconf[0]] 
                = CopyJSON(GetObj(SrcLabel)[OptionLabel][SrcOptionsLabel.magconf[0]]);
        }
    }
    if(!isgap){
        if(srcobj[OptionLabel].hasOwnProperty(SrcOptionsLabel.gaptbl[0])){
            delete srcobj[OptionLabel][SrcOptionsLabel.gaptbl[0]];
        }
        if(srcobj[OptionLabel].hasOwnProperty(SrcOptionsLabel.gaplink[0])){
            delete srcobj[OptionLabel][SrcOptionsLabel.gaplink[0]];
        }
    }

    let calcobj = {};
    calcobj[AccLabel] = accobj;
    calcobj[SrcLabel] = srcobj;
    calcobj[ConfigLabel] = confobj;
    calcobj[OutFileLabel] = CopyJSON(SPGUI.setting[OutFileLabel]);

    if(confobj[OptionLabel].hasOwnProperty(ConfigOptionsLabel.fmateri[0]) || 
            confobj[OptionLabel].hasOwnProperty(ConfigOptionsLabel.amateri[0])){
        if(SPGUI.curr_spjson.hasOwnProperty(FMaterialLabel)){
            if(SPGUI.curr_spjson[FMaterialLabel].toString() == "[object Object]"){
                calcobj[FMaterialLabel] = CopyJSON(SPGUI.curr_spjson[FMaterialLabel]);
            }
        }
    }

    if(scanobj != null){
        calcobj[ScanLabel] = scanobj;
    }
    if(IsFELReuse(calcid)){
        calcobj[FELBunchFactor] = CopyJSON(SPGUI.orginobj[FELBunchFactor]);
    }

    return calcobj;
}

function GetCurrent()
{
    let categs = [AccLabel, SrcLabel, ConfigLabel];

    let blnames = GetObjNames(BLLabel);
    for(let i = 0; i < blnames.length; i++){
        let currbl = SPGUI.curr_spjson[BLLabel][blnames[i]];
        for(let j = 0; j < categs.length; j++){
            let objname = currbl[categs[j]];
            if(SPGUI.curr_spjson[categs[j]].hasOwnProperty(objname)){
                continue;
            }
            let names = GetObjNames(categs[j]);
            currbl[categs[j]] = names[0];
        }    
    }

    let curr = SPGUI.curr_spjson[CurrentLabel];
    if(!SPGUI.curr_spjson[BLLabel].hasOwnProperty(curr)){
        SPGUI.curr_spjson[CurrentLabel] = Object.keys(SPGUI.curr_spjson[BLLabel])[0];
        curr = SPGUI.curr_spjson[CurrentLabel];
    }
    let currbl = SPGUI.curr_spjson[BLLabel][curr];

    return [
            curr,
            currbl[AccLabel],
            currbl[SrcLabel],
            currbl[ConfigLabel]
        ];
}

function CreatePrmSetMenu()
{
    let edmenutmp = CopyJSON(SelectPrmSetMenus);

    let labels = [BLLabel, AccLabel, SrcLabel, ConfigLabel];
    let objs = [];
    for(let j = 0; j < labels.length; j++){
        objs.push(SPGUI.curr_spjson[labels[j]]);
    }

    for(let n = 0; n < objs.length; n++){
        let keys = Object.keys(objs[n]);
        keys.sort();
        for(let j = 0; j < keys.length; j++){
            edmenutmp[n].submenu.push({label:keys[j]});
        }
    }
    let editmenu = SetMenuItems(edmenutmp, PrmSetLabel);
    let currsets = GetCurrent();

    let prmdiv = document.getElementById("prmset");
    prmdiv.innerHTML = "";
    prmdiv.innerText = "Parameter Set";
    let ids = CreateMenuList(editmenu, prmdiv);

    for(let i = 0; i < ids.length; i++){
        let menuitem = document.getElementById(ids[i]);
        menuitem.addEventListener("contextmenu", (e) => {
            let ppmenu = document.getElementById("prmedit");
            ppmenu.style.zIndex = 2;
            ppmenu.style.top = e.pageY+"px";
            ppmenu.style.left = e.pageX+"px";
            ppmenu.style.visibility = "visible";
            SPGUI.targetprmsetid = e.currentTarget.id;
            return false;   
        });
    }

    for(let j = 0; j < labels.length; j++){
        let id = GetIDFromItem(PrmSetLabel, GetIDFromItem(labels[j], currsets[j], -1), -1);
        let menuitem = document.getElementById(id);
        menuitem.className += " selected";
    }
}

function CreateProcess(dataname)
{
    let scanobj = null;
    if(SPGUI.targetid.hasOwnProperty("scanobj")){
        scanobj = CopyJSON(SPGUI.targetid.scanobj);
        scanobj[OptionLabel] = CopyJSON(SPGUI.setting[ScanOptionlabel])
    }
    SPGUI.targetid = {};
    
    if(SPGUI.process_event_origin_id == CreateProcessLabel 
            && SPGUI.calcprocess != null){
        SPGUI.calcprocess.AppendProcess(GetCalcObj(scanobj), dataname);
    }
    else if(SPGUI.calcprocess == null){
        let id = GetIDFromItem(CalcIDLabel, SPGUI.totalcalcs);
        SPGUI.totalcalcs++;
        let cprocdiv = document.getElementById("calcproc");
        let issgl = SPGUI.process_event_origin_id == StartCalcLabel;
        SPGUI.calcprocess = new CalcProcess(issgl, GetCalcObj(scanobj), dataname, id);
        cprocdiv.appendChild(SPGUI.calcprocess.GetList());
        document.getElementById("calcprocfld").classList.replace("none", "flexcont");
    }
    if(SPGUI.process_event_origin_id == StartCalcLabel){
        SPGUI.calcprocess.Start();
        SPGUI.calcprocess = null;
    }
}

function GetCurrentDataName(isfix = false)
{
    let folder = SPGUI.setting[OutFileLabel][OutputOptionsLabel.folder[0]];
    if(folder == "" && SPGUI.runEnvironment == "tauri"){
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.folder[0]] = SPGUI.wdname;
        let idoutdir = GetIDFromItem(OutFileLabel, OutputOptionsLabel.folder[0], -1);
        document.getElementById(idoutdir).value = SPGUI.wdname;
        folder = SPGUI.wdname;
    }
    if(folder != ""){
        let ua = window.navigator.userAgent.toLowerCase();
        if(ua.indexOf("windows") >= 0){
            folder += "\\";
        }
        else{
            folder += "/";
        }
    }
    let fname = SPGUI.setting[OutFileLabel][OutputOptionsLabel.prefix[0]];
    if(SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] >= 0){
        if(fname != ""){
            fname += "-";
        }
        fname += SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]].toString();
    }
    if(fname == ""){
        fname = "untitled";
    }
    fname += ".json";
    return folder+fname;
}

function IncrementSerialNumber()
{
    if(SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]] >= 0){
        SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]]++;
        let id = GetIDFromItem(OutFileLabel, OutputOptionsLabel.serial[0], -1);
        document.getElementById(id).value = 
            SPGUI.setting[OutFileLabel][OutputOptionsLabel.serial[0]];
    }
}

async function SaveParameters(datapath, data, ischng = true)
{
    let result = await window.__TAURI__.tauri.invoke("write_file", 
        { path: datapath, data: data});
    if (result == "" && ischng) {
        SPGUI.setting.prmfile = datapath;
        TWindow.appWindow.setTitle("SPECTRA "+FullVersionNumber+" - "+SPGUI.setting.prmfile);
    }
}

//----- should be consistent with the numerical code -----
function CanBundle(scan2d)
{
    let confobj = GetObj(ConfigLabel);
    let calctype = GetCurrentCalcID();
    if(calctype.indexOf(MenuLabels.XXpYYp) >= 0){
        return false;
    }
    if(calctype.indexOf(MenuLabels.XXpprj) >= 0 || calctype.indexOf(MenuLabels.YYpprj) >= 0){
        return confobj[OptionLabel][ConfigOptionsLabel.CMD[0]] == false;
    }
    else if(calctype.indexOf(MenuLabels.vpdens) >= 0){
        return scan2d == false;
    }
    else if(calctype.indexOf(MenuLabels.cohrad) >= 0){
        if(SPGUI.confopt.IsShown(ConfigOptionsLabel.fel[0])){
            return confobj[OptionLabel][ConfigOptionsLabel.fel[0]] == NoneLabel;
        }
    }
    return true;
}
//-------------------------------------------------------

function ArrangeCurrentPrmSet(eventid, items, newname)
{
    if(items.categ == BLLabel){
        if(SPGUI.curr_spjson[CurrentLabel] == items.item){
            SPGUI.curr_spjson[CurrentLabel] = newname;
        }
    }
    else{
        for(var el in SPGUI.curr_spjson[BLLabel]){
            let obj = SPGUI.curr_spjson[BLLabel][el];       
            if(obj[items.categ] == items.item 
                    && eventid == PrmSetRename){
                obj[items.categ] = newname;
            }
        }
        GetBL()[items.categ] = newname;
    }
    CreatePrmSetMenu();
    ArrangeDataImport(false);
    LoadJSON();
}

function ArrangeDataImport(isinit)
{
    let obj, fid, item;
    let ascii = SPGUI.importformats.fmt;
    let labels = SPGUI.importformats.label;

    let poconfig = {};
    for(let i = 0; i < labels.length; i++){
        fid = labels[i].id;
        item = GetItemFromID(fid);
        if(labels[i].isopt){
            obj = GetObj(item.categ)[OptionLabel];
        }
        else{
            obj = GetObj(item.categ);
        }   
        if(isinit){
            if(fid == GetIDFromItem(SrcLabel, SrcPrmsLabel.fvsz)){
                SPGUI.importdata[fid] = new PlotArea(fid, ascii[fid], FELSecIdxLabel);
            }
            else{
                SPGUI.importdata[fid] = new PlotArea(fid, ascii[fid]);
            }
        }
        SPGUI.importdata[fid].SetJSONObj(obj);
    }
}

function convert_filename2json(prmfile)
{
	let idx = prmfile.lastIndexOf(".prm");
	if(idx >= 0){
		return prmfile.substr(0, idx)+".json";
	}
	return prmfile+".json";
}

function Message(caption, msg, footers)
{
    // message modal dialog
    CreateModalInfDialog("msgDialog", caption, msg, footers);
    document.getElementById("msgDialog").showModal();
}

function MenuCommand(event)
{
    if(document.getElementById(event.currentTarget.id).className.indexOf("disable") >= 0){
        return;
    }
    document.getElementById("menu").focus();
    let menuitem = GetItemFromID(event.currentTarget.id);

    let options  = {
        title: "",
        filters: [{
            name: "JSON",
            extensions: ["json"]
          }]
    };

    if(event.currentTarget.id == StartCalcLabel || 
            event.currentTarget.id == CreateProcessLabel){
        SPGUI.process_event_origin_id = event.currentTarget.id;
        if(event.currentTarget.id == StartCalcLabel 
                && SPGUI.calcprocess != null){
            CreateProcess("");
        }
        else if(GetCurrentCalcID().indexOf(MenuLabels.fixed) 
                >= 0){
            CreateProcess("");
        }
        else{
            let datapath = GetCurrentDataName();
            if(datapath != ""){
                CreateProcess(datapath);
                IncrementSerialNumber();    
            }
        }
    }
    else if(menuitem.item == Scan1D 
            || menuitem.item == Scan2D1D
            || menuitem.item == Scan2D2D)
    {
        let calctype = GetCurrentCalcID();
        if(calctype == ""){
            window.Alert("Select the calculation type first.");
            return;
        }
        let item = GetItemFromID(SPGUI.targetid.id);
        if(!SPGUI.setting.hasOwnProperty(ScanLabel)){
            SPGUI.setting[ScanLabel] = {};
        }
        let obj = SPGUI.setting[ScanLabel];
        if(!obj.hasOwnProperty(item.categ)){
            obj[item.categ] = {};
        }
        let categ = obj[item.categ], values, currval, target;
        if(item.subitem == ""){
            if(!categ.hasOwnProperty(item.item)){
                categ[item.item] = {};
            }
            values = categ[item.item];
            currval = GetObj(item.categ)[item.item];
            target = item.item;
        }
        else{
            if(!categ.hasOwnProperty(OptionLabel)){
                categ[OptionLabel] = {};
            }
            let options = categ[OptionLabel];
            if(!options.hasOwnProperty(item.item)){
                options[item.item] = {};
            }
            let subitem = options[item.item];
            if(!subitem.hasOwnProperty(item.subitem)){
                subitem[item.subitem] = {};
            }
            values = subitem[item.subitem];
            currval = GetObj(item.categ)[OptionLabel][item.item][item.subitem];
            target = item.subitem;
        }
        if(!Array.isArray(values) || values.length < 4){
            if(item.jxy >= 0){
                if(typeof SPGUI.targetid.isscan == "number"){
                    values = 
                    [
                        [currval[0], currval[1]], 
                        [currval[0]+1, currval[1]+1], 
                        [1, 1],
                        [0, 0]
                    ];
                }
                else{
                    values = 
                    [
                        [currval[0]*0.9, currval[1]*0.9], 
                        [currval[0]*1.1, currval[1]*1.1], 
                        [11, 11],
                        [0, 0]
                    ];
                }
            }
            else{
                if(typeof SPGUI.targetid.isscan == "number"){
                    values = [currval, currval+1, 1, 0];
                }
                else{
                    values = [currval*0.9, currval*1.1, 11, 0];
                }
            }
        }
        let scanprmtypes = {};
        ScanLabelOrder.forEach(function(element){
            scanprmtypes[ScanPrmsLabel[element]] = 1;
        });
        if(menuitem.item == Scan1D){
            scanprmtypes[ScanPrmsLabel.initial2] = -1;
            scanprmtypes[ScanPrmsLabel.final2] = -1;
            scanprmtypes[ScanPrmsLabel.points2] = -1;
            scanprmtypes[ScanPrmsLabel.interv2] = -1;
            scanprmtypes[ScanPrmsLabel.iniser2] = -1;
            if(typeof SPGUI.targetid.isscan == "number"){
                scanprmtypes[ScanPrmsLabel.points] = -1;
            }
            else{
                scanprmtypes[ScanPrmsLabel.interv] = -1;
            }
            def_scan[ScanPrmsLabel.initial] = values[0];
            def_scan[ScanPrmsLabel.final] = values[1];
            def_scan[ScanPrmsLabel.points] = values[2];
            def_scan[ScanPrmsLabel.interv] = values[2];
            def_scan[ScanPrmsLabel.iniser] = values[3];
        }
        else{
            scanprmtypes[ScanPrmsLabel.initial] = -1;
            scanprmtypes[ScanPrmsLabel.final] = -1;
            scanprmtypes[ScanPrmsLabel.points] = -1;
            scanprmtypes[ScanPrmsLabel.interv] = -1;
            scanprmtypes[ScanPrmsLabel.iniser] = -1;
            if(typeof SPGUI.targetid.isscan == "number"){
                scanprmtypes[ScanPrmsLabel.points2] = -1;
            }
            else{
                scanprmtypes[ScanPrmsLabel.interv2] = -1;
            }
            def_scan[ScanPrmsLabel.initial2] = values[0];
            def_scan[ScanPrmsLabel.final2] = values[1];
            def_scan[ScanPrmsLabel.points2] = values[2];
            def_scan[ScanPrmsLabel.interv2] = values[2];
            def_scan[ScanPrmsLabel.iniser2] = values[3];
        }
        SPGUI.scanconftbl.JSONObj = def_scan;
        SPGUI.scanconftbl.SetPrmTypes(scanprmtypes);
        if(menuitem.item == Scan2D1D){
            let disables = [false, false];
            disables[1-item.jxy] = true;
            SPGUI.scanconftbl.DisableInput(ScanPrmsLabel.initial2, disables);
            SPGUI.scanconftbl.DisableInput(ScanPrmsLabel.final2, disables);
            SPGUI.scanconftbl.DisableInput(ScanPrmsLabel.iniser2, disables);
            if(typeof SPGUI.targetid.isscan == "number"){
                SPGUI.scanconftbl.DisableInput(ScanPrmsLabel.interv2, disables);
            }
            else{
                SPGUI.scanconftbl.DisableInput(ScanPrmsLabel.points2, disables);
            }
        }

        let scanopttypes = {};
        ScanOptionsOrder.forEach(function(element){
            scanopttypes[ScanOptionsLabel[element][0]] = true;
        });
        scanopttypes[ScanOptionsLabel.bundle[0]] = 
        calctype.indexOf(MenuLabels.fixed) < 0 && CanBundle(menuitem.item == Scan2D2D);
        if(!SPGUI.setting.hasOwnProperty(ScanOptionlabel)){
            SPGUI.setting[ScanOptionlabel] = CopyJSON(def_scan_option);
        }
        let scanoptobj = {};
        scanoptobj[OptionLabel] = SPGUI.setting[ScanOptionlabel];
        SPGUI.scanconfopt.JSONObj = scanoptobj;
        SPGUI.scanconfopt.SetValidity(scanopttypes, {}, {});

        SPGUI.targetid.menuid = menuitem.item;
        SPGUI.targetid.isinteger = typeof SPGUI.targetid.isscan == "number";
        document.getElementById("scanconftitle").innerHTML = item.categ+" - "+target;
        document.getElementById("scanDialog").showModal();
    }
    else if(event.currentTarget.id == ExportSettingLabel){
        let jsonstr;
        if(SPGUI.calcprocess != null){
            jsonstr = SPGUI.calcprocess.ExportProcesses();
        }
        else{
            let calcobj = GetCalcObj(null);
            jsonstr = FormatArray(JSON.stringify(calcobj, null, JSONIndent));
        }
        options.title = "Input a data name to export the parameters.";
        window.__TAURI__.dialog.save(options).then((datapath) => {
            if(datapath != null){
                SaveParameters(datapath, jsonstr, false);
            }
        });

    }
    else if(event.currentTarget.id == FileNew){
        SPGUI.curr_spjson = CopyJSON(def_spjson);
        SPGUI.setting.prmfile = "untitled";
        document.title = "SPECTRA "+FullVersionNumber+" - "+SPGUI.setting.prmfile;
        CreatePrmSetMenu();
        ArrangeDataImport(false);
        LoadJSON();
    }
    else if(event.currentTarget.id == FileOpen){
        options.title = "Open a SPECTRA parameter file.";
        window.__TAURI__.dialog.open(options).then((datapath) => {
            if(datapath != null){
                OpenParameterFile(datapath);
            }
        });
    }
    else if(event.currentTarget.id == FileOpenUp 
            || event.currentTarget.id == FileConvertUp){
        SPGUI.upload_menu = event.currentTarget.id;
        let suffix = event.currentTarget.id == FileOpenUp ? ".json" : "*.*";
        document.getElementById("open").accept = suffix;
        document.getElementById("open").click();
    }
    else if(event.currentTarget.id == FileOpenList){
        SPGUI.socketio.emit("getfiles", "prm");
    }
    else if(event.currentTarget.id == FileImport){
        document.getElementById("import").click();
    }
    else if(event.currentTarget.id == FileConvert){
        options.title = "Open parameter files for older (10 or before) versions.";
        options.filters = [{
            name: "All files",
            extensions: ["*"]
          }];
        options.multiple = true;
            
        window.__TAURI__.dialog.open(options).then((datapaths) => {
            if(datapaths != null){
                let commands = [];
                let jsonfiles = [];
                for(let n = 0; n < datapaths.length; n++){
                    jsonfiles.push(convert_filename2json(datapaths[n]));
                    commands.push(new window.__TAURI__.shell.Command("converter", [datapaths[n], jsonfiles[n]]));
                    commands[n].on("close", (data) => {
                        if(n == datapaths.length-1){
                            jsonfiles.push("hr");
                            Message("Conversion Completed", "New files generated.", jsonfiles);
                            OpenParameterFile(jsonfiles[0]);    
                        }
                        else{
                            commands[n+1].spawn();
                        }
                    });
                }
                commands[0].spawn();
            }
        });

    }
    else if(event.currentTarget.id == FileSave || event.currentTarget.id == FileSaveAs){
        let data = FormatArray(JSON.stringify(SPGUI.curr_spjson, null, JSONIndent));
        if(event.currentTarget.id == FileSaveAs){
            options.title = "Input a data name to save the parameters.";
            window.__TAURI__.dialog.save(options).then((datapath) => {
                if(datapath != null){
                    SaveParameters(datapath, data);
                }
            });
        }
        else{
            SaveParameters(SPGUI.setting.prmfile, data);
        }
    }
    else if(event.currentTarget.id == FileDownload || 
            event.currentTarget.id == SettingDownloadLabel){
        let jsonstr;
        if(event.currentTarget.id == SettingDownloadLabel){
            if(SPGUI.calcprocess != null){
                jsonstr = SPGUI.calcprocess.ExportProcesses();
            }
            else{
                let calcobj = GetCalcObj(null);
                jsonstr = FormatArray(JSON.stringify(calcobj, null, JSONIndent));
            }
        }
        else{
            jsonstr = FormatArray(JSON.stringify(SPGUI.curr_spjson, null, JSONIndent));
        }
        jsonstr = FormatArray(jsonstr);
        let blob = new Blob([jsonstr], {type:"application/json"});
        let link = document.createElement("a");
        link.href = window.URL.createObjectURL(blob);
        link.download = SPGUI.setting.prmfile;
        link.click();
        link.remove();
    }
    else if(event.currentTarget.id == FileExit){
        if(SPGUI.runEnvironment == "tauri"){
            BeforeExit().then((e) => {
                window.__TAURI__.process.exit(0);
            });
        }
        else{
            window.open('','_self').close();
        }
    }
    else if(event.currentTarget.id == EditAccuracy){
        SetAccuracyConfig();
        document.getElementById("accuDialog").showModal();
    }
    else if(event.currentTarget.id == EditMagnet){
        document.getElementById("magDialog").showModal();
    }
    else if(event.currentTarget.id == EditMaterial){
        document.getElementById("fmaterialDialog").showModal();
    }
    else if(event.currentTarget.id == EditUnits){
        document.getElementById("unitDialog").showModal();
    }
    else if(event.currentTarget.id == EditMPIConfig){
        document.getElementById("pcDialog").showModal();
    }
    else if(event.currentTarget.id == EditPhaseErrConfig){
        document.getElementById("perrDialog").showModal();
    }
    else if(event.currentTarget.id == EditFiltPlotConfig){
        document.getElementById("filtplotDialog").showModal();
    }
    else if(event.currentTarget.id == PrmSetDuplicate 
            || event.currentTarget.id == PrmSetRename 
            || event.currentTarget.id == PrmSetDelete)
    {
        if(SPGUI.targetprmsetid.indexOf(PrmSetLabel) == -1){
            Alert("Fatal Error: Valid ID not found in "+SPGUI.targetprmsetid+".");
            return;
        }
        let targetid = SPGUI.targetprmsetid.slice(PrmSetLabel.length+IDSeparator.length);
        let items = GetItemFromID(targetid);
        let obj = SPGUI.curr_spjson[items.categ];
        if(!obj.hasOwnProperty(items.item)){
            Alert("Fatal Error: Parameter set "+items.item+" not found.");
            return;
        }
        let currkeys = Object.keys(obj);
        if(event.currentTarget.id == PrmSetDelete){
            if(currkeys.length <= 1){
                Alert("At least one parameter set should be left.");
                return;
            }
            delete obj[items.item];
            let newname;
            for(let i = 0; i < currkeys.length; i++){
                if(currkeys[i] == items.item){
                    if(i == 0){
                        newname = currkeys[1];
                    }
                    else{
                        newname = currkeys[i-1];
                    }
                    break;
                }
            }
            ArrangeCurrentPrmSet(event.currentTarget.id, items, newname);
        }
        else{
            let psnameobj = {};
            SPGUI.setting[PSNameLabel][PSNameOptionsLabel.newname[0]] = items.item;
            psnameobj[OptionLabel] = SPGUI.setting[PSNameLabel];
            SPGUI.curritem = items;
            SPGUI.psnameopt.JSONObj = psnameobj;
            let pstypes = {[PSNameOptionsLabel.newname[0]]: true};
            SPGUI.psnameopt.SetValidity(pstypes, {}, {});
            SPGUI.eventid = event.currentTarget.id;
            document.getElementById("psDialog").showModal();
            return;
        }
    }
    else if(event.currentTarget.id.indexOf(PrmSetLabel) >= 0){
        let targetid = event.currentTarget.id.slice(
                PrmSetLabel.length+IDSeparator.length);
        let item = GetItemFromID(targetid);
        if(item.categ == BLLabel){
            SPGUI.curr_spjson[CurrentLabel] = item.item;
        }
        else{
            GetBL()[item.categ] = item.item;
        }
        CreatePrmSetMenu();
        ArrangeDataImport(false);
        LoadJSON();
    }
    else if(event.currentTarget.id.indexOf(HelpAbout) >= 0){
        document.getElementById("aboutDialog").showModal();
    }
    else if(event.currentTarget.id.indexOf(HelpReference) >= 0){
        if(SPGUI.runEnvironment ==  "tauri"){
            OpenReference();
        }
        else{
            window.open("help/reference.html");
        }
    }
    else{
        SelectCalculation(event);        
    }
}


