"use strict";

function SplitPath(pathname){
    let delim;
    if(pathname.indexOf("\\") >= 0){
        delim = "\\";
    }
    else if(pathname.indexOf("/") >= 0){
        delim = "/";
    }
    else{
        return {dir:"", name:pathname};
    }
    let items = pathname.split(delim);
    let name = items.pop();
    let dir = items.join(delim);

    return {dir:dir, name:name};
}

function GetAsciiData1D(plobj)
{
    let objs = plobj.data;
    let nplots = objs.length;
    let xarrays = [];
    let legends = [];
    for(let n = 0; n < nplots; n++){
        xarrays.push(objs[n].x);
        if(objs[n].hasOwnProperty("name")){
            legends.push("\""+objs[n].name+"\"");
        }
        else{
            legends.push("-");
        }
    }
    let xidx = [0];
    let xarrr = [xarrays[0]];
    for(let n = 1; n < nplots; n++){
        let xastrn = JSON.stringify(xarrays[n]);
        let mf = -1;
        for(let m = 0; m < n; m++){
            let xastrm = JSON.stringify(xarrays[m]);
            if(xastrn == xastrm){
                mf = m;
                break;
            }
        }
        if(mf >= 0){
            xidx.push(xidx[mf]);
        }
        else{
            xidx.push(xarrr.length);
            xarrr.push(xarrays[n]);
        }
    }
    let xtitle = plobj.layout.xaxis.title.text;
    let ytitle = plobj.layout.yaxis.title.text;

    let yindices = [];
    for(let n = 0; n < xarrr.length; n++){
        yindices.push([]);
    }
    for(let n = 0; n < nplots; n++){
        yindices[xidx[n]].push(n);
    }
    let titles = [];
    let ndata = 0;
    for(let j = 0; j < xarrr.length; j++){
        if(xarrr[j].length > ndata){
            ndata = xarrr[j].length;
        }
        titles.push("\""+xtitle+"\"");
        if(nplots == 1){
            titles.push("\""+ytitle+"\"");
        }
        else{
            for(let i = 0; i < yindices[j].length; i++){
                titles.push(legends[yindices[j][i]]);
            }       
        }
    }

    let lineval = [], values = [titles.join("\t")];
    lineval.length = titles.length;
    for(let n = 0; n < ndata; n++){
        let jt = -1;
        for(let j = 0; j < xarrr.length; j++){
            jt++;
            if(xarrr[j].length <= n){
                lineval[jt] = "-";
            }
            else{
                lineval[jt] = xarrr[j][n].toExponential(5);
            }
            for(let i = 0; i < yindices[j].length; i++){
                jt++;
                if(xarrr[j].length <= n){
                    lineval[jt] = "-";
                }
                else{
                    lineval[jt] = objs[yindices[j][i]].y[n].toExponential(5);
                }
            }       
        }
        values.push(lineval.join("\t"));
    }
    let data = values.join("\n");
    return data;
}

function GetAsciiData2D(plobj, ztitles)
{
    let objs = plobj.data;
    let nplots = objs.length;
    let xarr = objs[0].x;
    let yarr = objs[0].y;
    let zarrays = [];
    for(let n = 0; n < nplots; n++){
        zarrays.push(objs[n].z);
    }
    let xtitle, ytitle;
    if(objs[0].type == "surface"){
        xtitle = plobj.layout.scene.xaxis.title.text;
        ytitle = plobj.layout.scene.yaxis.title.text;
    }
    else{
        xtitle = plobj.layout.xaxis.title.text;
        ytitle = plobj.layout.yaxis.title.text;
    }

    let titles = ["\""+xtitle+"\"", "\""+ytitle+"\""];
    for(let j = 0; j < ztitles.length; j++){
        titles.push("\""+ztitles[j]+"\"");
    }

    let values = [titles.join("\t")];
    for(let ny = 0; ny < yarr.length; ny++){
        for(let nx = 0; nx < xarr.length; nx++){           
            let lineval = [xarr[nx].toExponential(5), yarr[ny].toExponential(5)];
            for(let j = 0; j < zarrays.length; j++){
                lineval.push(zarrays[j][ny][nx].toExponential(5));
            }
            values.push(lineval.join("\t"));
        }    
    }
    let data = values.join("\n");
    return data;
}
 
function GetAsciiData(dim, plobj, ztitles)
{
    let data;
    if(dim == 2){
        data = GetAsciiData2D(plobj, ztitles);
    }
    else{
        data = GetAsciiData1D(plobj);
    }
    return data;
}

function GetPlotConfig0D(obj, ytitles, mode = "lines")
{
    let yindices = [];
    for(let i = 0; i < ytitles.length; i++){
        let yindex = obj.titles.indexOf(ytitles[i]);
        if(yindex < 0){
            return null;
        }
        yindices.push(yindex);
    }
    let xdata = [];
    for(let i = 0; i < obj.data[yindices[0]].length; i++){
        xdata.push(i+1);
    }
    let data = [];
    for(let i = 0; i < ytitles.length; i++){
        data.push(
            {
                x:xdata,
                y:obj.data[yindices[i]],
                type:PlotlyScatterType,
                name:ytitles[i],
                mode: mode
            }
        )
    }
    return data;
}

function GetPlotConfig1D(obj, xtitle, ytitles, legends, mode = "lines")
{
    let xindex = obj.titles.indexOf(xtitle);
    if(xindex < 0){
        return null;
    }
    let yindices = [];
    for(let i = 0; i < ytitles.length; i++){
        let yindex = obj.titles.indexOf(ytitles[i]);
        if(yindex < 0){
            return null;
        }
        yindices.push(yindex);
    }
    let data = [];
    for(let i = 0; i < ytitles.length; i++){
        data.push(
            {
                x: obj.data[xindex],
                y: obj.data[yindices[i]],
                type: PlotlyScatterType,
                name: i >= legends.length ? ytitles[i] : legends[i],
                mode: mode
            }
        )
    }
    return data;
}

function GetLayout1D(axtitles, ranges, logranges, xyscale)
{
    let defaxis = {
            ticks:"inside", zeroline:false, type:"linear",
            showline:true, mirror:true, exponentformat:"power"};
    let xaxis = CopyJSON(defaxis);
    let yaxis = CopyJSON(defaxis);
    xaxis.title = {text:axtitles[0]};
    yaxis.title = {text:axtitles[1]};

    if(xyscale != null){
        xaxis.type = xyscale[0];
        yaxis.type = xyscale[1];
    }
    if(xaxis.type == "log"){
        xaxis.range = logranges[0];
    }
    else{
        if(ranges != null){
            xaxis.range = ranges[0];
        }    
    }
    if(yaxis.type == "log"){
        yaxis.range = logranges[1];
    }
    else{
        if(ranges != null){
            yaxis.range = ranges[1];
        }    
    }

    let legend, margin;
    legend = {x:0.02, y:0.98};
    margin = CopyJSON(PlotlyPrms.margin1d);
    let layout = {
        showlegend:true, 
        autosize:true,
        legend:legend,
        margin:margin,
        xaxis:xaxis,
        yaxis:yaxis,
        hovermode:"closest"
    };
    return layout;
}

function GetLayout2D(nplots, iscontour, xytitles, ztitles, xyranges, zrange)
{
    let rows, cols;
    if(nplots == 1){
        rows = cols = 1;
    }
    else if(nplots == 2){
        rows = 1;
        cols = 2;
    }
    else if(nplots == 3){
        rows = 1;
        cols = 3;
    }
    else{
        rows = Math.ceil(Math.sqrt(nplots));
        cols = rows-1;
        while(rows*cols < nplots){
            cols++;
        }
    }

    let layout;
    if(iscontour){
        let dx = 0.1/Math.max(1, cols-1);
        let dy = 0.1/Math.max(1, rows-1);
        layout = 
            {
                grid:
                {
                    rows:rows, 
                    columns:cols, 
                    pattern:"independent",
                    xgap:dx,
                    ygap:dy
                },
                xaxis:{title:{text:xytitles[0]}},
                yaxis:{title:{text:xytitles[1]}}
            };
            if(xyranges != null){
                layout.xaxis.range = xyranges[0];
                layout.yaxis.range = xyranges[1];
            }
        for(let i = 1; i < nplots; i++){
            let xlabel = "xaxis"+(i+1).toString();
            let ylabel = "yaxis"+(i+1).toString();
            layout[xlabel] = {title:{text:xytitles[0]}};
            layout[ylabel] = {title:{text:xytitles[1]}};
        }
        layout.margin = CopyJSON(PlotlyPrms.margin1d);    
    }
    else{
        let scenes = [];
        let domains = [];
        let serno = 0;
        for(let m = 0; m < rows; m++){
            let ypos = [1.0-(m+1.0)/rows, 1.0-m/rows];
            for(let n = 0; n < cols; n++){
                let xpos = [n/cols, (n+1.0)/cols];
                serno++;
                if(nplots > 1){
                    scenes.push("scene"+serno.toString());
                }
                else{
                    scenes.push("scene");
                }
                let scnobj = {domain:{x:xpos, y:ypos}};         
                scnobj.xaxis = {title:{text:xytitles[0]}};
                scnobj.yaxis = {title:{text:xytitles[1]}};
                scnobj.zaxis = {tickformat: ".2g", title:{text:ztitles[serno-1]}};
                if(zrange != null){
                    scnobj.zaxis.range = zrange;
                }
                scnobj.aspectratio = {x:1, y:1, z:1};
                domains.push(scnobj);
                if(scenes.length == nplots){
                    break;
                }
            }
            if(scenes.length == nplots){
                break;
            }
        }
        layout = {};
        for(let i = 0; i < nplots; i++){
            layout[scenes[i]] = domains[i];
        }
        layout.margin = CopyJSON(PlotlyPrms.margin2d);    
    }
    if(zrange != null){
        layout.zaxis = {range: zrange};
    }
    return layout;
}

function GetPlotConfig2D(obj, titles, axtitles, config, isdata2d)
{
    let xindex = obj[0].titles.indexOf(titles[0]);
    if(xindex < 0){
        return null;
    }
    let yindex = obj[0].titles.indexOf(titles[1]);
    if(yindex < 0){
        return null;
    }
    let nobjs = obj.length;
    let nplots = (titles.length-2)*nobjs;
    let zindices = [];
    for(let i = 0; i < nplots; i++){
        let objidx = i%nobjs;
        let tidx = Math.floor(i/nobjs);
        let zindex = obj[objidx].titles.indexOf(titles[tidx+2]);
        if(zindex < 0){
            return null;
        }
        zindices.push(zindex);
    }
    
    let layout = {
        "surface": GetLayout2D(nplots, false, axtitles, axtitles.slice(2), null, null),
        "heatmap": GetLayout2D(nplots, true, axtitles, axtitles.slice(2), null, null)
    };

    let data = [];
    for(let i = 0; i < nplots; i++){
        let objidx = i%nobjs;

        let zdata = [];
        if(isdata2d){
            zdata = obj[objidx].data[zindices[i]];
        }
        else{
            let mesh = [obj[objidx].data[xindex].length, obj[objidx].data[yindex].length];
            for(let ny = 0; ny < mesh[1]; ny++){
                let mvdidx = GetIndexMDV(mesh, [0, ny], 2);
                zdata.push(obj[objidx].data[zindices[i]].slice(mvdidx, mvdidx+mesh[0]));
            }
        }

        let scntitle = nplots > 1 ? "scene"+(i+1).toString() : "scene";
        let clbar = CopyJSON(PlotlyPrms.colorbar);
        clbar.title = titles[2];
        let sglobj = 
        {
            x: obj[objidx].data[xindex],
            y: obj[objidx].data[yindex],
            z: zdata,
            type: config.type2d,
            zsmooth: "best",
            scene:scntitle,
            showscale: nplots==1,
            colorbar: clbar,
            contours: CopyJSON(PlotlyPrms.heatmap),      
            colorscale: config.colorscale,
        };
        if(i > 0){
            sglobj.xaxis = "x"+(i+1).toString();
            sglobj.yaxis = "y"+(i+1).toString();
        }
        data.push(sglobj);
    }
    return {data:data, layout:layout};
}

function GetIndexMDV(mesh, n, dim)
{
	let index = n[dim-1];
	for(let nd = dim-2; nd >= 0; nd--){
		index = index*mesh[nd]+n[nd];
	}
	return index;
}

function ConvertPow2Super(unit)
{
    let cunit = unit.replace(/\^2/g, "<sup>2</sup>");
    return cunit.replace(/\^3/g, "<sup>3</sup>");
}

function RemoveSuffix(str, suf)
{
    let strlow = str.toLowerCase();
    let suflow = suf.toLowerCase();
    let idx = strlow.lastIndexOf(suflow);
    if(idx >= 0){
        return str.substr(0, idx)
    }
    return str;
}

function ApplySingleSelection(select)
{
    let selected = false;
    for(let i = 0; i < select.options.length; i++){
        if(select.options[i].selected){
            if(!selected){
                selected = true;
            }
            else{
                select.options[i].selected = false;
            }
        }
    }
}

class PostProcessor {
    constructor(){
        this.m_panel = document.createElement("div");
        this.m_panel.className = "flexcont flexcolumn flexstretch";
        this.m_panel.style.width = "200px";
        this.m_onrelayout = false;
        this.m_plotwindow = null;
        this.m_currindex = -1;

        this.m_dataselections = document.createElement("select");
        this.m_dataselections.setAttribute("size", "10");
        this.m_dataselections.addEventListener("change", 
            (e) => {
                this.m_currindex = e.currentTarget.selectedIndex;
                this.SaveFormerSettings();
                this.ShowItems(this.m_currindex);
                this.PlotCurrentResult();
            });

        this.m_datacaptions = document.createElement("div");
        this.m_datacaptions.style.width = "100%";
        this.m_datacaptions.style.whiteSpace = "normal";
        this.m_datacaptions.style.margin = "0";
        this.m_datacaptions.style.wordBreak = "break-all";

        this.m_comcaptions = document.createElement("div");
        this.m_comcaptions.style.width = "100%";
        this.m_comcaptions.style.whiteSpace = "normal";
        this.m_comcaptions.style.margin = "0 0 10px 0";
        this.m_comcaptions.style.overflow = "auto";
        this.m_framecurr = [0, 0];

        this.m_titles = [];
        this.m_units = [];
        this.m_dimensions = [];
        this.m_variables = [];
        this.m_scans = [];
        this.m_details = [];
        this.m_datasets = [];
        this.m_comments = [];
        this.m_datapaths = [];
        this.m_datanames = [];
        this.m_currztitles = "";
        this.m_freader = new FileReader();
        this.m_freader.addEventListener("load", 
            (e) =>{
                if(!this.LoadOutputFile(this.m_freader.result)){
                    SwitchSpinner(false);
                }
            });

        this.m_fdialog = CreateFileDialogElement("", "Import", true, true);
        this.m_fdialog.button.addEventListener("click",
            (e) => {
                this.m_fdialog.file.click();
            });
        this.m_fdialog.file.addEventListener("change",
            (e) => {
                this.SetImportFiles(e.target.files);
                SwitchSpinner(true);
                this.m_freader.readAsText(this.m_importedfiles.files[0]);
            });

        let clrbtn = document.createElement("input");
        clrbtn.setAttribute("type", "button");
        clrbtn.setAttribute("value", "Clear");
        clrbtn.addEventListener("click",
            (e) => {
                this.ClearData();
            });

        this.m_fdialog.body.className = "flexcont flexspace";
        this.m_fdialog.body.appendChild(clrbtn);

        this.m_panel.appendChild(this.m_fdialog.body);
        this.m_panel.appendChild(this.m_dataselections);
        this.m_panel.appendChild(this.m_datacaptions);    
        this.m_panel.appendChild(this.m_comcaptions);    

        let selbox = CreateSelectBox("x axis", 1, false, true);
        this.m_xdiv = selbox.body;
        this.m_xaxis = selbox.select;

        selbox = CreateSelectBox("x-y axis", 1, false, true);
        this.m_xydiv = selbox.body;
        this.m_xyaxis = selbox.select;
        
        selbox = CreateSelectBox("Items to Plot", 5, true, true);
        this.m_zdiv = selbox.body;
        this.m_zaxis = selbox.select;

        this.m_dlbtn = "";
        if(SPGUI.runEnvironment == "node"){
            this.m_dlbtn = document.createElement("button");
            this.m_dlbtn.innerHTML = "Download Output File";
            this.m_dlbtn.className = "none";
            this.m_dlbtn.addEventListener("click", (e)=>{
                let tgt = GetSelections(this.m_dataselections).text[0];
                SPGUI.socketio.emit("load", "data", tgt);
            });
        }
        this.m_expotbtn = document.createElement("button");
        this.m_expotbtn.className = "none";
        this.m_expotbtn.innerHTML = "Export as ASCII";
        this.m_expotbtn.addEventListener("click", (e)=>{
            this.m_plotwindow.ExportAsASCII();
        });

        this.m_newwinbtn = document.createElement("button");
        this.m_newwinbtn.className = "none";
        this.m_newwinbtn.innerHTML = "Open New Window";
        this.m_newwinbtn.addEventListener("click", (e)=>{
            let size = this.m_plotwindow.GetClientSize();
            SPGUI.plotobject = 
                {size:size, window:[this.m_plotwindow], object:null, ispostp:true}; 
            CreateNewplot();

        });

        this.m_xdiv.classList.replace("flexcont", "none");
        this.m_xydiv.classList.replace("flexcont", "none");
        this.m_zdiv.classList.replace("flexcont", "none");

        this.m_panel.appendChild(this.m_xdiv);
        this.m_panel.appendChild(this.m_xydiv);
        this.m_panel.appendChild(this.m_zdiv);
        if(SPGUI.runEnvironment == "node"){
            this.m_panel.appendChild(this.m_dlbtn);
        }

        this.m_panel.appendChild(this.m_expotbtn);
        this.m_panel.appendChild(this.m_newwinbtn);

        let hrdiv = document.createElement("hr");
        this.m_panel.appendChild(hrdiv);
       
        this.m_compenable = false;
        this.m_compdiv = document.createElement("div");
        this.m_compdiv.className = "none flexcolumn";
        this.m_panel.appendChild(this.m_compdiv);

        let chkdiv = document.createElement("div");
        chkdiv.className = "flexcont flexcenter"
        this.m_compchkbox = document.createElement("input");
        this.m_compchkbox.setAttribute("type", "checkbox");
        this.m_compchkbox.id = "comparative_check";
        let label = document.createElement("label");
        label.setAttribute("for", this.m_compchkbox.id);
        label.innerHTML = "Comparative Plot";
        chkdiv.appendChild(this.m_compchkbox);
        chkdiv.appendChild(label);

        this.m_compchkbox.addEventListener("click", (e) => {
            this.CheckComparable(e.currentTarget.checked);
        });
        selbox = CreateSelectBox("Data Available", 5, true, true);
        this.m_complist = selbox.select;
        this.m_complist.className = "none";
        this.m_compdiv.appendChild(chkdiv);
        this.m_compdiv.appendChild(this.m_complist);

        this.m_xaxis.addEventListener("change", 
            (e) => {
                this.PlotCurrentResult();
            });

        this.m_xyaxis.addEventListener("change", 
            (e) => {
                this.SwitchItemSelection(this.m_currindex);
                this.ArrangeComparative();
                this.PlotCurrentResult();
            });
           
        this.m_zaxis.addEventListener("change", 
            (e) => {
                this.ArrangeComparative();
                this.PlotCurrentResult();
            });

        this.m_complist.addEventListener("change", 
            (e) => {
                this.PlotCurrentResult();
            });
    }

    CheckComparable(checked, isclear = false){
        this.m_compenable = checked;
        if(this.m_compenable){
            this.m_complist.classList.remove("none");
        }
        else{
            this.m_complist.classList.add("none");
        }
        this.SwitchItemSelection(this.m_currindex);
        if(isclear){
            return;
        }
        this.PlotCurrentResult();
    }

    Refresh(){
        if(this.m_plotwindow != null){
            this.m_plotwindow.RefreshPlotObject();
        }
    }

    SetImportFiles(files){
        this.m_importedfiles = {files:files, index:0};
    }

    GetPanel(){
        return this.m_panel;
    }

    LoadData(dataobj, comment, dataname, datapath){
        this.m_titles.push(dataobj[DataTitlesLabel]);
        this.m_units.push(dataobj[UnitsLabel]);
        this.m_dimensions.push(dataobj[DataDimLabel]);
        if(dataobj.hasOwnProperty(VariablesLabel)){
            if(dataobj[VariablesLabel] == 0){
                this.m_variables.push(dataobj[DataDimLabel]);
                this.m_scans.push(0);
            }
            else{
                this.m_variables.push(dataobj[VariablesLabel]);
                this.m_scans.push(dataobj[DataDimLabel]-dataobj[VariablesLabel]);
            }
        }
        else{
            this.m_variables.push(dataobj[DataDimLabel]);
            this.m_scans.push(0);
        }
        if(dataobj.hasOwnProperty(DetailsLabel)){
            this.m_details.push(dataobj[DetailsLabel]);
        }
        else{
            this.m_details.push([]);
        }

        if(dataobj.hasOwnProperty(RelateDataLabel)){
            let reldata = dataobj[RelateDataLabel];
            if(comment != ""){
                reldata.Comment = comment;
            }
            this.m_comments.push(reldata);
        }
        else{
            this.m_comments.push(comment);
        }
        this.m_datasets.push(dataobj[DataLabel]);

        this.m_datapaths.push(datapath);
        this.m_datanames.push(dataname);
        let option = document.createElement("option");
        option.text = SplitPath(dataname).name;
        this.m_dataselections.appendChild(option);
    }

    DeleteItem(index){
        this.m_titles.splice(index, 1);
        this.m_units.splice(index, 1);
        this.m_dimensions.splice(index, 1);
        this.m_variables.splice(index, 1);
        this.m_scans.splice(index, 1);
        this.m_details.splice(index, 1);
        this.m_comments.splice(index, 1);
        this.m_datasets.splice(index, 1);
        this.m_datapaths.splice(index, 1);
        this.m_datanames.splice(index, 1);
        this.m_dataselections.remove(index);
    }

    LoadOutputFile(result){
        let fileobj;
        try{
            fileobj = JSON.parse(result);
        }
        catch (e){
            Alert("Error: Cannot load the output file.\n"+e.message);
            return false;
        }
        if(!fileobj.hasOwnProperty(OutputLabel)){
            Alert("No result is saved.");
            return false;
        }

        let comments  = [];
        let datanames = [];
        let dataobjs = [];
        let datapaths = [];

        let comment = "";
        if(fileobj.hasOwnProperty(InputLabel)){
            if(fileobj[InputLabel].hasOwnProperty(OutFileLabel)){
                let label = OutputOptionsLabel.comment[0];
                if(fileobj[InputLabel][OutFileLabel].hasOwnProperty(label)){
                    comment = fileobj[InputLabel][OutFileLabel][label];
                }
            }
        }
        let dataname = this.m_importedfiles.files[this.m_importedfiles.index].name;
        let outobj = fileobj[OutputLabel];
        
        comments.push(comment);
        datanames.push(dataname);
        datapaths.push(dataname);
        dataobjs.push(outobj);

        let cmderr = "";
        if(fileobj.hasOwnProperty(CMDResultLabel)){
            if(fileobj[CMDResultLabel].hasOwnProperty(CMDErrorLabel)){
                cmderr = fileobj[CMDResultLabel][CMDErrorLabel];
            }
        }

        let cmdoptions = [
                CMDModalFluxLabel, CMDFieldLabel, CMDIntensityLabel, CMDCompareIntLabel, 
                CMDCompareXLabel, CMDCompareYLabel, 
                FELCurrProfile, FELEtProfile, FELCurrProfileR56, FELEtProfileR56, FELBunchFactor, 
                FELPulseEnergy, FELEfield, FELInstPower, FELSpectrum];
        for(let j = 0; j < cmdoptions.length; j++){
            if(fileobj.hasOwnProperty(cmdoptions[j])){
                comments.push(cmderr);
                datanames.push(dataname+IDSeparator+cmdoptions[j]);
                datapaths.push(dataname);
                dataobjs.push(fileobj[cmdoptions[j]]);      
            }
        }
        for(let j = 0; j < dataobjs.length; j++){
            for(let i = this.m_datapaths.length; i >= 0; i--){
                if(this.m_datapaths[i] == datapaths[j] && 
                        this.m_datanames[i] == datanames[j]){
                    this.DeleteItem(i);
                }
            }
            this.LoadData(dataobjs[j], comments[j], datanames[j], datapaths[j]);
        }

        let options = this.m_dataselections.options;
        options[options.length-1].selected = true;
        this.m_currindex = this.m_titles.length-1;
        this.ShowItems(this.m_currindex);

        this.m_importedfiles.index++;
        // load until "m_importedfiles.index" reaches the number of files loaded
        if(this.m_importedfiles.index == this.m_importedfiles.files.length){
            this.m_fdialog.file.value = "";
            SwitchSpinner(false);
        }
        else{
            this.m_freader.readAsText(
                this.m_importedfiles.files[this.m_importedfiles.index]);
        }
        return true;
    }

    //-------->>>>>>>>
    RemoveData(value)
    {
        let idx = this.m_objnames.indexOf(value);
        if(idx < 0){
            return;
        }
        this.m_objnames.splice(idx, 1);
        this.m_dataobjs.splice(idx, 1);
        this.DeleteOutputFile(value);
        for(let n = this.m_allids.length-1; n >= 0; n--){
            if(this.m_allids[n].split("/")[0] == value){
                this.m_allids.splice(n, 1);
            }
        }
        if(this.m_objnames.length == 0){
            this.ClearData();
        }
        else{
            SetSelection(this.m_outfiles, this.m_objnames[0], false);
            this.ChangeDataname(this.m_objnames[0]);
        }
    }

    ClearData(){
        this.m_titles.length = 0;
        this.m_units.length = 0;
        this.m_dimensions.length = 0;
        this.m_variables.length = 0;
        this.m_scans.length = 0;
        this.m_details.length = 0;
        this.m_datasets.length = 0;
        this.m_comments.length = 0;
        this.m_datapaths.length = 0;
        this.m_datanames.length = 0;
        this.m_dataselections.innerHTML = "";
        this.m_comcaptions.innerHTML = "";
        this.m_datacaptions.innerHTML = "";
        this.m_xcurr = "";
        this.m_xycurr = "";
        this.m_zcurr = "";
        this.m_compenable = false;
        this.m_compchkbox.checked = false;
        this.CheckComparable(false, true);

        this.m_xdiv.classList.replace("flexcont", "none");
        this.m_xydiv.classList.replace("flexcont", "none");
        this.m_zdiv.classList.replace("flexcont", "none");
        this.m_compdiv.classList.replace("flexcont", "none");
        this.m_expotbtn.classList.add("none");
        this.m_newwinbtn.classList.add("none");
        if(SPGUI.runEnvironment == "node"){
            this.m_dlbtn.classList.add("none");
        }
                
        this.ShowItems(-1);
        this.m_currindex = -1;

        let pp_plot = document.getElementById("pp_plot");
        pp_plot.innerHTML = "";
    }

    GetPlotConfig(index){
        let dims = this.m_scans[index]+this.m_variables[index];
        if(this.m_details[index].length > 0){
            // plot x: variable, frame: scan
            if(dims == 2){
                this.m_plotdims = [
                    [[0], [1]],
                    // skip [[1], [0]] to avoid discrepancy in # variables
                    [[0, 1], []]
                ];
            }
            else if(dims == 3){
                this.m_plotdims = [
                    [[0], [1, 2]]
                ];
            }
            else{
                this.m_plotdims = [];
            }
        }
        else if(dims == 0 || dims > 4){
            this.m_plotdims = [];
            return false;
        }
        else if(dims == 4){
            this.m_plotdims = SPGUI.idxpair4d;
        }
        else if(dims == 3){
            this.m_plotdims = SPGUI.idxpair3d;
        }
        else if(dims == 2){
            if(this.m_scans[index] == 0){
                this.m_plotdims = [
                    [[0, 1], []]
                ];
            }
            else{
                this.m_plotdims = SPGUI.idxpair2d;
            }
        }
        else{
            this.m_plotdims = [];
        }
        return true;
    }

    ShowItems(index){
        this.m_xaxis.innerHTML = "";
        this.m_xyaxis.innerHTML = "";
        this.m_zaxis.innerHTML = "";
        this.m_complist.innerHTML = "";
        if(index < 0){
            return;
        }

        this.m_datacaptions.innerHTML = this.m_datapaths[index];
        this.m_comcaptions.innerHTML = "";
        if(typeof this.m_comments[index] == "object"){
            let keys = Object.keys(this.m_comments[index]);
            for(let j = 0; j < keys.length; j++){
                let cont = document.createElement("div");

                let comobj = this.m_comments[index][keys[j]];
                if(Array.isArray(comobj)){
                    if(comobj.length > 2){
                        let tmpobj = [];
                        for(let i = 0; i < 2; i++){
                            tmpobj.push(comobj[i].toString());
                        }
                        tmpobj.push("...");
                        comobj = tmpobj.join(",");
                    }
                }
                cont.innerHTML = keys[j]+": "+comobj;
                this.m_comcaptions.appendChild(cont);
            }
        }
        else if(this.m_comments[index] != ""){
            this.m_comcaptions.innerHTML = "\""+this.m_comments[index]+"\"";
        }

        if(!this.GetPlotConfig(index)){
            return;
        }

        let items = this.m_titles[index];
        if(this.m_plotdims.length > 0){
            for(let i = 0; i < this.m_plotdims.length; i++){
                let option = document.createElement("option");
                if(this.m_plotdims[i].length == 0){
                    continue;
                }
                else if(this.m_plotdims[i][0].length == 1){
                    option.text = items[this.m_plotdims[i][0][0]];
                }
                else{
                    option.text = "("+items[this.m_plotdims[i][0][0]]+", "+items[this.m_plotdims[i][0][1]]+")";
                }
                this.m_xyaxis.appendChild(option);
            }   
            this.m_xydiv.classList.replace("none", "flexcont");
        }
        else{
            this.m_xydiv.classList.replace("flexcont", "none");
        }
        this.SetTargetItems(index);
        this.SwitchItemSelection(index);
        this.RestoreFormerSettings();
        this.ArrangeComparative();
        this.PlotCurrentResult();
    }

    SaveFormerSettings(){
        this.m_xcurr = "";
        this.m_xycurr = "";
        this.m_zcurr = "";
        this.m_framecurr = [0, 0];
        if(this.m_xdiv.classList.contains("flexcont")){
            let sel = GetSelections(this.m_xaxis);
            this.m_xcurr = sel.text[0];
        }
        if(this.m_xydiv.classList.contains("flexcont")){
            let sel = GetSelections(this.m_xyaxis);
            this.m_xycurr = sel.text[0];            
        }
        if(this.m_zdiv.classList.contains("flexcont")){
            let sel = GetSelections(this.m_zaxis);
            this.m_zcurr = sel.text[0];                        
        }
    }

    RestoreFormerSettings(){
        if(this.m_xdiv.classList.contains("flexcont") && this.m_xcurr != ""){
            SetSelection(this.m_xaxis, this.m_xcurr, false);
        }
        if(this.m_xydiv.classList.contains("flexcont") && this.m_xycurr != ""){
            SetSelection(this.m_xyaxis, this.m_xycurr, false);
        }
        if(this.m_zdiv.classList.contains("flexcont") && this.m_zcurr != ""){
            SetSelection(this.m_zaxis, this.m_zcurr, false);
        }
    }

    SetTargetItems(index){
        let dim = this.GetFrameConfig();
        let items = this.m_titles[index];
        if(dim.plot == 2){
            for(let i = this.m_dimensions[index]; i < items.length; i++){
                let option = document.createElement("option");
                option.text = items[i];
                this.m_zaxis.appendChild(option);
            }
            this.m_xdiv.classList.replace("flexcont", "none");
            this.m_zdiv.classList.replace("none", "flexcont");
            this.m_zaxis.options[0].selected = true;
        }
        else{
            if(this.m_xydiv.classList.contains("flexcont")){
                this.m_xdiv.classList.replace("flexcont", "none");
                for(let i = this.m_dimensions[index]; i < items.length; i++){
                    let option = document.createElement("option");
                    option.text = items[i];
                    this.m_zaxis.appendChild(option);
                }
                this.m_zaxis.options[0].selected = true;
            }
            else{
                let fidx = -1, pidx = -1;
                this.m_xdiv.classList.replace("none", "flexcont");
                for(let i = 0; i < items.length; i++){
                    let option = document.createElement("option");
                    option.text = items[i];
                    this.m_xaxis.appendChild(option);
                    option = document.createElement("option");
                    option.text = items[i];
                    this.m_zaxis.appendChild(option);
                    if(fidx < 0 && items[i].indexOf("Flux") >= 0){
                        fidx = i;
                    }
                    if(pidx < 0 && items[i].indexOf("Power") >= 0){
                        pidx = i;
                    }
                }
                if(fidx >= 0){
                    this.m_zaxis.options[fidx].selected = true;
                }
                else if(pidx >= 0){
                    this.m_zaxis.options[pidx].selected = true;
                }
                else{
                    this.m_zaxis.options[this.m_dimensions[index]].selected = true;
                }
            }
            this.m_zdiv.classList.replace("none", "flexcont");
        }
        this.m_expotbtn.classList.remove("none");
        this.m_newwinbtn.classList.remove("none");
        if(SPGUI.runEnvironment == "node"){
            this.m_dlbtn.classList.remove("none");
        }
    }

    SwitchItemSelection(index){
        let dim = this.GetFrameConfig();
        let issingle = this.m_compenable || 
            (dim.frame > 0 && (dim.plot > 1 || this.m_details[index].length > 0));

        if(issingle){
            this.m_zaxis.size = 1;
            this.m_zaxis.multiple = false;
            ApplySingleSelection(this.m_zaxis);
        }
        else{
            this.m_zaxis.size = 5;
            this.m_zaxis.multiple = true;    
        }
    }

    IsComparable(index){
        if(this.m_scans[this.m_currindex] != this.m_scans[index]){
            return false;
        }
        if(this.m_variables[this.m_currindex] != this.m_variables[index]){
            return false;
        }
        let dim = this.m_scans[this.m_currindex]+this.m_variables[this.m_currindex];
        for(let j = 0; j < dim; j++){
            if(this.m_datasets[this.m_currindex][j].length 
                != this.m_datasets[index][j].length){
                    return false;
            }
        }
        return true;
    }

    ArrangeComparative(){
        let dim = this.GetFrameConfig();
        let iscomp = true;

        if(this.m_dataselections.length < 2){
            // number of data sets == 1
            iscomp = false;
        }
        else if(dim.frame > 0){
            if(dim.plot > 1){
                // 2d animation 
                iscomp = false;
            }
            else{
                // 1d animation & details
                iscomp = this.m_details[this.m_currindex].length == 0;
            }
        }

        if(!iscomp){
            this.m_compdiv.classList.replace("flexcont", "none");
            return;
        }
        else{
            this.m_compdiv.classList.replace("none", "flexcont");
        }

        let index = GetSelections(this.m_dataselections).index[0];
        let titles = [];
        if(dim.frame > 0){
            titles.push(GetSelections(this.m_xyaxis).text[0]);
        }
        else if(this.m_dimensions[index] == 1){
            titles.push(GetSelections(this.m_xaxis).text[0]);
        }
        else{
            titles.push(this.m_titles[index][0]);
            titles.push(this.m_titles[index][1]);
        }
        titles.push(GetSelections(this.m_zaxis).text[0]);
        this.m_compdataidx = []; 
        for(let i = 0; i < this.m_titles.length; i++){
            if(i == index){
                continue;
            }
            let isng = this.m_dimensions[i] > 2 || 
                this.m_dimensions[i] != this.m_variables[i];
            if(dim.frame > 0){
                if(!this.IsComparable(i) && isng){
                    continue;
                }
            }
            else if(isng){
                continue;
            }
            let found = true;
            for(let j = 0; j < titles.length; j++){
                if(titles[j] == "Energy"){
                    found = found && 
                        (this.m_titles[i].indexOf("Energy") >= 0 || 
                         this.m_titles[i].indexOf("Harmonic Energy") >= 0);
                }
                else{
                    found = found && this.m_titles[i].indexOf(titles[j]) >= 0;
                }
                if(!found){
                    break;
                }
            }
            if(found){
                this.m_compdataidx.push(i);
            }
        }
        this.m_complist.innerHTML = "";
        for(let i = 0; i < this.m_compdataidx.length; i++){
            let option = document.createElement("option");
            option.text = this.m_dataselections.options[this.m_compdataidx[i]].text;
            this.m_complist.appendChild(option);
        }
    }

    ApplytUnitZtitles(index){
        for(let j = 0; j < this.m_currztitles.length; j++){
            let tindex = this.m_titles[index].indexOf(this.m_currztitles[j]);
            if(tindex >= 0){
                let unit = this.m_units[index][tindex];
                if(unit != ""){
                    this.m_currztitles[j] = this.m_currztitles[j]+" ("+unit+")";
                }
            }
        }
    }

    PlotCurrentResult(){
        let pp_plot = document.getElementById("pp_plot");
        pp_plot.innerHTML = "";

        let primarydata = GetSelections(this.m_dataselections);
        let index = primarydata.index[0];
        if(index < 0){
            return;
        }
        let dataname = primarydata.text[0];

        let fconf = this.GetFrameConfig();     
        let titles;
        let axtitles = [];
        let axindices = [];
        let plobj = {};

        let frameindices = [];
        if(fconf.frame > 0){
            frameindices = [0, 0];       
            let axes = CopyJSON(this.m_plotdims[fconf.xyidx][0]);
            let frames = CopyJSON(this.m_plotdims[fconf.xyidx][1]);

            for(let j = 0; j < axes.length; j++){
                axindices.push(axes[j]);
            }

            if(frames.length == 1){
                frames.push(-1);
            }            
            if(axes.length == 1){
                axes.push(-1);
            }
            let itemidx, itemtexts = GetSelections(this.m_zaxis);
            itemidx = itemtexts.index;
            for(let j = 0; j < itemidx.length; j++){
                itemidx[j] += this.m_dimensions[index];
                axindices.push(itemidx[j]);
            }
            axtitles = this.GetAxisTitles(index, axindices, axes[1] < 0);

            this.m_currdetails = this.m_details[index];
            this.m_detvars = [];
            let comptarget = {index:[], text:[]};
            let companim = [this.m_currindex];
            let comptmp = GetSelections(this.m_complist);
            if(this.m_currdetails.length > 0){
                this.ArrangeSliceData(index, frames, itemidx[0]);
            }
            else if(this.m_compenable){
                for(let i = 0; i < comptmp.index.length; i++){
                    let compidx = comptmp.index[i];
                    if(this.IsComparable(this.m_compdataidx[compidx])){
                        companim.push(this.m_compdataidx[compidx]);
                    }
                    else{
                        comptarget.index.push(comptmp.index[i]);
                        comptarget.text.push(comptmp.text[i]);
                    }
                }
                if(companim.length == 1){
                    this.ArrangeSliceDataSingle(index, axes, frames, itemidx[0]);
                }
                else{
                    this.ArrangeSliceData(index, frames, itemidx[0], companim);
                }
            }
            else if(itemidx.length > 1){
                this.ArrangeSliceData(index, frames, itemidx);
            }
            else{
                this.ArrangeSliceDataSingle(index, axes, frames, itemidx[0]);
            }

            if(axes[1] < 0){
                if(companim.length > 1){
                    plobj.animlegend = [];
                    for(let n = 0; n < companim.length; n++){
                        plobj.animlegend.push(this.m_dataselections.options[companim[n]].text);
                    }
                }
                else if(this.m_currdetails.length > 0){
                    plobj.animlegend = this.m_currdetails;
                }
                else{
                    if(itemidx.length > 1){
                        plobj.animlegend = itemtexts.text;
                    }
                    else{
                        plobj.animlegend = itemtexts.text[0];
                    }
                }
                plobj.dimension = 1;

                this.m_compdata = [];
                if(this.m_compenable){
                    let xtl = this.m_titles[index][axindices[0]];
                    let ytl = [this.m_titles[index][axindices[1]]];
                    let legend, obj;
                    for(let i = 0; i < comptarget.index.length; i++){
                        let compidx = this.m_compdataidx[i];
                        if(this.m_details[compidx].length > 0){
                            for(let j = 0; j < this.m_details[compidx].length; j++){
                                if(this.m_titles[compidx].indexOf(xtl) < 0 && xtl == "Energy"){
                                    xtl = "Harmonic Energy";
                                }
                                obj = {data:this.m_datasets[compidx][j], titles:this.m_titles[compidx]};
                                legend = [this.m_details[compidx][j]];
                                let mdata = GetPlotConfig1D(obj, xtl, ytl, legend);
                                this.m_compdata.push(mdata[0]);
                            }
                        }
                        else{
                            obj = {data:this.m_datasets[compidx], titles:this.m_titles[compidx]};
                            legend = [comptarget.text[i]];
                            let mdata = GetPlotConfig1D(obj, xtl, ytl, legend);
                            this.m_compdata.push(mdata[0]);
                        }
                    }
                }
                plobj.compdata = this.m_compdata;
            }
            else{   
                let ztitles = GetSelections(this.m_zaxis);
                this.m_currztitles = CopyJSON(ztitles.text);
                this.ApplytUnitZtitles(index)
                plobj.dimension = 2;
                plobj.ztitle = ztitles.text[0];
            }
            plobj.detvars = this.m_detvars;
            plobj.vardata = this.m_vardata;
            plobj.data = this.m_slicedata;
            plobj.ftitles = this.m_ftitles;
            plobj.fdata = this.m_fdata;
            plobj.axtitles = axtitles;
            this.m_plotwindow = new PlotWindow(pp_plot, "plotcurrent", 
                plobj, SPGUI.setting[POLabel]["plotcurrent"], dataname, frameindices);
            return;
        }
        else if(fconf.plot == 1){
            plobj.dimension = 1;
            let xaxes = GetSelections(this.m_xaxis);
            let yaxes = GetSelections(this.m_zaxis);
            titles = xaxes.text.concat(yaxes.text);
            if(titles.length < 2){
                return;
            }            
            axindices.push(xaxes.index[0]);
            for(let j = 0; j < yaxes.text.length; j++){
                axindices.push(yaxes.index[j]);
            }
            axtitles = this.GetAxisTitles(index, axindices, true);
        }
        else{
            plobj.dimension = 2;
            let axes = [];
            for(let j = 0; j < 2; j++){
                axes.push(this.m_titles[index][j]);
                axindices.push(j);
            }
            let ztitles = GetSelections(this.m_zaxis);
            this.m_currztitles = CopyJSON(ztitles.text);
            this.ApplytUnitZtitles(index)
            for(let j = 0; j < ztitles.index.length; j++){
                axindices.push(ztitles.index[j]+2);
            }
            titles = axes.concat(ztitles.text);
            if(titles.length < 3){
                return;
            }
            axtitles = this.GetAxisTitles(index, axindices, false);
        }
        let obj = [];
        let legprefix = [];
        if(this.m_details[index].length > 0){
            for(let i = 0; i < this.m_details[index].length; i++){
                obj.push({data:this.m_datasets[index][i], titles:this.m_titles[index]});
                legprefix.push(this.m_details[index][i]);
                if(this.m_compenable){
                    legprefix[i] += ": "+primarydata.text[0]
                }
            }
        }
        else{
            obj.push({data:this.m_datasets[index], titles:this.m_titles[index]});
        }      
        if(this.m_compenable){
            if(legprefix.length == 0){                
                legprefix.push(RemoveSuffix(primarydata.text[0], ".json"));
            }
            let comptarget = GetSelections(this.m_complist);
            for(let i = 0; i < comptarget.index.length; i++){
                let compidx = this.m_compdataidx[comptarget.index[i]];
                if(this.m_details[compidx].length > 0){
                    for(let k = 0; k < this.m_details[compidx].length; k++){
                        obj.push({data:this.m_datasets[compidx][k], titles:this.m_titles[compidx]});
                        legprefix.push(this.m_details[compidx][k]);
                        legprefix[legprefix.length-1] += ": "+comptarget.text[i];
                    }
                }
                else{
                    obj.push({data:this.m_datasets[compidx], titles:this.m_titles[compidx]});
                    legprefix.push(RemoveSuffix(comptarget.text[i], ".json"));
                }
                if(plobj.dimension == 2){
                    axtitles.push(axtitles[axtitles.length-1]);
                }
            }
            if(plobj.dimension == 2 && legprefix.length > 1){
                this.m_currztitles = CopyJSON(legprefix);
                this.ApplytUnitZtitles(index)
            }
        }

        plobj.legprefix = legprefix;
        plobj.titles = titles;
        plobj.axtitles = axtitles;
        plobj.data = obj;
        plobj.dimension = this.m_dimensions[index];
        plobj.isdata2d = false;
        this.m_plotwindow = new PlotWindow(pp_plot, "plotcurrent", 
            plobj, SPGUI.setting[POLabel]["plotcurrent"], dataname, frameindices);
    }

    SetFrames(dataidx, frames){
        this.m_fdata = [];
        this.m_ftitles = [];
        let nframes = [];
        for(let j = 0; j < 2; j++){
            if(frames[j] < 0){
                nframes.push(1);
                this.m_fdata.push([0]);
                this.m_ftitles.push("");
            }
            else{
                if(this.m_currdetails.length > 0){
                    nframes.push(this.m_datasets[dataidx][0][frames[j]].length);
                    this.m_fdata.push(this.m_datasets[dataidx][0][frames[j]]);
                }
                else{
                    nframes.push(this.m_datasets[dataidx][frames[j]].length);
                    this.m_fdata.push(this.m_datasets[dataidx][frames[j]]);
                }
                this.m_ftitles.push(this.m_titles[dataidx][frames[j]]+"=");
            }
        }
        return nframes;
    }

    ArrangeSliceData(dataidx, frames, tgtindices, compidx = []){
        let targetidx, ncomps;
        let isarrtgt = Array.isArray(tgtindices);
        if(isarrtgt){
            ncomps = tgtindices.length;
            compidx = new Array(ncomps);
            for(let n = 0; n < ncomps; n++){
                compidx[n] = dataidx;
            }
        }
        else{
            ncomps = compidx.length;
            targetidx = tgtindices;            
        }

        let nframes = this.SetFrames(dataidx, frames);
        let ndetails = this.m_currdetails.length;
        let nvars = [];
        for(let j = 0; j < ncomps; j++){
            this.m_detvars.push(this.m_datasets[compidx[j]][0]);
            nvars.push(this.m_detvars[j].length);
        }
        for(let j = 0; j < ndetails; j++){
            this.m_detvars.push(this.m_datasets[dataidx][j][0]);
            nvars.push(this.m_detvars[j+ncomps].length);
        }
        if(nvars.length == 0){
            return;
        }
        
        this.m_slicedata = new Array(nframes[1]);
        for(let f1 = 0; f1 < nframes[1]; f1++){
            this.m_slicedata[f1] = new Array(nframes[0]);
            for(let f0 = 0; f0 < nframes[0]; f0++){
                this.m_slicedata[f1][f0] = new Array(ndetails+ncomps);
                for(let j = 0; j < ndetails+ncomps; j++){
                    this.m_slicedata[f1][f0][j] = new Array(nvars[j]);
                }    
            }
        }

        let nd = [0];
        for(let i = 1; i < 3; i++){
            if(i >= this.m_dimensions[dataidx]){
                nd.push(1);
            }
            else{
                // number of scans
                if(ncomps > 0){
                    nd.push(this.m_datasets[compidx[0]][i].length);
                }
                else{
                    nd.push(this.m_datasets[dataidx][0][i].length);
                }
            }
        }

        let n = [0,0,0], f1, f0, datar, ridx;

        for(n[2] = 0; n[2] < nd[2]; n[2]++){
            for(n[1] = 0; n[1] < nd[1]; n[1]++){
                for(let j = 0; j < ndetails+ncomps; j++){
                    if(isarrtgt){
                        targetidx = tgtindices[j];
                    }
                    for(n[0] = 0; n[0] < nvars[j]; n[0]++){
                        nd[0] = nvars[j];
                        ridx = GetIndexMDV(nd, n, 3);
                        f0 = n[frames[0]];
                        if(frames[1] < 0){
                            f1 = 0;
                        }
                        else{
                            f1 = n[frames[1]];
                        }
                        if(j < ncomps){
                            datar = this.m_datasets[compidx[j]][targetidx][ridx];
                        }
                        else{
                            datar = this.m_datasets[dataidx][j-ncomps][targetidx][ridx];
                        }
                        this.m_slicedata[f1][f0][j][n[0]] = datar;
                    }
                }
            }    
        }
    }

    ArrangeSliceDataSingle(dataidx, axes, frames, targetidx){
        let nframes = this.SetFrames(dataidx, frames);
        this.m_vardata = [];
        let mesh = [];
        for(let j = 0; j < 2; j++){
            if(axes[j] < 0){
                mesh.push(1);
                this.m_vardata.push([0]);
            }
            else{
                mesh.push(this.m_datasets[dataidx][axes[j]].length);    
                this.m_vardata.push(this.m_datasets[dataidx][axes[j]]);
            }
        }
        let nd = [];
        for(let i = 0; i < 4; i++){
            if(i >= this.m_dimensions[dataidx]){
                nd.push(1);
            }
            else{
                nd.push(this.m_datasets[dataidx][i].length);
            }
        }

        this.m_slicedata = new Array(nframes[1]);
        for(let f1 = 0; f1 < nframes[1]; f1++){
            this.m_slicedata[f1] = new Array(nframes[0]);
            for(let f0 = 0; f0 < nframes[0]; f0++){
                this.m_slicedata[f1][f0] = new Array(mesh[1]);
                for(let m1 = 0; m1 < mesh[1]; m1++){
                    this.m_slicedata[f1][f0][m1] = new Array(mesh[0]);
                }    
            }
        }

        let n = [0,0,0,0], f1, f0, m1, m0, datar, ridx;
        for(n[3] = 0; n[3] < nd[3]; n[3]++){
            for(n[2] = 0; n[2] < nd[2]; n[2]++){
                for(n[1] = 0; n[1] < nd[1]; n[1]++){
                    for(n[0] = 0; n[0] < nd[0]; n[0]++){
                        f0 = n[frames[0]];
                        m0 = n[axes[0]];
                        ridx = GetIndexMDV(nd, n, 4);
                        if(frames[1] < 0){
                            f1 = 0;
                        }
                        else{
                            f1 = n[frames[1]];
                        }
                        if(axes[1] < 0){
                            m1 = 0;
                        }
                        else{
                            m1 = n[axes[1]];
                        }
                        datar = this.m_datasets[dataidx][targetidx][ridx];
                        this.m_slicedata[f1][f0][m1][m0] = datar;
                    }
                }                    
            }   
        }
    }

    GetFrameConfig(){
        let plotdim, framedim, xyidx;
        if(this.m_xydiv.classList.contains("flexcont")){
            xyidx = GetSelections(this.m_xyaxis).index[0];
            plotdim = this.m_plotdims[xyidx][0].length;
            framedim = this.m_plotdims[xyidx][1].length;   
        }
        else{
            xyidx = -1;
            if(this.m_plotdims.length == 0){
                plotdim = 1;
            }
            else{
                plotdim = this.m_plotdims[0].length;
            }
            framedim = 0;
        }
        return {plot:plotdim, frame:framedim, xyidx:xyidx};
    }

    GetAxisTitles(dataidx, axindices, is1d)
    {
        let axtitles = [];
        let unitless = is1d && axindices.length > 2;     
        for(let j = 0; j < axindices.length; j++){
            let i = axindices[j];
            if(unitless && j > 1){
                axtitles[1] += ", "+this.m_titles[dataidx][i];
                continue;
            }
            let unit = "";
            if(!unitless || j == 0){
                unit = ConvertPow2Super(this.m_units[dataidx][i]);
            }
            if(unit != ""){
                axtitles.push(this.m_titles[dataidx][i]+" ("+unit+")");
            }
            else{
                axtitles.push(this.m_titles[dataidx][i]);
            }
        }
        return axtitles;
    }
}
