"use strict";

// ---------- functions to arrange GUI ----------
function Is3DField(srcobj)
{
    let srctype = srcobj[TypeLabel];
    switch(srctype){
        case SrcTypeLabel.WLEN_SHIFTER:
        case SrcTypeLabel.BM:
        case SrcTypeLabel.WIGGLER:
        case SrcTypeLabel.EMPW:
            return false;
        case SrcTypeLabel.FIELDMAP3D:
            return true;
        case SrcTypeLabel.CUSTOM:
            return srcobj[OptionLabel][SrcOptionsLabel.natfocus[0]] != NoneLabel;
    }
    if(srcobj[OptionLabel][SrcOptionsLabel.segment_type[0]] != NoneLabel){
        if(srcobj[OptionLabel][SrcOptionsLabel.perlattice[0]]){
            return true;
        }
    }
    if(srcobj[OptionLabel][SrcOptionsLabel.natfocus[0]] != NoneLabel){
        return true;
    }
    return false;
}

function Output4CMD(spjson)
{
    let cmdcont = {wigner:false, result:false, id:""};
    if(spjson.hasOwnProperty(InputLabel) == false){
        return cmdcont;
    }
    if(spjson.hasOwnProperty(OutputLabel) == false){
        return cmdcont;
    }
    if(spjson[InputLabel][ConfigLabel].hasOwnProperty(OrgTypeLabel)){
        cmdcont.id = spjson[InputLabel][ConfigLabel][OrgTypeLabel];
    }
    else{
        cmdcont.id = spjson[InputLabel][ConfigLabel][TypeLabel];
    }
    if(cmdcont.id.indexOf(MenuLabels.XXpprj) >= 0 || 
        cmdcont.id.indexOf(MenuLabels.YYpprj) >= 0 || 
        cmdcont.id.indexOf(MenuLabels.XXpYYp) >= 0)
    {
        cmdcont.wigner = true;
    }
    if(cmdcont.wigner){
        cmdcont.result = spjson.hasOwnProperty(CMDResultLabel);
    }
    return cmdcont;
}

function IsFELBunchFactor(spjson)
{
    return spjson.hasOwnProperty(FELBunchFactor);
}

// ---------- Accelerator ----------
function AccPrmTypes(acctype, bunchtype)
{
    let accprmtypes = {};
    if(bunchtype == ImportPaticleLabel){
        AccLabelOrder.forEach(function(element){
            accprmtypes[AccPrmsLabel[element]] = -1;
        });    
        accprmtypes[AccPrmsLabel.eGeV] = 1;
        accprmtypes[AccPrmsLabel.gaminv] = 0;
        accprmtypes[AccPrmsLabel.bunchcharge] = 1;
        if(acctype == AccTypeLabel.RING){
            accprmtypes[AccPrmsLabel.imA] = 1;
        }
        else{
            accprmtypes[AccPrmsLabel.pulsepps] = 1;
        }
        return accprmtypes;
    }
    AccLabelOrder.forEach(function(element){
        accprmtypes[AccPrmsLabel[element]] = 1;
    });

    accprmtypes[AccPrmsLabel.cirm] = 2;
    accprmtypes[AccPrmsLabel.bunches] = 2;
    accprmtypes[AccPrmsLabel.peakcurr] = 0;
    accprmtypes[AccPrmsLabel.epsilon] = 0;
    accprmtypes[AccPrmsLabel.sigma] = 0;
    accprmtypes[AccPrmsLabel.sigmap] = 0;
    accprmtypes[AccPrmsLabel.gaminv] = 0;
    accprmtypes[AccPrmsLabel.minsize] = -1; // hidden parameter
    if(acctype == AccTypeLabel.RING){
        accprmtypes[AccPrmsLabel.aimA] = -1;
        accprmtypes[AccPrmsLabel.pulsepps] = -1;
        accprmtypes[AccPrmsLabel.bunchcharge] = -1;
    }
    else{
        accprmtypes[AccPrmsLabel.aimA] = 0;
        accprmtypes[AccPrmsLabel.imA] = -1;
        accprmtypes[AccPrmsLabel.cirm] = -1;
        accprmtypes[AccPrmsLabel.bunches] = -1;
    }

    if(bunchtype == CurrProfLabel || bunchtype == EtProfLabel){
        accprmtypes[AccPrmsLabel.aimA] = -1;
        accprmtypes[AccPrmsLabel.cirm] = -1;
        accprmtypes[AccPrmsLabel.bunches] = -1;
        accprmtypes[AccPrmsLabel.bunchlength] = -1;
        accprmtypes[AccPrmsLabel.bunchcharge] = -1;
        accprmtypes[AccPrmsLabel.peakcurr] = -1;
    }
    if(bunchtype == EtProfLabel){
        accprmtypes[AccPrmsLabel.espread] = -1;
    }
    else{
        accprmtypes[AccPrmsLabel.R56add] = -1;
    }

    return accprmtypes;
}

function AccOptTypes(accobj, srcobj, confobj)
{
    let accopttypes = {};

    AccOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){            
            accopttypes[AccOptionsLabel[element][0]] = true;
        }
    });
    
    let bunchtype = 
        accobj[OptionLabel][AccOptionsLabel.bunchtype[0]];
    let injectype = 
        accobj[OptionLabel][AccOptionsLabel.injectionebm[0]];
    accopttypes[AccOptionsLabel.bunchdata[0]] 
        = bunchtype == ImportPaticleLabel;
    accopttypes[AccOptionsLabel.currdata[0]] 
        = bunchtype == CurrProfLabel;
    accopttypes[AccOptionsLabel.Etdata[0]] 
        = bunchtype == EtProfLabel;

    let injopt = true;
    if(confobj[TypeLabel] != ""){
        injopt = confobj[TypeLabel].indexOf(MenuLabels.far) < 0;
    }

    accopttypes[AccOptionsLabel.injectionebm[0]] = injopt;
    accopttypes[AccOptionsLabel.injecprm[0]] 
        = injopt && injectype == CustomLabel;
                    
    // single e-, zero emittance, energy spread
    let calcid = "";
    if(confobj != "" && confobj.hasOwnProperty(TypeLabel)){
        calcid = confobj[TypeLabel];
    }
    if(typeof calcid == "undefined"){
        calcid = "";
    }

    let isfel = false;
    if(calcid.indexOf(MenuLabels.cohrad) >= 0 && 
        confobj[OptionLabel][ConfigOptionsLabel.fel[0]] != NoneLabel)
    {
        let conftypes = ConfigPrmOptTypes(accobj, srcobj, confobj);
        isfel = conftypes[1][ConfigOptionsLabel.fel[0]];
    }

    if(calcid == "" || isfel || accopttypes[AccOptionsLabel.bunchdata[0]]){
        accopttypes[AccOptionsLabel.zeroemitt[0]]  = false;
        accopttypes[AccOptionsLabel.singlee[0]] = false;
        accopttypes[AccOptionsLabel.zerosprd[0]] = false;
    }
    else{
        if(calcid.indexOf(MenuLabels.cohrad) >= 0 && 
        (
            calcid.indexOf(MenuLabels.efield) >= 0
            || calcid.indexOf(MenuLabels.camp) >= 0
        ))
        {
            accopttypes[AccOptionsLabel.singlee[0]] = true;
        }
        else{
            accopttypes[AccOptionsLabel.singlee[0]] = false;
        }
        if(calcid.indexOf(MenuLabels.pdensa) >= 0 || 
            calcid.indexOf(MenuLabels.pdenss) >= 0 ||
            calcid.indexOf(MenuLabels.ppower)
        ){
            accopttypes[AccOptionsLabel.zerosprd[0]] = false;
        }
        if(calcid.indexOf(MenuLabels.far) >= 0 
                || calcid.indexOf(MenuLabels.Kvalue) >= 0){
            accopttypes[AccOptionsLabel.bunchtype[0]] = false;
            accopttypes[AccOptionsLabel.currdata[0]] = false;
    
        }
        accopttypes[AccOptionsLabel.zeroemitt[0]] 
            = calcid.indexOf(MenuLabels.tflux)  < 0;
        
        if(accobj[OptionLabel][AccOptionsLabel.bunchtype[0]] == EtProfLabel ||
            accobj[OptionLabel][AccOptionsLabel.bunchtype[0]] == ImportPaticleLabel
        )
        {
            accopttypes[AccOptionsLabel.zerosprd[0]] = false;
        }
        else{
            if(srcobj[TypeLabel] == SrcTypeLabel.CUSTOM){
                accopttypes[AccOptionsLabel.zerosprd[0]] = 
                    confobj[OptionLabel][ConfigOptionsLabel.wiggapprox[0]] == false && 
                    confobj[OptionLabel][ConfigOptionsLabel.esmooth[0]] == false;
            }
            else{
                accopttypes[AccOptionsLabel.zerosprd[0]] = 
                calcid.indexOf(MenuLabels.cohrad) >= 0 ||
                (
                    srcobj[TypeLabel] != SrcTypeLabel.BM &&
                    srcobj[TypeLabel] != SrcTypeLabel.WIGGLER &&
                    srcobj[TypeLabel] != SrcTypeLabel.EMPW &&
                    srcobj[TypeLabel] != SrcTypeLabel.WLEN_SHIFTER
                );    
            }
        }
        if(accopttypes[AccOptionsLabel.singlee[0]] 
                && accobj[OptionLabel][AccOptionsLabel.singlee[0]]){
            accopttypes[AccOptionsLabel.zeroemitt[0]]  = false;
            accopttypes[AccOptionsLabel.zerosprd[0]] = false;
        }
    }
    return accopttypes;
}

function InjectionPrmTypes()
{
    let injprmtypes = {};
    for(let j = 0; j < InjectionLabelOrder.length; j++){
        injprmtypes[InjectionPrmsLabel[InjectionLabelOrder[j]]] = 1;
    }
    return injprmtypes;
}

// ---------- Light Source ----------
function GetSrcContents(srcobj)
{
    let srctype = srcobj[TypeLabel];
    let isgap, isgaplink, isbxy, isb, islu, isper, isapple, isKxy, isK;
    let ise1st, isund, iswiggler;

    isgap = ise1st = 3; // scan possible, but not an input parameter
    islu = isb = isK = 1;
    isund = isper = 0;
    isbxy = isapple = isKxy = iswiggler = -1;
    switch(srctype){
        case SrcTypeLabel.BM:
        case SrcTypeLabel.WLEN_SHIFTER:
        case SrcTypeLabel.CUSTOM:
        case SrcTypeLabel.FIELDMAP3D:
            isgap = islu = isK = ise1st = isund = isper = -1;
            break;
        case SrcTypeLabel.LIN_UND:
        case SrcTypeLabel.VERTICAL_UND:
        case SrcTypeLabel.HELICAL_UND:
            isb  = 2;
            break;
        case SrcTypeLabel.ELLIPTIC_UND:
        case SrcTypeLabel.FIGURE8_UND:
        case SrcTypeLabel.VFIGURE8_UND:
            isbxy = 2;
            isKxy = 1;
            isb = isK = -1;
            break;
        case SrcTypeLabel.MULTI_HARM_UND:
            isgap = isb = isK = -1;
            isKxy = 1;
            break;
        case SrcTypeLabel.EMPW:
            isb = isK = ise1st = isund = -1;
            isbxy = isKxy = 1;
            iswiggler = 0;
            break;
        case SrcTypeLabel.WIGGLER:
            ise1st = isund = -1;
            iswiggler = 0;
            break;
        case SrcTypeLabel.CUSTOM_PERIODIC:
            isgap = isb = isK = -1;
            islu = isKxy = ise1st = 0;
            break;
    }
    if(srcobj[OptionLabel][SrcOptionsLabel.apple[0]] 
            && srctype == SrcTypeLabel.ELLIPTIC_UND)
    {
        isbxy = isKxy = 0;
        isapple = 1;
    }
    isgaplink = isgap;
    if(srcobj[OptionLabel][SrcOptionsLabel.gaplink[0]] == NoneLabel){
        isgap = -1;
    }

    let srccont = {
        isgap:isgap, isgaplink:isgaplink, isbxy:isbxy, isb:isb, islu:islu, isper:isper, 
        isapple:isapple, isKxy:isKxy, isK:isK, ise1st:ise1st, isund:isund, iswiggler:iswiggler
    };
    return srccont;
}

function SrcPrmTypes(accobj, srcobj, srccont)
{
    let srctype = srcobj[TypeLabel];
    let twinbm = srcobj[OptionLabel][SrcOptionsLabel.bmtandem[0]];
    let srcprmtypes = {};

    let bunchtype = accobj[OptionLabel][AccOptionsLabel.bunchtype[0]];
    let iAitem = 0;
    let emittitem = 0;
    if(bunchtype == ImportPaticleLabel){
        emittitem = -1;
    }
    if(accobj[TypeLabel] == AccTypeLabel.LINAC 
            && bunchtype != GaussianLabel){
        iAitem = -1;
    }

    if(srctype == SrcTypeLabel.BM 
            || srctype == SrcTypeLabel.WLEN_SHIFTER
            || srctype == SrcTypeLabel.CUSTOM
            || srctype == SrcTypeLabel.FIELDMAP3D)
    {
        for(let j = 0; j < SrcLabelOrder.length; j++){
            srcprmtypes[SrcPrmsLabel[SrcLabelOrder[j]]] = -1;
        }
        if(srctype == SrcTypeLabel.CUSTOM){
            srcprmtypes[SrcPrmsLabel.fvsz] = PlotObjLabel;
            srcprmtypes[SrcPrmsLabel.tpower] = iAitem;
            return srcprmtypes;
        }
        else if(srctype == SrcTypeLabel.FIELDMAP3D){
            srcprmtypes[SrcPrmsLabel.fmap] = FileLabel;
            return srcprmtypes;
        }
        srcprmtypes[SrcPrmsLabel.ec] = 0;
        srcprmtypes[SrcPrmsLabel.lc] = 0;
        if(srctype == SrcTypeLabel.BM){
            srcprmtypes[SrcPrmsLabel.tpowerrev] = iAitem;
            srcprmtypes[SrcPrmsLabel.linpower] = iAitem;
            srcprmtypes[SrcPrmsLabel.b] = 1;
            srcprmtypes[SrcPrmsLabel.radius] = 1;
            srcprmtypes[SrcPrmsLabel.bendlength] = 1;
            srcprmtypes[SrcPrmsLabel.fringelen] = 1;
            srcprmtypes[SrcPrmsLabel.bminterv] = twinbm?1:-1;
        }
        else{
            srcprmtypes[SrcPrmsLabel.tpower] = iAitem;
            srcprmtypes[SrcPrmsLabel.bmain] = 1;
            srcprmtypes[SrcPrmsLabel.mplength] = 1;
            srcprmtypes[SrcPrmsLabel.subpolel] = 1;
            srcprmtypes[SrcPrmsLabel.subpoleb] = 1;
        }
        return srcprmtypes;
    }

    let idxsigr = SrcLabelOrder.indexOf("sigmar");
    for(let j = 0; j < SrcLabelOrder.length; j++){
        srcprmtypes[SrcPrmsLabel[SrcLabelOrder[j]]] 
            = j < idxsigr ? 1 : 0;
    }
    srcprmtypes[SrcPrmsLabel.fmap] = -1;
    srcprmtypes[SrcPrmsLabel.fvsz] = -1;
    srcprmtypes[SrcPrmsLabel.bminterv] = -1;

    let ismulth = srctype == SrcTypeLabel.MULTI_HARM_UND;

    srcprmtypes[SrcPrmsLabel.gap] = srccont.isgap;
    srcprmtypes[SrcPrmsLabel.bxy] = srccont.isbxy;
    srcprmtypes[SrcPrmsLabel.b] = srccont.isb;
    srcprmtypes[SrcPrmsLabel.bmain] = -1;
    srcprmtypes[SrcPrmsLabel.lu] = srccont.islu;
    srcprmtypes[SrcPrmsLabel.devlength] = 1;
    srcprmtypes[SrcPrmsLabel.reglength] = 0;
    srcprmtypes[SrcPrmsLabel.periods] = srccont.isper;
    srcprmtypes[SrcPrmsLabel.Kxy0] = srccont.isapple;
    srcprmtypes[SrcPrmsLabel.phase] = srccont.isapple;
    srcprmtypes[SrcPrmsLabel.Kxy] = srccont.isKxy;
    srcprmtypes[SrcPrmsLabel.K] = srccont.isK;
    srcprmtypes[SrcPrmsLabel.Kperp] = srccont.isKxy >= 0 ? 0 : -1;
    srcprmtypes[SrcPrmsLabel.e1st] = srccont.ise1st;
    srcprmtypes[SrcPrmsLabel.radius] = -1;
    srcprmtypes[SrcPrmsLabel.bendlength] = -1;
    srcprmtypes[SrcPrmsLabel.mplength] = -1;
    srcprmtypes[SrcPrmsLabel.fringelen] = -1;
    srcprmtypes[SrcPrmsLabel.subpolel] = -1;
    srcprmtypes[SrcPrmsLabel.subpoleb] = -1;
    srcprmtypes[SrcPrmsLabel.csrorg] = -1;
    srcprmtypes[SrcPrmsLabel.fvsz1per] = 
        srctype == SrcTypeLabel.CUSTOM_PERIODIC ? PlotObjLabel : -1;
    srcprmtypes[SrcPrmsLabel.multiharm] = ismulth ? GridLabel : -1;

    srcprmtypes[SrcPrmsLabel.sigmar] 
        = srccont.isund;
    srcprmtypes[SrcPrmsLabel.sigmarx] 
        = srccont.iswiggler;
    srcprmtypes[SrcPrmsLabel.sigmary] 
        = srccont.iswiggler;
    srcprmtypes[SrcPrmsLabel.Sigmax] = emittitem;
    srcprmtypes[SrcPrmsLabel.Sigmay] = emittitem;
    srcprmtypes[SrcPrmsLabel.lambda1] = srccont.isund;
    srcprmtypes[SrcPrmsLabel.fd] = -1;
    srcprmtypes[SrcPrmsLabel.flux] 
        = iAitem < 0 ? -1 : srccont.isund;
    srcprmtypes[SrcPrmsLabel.brill] 
        = iAitem < 0 || emittitem < 0 ? -1 : srccont.isund;
    srcprmtypes[SrcPrmsLabel.pkbrill] 
        = iAitem < 0 || emittitem < 0 ? -1 : srccont.isund;
    srcprmtypes[SrcPrmsLabel.degener] 
        = iAitem < 0 || emittitem < 0 ? -1 : srccont.isund;
    srcprmtypes[SrcPrmsLabel.ec] = srccont.iswiggler;
    srcprmtypes[SrcPrmsLabel.lc] = srccont.iswiggler;
    srcprmtypes[SrcPrmsLabel.tpower] = iAitem;
    srcprmtypes[SrcPrmsLabel.tpowerrev] = -1;
    srcprmtypes[SrcPrmsLabel.linpower] = -1;

    return srcprmtypes;
}

function SrcOptTypes(accobj, srcobj, calcid)
{
    let srcoptjson = srcobj[OptionLabel];
    let srcprmtypes = [];
    let srccont = GetSrcContents(srcobj);
    srcprmtypes = SrcPrmTypes(accobj, srcobj, srccont);

    let srcopttypes = {};
    SrcOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){            
            srcopttypes[SrcOptionsLabel[element][0]] = true;
        }
    });
    
    let isubuitin = srccont.ise1st > 0
    let isund = srccont.isund >= 0;
    let gaplabel = SrcOptionsLabel.gaplink[0];
    let isseg = srcoptjson[SrcOptionsLabel.segment_type[0]] != NoneLabel;
    let isperr = srcoptjson[SrcOptionsLabel.phaseerr[0]];
    srcopttypes[gaplabel] = srccont.isgaplink >= 0;
    srcopttypes[SrcOptionsLabel.gaptbl[0]]
        = srccont.isgap >= 0 && srcoptjson[gaplabel] == ImpGapTableLabel;

    srcopttypes[SrcOptionsLabel.apple[0]]
        = srcobj[TypeLabel] == SrcTypeLabel.ELLIPTIC_UND;

    srcopttypes[SrcOptionsLabel.field_str[0]] 
        = srcprmtypes[SrcPrmsLabel.lu] >= 0 
        && srcobj[TypeLabel] != SrcTypeLabel.CUSTOM_PERIODIC;
    srcopttypes[SrcOptionsLabel.endmag[0]] = srcprmtypes[SrcPrmsLabel.lu] >= 0;

    if(calcid.indexOf(MenuLabels.far) >= 0 
            || calcid.indexOf(MenuLabels.Kvalue) >= 0){
        srcopttypes[SrcOptionsLabel.natfocus[0]] = false;
        srcopttypes[SrcOptionsLabel.fielderr[0]] = false;
        srcopttypes[SrcOptionsLabel.ferrprm[0]] = false; 
        srcopttypes[SrcOptionsLabel.phaseerr[0]] = false;
        srcopttypes[SrcOptionsLabel.perrprm[0]] = false;
        srcopttypes[SrcOptionsLabel.bmtandem[0]] = false;
    }
    else{
        srcopttypes[SrcOptionsLabel.natfocus[0]] = isund;
        srcopttypes[SrcOptionsLabel.fielderr[0]] = isund;
        srcopttypes[SrcOptionsLabel.ferrprm[0]] 
            = isund && srcoptjson[SrcOptionsLabel.fielderr[0]];
        srcopttypes[SrcOptionsLabel.phaseerr[0]] = isund && !isseg;
        srcopttypes[SrcOptionsLabel.perrprm[0]] 
            = isund && !isseg && srcoptjson[SrcOptionsLabel.phaseerr[0]];
        srcopttypes[SrcOptionsLabel.bmtandem[0]]
            = srcobj[TypeLabel] == SrcTypeLabel.BM;
    }

    if(calcid.indexOf(MenuLabels.simpcalc) >= 0){
        srcopttypes[SrcOptionsLabel.segment_type[0]] = false;
        srcopttypes[SrcOptionsLabel.perlattice[0]] = false;
    }
    else if(calcid.indexOf(MenuLabels.Kvalue) >= 0){
        srcopttypes[SrcOptionsLabel.perlattice[0]] = false;
    }
    else{
        srcopttypes[SrcOptionsLabel.segment_type[0]] = isubuitin && !isperr;
        srcopttypes[SrcOptionsLabel.perlattice[0]] = 
            isubuitin && !isperr && srcoptjson[SrcOptionsLabel.segment_type[0]] != NoneLabel;
    }
    srcopttypes[SrcOptionsLabel.segprm[0]] = srcopttypes[SrcOptionsLabel.segment_type[0]] 
        && srcoptjson[SrcOptionsLabel.segment_type[0]] != NoneLabel;

    if(calcid.indexOf(MenuLabels.vpdens) >= 0){
        srcopttypes[SrcOptionsLabel.natfocus[0]] = false;
        srcopttypes[SrcOptionsLabel.perlattice[0]] = false;
    }

    srcopttypes[SrcOptionsLabel.magconf[0]] = false;
        // do not show in the parameter table panel

    return srcopttypes;
}

function FerrPrmTypes()
{
    let ferrprmtypes = {};
    for(let j = 0; j < FerrLabelOrder.length; j++){
        ferrprmtypes[FerrPrmsLabel[FerrLabelOrder[j]]] = 1;
    }
    return ferrprmtypes;
}

function PerrPrmTypes()
{
    let perrprmtypes = {};
    for(let j = 0; j < PerrLabelOrder.length; j++){
        perrprmtypes[PerrPrmsLabel[PerrLabelOrder[j]]] = 1;
    }
    return perrprmtypes;
}

function SegPrmTypes(srcobj)
{
    let segprmtypes = {};
    for(let j = 0; j < SegLabelOrder.length; j++){
        segprmtypes[SegPrmsLabel[SegLabelOrder[j]]] = -1;
    }
    segprmtypes[SegPrmsLabel.pslip] = 0;
    segprmtypes[SegPrmsLabel.interval] = 1;
    let segtype = srcobj[OptionLabel][SrcOptionsLabel.segment_type[0]];
    if(segtype == IdenticalLabel){
        segprmtypes[SegPrmsLabel.segments] = 1;
        segprmtypes[SegPrmsLabel.phi0] = 1;
    }
    else{
        segprmtypes[SegPrmsLabel.hsegments] = 1;
        segprmtypes[SegPrmsLabel.phi12] = 1;
    }
    if(srcobj[OptionLabel][SrcOptionsLabel.perlattice[0]]){
        segprmtypes[SegPrmsLabel.mdist] = 1;
    }
    return segprmtypes;
}

//---------- Configurations ----------
function CustomAccuracyOpt(accobj, srcobj, confobj, acuobj)
{
    let loadebm = accobj[OptionLabel][AccOptionsLabel.bunchtype[0]] == ImportPaticleLabel;
    let is3dsrc = Is3DField(srcobj);
    let srccont = GetSrcContents(srcobj);
    let conftypes = ConfigPrmOptTypes(accobj, srcobj, confobj)
    let isfel = conftypes[1][ConfigOptionsLabel.fel[0]] 
        && confobj[OptionLabel][ConfigOptionsLabel.fel[0]] != NoneLabel;
    let calcid = confobj[TypeLabel];
    let iscoh = calcid.indexOf(MenuLabels.cohrad) >= 0;
    let isfar = calcid.indexOf(MenuLabels.far) >= 0;
    let issrcpoint = calcid.indexOf(MenuLabels.srcpoint)  >= 0;
    let iswigner = calcid.indexOf(MenuLabels.wigner) >= 0;
    let srctype = srcobj[TypeLabel];
    let iswigapprox = conftypes[1][ConfigOptionsLabel.wiggapprox[0]] && 
        confobj[OptionLabel][ConfigOptionsLabel.wiggapprox[0]];

    let accutypes = {};
    AccuracyOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){            
            accutypes[AccuracyOptionsLabel[element][0]] = true;
        }
    });  

    accutypes[AccuracyOptionsLabel.accdisctra[0]] = 
        !isfar || srctype != SrcTypeLabel.CUSTOM_PERIODIC || iswigapprox;

    accutypes[AccuracyOptionsLabel.accineE[0]] = iscoh;
    accutypes[AccuracyOptionsLabel.acclimeE[0]] = iscoh;
    accutypes[AccuracyOptionsLabel.accconvMCcoh[0]] = iscoh && !isfel;
    accutypes[AccuracyOptionsLabel.accconvMC[0]] = 
        (is3dsrc || loadebm) && !accutypes[AccuracyOptionsLabel.accconvMCcoh[0]];
    accutypes[AccuracyOptionsLabel.acclimMCpart[0]] = is3dsrc || loadebm || (iscoh && !isfel);
    accutypes[AccuracyOptionsLabel.accMCpart[0]] = 
        accutypes[AccuracyOptionsLabel.acclimMCpart[0]]
        && acuobj[AccuracyOptionsLabel.acclimMCpart[0]];
    accutypes[AccuracyOptionsLabel.accconvharm[0]] = isfar && srccont.ise1st >= 0;

    accutypes[AccuracyOptionsLabel.accinsrc[0]] = issrcpoint || isfel;
    accutypes[AccuracyOptionsLabel.acclimsrc[0]] = issrcpoint || isfel;
    accutypes[AccuracyOptionsLabel.accdiscsrc[0]] = (issrcpoint && !iswigner) || isfel;

    accutypes[AccuracyOptionsLabel.accinobs[0]] = !issrcpoint;
    accutypes[AccuracyOptionsLabel.acclimobs[0]] = !issrcpoint;

    accutypes[AccuracyOptionsLabel.accEcorr[0]] = isfel;
        
    return accutypes;
}

function ConfigPrmOptTypes(accobj, srcobj, confobj)
{
    let confprmtypes = {};
    let confopttypes = {};
    ConfigLabelOrder.forEach(function(element){
        confprmtypes[ConfigPrmsLabel[element]] = -1;
    });
    ConfigOptionsOrder.forEach(function(element){
        if(element != SeparatorLabel){            
            confopttypes[ConfigOptionsLabel[element][0]] = false;
        }
    });

    let calcid = confobj[TypeLabel];
    if(typeof calcid == "undefined"){
        calcid = "";
    }
    if(calcid == ""){
        return [confprmtypes, confopttypes];
    } 

    if(calcid.indexOf(MenuLabels.CMD) >= 0){
        let cmdcont = Output4CMD(SPGUI.orginobj);
        calcid = cmdcont.id;
    }

    let srccont = GetSrcContents(srcobj);
    let iswigner = calcid.indexOf(MenuLabels.wigner) >= 0;
    let istf = calcid.indexOf(MenuLabels.tflux)  >= 0;
    let isspd = calcid.indexOf(MenuLabels.spdens)  >= 0;
    let iscoh = calcid.indexOf(MenuLabels.cohrad)  >= 0;
    let istime = calcid.indexOf(MenuLabels.temporal)  >= 0;
    let isFourier = iscoh && istime && 
            confobj[OptionLabel][ConfigOptionsLabel.fouriep[0]];
    let issrcpoint = calcid.indexOf(MenuLabels.sprof)  >= 0;
    let isedep = calcid.indexOf(MenuLabels.energy) >= 0;
    let isvoldens = calcid.indexOf(MenuLabels.vpdens) >= 0;
    let isfixedp = calcid.indexOf(MenuLabels.fixed) >= 0;
    let isbmlike = false;
    let issimple = calcid.indexOf(MenuLabels.simpcalc)  >= 0;
    let isKdep = calcid.indexOf(MenuLabels.Kvalue)  >= 0;
    let isallharm = calcid.indexOf(MenuLabels.allharm) >= 0;

    // boolean for show info.
    let isshow = {};
   
    let iscustomsrc = false;
    let isbm = false;
    switch(srcobj[TypeLabel]){
        case SrcTypeLabel.BM:
            isbm = true;
        case SrcTypeLabel.WIGGLER:
        case SrcTypeLabel.EMPW:
        case SrcTypeLabel.WLEN_SHIFTER:
            isbmlike = true;
            break;
        case SrcTypeLabel.CUSTOM:
            iscustomsrc = true;
            isbmlike = 
                confobj[OptionLabel][ConfigOptionsLabel.wiggapprox[0]] ||
                confobj[OptionLabel][ConfigOptionsLabel.esmooth[0]];
            break;
    } 
    let iswiggler = isbmlike && 
        srcobj[TypeLabel] != SrcTypeLabel.WLEN_SHIFTER;
    let iswigapprox = calcid.indexOf(MenuLabels.far) >= 0 
        && srccont.isund >= 0 && confobj[OptionLabel][ConfigOptionsLabel.wiggapprox[0]];

    //--- parameters ---
    // z positions
    confprmtypes[ConfigPrmsLabel.slit_dist] = 1;
    if(iswigner || istf || (isspd && !isfixedp) || isFourier || issrcpoint || issimple){
        confprmtypes[ConfigPrmsLabel.slit_dist] = -1;
    }
    if(isspd && !isfixedp){
        confprmtypes[ConfigPrmsLabel.zrange] = 1;
        confprmtypes[ConfigPrmsLabel.zmesh] = 1;
    }

    // energy 
    if(isedep){
        confprmtypes[ConfigPrmsLabel.erange] = 1;
        if(isbmlike || iswigapprox){
            confprmtypes[ConfigPrmsLabel.emesh] = 1;
        }
        else{
            confprmtypes[ConfigPrmsLabel.de] = 1;
        }
    }
    if(isallharm){
        confprmtypes[ConfigPrmsLabel.erange] = 1;
        confprmtypes[ConfigPrmsLabel.emesh] = 1;
    }

    if(isvoldens && srcobj[TypeLabel] == SrcTypeLabel.CUSTOM && isbmlike){
        confprmtypes[ConfigPrmsLabel.epitch] = 1;
    }

    let isefix, isdetune = false;
    if(calcid.indexOf(MenuLabels.spatial) >= 0){
        isefix = calcid.indexOf(MenuLabels.fdensa) >= 0 
            || calcid.indexOf(MenuLabels.camp) >= 0
            || calcid.indexOf(MenuLabels.fdenss) >= 0;
    }
    else if(isfixedp){
        isefix = iswigner
            || calcid.indexOf(MenuLabels.fdensa) >= 0 
            || calcid.indexOf(MenuLabels.fdenss) >= 0
            || calcid.indexOf(MenuLabels.pflux) >= 0
            || istf
            || (srccont.isund < 0 && issimple);
    }
    else{
        isefix = calcid.indexOf(MenuLabels.fluxfix) >= 0
            || calcid.indexOf(MenuLabels.phasespace) >= 0;
    }
    if(srccont.isund >= 0 && iswigner){
        isdetune = !isedep;
    }
    let normenergy = srccont.isund >= 0 && 
        isefix && !isKdep && !iswigner && !issrcpoint;  

    if(isefix){
        confprmtypes[ConfigPrmsLabel.efix] = 1;
        if(srccont.isund >= 0 && normenergy
                && confobj[OptionLabel][ConfigOptionsLabel.normenergy[0]])
        {
            confprmtypes[ConfigPrmsLabel.efix] = 0;
            confprmtypes[ConfigPrmsLabel.nefix] = 1;
        }
    }
    if(isdetune){
        if(isefix){
            confprmtypes[ConfigPrmsLabel.efix] = 0;
        }
        confprmtypes[ConfigPrmsLabel.nefix] = -1;
        confprmtypes[ConfigPrmsLabel.detune] = 1;
    }
    if(issrcpoint){
        if(srccont.isund >= 0){
            confprmtypes[ConfigPrmsLabel.efix] = 0;
            confprmtypes[ConfigPrmsLabel.detune] = 1;
        }
        else{
            confprmtypes[ConfigPrmsLabel.efix] = 1;
        }
    }

    // obs. points
    let isangle = isFourier || 
        confobj[OptionLabel][ConfigOptionsLabel.defobs[0]] == ObsPointAngle;
    if(isedep || istime || isfixedp){
        let isdens = calcid.indexOf(MenuLabels.fdensa)  >= 0
            || calcid.indexOf(MenuLabels.fdenss) >= 0
            || calcid.indexOf(MenuLabels.pdensa) >= 0
            || calcid.indexOf(MenuLabels.pdenss) >= 0
            || calcid.indexOf(MenuLabels.efield) >= 0;
        let ispartial = calcid.indexOf(MenuLabels.pflux)  >= 0
            || calcid.indexOf(MenuLabels.ppower) >= 0;
        if(isdens){
            if(isangle){
                confprmtypes[ConfigPrmsLabel.qxyfix] = 1;
            }
            else{
                confprmtypes[ConfigPrmsLabel.xyfix] = 1;
            }
        }
        if(isspd){
            confprmtypes[ConfigPrmsLabel.xyfix] = 1;
            confprmtypes[ConfigPrmsLabel.Qnorm] = 1;
            confprmtypes[ConfigPrmsLabel.Phinorm] = 1;
        }
    }
    else if(isspd){
        if(calcid.indexOf(MenuLabels.xzplane) >= 0){
            confprmtypes[ConfigPrmsLabel.spdyfix] = 1;
        }
        else if(calcid.indexOf(MenuLabels.yzplane) >= 0){
            confprmtypes[ConfigPrmsLabel.spdxfix] = 1;
        }
        else if(calcid.indexOf(MenuLabels.pipe) >= 0){
            confprmtypes[ConfigPrmsLabel.spdrfix] = 1;
        }
    }
  
    // slit
    isshow.size = false;
    isshow.div = false;
    let isrect = calcid.indexOf(MenuLabels.slitrect) >= 0;
    let aptopt = false;
    let isnorm = false;
    if(isedep || istime || isfixedp || (isKdep && !issimple)){
        let iscirc = calcid.indexOf(MenuLabels.slitcirc) >= 0;
        if(isrect || iscirc){
            if(isangle){
                confprmtypes[ConfigPrmsLabel.qslitpos] = 
                    calcid.indexOf(MenuLabels.fluxpeak) < 0 ? 1 : -1;
                isshow.div = true;
            }
            else{
                confprmtypes[ConfigPrmsLabel.slitpos] = 
                    calcid.indexOf(MenuLabels.fluxpeak) < 0 ? 1 : -1;
                isshow.size = true;
            }
        }
        if(iscirc){
            if(isangle){
                confprmtypes[ConfigPrmsLabel.slitq] = 1;
            }
            else{
                confprmtypes[ConfigPrmsLabel.slitr] = 1;
            }
        }
        else if(isrect){
            if((isedep && srccont.isund >= 0) 
                    || (isKdep && calcid.indexOf(MenuLabels.fluxfix) < 0)){
                aptopt = true;
                isnorm = 
                    confobj[OptionLabel][ConfigOptionsLabel.aperture[0]] == NormSlitLabel;
            }
            if(isnorm){
                if(isangle){
                    confprmtypes[ConfigPrmsLabel.qslitapt] = 0;
                }
                else{
                    confprmtypes[ConfigPrmsLabel.slitapt] = 0;
                }
                confprmtypes[ConfigPrmsLabel.nslitapt] = 1;
                if(isKdep && calcid.indexOf(MenuLabels.ppower) < 0){
                    confprmtypes[ConfigPrmsLabel.pplimit] = 
                        confobj[OptionLabel][ConfigOptionsLabel.powlimit[0]] ? 1 : -1;
                }
            }
            else{
                if(isangle){
                    confprmtypes[ConfigPrmsLabel.qslitapt] = 1;
                }
                else{
                    confprmtypes[ConfigPrmsLabel.slitapt] = 1;
                }
            }
        }
    }
    
    // obs. range
    if(calcid.indexOf(MenuLabels.along) >= 0 
            || calcid.indexOf(MenuLabels.meshxy) >= 0)
    {
        if(isangle){
            confprmtypes[ConfigPrmsLabel.qxrange] = 1;
            confprmtypes[ConfigPrmsLabel.qyrange] = 1;
            isshow.div = true;
        }
        else{
            confprmtypes[ConfigPrmsLabel.xrange] = 1;
            confprmtypes[ConfigPrmsLabel.yrange] = 1;
            isshow.size = true;
        }
        confprmtypes[ConfigPrmsLabel.xmesh] = 1;
        confprmtypes[ConfigPrmsLabel.ymesh] = 1;
    } 
    if(calcid.indexOf(MenuLabels.meshrphi) >= 0 ){
        if(isangle){
            confprmtypes[ConfigPrmsLabel.qrange] = 1;
            confprmtypes[ConfigPrmsLabel.qphimesh] = 1;    
            isshow.div = true;
        }
        else{
            confprmtypes[ConfigPrmsLabel.rrange] = 1;
            confprmtypes[ConfigPrmsLabel.rphimesh] = 1;    
            isshow.size = true;
        }
        confprmtypes[ConfigPrmsLabel.phirange] = 1;
        confprmtypes[ConfigPrmsLabel.phimesh] = 1;
    }
    if(isspd){
        if(calcid.indexOf(MenuLabels.xzplane) >= 0){
            if(isangle){
                confprmtypes[ConfigPrmsLabel.qxrange] = 1;
            }
            else{
                confprmtypes[ConfigPrmsLabel.xrange] = 1;
            }
            confprmtypes[ConfigPrmsLabel.xmesh] = 1;
        }
        else if(calcid.indexOf(MenuLabels.yzplane) >= 0){
            if(isangle){
                confprmtypes[ConfigPrmsLabel.qyrange] = 1;
            }
            else{
                confprmtypes[ConfigPrmsLabel.yrange] = 1;
            }
            confprmtypes[ConfigPrmsLabel.ymesh] = 1;
        }
        else if(calcid.indexOf(MenuLabels.pipe) >= 0){
            confprmtypes[ConfigPrmsLabel.phirange] = 1;
            confprmtypes[ConfigPrmsLabel.phimesh] = 1;
        }
    }
    if(isvoldens)
    {
        confprmtypes[ConfigPrmsLabel.xrange] = 1;
        confprmtypes[ConfigPrmsLabel.xmesh] = 1;
        confprmtypes[ConfigPrmsLabel.yrange] = 1;
        confprmtypes[ConfigPrmsLabel.ymesh] = 1;
        if(confobj[OptionLabel][ConfigOptionsLabel.dstep[0]] 
                != ArbPositionsLabel){
            confprmtypes[ConfigPrmsLabel.drange] = 1;
            confprmtypes[ConfigPrmsLabel.dmesh] = 1;   
        }
        confprmtypes[ConfigPrmsLabel.Qgl] = 1;
        confprmtypes[ConfigPrmsLabel.Phiinc] = 1;
        confprmtypes[ConfigPrmsLabel.qslitapt] = 1;
        confprmtypes[ConfigPrmsLabel.illumarea] = 0;
    }

    // K, harmonic
    if(isKdep){
        if(srccont.isbxy >= 0){
            confprmtypes[ConfigPrmsLabel.krange] = -1;
            confprmtypes[ConfigPrmsLabel.ckrange] = 1;
        }
        else{
            confprmtypes[ConfigPrmsLabel.krange] = 1;
            confprmtypes[ConfigPrmsLabel.ckrange] = -1;
        }
        confprmtypes[ConfigPrmsLabel.e1strange] = 0;
        if(!isallharm){
            confprmtypes[ConfigPrmsLabel.kmesh] = 1;
        }
        if(calcid.indexOf(MenuLabels.tgtharm) >= 0){
            confprmtypes[ConfigPrmsLabel.hrange] = 1;
        }
    }
    if(iswigner && isallharm){
        if(srccont.isbxy >= 0){
            confprmtypes[ConfigPrmsLabel.krange] = -1;
            confprmtypes[ConfigPrmsLabel.ckrange] = 1;
        }
        else{
            confprmtypes[ConfigPrmsLabel.krange] = 1;
            confprmtypes[ConfigPrmsLabel.ckrange] = -1;
        }
    }
    let isharmfix = srccont.isund >= 0 
        && (calcid.indexOf(MenuLabels.pdensr) >= 0 
        || (isfixedp && issimple)
        || issrcpoint 
        || (iswigner && calcid.indexOf(MenuLabels.Kvalue) < 0));
    if(isharmfix){
        confprmtypes[ConfigPrmsLabel.hfix] = 1;
    }
    if(isallharm){
        confprmtypes[ConfigPrmsLabel.hmax] = 1;
    }

    // temporal
    if(istime){
        confprmtypes[ConfigPrmsLabel.trange] = 1;
        confprmtypes[ConfigPrmsLabel.tmesh] = 1;
    }

    // Wigner
    if(iswigner){
        if(calcid.indexOf(MenuLabels.XXpslice) >= 0){
            confprmtypes[ConfigPrmsLabel.Yfix] = 1;
            confprmtypes[ConfigPrmsLabel.Ypfix] = 1;
        }
        else if(calcid.indexOf(MenuLabels.YYpslice) >= 0){
            confprmtypes[ConfigPrmsLabel.Xfix] = 1;
            confprmtypes[ConfigPrmsLabel.Xpfix] = 1;
        }
        if(calcid.indexOf(MenuLabels.XXpslice) >= 0 
                || calcid.indexOf(MenuLabels.XXpprj) >= 0){
            confprmtypes[ConfigPrmsLabel.Xrange] = 1;
            confprmtypes[ConfigPrmsLabel.Xprange] = 1;
            confprmtypes[ConfigPrmsLabel.Xmesh] = 1;
            confprmtypes[ConfigPrmsLabel.Xpmesh] = 1;
        }
        else if(calcid.indexOf(MenuLabels.YYpslice) >= 0 
                || calcid.indexOf(MenuLabels.YYpprj) >= 0){
            confprmtypes[ConfigPrmsLabel.Yrange] = 1;
            confprmtypes[ConfigPrmsLabel.Yprange] = 1;
            confprmtypes[ConfigPrmsLabel.Ymesh] = 1;
            confprmtypes[ConfigPrmsLabel.Ypmesh] = 1;
        }
        else if(calcid.indexOf(MenuLabels.XXpYYp) >= 0 || 
                calcid.indexOf(MenuLabels.Wrel) >= 0){
            confprmtypes[ConfigPrmsLabel.Xrange] = 1;
            confprmtypes[ConfigPrmsLabel.Xmesh] = 1;
            confprmtypes[ConfigPrmsLabel.Xprange] = 1;
            confprmtypes[ConfigPrmsLabel.Xpmesh] = 1;
            confprmtypes[ConfigPrmsLabel.Yrange] = 1;
            confprmtypes[ConfigPrmsLabel.Ymesh] = 1;
            confprmtypes[ConfigPrmsLabel.Yprange] = 1;
            confprmtypes[ConfigPrmsLabel.Ypmesh] = 1;
        }
        else if(calcid.indexOf(MenuLabels.Wslice) >= 0
                 || isedep || isKdep){
            confprmtypes[ConfigPrmsLabel.Xfix] = 1;
            confprmtypes[ConfigPrmsLabel.Xpfix] = 1;
            confprmtypes[ConfigPrmsLabel.Yfix] = 1;
            confprmtypes[ConfigPrmsLabel.Ypfix] = 1;
        }
        else if(calcid.indexOf(MenuLabels.WprjX) >= 0){
            confprmtypes[ConfigPrmsLabel.Yfix] = 1;
            confprmtypes[ConfigPrmsLabel.Ypfix] = 1;
        }
        else if(calcid.indexOf(MenuLabels.WprjY) >= 0){
            confprmtypes[ConfigPrmsLabel.Xfix] = 1;
            confprmtypes[ConfigPrmsLabel.Xpfix] = 1;
        }
    }
    if(issrcpoint){
        confprmtypes[ConfigPrmsLabel.Xrange] = 1;
        confprmtypes[ConfigPrmsLabel.Xmesh] = 1;
        confprmtypes[ConfigPrmsLabel.Yrange] = 1;
        confprmtypes[ConfigPrmsLabel.Ymesh] = 1;
    }

    // aperture limit for comp. amp.
    if(iswigner || issrcpoint){
        if(iscustomsrc){
            confprmtypes[ConfigPrmsLabel.gtacc] = 1;
        }
        if(isbm || iswiggler){
            confprmtypes[ConfigPrmsLabel.horizacc] = 1;
        }    
    }

    // horizontal aperture for GABrill & total flux (BM)
    if(isbm && 
            (calcid.indexOf(MenuLabels.fdensa) >= 0 ||
            istf || issimple))
    {
        confprmtypes[ConfigPrmsLabel.horizacc] = 1;
    }

    // size, divergence
    let ispower = calcid.indexOf(MenuLabels.pdensa) >= 0 
                || calcid.indexOf(MenuLabels.pdenss) >= 0
                || calcid.indexOf(MenuLabels.ppower) >= 0;
    if(isshow.size){
        confprmtypes[ConfigPrmsLabel.psize] = ispower ? 0 : -1;
        confprmtypes[ConfigPrmsLabel.fsize] = ispower ? -1 : 0;
    }
    if(isshow.div){
        confprmtypes[ConfigPrmsLabel.pdiv] = ispower ? 0 : -1;
        confprmtypes[ConfigPrmsLabel.fdiv] = ispower ? -1 : 0;
    }
    if(srccont.isund < 0 && srccont.iswiggler < 0){
        confprmtypes[ConfigPrmsLabel.psize] = -1;
        confprmtypes[ConfigPrmsLabel.fsize] = -1;
        confprmtypes[ConfigPrmsLabel.pdiv] = -1;
        confprmtypes[ConfigPrmsLabel.fdiv] = -1;
    }

    //--- options ---
    // filtering    
    if(iswigner || issimple || iscoh){
        confopttypes[ConfigOptionsLabel.filter[0]] = false;
    }
    else if(isedep || isvoldens){
        confopttypes[ConfigOptionsLabel.filter[0]] = true;
    }
    else if(calcid.indexOf(MenuLabels.pdensr) >= 0){
        confopttypes[ConfigOptionsLabel.filter[0]] = false;
    }
    else if(calcid.indexOf(MenuLabels.pdensa) >= 0
        || calcid.indexOf(MenuLabels.pdenss) >= 0
        || calcid.indexOf(MenuLabels.ppower) >= 0)
    {
        if(srcobj[TypeLabel] == SrcTypeLabel.FIELDMAP3D){
            confopttypes[ConfigOptionsLabel.filter[0]] = false;
        }
        else if(srccont.isund >= 0 && 
            srcobj[SrcOptionsLabel.segment_type[1]] != NoneLabel &&
            srcobj[SrcOptionsLabel.perlattice[1]])
        {
            confopttypes[ConfigOptionsLabel.filter[0]] = false;
        }
        else if(srcobj[TypeLabel] == SrcTypeLabel.WIGGLER 
            || srcobj[TypeLabel] == SrcTypeLabel.EMPW)
        {
            confopttypes[ConfigOptionsLabel.filter[0]] 
                = calcid.indexOf(MenuLabels.far) >= 0;
        }
        else{
            confopttypes[ConfigOptionsLabel.filter[0]] = true;
        }     
    }
    if(confopttypes[ConfigOptionsLabel.filter[0]]){
        let ftype = confobj[OptionLabel][ConfigOptionsLabel.filter[0]];
        if(ftype == GenFilterLabel){            
            confopttypes[ConfigOptionsLabel.fmateri[0]] = true;
        }
        else if(ftype == BPFGaussianLabel || ftype == BPFBoxCarLabel){
            confopttypes[ConfigOptionsLabel.bpfprms[0]] = true;        
        }
        else if(ftype == CustomLabel){
            confopttypes[ConfigOptionsLabel.fcustom[0]] = true;        
        }
    }
    if(isvoldens){
        confopttypes[ConfigOptionsLabel.amateri[0]] = true;
    }
    let isfilter = confopttypes[ConfigOptionsLabel.filter[0]]
        && confobj[OptionLabel][ConfigOptionsLabel.filter[0]] != NoneLabel;

    // energy step
    if(isallharm || (isedep && (iswigapprox || isbmlike))){
        confopttypes[ConfigOptionsLabel.estep[0]] = true;
    }

    // aperture, powerlimit
    if(aptopt){
        confopttypes[ConfigOptionsLabel.aperture[0]] = true;
        confopttypes[ConfigOptionsLabel.powlimit[0]] = 
            isnorm && calcid.indexOf(MenuLabels.fluxpeak) >= 0;
    }

    // depth step
    confopttypes[ConfigOptionsLabel.dstep[0]] 
        = calcid.indexOf(MenuLabels.vpdens) >= 0;

    // depth step data
    confopttypes[ConfigOptionsLabel.depthdata[0]] 
        = calcid.indexOf(MenuLabels.vpdens) >= 0 && 
        confobj[OptionLabel][ConfigOptionsLabel.dstep[0]] == ArbPositionsLabel;

    // energy representation
    confopttypes[ConfigOptionsLabel.normenergy[0]] = normenergy;

    // defobs
    if(calcid.indexOf(MenuLabels.temporal) >= 0){
        confopttypes[ConfigOptionsLabel.defobs[0]] 
            = confobj[OptionLabel][ConfigOptionsLabel.fouriep[0]] == false;
    }
    else if(calcid.indexOf(MenuLabels.srcpoint) >= 0
        || calcid.indexOf(MenuLabels.wigner) >= 0
        || calcid.indexOf(MenuLabels.spdens) >= 0
        || calcid.indexOf(MenuLabels.vpdens) >= 0
    )
    {
        confopttypes[ConfigOptionsLabel.defobs[0]] = false;
    }
    else if(isKdep){
        confopttypes[ConfigOptionsLabel.defobs[0]] 
            = calcid.indexOf(MenuLabels.pflux) >= 0 
                || calcid.indexOf(MenuLabels.ppower) >= 0;
    }
    else if(istf || issimple)
    {
        confopttypes[ConfigOptionsLabel.defobs[0]] 
            = srcobj[TypeLabel] == SrcTypeLabel.BM;
    }
    else{
        confopttypes[ConfigOptionsLabel.defobs[0]] = true;
    }

    // optDx, xsmooth
    if(iswigner && isbmlike){
        let snglbm = isbm && !srcobj[OptionLabel][SrcOptionsLabel.bmtandem[0]];
        confopttypes[ConfigOptionsLabel.optDx[0]] = snglbm;
        confopttypes[ConfigOptionsLabel.xsmooth[0]] = !isbm || snglbm;
    }

    // Fourier plane
    confopttypes[ConfigOptionsLabel.fouriep[0]] 
        = calcid.indexOf(MenuLabels.temporal) >= 0;

    let ispowsimple = !isvoldens && !isfilter && (
            calcid.indexOf(MenuLabels.ppower) >= 0 ||
            calcid.indexOf(MenuLabels.pdenss) >= 0 ||
            calcid.indexOf(MenuLabels.spdens) >= 0
        );

    // Wiggler approximation
    confopttypes[ConfigOptionsLabel.wiggapprox[0]] = 
        (isedep && srccont.isund >= 0 
            && calcid.indexOf(MenuLabels.far) >= 0) ||
        (srcobj[TypeLabel] == SrcTypeLabel.CUSTOM && !ispowsimple
            && calcid.indexOf(MenuLabels.near) >= 0);
    
    // Energy smoothing
    confopttypes[ConfigOptionsLabel.esmooth[0]] = 
        calcid.indexOf(MenuLabels.far) < 0 &&
        confopttypes[ConfigOptionsLabel.wiggapprox[0]];
    if(confopttypes[ConfigOptionsLabel.wiggapprox[0]]){
        if(confobj[OptionLabel][ConfigOptionsLabel.esmooth[0]]){
            confopttypes[ConfigOptionsLabel.wiggapprox[0]] = false;
            confopttypes[ConfigOptionsLabel.esmooth[0]] = true;
        }
        if(confobj[OptionLabel][ConfigOptionsLabel.wiggapprox[0]]){
            confopttypes[ConfigOptionsLabel.wiggapprox[0]] = true;
            confopttypes[ConfigOptionsLabel.esmooth[0]] = false;
        }
    }
    confopttypes[ConfigOptionsLabel.smoothwin[0]] = 
        confopttypes[ConfigOptionsLabel.esmooth[0]] &&
        confobj[OptionLabel][ConfigOptionsLabel.esmooth[0]]

    // accuracy
    let cmdpp = confobj[TypeLabel].indexOf(MenuLabels.CMDPP) >= 0;

    confopttypes[ConfigOptionsLabel.acclevel[0]] = false;
    confopttypes[ConfigOptionsLabel.accuracy[0]] = !cmdpp;

    // CMD
    let cmdenable = 
        calcid.indexOf(MenuLabels.XXpprj) >= 0 ||
        calcid.indexOf(MenuLabels.YYpprj) >= 0 ||
        calcid.indexOf(MenuLabels.XXpYYp) >= 0;
    let cmdon = confobj[OptionLabel][ConfigOptionsLabel.CMD[0]];

    let showgs = false;
    if(confobj[TypeLabel].indexOf(MenuLabels.CMD) < 0){
        confopttypes[ConfigOptionsLabel.CMD[0]] = cmdenable;
        confopttypes[ConfigOptionsLabel.CMDprms[0]] = cmdenable && cmdon;
        confopttypes[ConfigOptionsLabel.CMDfld[0]] = cmdenable && cmdon;
        confopttypes[ConfigOptionsLabel.CMDint[0]] = cmdenable && cmdon;
        confopttypes[ConfigOptionsLabel.CMDcmp[0]] = cmdenable && cmdon;
        confopttypes[ConfigOptionsLabel.CMDcmpint[0]] = cmdenable && cmdon;
        showgs = cmdenable && cmdon;
    }
    else{
        confopttypes[ConfigOptionsLabel.CMD[0]] = false;
        confopttypes[ConfigOptionsLabel.CMDprms[0]] = true;
        confopttypes[ConfigOptionsLabel.CMDfld[0]] = true;
        confopttypes[ConfigOptionsLabel.CMDint[0]] = true;
        confopttypes[ConfigOptionsLabel.CMDcmp[0]] = true;
        confopttypes[ConfigOptionsLabel.CMDcmpint[0]] = true;
        showgs = !cmdpp;
    }

    confopttypes[ConfigOptionsLabel.GSModel[0]] = false;
    confopttypes[ConfigOptionsLabel.GSModelXY[0]] = false;
    if(showgs){
        if(calcid.indexOf(MenuLabels.XXpYYp) >= 0){
            confopttypes[ConfigOptionsLabel.GSModelXY[0]] = true;
        }
        else{
            confopttypes[ConfigOptionsLabel.GSModel[0]] = true;
        }    
    }

    // FEL
    confopttypes[ConfigOptionsLabel.fel[0]] 
    = iscoh && (srccont.isund >= 0 || iscustomsrc) 
        && accobj[OptionLabel][AccOptionsLabel.bunchtype[0]] != ImportPaticleLabel;
    
    confopttypes[ConfigOptionsLabel.seedspec[0]] = confopttypes[ConfigOptionsLabel.fel[0]]
        && confobj[OptionLabel][ConfigOptionsLabel.fel[0]] == FELSeedCustomLabel;    
    confopttypes[ConfigOptionsLabel.exportInt[0]] = confopttypes[ConfigOptionsLabel.fel[0]]
        && confobj[OptionLabel][ConfigOptionsLabel.fel[0]] != NoneLabel
        && confobj[OptionLabel][ConfigOptionsLabel.fel[0]] != FELReuseLabel;
    confopttypes[ConfigOptionsLabel.R56Bunch[0]] = confopttypes[ConfigOptionsLabel.exportInt[0]];
    confopttypes[ConfigOptionsLabel.FELprms[0]] = confopttypes[ConfigOptionsLabel.fel[0]]
        && confobj[OptionLabel][ConfigOptionsLabel.fel[0]] != NoneLabel
    confopttypes[ConfigOptionsLabel.exportEt[0]] = confopttypes[ConfigOptionsLabel.exportInt[0]]
        && (confobj[OptionLabel][ConfigOptionsLabel.R56Bunch[0]] || confobj[OptionLabel][ConfigOptionsLabel.exportInt[0]]);

    return [confprmtypes, confopttypes];
}

function BPFPrmOpt(confobj)
{
    let bpfprmtypes = {};

    for(let j = 0; j < BPFLabelOrder.length; j++){
        bpfprmtypes[BPFPrmsLabel[BPFLabelOrder[j]]] = 1;
    }   
    bpfprmtypes[BPFPrmsLabel.bpfwidth] = 
        confobj[OptionLabel][ConfigOptionsLabel.filter[0]] == BPFBoxCarLabel ? 1 : -1;
    bpfprmtypes[BPFPrmsLabel.bpfsigma] = 
        confobj[OptionLabel][ConfigOptionsLabel.filter[0]] == BPFGaussianLabel ? 1 : -1;

    return bpfprmtypes;
}

function CMDPrmOpt(confobj)
{
    let cmdprmtypes = {};

    for(let j = 0; j < CMDLabelOrder.length; j++){
        cmdprmtypes[CMDPrmsLabel[CMDLabelOrder[j]]] = -1;
    }
    cmdprmtypes[CMDPrmsLabel.maxmode] = 1;
    cmdprmtypes[CMDPrmsLabel.fcutoff] = 1;
    cmdprmtypes[CMDPrmsLabel.cutoff] = 1;
    let isexport = 
        confobj[OptionLabel][ConfigOptionsLabel.CMDfld[0]] != NoneLabel ||
        confobj[OptionLabel][ConfigOptionsLabel.CMDint[0]] 
        ? 1 : -1;
    let calcid;
    let cmdpp = confobj[TypeLabel].indexOf(MenuLabels.CMDPP) >= 0;
    if(confobj[TypeLabel].indexOf(MenuLabels.CMD) >= 0){
        let cmdcont = Output4CMD(SPGUI.orginobj);
        calcid = cmdcont.id;
    }
    else{
        calcid = confobj[TypeLabel];
    }
    if(calcid.indexOf(MenuLabels.XXpprj) >= 0){
        cmdprmtypes[CMDPrmsLabel.HGorderx] = cmdpp?-1:1;
        cmdprmtypes[CMDPrmsLabel.maxHGorderx] = cmdpp?0:-1;
        cmdprmtypes[CMDPrmsLabel.fieldgridx] = isexport;        
    }
    else if(calcid.indexOf(MenuLabels.YYpprj) >= 0){
        cmdprmtypes[CMDPrmsLabel.HGordery] = cmdpp?-1:1;
        cmdprmtypes[CMDPrmsLabel.maxHGordery] = cmdpp?0:-1;
        cmdprmtypes[CMDPrmsLabel.fieldgridy] = isexport;                
    }
    else if(calcid.indexOf(MenuLabels.XXpYYp) >= 0){
        cmdprmtypes[CMDPrmsLabel.HGorderxy] = cmdpp?-1:1;
        cmdprmtypes[CMDPrmsLabel.maxHGorderxy] = cmdpp?0:-1;
        cmdprmtypes[CMDPrmsLabel.fieldgridxy] = isexport;
    }

    return cmdprmtypes;
}

function FELPrmOpt(srcobj, confobj)
{
    let isfelsec = false;
    if(srcobj[TypeLabel] == SrcTypeLabel.CUSTOM){
        if(srcobj.hasOwnProperty(SrcPrmsLabel.fvsz)){
            if(srcobj[SrcPrmsLabel.fvsz].hasOwnProperty(FELSecIdxLabel)){
                if(Array.isArray(srcobj[SrcPrmsLabel.fvsz][FELSecIdxLabel])){
                    isfelsec = srcobj[SrcPrmsLabel.fvsz][FELSecIdxLabel].length > 0;
                }
            }
        }        
    }

    let felprmtypes = {};
    for(let j = 0; j < FELLabelOrder.length; j++){
        felprmtypes[FELPrmsLabel[FELLabelOrder[j]]] = -1;
    }
    let felopt = confobj[OptionLabel][ConfigOptionsLabel.fel[0]];
    if(felopt == FELSeedLabel || felopt == FELCPSeedLabel || felopt == FELSeedCustomLabel){
        felprmtypes[FELPrmsLabel.pulseE] = 1;
        if(felopt != FELSeedCustomLabel){
            felprmtypes[FELPrmsLabel.wavelen] = 1;
            if(felopt == FELSeedLabel){
                felprmtypes[FELPrmsLabel.pulselen] = 1;
            }
            else{
                felprmtypes[FELPrmsLabel.tlpulselen] = 1;
            }    
        }
        felprmtypes[FELPrmsLabel.srcsize] = 1;
        felprmtypes[FELPrmsLabel.waistpos] = 1;
        felprmtypes[FELPrmsLabel.timing] = 1;
        if(felopt == FELCPSeedLabel || felopt == FELSeedCustomLabel){
            felprmtypes[FELPrmsLabel.gdd] = 1;
            felprmtypes[FELPrmsLabel.tod] = 1;
        }
    }
    else if(felopt == FELDblSeedLabel){
        felprmtypes[FELPrmsLabel.pulseE_d] = 1;
        felprmtypes[FELPrmsLabel.wavelen_d] = 1;
        felprmtypes[FELPrmsLabel.tlpulselen_d] = 1;
        felprmtypes[FELPrmsLabel.srcsize_d] = 1;
        felprmtypes[FELPrmsLabel.waistpos_d] = 1;
        felprmtypes[FELPrmsLabel.timing_d] = 1;
        felprmtypes[FELPrmsLabel.gdd_d] = 1;
        felprmtypes[FELPrmsLabel.tod_d] = 1;
    }
    felprmtypes[FELPrmsLabel.eproi] = 1;
    felprmtypes[FELPrmsLabel.particles] = 1;
    felprmtypes[FELPrmsLabel.svstep] = isfelsec ? -1 : 1;
    felprmtypes[FELPrmsLabel.radstep] = 1;
    felprmtypes[FELPrmsLabel.edevstep] = 
        (confobj[OptionLabel][ConfigOptionsLabel.exportEt[0]] && (
            confobj[OptionLabel][ConfigOptionsLabel.R56Bunch[0]] || 
            confobj[OptionLabel][ConfigOptionsLabel.exportInt[0]] )
        ) ? 1 : -1;
    felprmtypes[FELPrmsLabel.R56] = 
        confobj[OptionLabel][ConfigOptionsLabel.R56Bunch[0]] ? 1 : -1;

    return felprmtypes;
}

// ---------- classes ----------
// ---------- Accelerator ----------
class Accelerator extends PrmTable {
    constructor(){
        let prmlabel = [];
        for(let i = 0; i < AccLabelOrder.length; i++){
            prmlabel.push(AccPrmsLabel[AccLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = AccPrmsFormat[AccLabelOrder[i]];
            isscan[prmlabel[i]] = AccIsScan[AccLabelOrder[i]];
        }
        super(AccLabel, prmlabel, format, def_acc, isscan);
    }
}

class AccOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < AccOptionsOrder.length; i++){
            if(AccOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{
                optionlabel.push(AccOptionsLabel[AccOptionsOrder[i]]);
            }
        }
        super(AccLabel, optionlabel, def_acc[OptionLabel]);
    }
}

class InjectionCondition extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(AccLabel, AccOptionsLabel.injecprm[0]);
        let prmlabel = [];
        for(let i = 0; i < InjectionLabelOrder.length; i++){
            prmlabel.push(InjectionPrmsLabel[InjectionLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = InjectionPrmsFormat[InjectionLabelOrder[i]];
            isscan[prmlabel[i]] = InjectionIsScan[InjectionLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_injecprm, isscan);
    }
}

// ---------- Light Source ----------
class LightSource extends PrmTable {
    constructor(){
        let prmlabel = [];
        for(let i = 0; i < SrcLabelOrder.length; i++){
            prmlabel.push(SrcPrmsLabel[SrcLabelOrder[i]]);
        }       
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = SrcPrmsFormat[SrcLabelOrder[i]];
            isscan[prmlabel[i]] = SrcIsScan[SrcLabelOrder[i]];
        }
        super(SrcLabel, prmlabel, format, def_src, isscan);
    }
}

class SrcOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < SrcOptionsOrder.length; i++){
            if(SrcOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{
                optionlabel.push(SrcOptionsLabel[SrcOptionsOrder[i]]);
            }
        }
        super(SrcLabel, optionlabel, def_src[OptionLabel]);
    }
}

class MagnetConfiguration extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(SrcLabel, SrcOptionsLabel.magconf[0]);
        let prmlabel = [];
        for(let i = 0; i < MagnetLabelOrder.length; i++){
            prmlabel.push(MagnetPrmsLabel[MagnetLabelOrder[i]]);
        }       
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = MagnetPrmsFormat[MagnetLabelOrder[i]];
            isscan[prmlabel[i]] = MagnetIsScan[MagnetLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_magconf, isscan);
    }
}

class FieldError extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(SrcLabel, SrcOptionsLabel.ferrprm[0]);
        let prmlabel = [];
        for(let i = 0; i < FerrLabelOrder.length; i++){
            prmlabel.push(FerrPrmsLabel[FerrLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = FerrPrmsFormat[FerrLabelOrder[i]];
            isscan[prmlabel[i]] = FerrIsScan[FerrLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_ferrprm, isscan);
    }
}

class PhaseError extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(SrcLabel, SrcOptionsLabel.perrprm[0]);
        let prmlabel = [];
        for(let i = 0; i < PerrLabelOrder.length; i++){
            prmlabel.push(PerrPrmsLabel[PerrLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = PerrPrmsFormat[PerrLabelOrder[i]];
            isscan[prmlabel[i]] = PerrIsScan[PerrLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_perrprm, isscan);
    }
}

class Segmentation extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(SrcLabel, SrcOptionsLabel.segprm[0]);
        let prmlabel = [];
        for(let i = 0; i < SegLabelOrder.length; i++){
            prmlabel.push(SegPrmsLabel[SegLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = SegPrmsFormat[SegLabelOrder[i]];
            isscan[prmlabel[i]] = SegIsScan[SegLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_segprm, isscan);
    }
}

//---------- Configurations ----------
class Configuration extends PrmTable {
    constructor(){
        let prmlabel = [];
        for(let i = 0; i < ConfigLabelOrder.length; i++){
            prmlabel.push(ConfigPrmsLabel[ConfigLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = ConfigPrmsFormat[ConfigLabelOrder[i]];
            isscan[prmlabel[i]] = ConfigPrmsIsScan[ConfigLabelOrder[i]];
        }
        super(ConfigLabel, prmlabel, format, def_conf, isscan);
    }
}

class ConfigOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < ConfigOptionsOrder.length; i++){
            if(ConfigOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(ConfigOptionsLabel[ConfigOptionsOrder[i]]);
            }
        }
        super(ConfigLabel, optionlabel, def_conf[OptionLabel]);
    }
}

class BPFPrms extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.bpfprms[0]);
        let prmlabel = [];
        for(let i = 0; i < BPFLabelOrder.length; i++){
            prmlabel.push(BPFPrmsLabel[BPFLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = BPFPrmsFormat[BPFLabelOrder[i]];
            isscan[prmlabel[i]] = BPFIsScan[BPFLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_bpfprm, isscan);
    }
}

class CMDPrms extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.CMDprms[0]);
        let prmlabel = [];
        for(let i = 0; i < CMDLabelOrder.length; i++){
            prmlabel.push(CMDPrmsLabel[CMDLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = CMDPrmsFormat[CMDLabelOrder[i]];
            isscan[prmlabel[i]] = CMDIsScan[CMDLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_CMDprm, isscan);
    }
}

class FELPrms extends PrmTable {
    constructor(){
        let categ = GetIDFromItem(ConfigLabel, ConfigOptionsLabel.FELprms[0]);
        let prmlabel = [];
        for(let i = 0; i < FELLabelOrder.length; i++){
            prmlabel.push(FELPrmsLabel[FELLabelOrder[i]]);
        }
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = FELPrmsFormat[FELLabelOrder[i]];
            isscan[prmlabel[i]] = FELIsScan[FELLabelOrder[i]];
        }
        super(categ, prmlabel, format, def_FELprm, isscan);
    }
}


class OutputOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < OutputOptionsOrder.length; i++){
            if(OutputOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(OutputOptionsLabel[OutputOptionsOrder[i]]);
            }
        }
        super(OutFileLabel, optionlabel, def_output_configs);
    }

    ForceFormatJSON(isforce){
        if(isforce){
            this.m_jsonobj[OptionLabel][OutputOptionsLabel.format[0]] = JSONOnly;
            let id = GetIDFromItem(OutFileLabel, OutputOptionsLabel.format[0], -1);
            let select = document.getElementById(id);
            select.value = JSONOnly;
        }
        this.DisableInput(OutputOptionsLabel.format[0], isforce);
    }
}


class ScanConfiguration extends PrmTable {
    constructor(){
        let prmlabel = [];
        for(let i = 0; i < ScanLabelOrder.length; i++){
            prmlabel.push(ScanPrmsLabel[ScanLabelOrder[i]]);
        }       
        let format  = {};
        let isscan = {};
        for(let i = 0; i < prmlabel.length; i++){
            format[prmlabel[i]] = ScanPrmsFormat[ScanLabelOrder[i]];
            isscan[prmlabel[i]] = false;
        }
        super(ScanLabel, prmlabel, format, def_scan, isscan);
        this.m_skipupdate = true;
    }
}

class ScanOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < ScanOptionsOrder.length; i++){
            if(ScanOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(ScanOptionsLabel[ScanOptionsOrder[i]]);
            }
        }
        super(ScanOptionlabel, optionlabel, def_scan_option);
    }
}

class AccuracyOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < AccuracyOptionsOrder.length; i++){
            if(AccuracyOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(AccuracyOptionsLabel[AccuracyOptionsOrder[i]]);
            }
        }
        super(AccuracyLabel, optionlabel, def_accuracy);
    }
}

class UnitOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < UnitOptionsOrder.length; i++){
            if(UnitOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(UnitOptionsLabel[UnitOptionsOrder[i]]);
            }
        }
        super(UnitLabel, optionlabel, def_units);
    }
}

class PCOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < PCOptionsOrder.length; i++){
            if(PCOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(PCOptionsLabel[PCOptionsOrder[i]]);
            }
        }
        super(PCLabel, optionlabel, def_pc_configs);
    }
}

class PhaseErrOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < PhaseErrOptionsOrder.length; i++){
            if(PhaseErrOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{
                optionlabel.push(PhaseErrOptionsLabel[PhaseErrOptionsOrder[i]]);
            }
        }
        super(PhaseErrLabel, optionlabel, def_perr_configs);
    }
}

class FilterPlotOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < FilterPlotOptionsOrder.length; i++){
            if(FilterPlotOptionsLabel[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{
                optionlabel.push(FilterPlotOptionsLabel[FilterPlotOptionsOrder[i]]);
            }
        }
        super(PlotFilterLabel, optionlabel, def_filtplot_configs);
    }
}

class PlotOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < PlotOptionsOrder.length; i++){
            if(PlotOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(PlotOptionsLabel[PlotOptionsOrder[i]]);
            }
        }
        super(POLabel, optionlabel, def_plot_configs);
    }
}

class PSNameOption extends OptionList {
    constructor(){
        let optionlabel = [];
        for(let i = 0; i < PSNameOptionsOrder.length; i++){
            if(PSNameOptionsOrder[i] == SeparatorLabel){
                optionlabel.push(SeparatorLabel);
            }
            else{                
                optionlabel.push(PSNameOptionsLabel[PSNameOptionsOrder[i]]);
            }
        }
        super(PSNameLabel, optionlabel, def_psname_configs);
    }
}
