"use strict";

// class to handle gap-field table
class GapFieldTable {
    constructor(){
    }

    IsApple(){
        return this.m_isapple;
    }
    
    SetSourceOption(){
        if(this.m_srcobj.hasOwnProperty(SrcPrmsLabel.phase) == false){
            this.m_srcobj[SrcPrmsLabel.phase] = 0;
        }
        if(this.m_srcobj.hasOwnProperty(SrcPrmsLabel.Kxy0) == false){
            this.m_srcobj[SrcPrmsLabel.Kxy0] = [0.7, 1.0];
        }        
        if(this.m_srcobj[OptionLabel].hasOwnProperty(SrcOptionsLabel.apple[0]) == false){
            this.m_srcobj[OptionLabel][SrcOptionsLabel.apple[0]] = false;
        }        
        let srccont = GetSrcContents(this.m_srcobj);
        this.m_isbxy = srccont.isbxy >= 0;
        this.m_isapple = this.m_srcobj[TypeLabel] == SrcTypeLabel.ELLIPTIC_UND
                        && this.m_srcobj[OptionLabel][SrcOptionsLabel.apple[0]];

        this.m_mainjxy = 1;
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VERTICAL_UND || 
                this.m_srcobj[TypeLabel] == SrcTypeLabel.VFIGURE8_UND){
            this.m_mainjxy = 0;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.FIGURE8_UND){
            this.m_mainjxy = 1;
        }
        else if(this.m_isbxy){
            this.m_mainjxy = 
                this.m_srcobj[SrcPrmsLabel.Kxy][0] > this.m_srcobj[SrcPrmsLabel.Kxy][1] ? 0 : 1;
        }
        if(this.m_isapple){
            let csn = this.GetPhaseFactor();
            this.m_mainjxy = Math.abs(csn[0]) > Math.abs(csn[1]) ? 0 : 1;
        }
    }

    SetSrcObj(srcobj){
        this.m_srcobj = srcobj;
        this.SetSourceOption();
        this.m_gap = [];
        this.m_bxy = [[], []];
        let gaptbl = this.m_srcobj[OptionLabel][SrcOptionsLabel.gaptbl[0]];

        if(typeof gaptbl == "undefined"){
            return;
        }
        if(Array.isArray(gaptbl.data) && gaptbl.data.length > 1){
            this.m_gap = [].concat(gaptbl.data[0]);
            this.m_bxy[0] = [].concat(gaptbl.data[1]);
            this.m_bxy[1] = [].concat(gaptbl.data[2]);
            this.SetBSquared();
        }
        if(this.m_gap.length > 0){
            this.m_mingap = this.m_gap.reduce((red,curr)=>curr<red?curr:red);
            this.m_maxgap = this.m_gap.reduce((red,curr)=>curr>red?curr:red);
        }
    }

    GetPhaseFactor(){
        let phi = Math.PI*2.0*this.m_srcobj[SrcPrmsLabel.phase]
            /this.m_srcobj[SrcPrmsLabel.lu];
        let csn = [0,0];
        csn[0] = Math.sin(phi);
        csn[1] = Math.cos(phi);
        for(let j = 0; j < 2; j++){
            if(Math.abs(csn[j]) < 1.0e-10){
                csn[j] = 0;
            }    
        }
        return csn;
    }

    SetBSquared(){
        let csn = [1, 1];
        if(this.m_isapple){
            csn = this.GetPhaseFactor();
        }
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.FIGURE8_UND){
            csn[0] = 2;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VFIGURE8_UND){
            csn[1] = 2;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.LIN_UND){
            csn[0] = 0;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VERTICAL_UND){
            csn[1] = 0;
        }

        let bxy, bsq;
        this.m_bsq = [];
        bsq = this.m_bsq;
        bxy = this.m_bxy;
        for(let n = 0; n < bxy[0].length; n++){
            if(this.m_srcobj[TypeLabel] == SrcTypeLabel.HELICAL_UND){
                bsq.push(bxy[1][n]*Math.sqrt(2));
            }
            else{
                bsq.push(Math.hypot(bxy[0][n]*csn[0], bxy[1][n]*csn[1]));
            }
        }
    }

    GetKCoef(){
        let kcoef = this.m_srcobj[SrcPrmsLabel.lu]/1000.0*COEF_K_VALUE;
        let kcoefa = [kcoef, kcoef];
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.FIGURE8_UND){
            kcoefa[0] *= 2.0;
        }
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VFIGURE8_UND){
            kcoefa[1] *= 2.0;
        }
        return kcoefa;
    }

    GetBpeak(){
        let magconf = GetObj(SrcLabel)[OptionLabel][SrcOptionsLabel.magconf[0]];
        if(magconf == undefined){
            magconf = def_magconf;
        }
        let Br = parseFloat(magconf[MagnetPrmsLabel.br]);
        if(isNaN(Br) || Br <= 0){
            Br = 1.2;
        }

        let bpeak = [];
        for(let j = 0; j < 2; j++){
            let geo = parseFloat(magconf[MagnetPrmsLabel.geofactor][j]);
            bpeak.push(COEF_BPEAK*Br*geo);
        }
        return bpeak;
    }

    GetGapKxy(item, jxy, tgtvalue)
    {
        let lu_mm = [this.m_srcobj[SrcPrmsLabel.lu], this.m_srcobj[SrcPrmsLabel.lu]];
        let if8 = -1;
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.FIGURE8_UND){
            lu_mm[0] *= 2;
            if8 = 0;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VFIGURE8_UND){
            lu_mm[1] *= 2;
            if8 = 1;
        }
        let Kpeak = [];
        let kcoef = this.GetKCoef();
        let bpeak = this.GetBpeak();
        for(let j = 0; j < 2; j++){
            Kpeak.push(bpeak[j]*kcoef[j]);
        }
        if(this.m_srcobj[TypeLabel] == SrcTypeLabel.LIN_UND || 
                this.m_srcobj[TypeLabel] == SrcTypeLabel.WIGGLER){
            Kpeak[0] = 0;
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.HELICAL_UND){
            Kpeak[0] = Kpeak[1];
        }
        else if(this.m_srcobj[TypeLabel] == SrcTypeLabel.VERTICAL_UND){
            Kpeak[1] = 0;
        }

        let csn;
        if(this.m_isapple){
            csn = this.GetPhaseFactor();
        }

        if(item == SrcPrmsLabel.gap){
            for(let j = 0; j < 2; j++){
                if(this.m_isapple){
                    this.m_srcobj[SrcPrmsLabel.Kxy0][j] 
                        = Kpeak[j]*Math.exp(-Math.PI*tgtvalue/lu_mm[j]);
                    this.m_srcobj[SrcPrmsLabel.Kxy][j] 
                        = this.m_srcobj[SrcPrmsLabel.Kxy0][j]*csn[j];
                }
                else{
                    this.m_srcobj[SrcPrmsLabel.Kxy][j] 
                        = Kpeak[j]*Math.exp(-Math.PI*tgtvalue/lu_mm[j]);
                }
            }
            return;
        }
        else{
            if(tgtvalue == 0){
                this.m_srcobj[SrcPrmsLabel.gap] = Infinity;
                if(this.m_isapple){
                    this.m_srcobj[SrcPrmsLabel.Kxy0][1-jxy] = 0; 
                }
                else{
                    this.m_srcobj[SrcPrmsLabel.Kxy][1-jxy] = 0; 
                }
                return;
            }
            else if(item == SrcPrmsLabel.e1st){
                let Kpeak0;
                if(this.m_isapple){
                    Kpeak0 = Math.hypot(Kpeak[0]*csn[0], Kpeak[1]*csn[1]);
                }
                else{
                    Kpeak0 = Math.hypot(Kpeak[0], Kpeak[1]);
                }
                if(if8 >= 0){
                    let K0 = Kpeak[if8];
                    let K1 = Kpeak[1-if8];
                    let frac = (-K0*K0+Math.sqrt(K0**4+4*(K1*tgtvalue)**2))/2/K1/K1;
                    this.m_srcobj[SrcPrmsLabel.gap] = -Math.log(frac)*lu_mm[1-if8]/Math.PI;
                }
                else{
                    this.m_srcobj[SrcPrmsLabel.gap] = -Math.log(tgtvalue/Kpeak0)*lu_mm[0]/Math.PI;
                }
                if(this.m_isapple){
                    this.m_srcobj[SrcPrmsLabel.Kxy0][jxy] 
                        = Kpeak[jxy]*Math.exp(-Math.PI*this.m_srcobj[SrcPrmsLabel.gap]/lu_mm[jxy]);
                }
                else{
                    this.m_srcobj[SrcPrmsLabel.Kxy][jxy] 
                        = Kpeak[jxy]*Math.exp(-Math.PI*this.m_srcobj[SrcPrmsLabel.gap]/lu_mm[jxy]);
                }
            }
            else{
                this.m_srcobj[SrcPrmsLabel.gap] = -Math.log(tgtvalue/Kpeak[jxy])*lu_mm[jxy]/Math.PI;            
            }    
        }

        if(this.m_isapple){
            this.m_srcobj[SrcPrmsLabel.Kxy0][1-jxy] 
                = Kpeak[1-jxy]*Math.exp(-Math.PI*this.m_srcobj[SrcPrmsLabel.gap]/lu_mm[1-jxy]);
            for(let j = 0; j < 2; j++){
                this.m_srcobj[SrcPrmsLabel.Kxy][j] 
                    = this.m_srcobj[SrcPrmsLabel.Kxy0][j]*csn[j];
            }
        }
        else{
            this.m_srcobj[SrcPrmsLabel.Kxy][1-jxy] 
                = Kpeak[1-jxy]*Math.exp(-Math.PI*this.m_srcobj[SrcPrmsLabel.gap]/lu_mm[1-jxy]);
        }
    }

    GetGapKxyFromTable(item, jxy, tgtvalue)
    {
        let bsqarr, gaparr, bxyarr;
        bsqarr = this.m_bsq;
        gaparr = this.m_gap;
        bxyarr = this.m_bxy;

        let Kxy = [0, 0], gap;
        let kcoef = this.GetKCoef();
        if(item == SrcPrmsLabel.gap){
            gap = tgtvalue;
        }
        else if(item == SrcPrmsLabel.e1st){ // tgtvalue = sqrt(Kx^2+Ky^2)
            gap = SolveParabolic(gaparr, bsqarr, tgtvalue/kcoef[this.m_mainjxy]);
            this.m_srcobj[SrcPrmsLabel.gap] = gap;
        }
        if(item == SrcPrmsLabel.gap || item == SrcPrmsLabel.e1st){
            for(let j = 0; j < 2; j++){
                Kxy[j] = GetParabolic(gaparr, bxyarr[j], gap)*kcoef[j];
            }
        }
        else{ // tgtvalue = Kx,y
            gap = SolveParabolic(gaparr, bxyarr[jxy], tgtvalue/kcoef[jxy]);
            if(isNaN(gap)){
                tgtvalue = NaN;            
            }
            this.m_srcobj[SrcPrmsLabel.gap] = gap;
            Kxy[jxy] = tgtvalue;
            if(this.m_isbxy){
                Kxy[1-jxy] = GetParabolic(gaparr, bxyarr[1-jxy], gap)*kcoef[1-jxy];
            }
        }

        let dround = 1.0+1e-6;
        let isok = gap >= this.m_mingap/dround && gap <= this.m_maxgap*dround;
        if(this.m_isapple){
            let csn = this.GetPhaseFactor();
            for(let j = 0; j < 2; j++){
                this.m_srcobj[SrcPrmsLabel.Kxy0][j] = Kxy[j];
                this.m_srcobj[SrcPrmsLabel.Kxy][j] = Kxy[j]*csn[j];
            }
        }
        else{
            for(let j = 0; j < 2; j++){
                this.m_srcobj[SrcPrmsLabel.Kxy][j] = Kxy[j];
            }
        }
        return isok;
    }

    UpdateGapField(item, jxy)
    {   
        let tgtvalue;
        let kcoef = this.GetKCoef();
        let srctype = this.m_srcobj[TypeLabel];

        jxy = jxy < 0 ? this.m_mainjxy : jxy;   
    
        let isauto = this.m_gap.length == 0 || 
            this.m_srcobj[OptionLabel][SrcOptionsLabel.gaplink[0]] == AutomaticLabel;

        switch(item){
            case SrcPrmsLabel.gap:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.gap];
                break;
            case SrcPrmsLabel.b:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.b]*kcoef[this.m_mainjxy];
                this.m_srcobj[SrcPrmsLabel.Kxy][jxy] = tgtvalue;
                break;
            case SrcPrmsLabel.bxy:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.bxy][jxy]*kcoef[jxy];
                this.m_srcobj[SrcPrmsLabel.Kxy][jxy] = tgtvalue;
                break;
            case SrcPrmsLabel.Kxy:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.Kxy][jxy];
                break;
            case SrcPrmsLabel.Kxy0:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.Kxy0][jxy];
                break;
            case SrcPrmsLabel.e1st:
                tgtvalue = this.m_srcobj[SrcPrmsLabel.Kperp];
                break;
        }

        let isok = true;
        if(item == SrcPrmsLabel.phase){
            let csn = this.GetPhaseFactor();
            for(let j = 0; j < 2; j++){
                this.m_srcobj[SrcPrmsLabel.Kxy][j] = this.m_srcobj[SrcPrmsLabel.Kxy0][j]*csn[j];
            }
        }
        else if(this.m_srcobj[OptionLabel][SrcOptionsLabel.gaplink[0]] == NoneLabel){
            if(item == SrcPrmsLabel.b){
                this.m_srcobj[SrcPrmsLabel.Kxy][this.m_mainjxy] 
                    = this.m_srcobj[SrcPrmsLabel.b]*kcoef[this.m_mainjxy];
                return;
            }
            else if(item == SrcPrmsLabel.bxy){
                this.m_srcobj[SrcPrmsLabel.Kxy][jxy] 
                    = this.m_srcobj[SrcPrmsLabel.bxy][jxy]*kcoef[jxy];
                return;
            }
            else if(item == SrcPrmsLabel.Kxy){
                this.m_srcobj[SrcPrmsLabel.bxy][jxy] 
                    = this.m_srcobj[SrcPrmsLabel.Kxy][jxy]/kcoef[jxy];
            }
            else if(item == SrcPrmsLabel.Kxy0){
                let csn = this.GetPhaseFactor();
                this.m_srcobj[SrcPrmsLabel.Kxy][jxy] 
                    = this.m_srcobj[SrcPrmsLabel.Kxy0][jxy]*csn[jxy];
            }
            else if(item == SrcPrmsLabel.e1st){
                let kxyr;
                if(srctype == SrcTypeLabel.LIN_UND){
                    kxyr = [0, 1];
                }
                else if(srctype == SrcTypeLabel.VERTICAL_UND){
                    kxyr = [1, 0];
                }
                else if(srctype == SrcTypeLabel.HELICAL_UND){
                    kxyr = [1/Math.sqrt(2), 1/Math.sqrt(2)];
                }
                else{
                    let phi = Math.atan2(this.m_srcobj[SrcPrmsLabel.Kxy][1], 
                        this.m_srcobj[SrcPrmsLabel.Kxy][0]);
                    kxyr = [Math.cos(phi), Math.sin(phi)];
                }       

                let Kperp = this.m_srcobj[SrcPrmsLabel.Kperp];
                this.m_srcobj[SrcPrmsLabel.b] = Kperp/kcoef[this.m_mainjxy];
                for(let j = 0; j < 2; j++){
                    this.m_srcobj[SrcPrmsLabel.Kxy][j] = Kperp*kxyr[j];
                }
                if(this.m_isapple){
                    let csn = this.GetPhaseFactor();
                    let islin = false;
                    for(let j = 0; j < 2; j++){
                        if(Math.abs(csn[j]) < 1.0e-10){
                            this.m_srcobj[SrcPrmsLabel.Kxy][j] = 
                            this.m_srcobj[SrcPrmsLabel.Kxy0][j] = Kperp;
                            islin = true;
                        }    
                    }
                    if(islin == false){
                        this.m_srcobj[SrcPrmsLabel.Kxy0][0] = 
                            this.m_srcobj[SrcPrmsLabel.Kxy][0]/csn[0];
                        this.m_srcobj[SrcPrmsLabel.Kxy0][1] = 
                            this.m_srcobj[SrcPrmsLabel.Kxy][0]/csn[1];
                    }
                }
            }
        }
        else if(isauto){
            this.GetGapKxy(item, jxy, tgtvalue);
        }
        else{
            isok = this.GetGapKxyFromTable(item, jxy, tgtvalue);
        }

        if(srctype == SrcTypeLabel.LIN_UND){
            this.m_srcobj[SrcPrmsLabel.Kxy][0] = 0;
        }
        else if(srctype == SrcTypeLabel.VERTICAL_UND){
            this.m_srcobj[SrcPrmsLabel.Kxy][1] = 0;
        }
        else if(srctype == SrcTypeLabel.HELICAL_UND){
            this.m_srcobj[SrcPrmsLabel.Kxy][0] = this.m_srcobj[SrcPrmsLabel.Kxy][1];
        }
    
        this.m_srcobj[SrcPrmsLabel.b] = 
            this.m_srcobj[SrcPrmsLabel.Kxy][this.m_mainjxy]/kcoef[this.m_mainjxy];
        for(let j = 0; j < 2; j++){
            this.m_srcobj[SrcPrmsLabel.bxy][j] = this.m_srcobj[SrcPrmsLabel.Kxy][j]/kcoef[j];
        }
        return isok;
    }
}

// functions for updating paramters
function GetParabolicSingle(xarr, yarr, x)
{
    let sum;
	sum  = (x-xarr[1])/(xarr[0]-xarr[1])*(x-xarr[2])/(xarr[0]-xarr[2])*yarr[0];
	sum += (x-xarr[2])/(xarr[1]-xarr[2])*(x-xarr[0])/(xarr[1]-xarr[0])*yarr[1];
	sum += (x-xarr[0])/(xarr[2]-xarr[0])*(x-xarr[1])/(xarr[2]-xarr[1])*yarr[2];
    return sum;
}

function GetParabolic(xv, yv, x)
{
	let xarr = [], yarr = [];
	let index = Math.min(xv.length-3, SearchIndex(xv, x));	

	for(let n = 0; n < 3; n++){
		xarr.push(xv[n+index]);
		yarr.push(yv[n+index]);
	}
    return GetParabolicSingle(xarr, yarr, x);
}

function ParabloicPeak(x0, x1, x2, f0, f1, f2)
{
    let a, b, peak = {};

    a = f0/(x0-x1)/(x0-x2)+f1/(x1-x0)/(x1-x2)+f2/(x2-x1)/(x2-x0);
    b = -(x1+x2)*f0/(x0-x1)/(x0-x2)-(x0+x2)*f1/(x1-x0)/(x1-x2)-(x1+x0)*f2/(x2-x1)/(x2-x0);
    if(a == 0.0){
        peak.x = x1;
        peak.y = f1;
    }
    else{
        peak.x = -b/2.0/a;
        peak.y = GetParabolicSingle([x0, x1, x2], [f0, f1, f2], peak.x);
    }
    return peak;
}

function GetParabolicPeak(x, y, peak, init, ihist)
{
    let i = init, j;
    let flag = false;

    while(i+2 < x.length){
        if((y[i]-y[i+1])*(y[i+1]-y[i+2]) < 0.0 ||
                (y[i] == y[i+1] && y[i+1] != y[i+2])){
            flag =  true;
            for(j = i+1; j >= i-ihist && j > 2; j--){
                if((y[j]-y[j-1])*(y[j-1]-y[j-2]) < 0.0){
                    flag = false;
                    break;
                }
            }
            for(j = i+1; flag && j <= i+ihist && j+2 < x.length; j++){
                if((y[j]-y[j+1])*(y[j+1]-y[j+2]) < 0.0){
                    flag = false;
                    break;
                }
            }
            if(flag){
                break;
            }
        }
        i++;
    }
    if(!flag) return -1;
    let pkobj = ParabloicPeak(x[i], x[i+1], x[i+2], y[i], y[i+1], y[i+2]);
    peak.x = pkobj.x;
    peak.y = pkobj.y;
    return i+1;
}

function SolveParabolic(xv, yv, y)
{
	let xarr = [], yarr = [], a = 0, b = 0, c;
	let index = Math.min(xv.length-3, SearchIndex(yv, y));

	for(let n = 0; n < 3; n++){
		xarr.push(xv[n+index]);
		yarr.push(yv[n+index]);
    }
    
    let ia = [1,2,0], ib = [2,0,1], dx = [];
    for(let i = 0; i < 3; i++){
        dx.push(xarr[ia[i]]-xarr[ib[i]]);
    }
    c = dx[0]*dx[1]*dx[2]*y;
    for(let i = 0; i < 3; i++){
        a += dx[i]*yarr[i];
        b -= (xarr[ia[i]]**2-xarr[ib[i]]**2)*yarr[i];
        c += xarr[ia[i]]*xarr[ib[i]]*(xarr[ia[i]]-xarr[ib[i]])*yarr[i];
    }

    let D = b**2-4*a*c;
    if(D < 0 || a == 0){
        return NaN;
    }
    let Dsqrt = Math.sqrt(D);

    let x1 = (-b+Dsqrt)/2/a;
    let x2 = (-b-Dsqrt)/2/a;
    if(Math.abs(x1-xarr[1]) < Math.abs(x2-xarr[1])){
        return x1;
    }
    return x2;
}

function Integrate(z, Bxy, coef, iscorr)
{
    let npoints = z.length;
    let items = Bxy.length;
    let zmid, bmid, nt;
    let zarr = new Array(3);
    let barr = [];
    let Ixy = [];

    for(let j = 0; j < items; j++){
        barr.push(new Array(3));
        Ixy.push(new Array(npoints));
        Ixy[j][0] = 0;
    }
    for(let n = 1; n < npoints; n++){
        zmid = (z[n-1]+z[n])*0.5;
        if(n == npoints-1){
            nt = n-2;
        }
        else{
            nt = n-1;
        }
        for(let na = 0; na < 3; na++){
            zarr[na] = z[nt+na];
            for(let j = 0; j < items; j++){
                barr[j][na] = Bxy[j][nt+na];
            }
        }
        for(let j = 0; j < items; j++){
            bmid = GetParabolicSingle(zarr, barr[j], zmid);
            Ixy[j][n] = Ixy[j][n-1]+
                coef*(Bxy[j][n-1]+bmid*4.0+Bxy[j][n])/6.0*(z[n]-z[n-1]);
        }
    }
    if(iscorr){
        let zrange = z[npoints-1]-z[0];
        let Islope = [0, 0], Ioffset = [0, 0];
        for(let j = 0; j < 2; j++){
            Islope[j] = Ixy[j][npoints-1]/zrange;
        }
        for(let n = 1; n < npoints; n++){
            for(let j = 0; j < 2; j++){
                Ixy[j][n] -= Islope[j]*(z[n]-z[0]);
                Ioffset[j] += Ixy[j][n];
            }    
        }
        for(let j = 0; j < 2; j++){
            Ioffset[j] /= npoints-1;
            for(let n = 0; n < npoints; n++){
                Ixy[j][n] -= Ioffset[j];
            }
        }
    }
    return Ixy;
}

function SearchIndex(xarr, x)
{
    let km, kp, k, n = xarr.length;

    km = 0;
    kp = n-1;
    while (kp-km > 1){
        k = (kp+km) >> 1;
        if((xarr[0] < xarr[1] && xarr[k] > x)
            || (xarr[0] > xarr[1] && xarr[k] < x)){
            kp = k;
        }
        else{
            km = k;
        }
    }
    return km;
}

function BesselJ1(x)
{
    let ax, z, xx, y, ans, ans1, ans2;

    if((ax = Math.abs(x)) < 8.0){
        y = x*x;
        ans1=x*(72362614232.0+y*(-7895059235.0+y*(242396853.1
                +y*(-2972611.439+y*(15704.48260+y*(-30.16036606))))));
        ans2=144725228442.0+y*(2300535178.0+y*(18583304.74
                +y*(99447.43394+y*(376.9991397+y*1.0))));
        ans=ans1/ans2;
    }
    else {
        z = 8.0/ax;
        y = z*z;
        xx = ax-2.356194491;
        ans1 = 1.0+y*(0.183105e-2+y*(-0.3516396496e-4
                +y*(0.2457520174e-5+y*(-0.240337019e-6))));
        ans2 = 0.04687499995+y*(-0.2002690873e-3
                +y*(0.8449199096e-5+y*(-0.88228987e-6
                +y*0.105787412e-6)));
        ans = Math.sqrt(0.636619772/ax)*(Math.cos(xx)*ans1-z*Math.sin(xx)*ans2);
        if (x < 0.0) ans = -ans;
    }
    return ans;
}

function BesselJ0(x)
{
    let ax, z, xx, y, ans, ans1, ans2;

    if((ax = Math.abs(x)) < 8.0){
        y = x*x;
        ans1 = 57568490574.0+y*(-13362590354.0+y*(651619640.7
            +y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
        ans2 = 57568490411.0+y*(1029532985.0+y*(9494680.718
            +y*(59272.64853+y*(267.8532712+y*1.0))));
        ans = ans1/ans2;
    }
    else{
        z = 8.0/ax;
        y = z*z;
        xx = ax-0.785398164;
        ans1 = 1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
                +y*(-0.2073370639e-5+y*0.2093887211e-6)));
        ans2 = -0.1562499995e-1+y*(0.1430488765e-3
                +y*(-0.6911147651e-5+y*(0.7621095161e-6
                -y*0.934935152e-7)));
        ans = Math.sqrt(0.636619772/ax)*(Math.cos(xx)*ans1-z*Math.sin(xx)*ans2);
    }
    return ans;
}

function UpdateConfig(id)
{
    let accobj = GetObj(AccLabel);
    let srcobj = GetObj(SrcLabel);
    let confobj = GetObj(ConfigLabel);
    let indices = [];

    let srccont = GetSrcContents(srcobj);
    let conftypes = ConfigPrmOptTypes(accobj, srcobj, confobj);
    let confprmtypes = conftypes[0];
    let confopttypes = conftypes[1];

    let posprm = [ConfigPrmsLabel.de, ConfigPrmsLabel.epitch];
    for(let i = 0; i < posprm.length; i++){
        if(confprmtypes[posprm[i]] >= 0){
            if(confobj[posprm[i]] <= 0){
                confobj[posprm[i]] = 1;
            }
            indices.push(GetIDFromItem(ConfigLabel, posprm[i], -1));
        }    
    }
    
    if(srccont.ise1st >= 0){
        if(confprmtypes[ConfigPrmsLabel.detune] >= 0){
            confobj[ConfigPrmsLabel.efix] = confobj[ConfigPrmsLabel.hfix]*
            (1+confobj[ConfigPrmsLabel.detune])*srcobj[SrcPrmsLabel.e1st];
            indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.efix, -1));
        }
        else if(confopttypes[ConfigOptionsLabel.normenergy[0]]){
            if(confobj[OptionLabel][ConfigOptionsLabel.normenergy[0]]){
                confobj[ConfigPrmsLabel.efix] = 
                srcobj[SrcPrmsLabel.e1st]*confobj[ConfigPrmsLabel.nefix];        
                indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.efix, -1));
            }
        }
    }

    let ldist = confobj[ConfigPrmsLabel.slit_dist];
    if(confprmtypes[ConfigPrmsLabel.fsize] >= 0
        || confprmtypes[ConfigPrmsLabel.fdiv] >= 0){
        let size = [srcobj[SrcPrmsLabel.Sigmax][0], srcobj[SrcPrmsLabel.Sigmay][0]];
        let div = [srcobj[SrcPrmsLabel.Sigmax][1], srcobj[SrcPrmsLabel.Sigmay][1]];
        if(confobj.hasOwnProperty(ConfigPrmsLabel.fsize) == false){
            confobj[ConfigPrmsLabel.fsize] = [0, 0];
        }
        if(confobj.hasOwnProperty(ConfigPrmsLabel.fdiv) == false){
            confobj[ConfigPrmsLabel.fdiv] = [0, 0];
        }
        for(let j = 0; j < 2; j++){
            confobj[ConfigPrmsLabel.fsize][j] 
                = Math.hypot(size[j], div[j]*ldist);
            confobj[ConfigPrmsLabel.fdiv][j] = 
                confobj[ConfigPrmsLabel.fsize][j]/ldist
            if(confprmtypes[ConfigPrmsLabel.fsize] >= 0){
                indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.fsize, j));
            }
            if(confprmtypes[ConfigPrmsLabel.fdiv] >= 0){
                indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.fdiv, j));
            }
        }
    }
    if(confprmtypes[ConfigPrmsLabel.slitapt] == 0){
        for(let j = 0; j < 2; j++){
            confobj[ConfigPrmsLabel.slitapt][j] 
            = confobj[ConfigPrmsLabel.fsize][j]*confobj[ConfigPrmsLabel.nslitapt][j];
            indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.slitapt, j));
        }
    }
    if(confprmtypes[ConfigPrmsLabel.qslitapt] == 0){
        for(let j = 0; j < 2; j++){
            confobj[ConfigPrmsLabel.qslitapt][j] 
            = confobj[ConfigPrmsLabel.fdiv][j]*confobj[ConfigPrmsLabel.nslitapt][j];
            indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.qslitapt, j));
        }
    }
    if(confprmtypes[ConfigPrmsLabel.psize] >= 0){
        let pdiv = [0, 0];
        if(confobj.hasOwnProperty(ConfigPrmsLabel.psize) == false){
            confobj[ConfigPrmsLabel.fsize] = [0, 0];
        }
        if(confobj.hasOwnProperty(ConfigPrmsLabel.pdiv) == false){
            confobj[ConfigPrmsLabel.pdiv] = [0, 0];
        }
        for(let j = 0; j < 2; j++){
            let k2 = srcobj[SrcPrmsLabel.Kxy][1-j];
            k2 = 1+k2*k2*0.5;
            confobj[ConfigPrmsLabel.pdiv][j]
                = Math.sqrt(k2)/Math.SQRT2*accobj[AccPrmsLabel.gaminv];
            confobj[ConfigPrmsLabel.psize][j] 
                = confobj[ConfigPrmsLabel.pdiv][j]*ldist;
            if(confprmtypes[ConfigPrmsLabel.psize] >= 0){
                indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.psize, j));
            }
            if(confprmtypes[ConfigPrmsLabel.pdiv] >= 0){
                indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.pdiv, j));
            }
        }
    }
    if(confprmtypes[ConfigPrmsLabel.illumarea] >= 0){
        let stheta = Math.sin(Math.PI*confobj[ConfigPrmsLabel.Qgl]/180);
        let csphi = [Math.cos(Math.PI*confobj[ConfigPrmsLabel.Phiinc]/180), 
                    Math.sin(Math.PI*confobj[ConfigPrmsLabel.Phiinc]/180)];
        let xyR = [];
        for(let j = 0; j < 2; j++){
            xyR.push(confobj[ConfigPrmsLabel.qslitapt][j]*ldist);
            indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.illumarea, j));
        }
        let MM = [
            [csphi[0]**2/stheta+csphi[1]**2, csphi[0]*csphi[1]*(1/stheta-1)],
            [csphi[0]*csphi[1]*(1/stheta-1), csphi[1]**2/stheta+csphi[0]**2]
        ];
        for(let j = 0; j < 2; j++){
            confobj[ConfigPrmsLabel.illumarea][j] 
                = xyR[0]*MM[j][0]+xyR[1]*MM[j][1];
        }
    }
    if(confprmtypes[ConfigPrmsLabel.e1strange] >= 0){
        let lu = srcobj[[SrcPrmsLabel.lu]]*0.001; // mm -> m
        let eGeV = accobj[AccPrmsLabel.eGeV]
        let kidx = srccont.isbxy >= 0 ? ConfigPrmsLabel.ckrange : ConfigPrmsLabel.krange;
        let krange = CopyJSON(confobj[kidx]);
        if(srcobj[TypeLabel] == SrcTypeLabel.HELICAL_UND){
            for(let j = 0; j < krange.length; j++){
                krange[j] *= Math.SQRT2;
            }
        }
        confobj[ConfigPrmsLabel.e1strange] = [];
        for(let j = 0; j < krange.length; j++){
            let e1str = COEF_E1ST*(eGeV**2)/lu/(1+krange[j]**2/2);    
            confobj[ConfigPrmsLabel.e1strange].push(e1str); 
            indices.push(GetIDFromItem(ConfigLabel, ConfigPrmsLabel.e1strange, j));
        }
    }
    return indices;
}

function UpdateLightSrc(id)
{
    let srcobj = GetObj(SrcLabel);
    let accobj = GetObj(AccLabel);
    let srctype = srcobj[TypeLabel];
    let eGeV = accobj[AccPrmsLabel.eGeV];
    let gaminv_mrad = accobj[AccPrmsLabel.gaminv];
    let iA = accobj[TypeLabel] == AccTypeLabel.RING ? 
                accobj[AccPrmsLabel.imA] : accobj[AccPrmsLabel.aimA];
    let nharm = (srctype == SrcTypeLabel.FIGURE8_UND 
                || srctype == SrcTypeLabel.VFIGURE8_UND) ? 0.5 : 1.0;
    iA /= 1000.0;

    let isready = IsDataReady(srcobj);
    let uprms = {power:0, flux:0, Kxy:[0,0]};

    let isconf = false;
    let idc = GetItemFromID(id);
    let prmlabel = idc.item;
    let jxy = idc.jxy;
    let indices = [];
    if(prmlabel == ALL_UPDATE){ // add properties with an array
        srcobj[SrcPrmsLabel.sigmar] = [0,0];
        srcobj[SrcPrmsLabel.sigmarx] = [0,0];
        srcobj[SrcPrmsLabel.sigmary] = [0,0];
        srcobj[SrcPrmsLabel.Sigmax] = [0,0];
        srcobj[SrcPrmsLabel.Sigmay] = [0,0];
    }
    if(srctype == SrcTypeLabel.CUSTOM){
        if(isready){
            uprms = GetUprmArb(srcobj);
        }
        srcobj[SrcPrmsLabel.tpower] = 
            COEF_TOTAL_POWER_ID*(eGeV**2)*iA
            *(COEF_K_VALUE*COEF_K_VALUE*4.0*uprms.power);
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.tpower, -1));
        return indices;
    }
    if(srctype == SrcTypeLabel.BM){
        if(prmlabel == SrcPrmsLabel.bendlength 
                || prmlabel == SrcPrmsLabel.fringelen){
            return indices;
        }
        if(prmlabel == SrcPrmsLabel.b || prmlabel == ALL_UPDATE){
            srcobj[SrcPrmsLabel.radius] 
            = COEF_BM_RADIUS*eGeV/srcobj[SrcPrmsLabel.b];
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.radius, -1));
        }
        else{
            srcobj[SrcPrmsLabel.b]
            = COEF_BM_RADIUS*eGeV/srcobj[SrcPrmsLabel.radius];
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.b, -1));
        }
        srcobj[SrcPrmsLabel.ec] = COEF_EC*(eGeV**2)*srcobj[SrcPrmsLabel.b];
        srcobj[SrcPrmsLabel.lc] = ONE_ANGSTROM_eV/srcobj[SrcPrmsLabel.ec]*0.1;
        srcobj[SrcPrmsLabel.linpower]
            = COEF_LINPWD*iA*(eGeV**3)*srcobj[SrcPrmsLabel.b];
        srcobj[SrcPrmsLabel.tpowerrev] 
            = srcobj[SrcPrmsLabel.linpower]*Math.PI*2.0*1000.0;
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.ec, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.lc, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.tpowerrev, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.linpower, -1));
        return indices;
    }
    if(srctype == SrcTypeLabel.WLEN_SHIFTER){
        if(prmlabel == SrcPrmsLabel.bmain || 
                prmlabel == SrcPrmsLabel.mplength ||
                prmlabel == SrcPrmsLabel.subpolel || prmlabel == ALL_UPDATE)
        {
            srcobj[SrcPrmsLabel.subpoleb] = 
                srcobj[SrcPrmsLabel.bmain]*srcobj[SrcPrmsLabel.mplength]
                /(2*srcobj[SrcPrmsLabel.subpolel]);
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.subpoleb, -1));
        }
        else if(prmlabel == SrcPrmsLabel.subpoleb){
            srcobj[SrcPrmsLabel.subpolel] = 
                srcobj[SrcPrmsLabel.bmain]*srcobj[SrcPrmsLabel.mplength]
                /(2*srcobj[SrcPrmsLabel.subpoleb]);
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.subpolel, -1));
        }
        srcobj[SrcPrmsLabel.tpower] = (COEF_K_VALUE**2)*4.0*(
            (srcobj[SrcPrmsLabel.bmain]**2)*srcobj[SrcPrmsLabel.mplength]+
            (srcobj[SrcPrmsLabel.subpoleb]**2)*srcobj[SrcPrmsLabel.subpolel]*2.0)
            *COEF_TOTAL_POWER_ID*(eGeV**2)*iA;
        srcobj[SrcPrmsLabel.ec] = COEF_EC*(eGeV**2)*srcobj[SrcPrmsLabel.bmain];
        srcobj[SrcPrmsLabel.lc] = ONE_ANGSTROM_eV/srcobj[SrcPrmsLabel.ec]*0.1;
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.tpower, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.ec, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.lc, -1));

        return indices;
    }

    let iscustomu = 
        srctype == SrcTypeLabel.FIGURE8_UND ||
        srctype == SrcTypeLabel.VFIGURE8_UND ||
        srctype == SrcTypeLabel.CUSTOM_PERIODIC ||
        srctype == SrcTypeLabel.MULTI_HARM_UND;

    let srccont = GetSrcContents(srcobj);

    if(srctype == SrcTypeLabel.CUSTOM_PERIODIC){
        if(isready){
            if(prmlabel == SrcPrmsLabel.fvsz1per){
                indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.lu, -1));
                indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kperp, -1));
            }
            let np = srcobj[SrcPrmsLabel.fvsz1per].data[0].length;
            srcobj[SrcPrmsLabel.lu] = 
                (srcobj[SrcPrmsLabel.fvsz1per].data[0][np-1]
                -srcobj[SrcPrmsLabel.fvsz1per].data[0][0])*1000.0; // m -> mm
        }
        else{
            srcobj[SrcPrmsLabel.lu] = def_src[SrcPrmsLabel.lu];
        }
    }

    let islength = false;
    let lu = srcobj[SrcPrmsLabel.lu]/1000.0;
    if(prmlabel == SrcPrmsLabel.devlength 
            || prmlabel == SrcPrmsLabel.lu
            || prmlabel == SrcOptionsLabel.segprm[0]
            || prmlabel == SrcPrmsLabel.fvsz1per
            || prmlabel == ALL_UPDATE){

        //!! definition of N should be consistent with solver
        if(srccont.iswiggler >= 0){
            let poles = Math.floor(srcobj[SrcPrmsLabel.devlength]/(lu/2.0)+1.0e-6);
            srcobj[SrcPrmsLabel.periods] = poles/2;
        }
        else{
            srcobj[SrcPrmsLabel.periods]
                = Math.floor(srcobj[SrcPrmsLabel.devlength]/(lu/nharm)+1.0e-6);
            }
        if(srcobj[OptionLabel][SrcOptionsLabel.endmag[0]]){
            srcobj[SrcPrmsLabel.periods] -= 2;
        }
        else{
            srcobj[SrcPrmsLabel.periods] -= 1;
        }
        if(srccont.iswiggler < 0){
            srcobj[SrcPrmsLabel.periods] = Math.floor(srcobj[SrcPrmsLabel.periods]/nharm);
        }
        srcobj[SrcPrmsLabel.reglength]
            = srcobj[SrcPrmsLabel.periods]*lu;
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.reglength, -1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.periods, -1));
        islength = true;
    }

    if(prmlabel == SrcOptionsLabel.magconf[0]){
        prmlabel = SrcPrmsLabel.gap;
    }

    let isfield = false;
    switch(prmlabel){
        case SrcPrmsLabel.gap:
        case SrcPrmsLabel.bxy:
        case SrcPrmsLabel.b:
        case SrcPrmsLabel.lu:
        case SrcPrmsLabel.Kxy:
        case SrcPrmsLabel.K:
        case SrcPrmsLabel.Kxy0:
        case SrcPrmsLabel.phase:
        case SrcPrmsLabel.e1st:
        case SrcPrmsLabel.multiharm:
        case SrcPrmsLabel.fvsz1per:
        case ALL_UPDATE:
            isfield = true;
            isconf = true;
            break;
    }

    if(prmlabel == SrcPrmsLabel.phase){
        SPGUI.gapfieldtable.SetBSquared();
    }
    if(prmlabel == SrcPrmsLabel.K || prmlabel == ALL_UPDATE || prmlabel == SrcPrmsLabel.lu){
        if(SPGUI.gapfieldtable.IsApple()){
            prmlabel = SrcPrmsLabel.Kxy0;
        }
        else{
            prmlabel = SrcPrmsLabel.Kxy;
        }
        if(srctype == SrcTypeLabel.LIN_UND ||  
                srctype == SrcTypeLabel.WIGGLER){
            srcobj[SrcPrmsLabel.Kxy][0] = 0;
            srcobj[SrcPrmsLabel.Kxy][1] = srcobj[SrcPrmsLabel.K];
        }
        else if(srctype == SrcTypeLabel.VERTICAL_UND){
            srcobj[SrcPrmsLabel.Kxy][0] = srcobj[SrcPrmsLabel.K];
            srcobj[SrcPrmsLabel.Kxy][1] = 0;
        }
        else if(srctype == SrcTypeLabel.HELICAL_UND){
            srcobj[SrcPrmsLabel.Kxy][0] = srcobj[SrcPrmsLabel.K];
            srcobj[SrcPrmsLabel.Kxy][1] = srcobj[SrcPrmsLabel.K];
        }
    }
    if(prmlabel == SrcPrmsLabel.e1st){
        let K2 = COEF_E1ST*(eGeV**2)/srcobj[SrcPrmsLabel.e1st]/lu-1.0;
        if(K2 <= 0){
            srcobj[SrcPrmsLabel.Kperp] = 0;
        }
        else{
            srcobj[SrcPrmsLabel.Kperp] = Math.sqrt(2*K2);
        }
        if(srctype == SrcTypeLabel.MULTI_HARM_UND){
            let karg;
            if(Math.hypot(srcobj[SrcPrmsLabel.Kxy][1], srcobj[SrcPrmsLabel.Kxy][0]) == 0){
                karg = 0;
            }
            else{
                karg = Math.atan2(srcobj[SrcPrmsLabel.Kxy][1], srcobj[SrcPrmsLabel.Kxy][0]);
            }
            srcobj[SrcPrmsLabel.Kxy][0] = srcobj[SrcPrmsLabel.Kperp]*Math.cos(karg);
            srcobj[SrcPrmsLabel.Kxy][1] = srcobj[SrcPrmsLabel.Kperp]*Math.sin(karg);
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 1));
        }
    }
    let Nper = srcobj[SrcPrmsLabel.periods];
    if(srctype == SrcTypeLabel.FIGURE8_UND ||
            srctype == SrcTypeLabel.VFIGURE8_UND){
        Nper /= 2;
    }
    let segtype = srcobj[OptionLabel][SrcOptionsLabel.segment_type[0]];
    let Mseg = 1;
    if(srccont.iswiggler < 0 && segtype != NoneLabel){
        if(segtype == IdenticalLabel){
            Mseg = srcobj[OptionLabel][SrcOptionsLabel.segprm[0]][SegPrmsLabel.segments];
        }
        else{
            Mseg = 2*srcobj[OptionLabel][SrcOptionsLabel.segprm[0]][SegPrmsLabel.hsegments];
        }
        Nper *= Mseg;
    }

    if(isfield || islength || prmlabel == UpdateBeam){
        if(iscustomu){
            if(isfield){
                if(srctype == SrcTypeLabel.FIGURE8_UND 
                        || srctype == SrcTypeLabel.VFIGURE8_UND){
                    let isok = SPGUI.gapfieldtable.UpdateGapField(prmlabel, jxy);
                    indices.push([GetIDFromItem(SrcLabel, SrcPrmsLabel.gap, -1), isok?"black":"red"]);
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.bxy, 0));
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.bxy, 1));
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 0));
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 1));
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kperp, -1));    
                }
                else if(srctype == SrcTypeLabel.MULTI_HARM_UND){
                    indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kperp, -1));
                }
            }
            if(isready){
                uprms = GetUprmArb(srcobj);
            }
            if(srctype == SrcTypeLabel.CUSTOM_PERIODIC){
                srcobj[SrcPrmsLabel.Kxy] = uprms.Kxy;
                indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 0));
                indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 1));
            }
            srcobj[SrcPrmsLabel.tpower] = 
                COEF_TOTAL_POWER_ID*(eGeV**2)*iA
                *(COEF_K_VALUE*COEF_K_VALUE*4.0*uprms.power*Nper);
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.tpower, -1));
        }
        else if(isfield){
            let isok = SPGUI.gapfieldtable.UpdateGapField(prmlabel, jxy);
            indices.push([GetIDFromItem(SrcLabel, SrcPrmsLabel.gap, -1), isok?"black":"red"]);
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.bxy, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.bxy, 1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.b, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy0, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy0, 1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kxy, 1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Kperp, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.K, -1));    
        }
        if(srctype == SrcTypeLabel.LIN_UND ||
                srctype == SrcTypeLabel.HELICAL_UND ||
                srctype == SrcTypeLabel.WIGGLER){
            srcobj[SrcPrmsLabel.K] = srcobj[SrcPrmsLabel.Kxy][1];
        }
        else if(srctype == SrcTypeLabel.VERTICAL_UND){
            srcobj[SrcPrmsLabel.K] = srcobj[SrcPrmsLabel.Kxy][0];
        }
        let Kx =  Math.abs(srcobj[SrcPrmsLabel.Kxy][0]);
        let Ky =  Math.abs(srcobj[SrcPrmsLabel.Kxy][1]);
        if(prmlabel != SrcPrmsLabel.e1st){
            srcobj[SrcPrmsLabel.Kperp] = Math.hypot(Kx, Ky);
        }
        if(srccont.ise1st >= 0){
            srcobj[SrcPrmsLabel.e1st] 
                = COEF_E1ST*(eGeV**2)/lu/(1+srcobj[SrcPrmsLabel.Kperp]**2/2);
            srcobj[SrcPrmsLabel.lambda1] 
                = ONE_ANGSTROM_eV/srcobj[SrcPrmsLabel.e1st]*0.1;

            let lambda = srcobj[SrcPrmsLabel.lambda1]*1e-9;
            let srcdiv = Math.sqrt(lambda/2/(Mseg*srcobj[SrcPrmsLabel.reglength])); 
            let srcsize = lambda/4/Math.PI/srcdiv;
            srcobj[SrcPrmsLabel.sigmar] = [srcsize, srcdiv];
            srcobj[SrcPrmsLabel.sigmar][0] *= 1000.0;
            srcobj[SrcPrmsLabel.sigmar][1] *= 1000.0;

            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.e1st, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmar, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmar, 1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.lambda1, -1));
        }
        let sigrx, sigry;
        if(srccont.iswiggler >= 0){
            if(srctype == SrcTypeLabel.WIGGLER){
                srcobj[SrcPrmsLabel.ec] = COEF_EC*(eGeV**2)*srcobj[SrcPrmsLabel.b];
            }
            else{
                srcobj[SrcPrmsLabel.ec] = COEF_EC*(eGeV**2)*srcobj[SrcPrmsLabel.bxy][1];
            }
            srcobj[SrcPrmsLabel.lc] = ONE_ANGSTROM_eV/srcobj[SrcPrmsLabel.ec]*0.1;

            let tanh1 = Math.tanh(1.0);
            let xid = Ky*gaminv_mrad*(0.0863*Math.pow(Kx, 2.13)+0.633)*tanh1;
            let xi = Ky*gaminv_mrad*(0.0194*Math.pow(Kx, 2.25)+0.283)*tanh1;
            srcobj[SrcPrmsLabel.sigmarx][1] = xid;
            srcobj[SrcPrmsLabel.sigmarx][0] = lu*Math.hypot(xi, Nper*xid/Math.sqrt(12));
            sigrx = [srcobj[SrcPrmsLabel.sigmarx][0], srcobj[SrcPrmsLabel.sigmarx][1]];

            let eta = 8.89e-2/Ky*gaminv_mrad;
            let etad = Math.sqrt(0.356+Kx*Kx)*gaminv_mrad;
            srcobj[SrcPrmsLabel.sigmary][1] = etad;
            srcobj[SrcPrmsLabel.sigmary][0] = lu*Math.hypot(eta, Nper*etad/Math.sqrt(12));
            sigry = [srcobj[SrcPrmsLabel.sigmary][0], srcobj[SrcPrmsLabel.sigmary][1]];

            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.ec, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.lc, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmarx, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmarx, 1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmary, 0));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.sigmary, 1));
        }
        else{
            sigrx = [srcobj[SrcPrmsLabel.sigmar][0], srcobj[SrcPrmsLabel.sigmar][1]];
            sigry = [srcobj[SrcPrmsLabel.sigmar][0], srcobj[SrcPrmsLabel.sigmar][1]];
        }
        let Sigmax = [accobj[AccPrmsLabel.sigma][0], accobj[AccPrmsLabel.sigmap][0]];
        let Sigmay = [accobj[AccPrmsLabel.sigma][1], accobj[AccPrmsLabel.sigmap][1]];
        for(let j = 0; j < 2; j++){
            srcobj[SrcPrmsLabel.Sigmax][j] = Math.hypot(sigrx[j], Sigmax[j]);
            srcobj[SrcPrmsLabel.Sigmay][j] = Math.hypot(sigry[j], Sigmay[j]);
        }
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Sigmax, 0));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Sigmax, 1));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Sigmay, 0));
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.Sigmay, 1));
    }

    if(srccont.ise1st >= 0){
        if(isfield){
            if(iscustomu){
                srcobj[SrcPrmsLabel.fd] = uprms.flux;
            }
            else{
                let gsi = 1/(1+srcobj[SrcPrmsLabel.Kperp]**2/2);
                let besarg = gsi/4*
                    (srcobj[SrcPrmsLabel.Kxy][1]**2-srcobj[SrcPrmsLabel.Kxy][0]**2);
                let fx = srcobj[SrcPrmsLabel.Kxy][1]*gsi*(BesselJ0(besarg)-BesselJ1(besarg));
                let fy = srcobj[SrcPrmsLabel.Kxy][0]*gsi*(BesselJ0(besarg)+BesselJ1(besarg));
                srcobj[SrcPrmsLabel.fd] = fx**2+fy**2;
            }
        }     
        if(isfield || islength || prmlabel == UpdateBeam || prmlabel == UpdateCurr){
            let fdcoef = COEF_FLUX_UND*(eGeV**2)*iA*(Nper**2);
            srcobj[SrcPrmsLabel.flux] = fdcoef*srcobj[SrcPrmsLabel.fd]
                *(Math.PI*2)*(srcobj[SrcPrmsLabel.sigmar][1]**2);
            srcobj[SrcPrmsLabel.brill] = srcobj[SrcPrmsLabel.flux]/4/(Math.PI**2)
                /(srcobj[SrcPrmsLabel.Sigmax][0]*srcobj[SrcPrmsLabel.Sigmax][1])
                /(srcobj[SrcPrmsLabel.Sigmay][0]*srcobj[SrcPrmsLabel.Sigmay][1]);
            srcobj[SrcPrmsLabel.pkbrill] = 
                srcobj[SrcPrmsLabel.brill]/iA*accobj[AccPrmsLabel.peakcurr];
            srcobj[SrcPrmsLabel.degener] = 
                srcobj[SrcPrmsLabel.pkbrill]*
                (srcobj[SrcPrmsLabel.lambda1]**3)*1.0e-12/8.0/CC;
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.flux, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.brill, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.pkbrill, -1));
            indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.degener, -1));
        }

        if(srcobj[OptionLabel][SrcOptionsLabel.segment_type[0]] != NoneLabel){
            let ldrift = srcobj[OptionLabel][SrcOptionsLabel.segprm[0]][SegPrmsLabel.interval]
                            -srcobj[SrcPrmsLabel.periods]*lu;
            let gamma2 = 1.0e+6/(gaminv_mrad*gaminv_mrad);
            let pslip = ldrift/(srcobj[SrcPrmsLabel.lambda1]*1.0e-9*2.0*gamma2);
            srcobj[OptionLabel][SrcOptionsLabel.segprm[0]][SegPrmsLabel.pslip]
                = Math.max(0, Math.floor(pslip));            
            let subcateg = GetIDFromItem(SrcLabel, SrcOptionsLabel.segprm[0], -1);
            indices.push(GetIDFromItem(subcateg, SegPrmsLabel.pslip, -1));
        }
    }

    if(!iscustomu){
        let K2 = srcobj[SrcPrmsLabel.Kperp]**2;
        srcobj[SrcPrmsLabel.tpower] 
            = COEF_TOTAL_POWER_ID*(eGeV**2)*iA*2*K2/lu*Nper;   
        indices.push(GetIDFromItem(SrcLabel, SrcPrmsLabel.tpower, -1));
    }

    if(isconf){
        let confindices;
        confindices = UpdateConfig(ALL_UPDATE);
        Array.prototype.push.apply(indices, confindices);
    }
    
    return indices;
}

function UpdateAcc(id)
{
    let issrc = false, isconf = false;
    let idc = GetItemFromID(id);
    let prmlabel = idc.item;
    
    switch(prmlabel){
        case AccPrmsLabel.eGeV:
        case AccPrmsLabel.emitt:
        case AccPrmsLabel.coupl:
        case AccPrmsLabel.espread:
        case AccPrmsLabel.beta:
        case AccPrmsLabel.alpha:
        case AccPrmsLabel.eta:
        case AccPrmsLabel.etap:
        case ALL_UPDATE:
			isconf = true;
        case AccPrmsLabel.imA:
        case AccPrmsLabel.cirm:
        case AccPrmsLabel.bunches:
        case AccPrmsLabel.pulsepps:
        case AccPrmsLabel.bunchlength:
        case AccPrmsLabel.bunchcharge:
        	issrc = true;
        	break;
    }
    
    let indices = [];
    let accobj = GetObj(AccLabel);
    if(prmlabel == ALL_UPDATE){ // add properties with an array
        accobj[AccPrmsLabel.epsilon] = [0,0];
        accobj[AccPrmsLabel.sigma] = [0,0];
        accobj[AccPrmsLabel.sigmap] = [0,0];
    }
    if(prmlabel == AccPrmsLabel.eGeV || prmlabel == ALL_UPDATE){
        accobj[AccPrmsLabel.gaminv] = 
            1.0/(accobj[AccPrmsLabel.eGeV]/COEF_GAMMA_EMEV);
        indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.gaminv, -1));
    }
    switch(prmlabel){
        case AccPrmsLabel.emitt:
        case AccPrmsLabel.coupl:
        case AccPrmsLabel.espread:
        case AccPrmsLabel.beta:
        case AccPrmsLabel.alpha:
        case AccPrmsLabel.eta:
        case AccPrmsLabel.etap:
        case ALL_UPDATE:
            accobj[AccPrmsLabel.epsilon][0] 
                = accobj[AccPrmsLabel.emitt]/(1.0+accobj[AccPrmsLabel.coupl]);
            accobj[AccPrmsLabel.epsilon][1] 
                = accobj[AccPrmsLabel.epsilon][0]*accobj[AccPrmsLabel.coupl];
            let espread = accobj[AccPrmsLabel.espread];
            for(let j = 0; j < 2; j++){
                let nemitt = accobj[AccPrmsLabel.epsilon][j];
                let alpha = accobj[AccPrmsLabel.alpha][j];
                let beta = accobj[AccPrmsLabel.beta][j];
                let eta =  accobj[AccPrmsLabel.eta][j];
                let etap = accobj[AccPrmsLabel.etap][j];
                accobj[AccPrmsLabel.sigma][j]
                    = Math.sqrt(beta*nemitt+(espread*eta)**2)*1000.0;
                accobj[AccPrmsLabel.sigmap][j] 
                	= Math.sqrt((1.0+alpha**2)*nemitt/beta+(espread*etap)**2)*1000.0;
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.epsilon, j));
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.sigma, j));
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.sigmap, j));
            }
            break;
    }
    let charge;
    if(accobj[TypeLabel] == AccTypeLabel.RING){
        switch(prmlabel){
            case AccPrmsLabel.imA:
            case AccPrmsLabel.cirm:
            case AccPrmsLabel.bunches:
            case AccPrmsLabel.bunchlength:
            case ALL_UPDATE:
                let pulserate = 
                        CC/accobj[AccPrmsLabel.cirm]*accobj[AccPrmsLabel.bunches];
                charge = accobj[AccPrmsLabel.imA]*1.0e-3/pulserate;
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.peakcurr, -1));
                break;
        }        
    }
    else{
        switch(prmlabel){
            case AccPrmsLabel.pulsepps:
            case AccPrmsLabel.bunchcharge:
            case AccPrmsLabel.bunchlength:
            case ALL_UPDATE:
                charge = accobj[AccPrmsLabel.bunchcharge]*1.0e-9;
                accobj[AccPrmsLabel.aimA] = charge*accobj[AccPrmsLabel.pulsepps]*1000.0;
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.peakcurr, -1));
                indices.push(GetIDFromItem(AccLabel, AccPrmsLabel.aimA, -1));
                break;
        }    
    }

    if(indices.indexOf(GetIDFromItem(AccLabel, AccPrmsLabel.peakcurr, -1)) >= 0){
        accobj[AccPrmsLabel.peakcurr] = 
            charge/Math.sqrt(2.0*Math.PI)/
            (accobj[AccPrmsLabel.bunchlength]*1.0e-3/CC);
    }

    if(issrc){
        let srcindices, idsrc;
        if(prmlabel == ALL_UPDATE){
            idsrc = GetIDFromItem(SrcLabel, ALL_UPDATE, -1);
        }
        else if(isconf){
            idsrc = GetIDFromItem(SrcLabel, UpdateBeam, -1);
        }
        else{
            idsrc = GetIDFromItem(SrcLabel, UpdateCurr, -1);
        }
        srcindices = UpdateLightSrc(idsrc);
        Array.prototype.push.apply(indices, srcindices);
    }
    if(isconf && prmlabel != ALL_UPDATE){
        let confindices;
        confindices = UpdateConfig(ALL_UPDATE);
        Array.prototype.push.apply(indices, confindices);
    }
    return indices;
}

function UpdatePrms(id)
{
    let indices = [];
    let idc = GetItemFromID(id);

    if(id == ALL_UPDATE){
        let ida = GetIDFromItem(AccLabel, ALL_UPDATE, -1);
        indices = UpdateAcc(ida);
    }
    else if(idc.categ == AccLabel){
        indices = UpdateAcc(id);
    }
    else if(idc.categ == SrcLabel){
        if(idc.item == SrcOptionsLabel.gaptbl[0]){
            SPGUI.gapfieldtable.SetSrcObj(GetObj(SrcLabel));
            id = GetIDFromItem(SrcLabel, SrcPrmsLabel.gap, -1);
        }
        indices = UpdateLightSrc(id);
    }
    else if(idc.categ == ConfigLabel){
        indices = UpdateConfig(id);
    }
    RefreshGUIPrms(indices);
}

function RefreshGUIPrms(indices)
{
    for(let i = 0; i < indices.length; i++){
        let id, color = "";
        if(Array.isArray(indices[i])){
            id = indices[i][0];
            color = indices[i][1];
        }
        else{
            id = indices[i];
        }
        let item = document.getElementById(id);
        if(item == null){
            continue;
        }      
        let idc = GetItemFromID(id);
        let jsonobj = GetObj(idc.categ);
        if(idc.subitem != ""){
            jsonobj = GetObj(idc.categ)[OptionLabel][idc.item];
            idc.item = idc.subitem;
        }

        let val;
        if(idc.jxy < 0){
            val = ToPrmString(jsonobj[idc.item], 5);
        }
        else{
            val = ToPrmString(jsonobj[idc.item][idc.jxy], 3);
        }
        if(item.tagName.toLowerCase() == "input"){
            item.setAttribute("value", val);
            item.value = val;
        }
        else{
            item.innerHTML = val;
        }
        if(color != ""){
            item.style.color = color;
        }
    }
}

function IsDataReady(srcobj)
{
    let srctype = srcobj[TypeLabel];
    if(srctype == SrcTypeLabel.CUSTOM 
        ||  srctype == SrcTypeLabel.CUSTOM_PERIODIC)
    {
        let prmlabel;
        if(srctype == SrcTypeLabel.CUSTOM){
            prmlabel = SrcPrmsLabel.fvsz;
        }
        else{
            prmlabel = SrcPrmsLabel.fvsz1per;
        }
        if(!srcobj.hasOwnProperty(prmlabel)){
            return false;
        }
        if(!srcobj[prmlabel].hasOwnProperty("data")){
            return false;
        }
        let data = srcobj[prmlabel].data;
        if(!Array.isArray(data)){
            return false;
        }
        if(data.length < 3){
            return false;
        }
        for(let j = 0; j < 3; j++){
            if(data[j].length < 2){
                return false;
            }
        }
        return true;
    }
    else{
        return true;
    }
}

function GetUprmArb(srcobj)
{
    let POINTSPERPER = 32;

    let srctype = srcobj[TypeLabel];
    let z, Bxy, npoints, lu, ku;
    let isf8 = 
        srctype == SrcTypeLabel.FIGURE8_UND ||
        srctype == SrcTypeLabel.VFIGURE8_UND;

    if(isf8 || srctype == SrcTypeLabel.MULTI_HARM_UND){
        let grid = [];
        if(srctype == SrcTypeLabel.FIGURE8_UND){
            grid = [
                ["1", "0", "0", "0"],
                ["0", "0", "1", "0"]
            ];
        }
        else if(srctype == SrcTypeLabel.VFIGURE8_UND){
            grid = [
                ["0", "0", "1", "0"],
                ["1", "0", "0", "0"]
            ];
        }
        else{
            if(Array.isArray(srcobj[SrcPrmsLabel.multiharm])){
                grid = srcobj[SrcPrmsLabel.multiharm];
            }
            if(grid.length == 0){
                grid = [["1", "90", "1", "0"]];
                // elliptic undulator 
            }
        }
        let kxy = [], phase = [], isnon0 = [true, true];
        for(let nh = 0; nh < grid.length; nh++){
            if(nh == 0){
                kxy.push([1, 1]);
            }
            else{
                kxy.push([0, 0]);
            }
            phase.push([0, 0]);
            for(let j = 0; j < 2; j++){
                if(grid[nh][2*j] != ""){
                    let val = parseFloat(grid[nh][2*j]);
                    if(!val.isNan){
                        kxy[nh][j] = Math.abs(val);
                    }
                }
                if(grid[nh][2*j+1] != ""){
                    let val = parseFloat(grid[nh][2*j+1]);
                    if(!val.isNan){
                        phase[nh][j] = val/180.0*Math.PI;
                    }
                }
                if(kxy[nh][j] > 0){
                    isnon0[j] = false;
                }
            }
        }
        for(let j = 0; j < 2; j++){
            if(isnon0[j]){
                kxy[0][j] = 1;
            }
        }
        let maxnh = 0;
        let ksum = [0, 0];
        for(let nh = 0; nh < grid.length; nh++){
            if(Math.hypot(kxy[nh][0], kxy[nh][1]) > 0){
                maxnh = nh;
            }
            for(let j = 0; j < 2; j++){
                ksum[j] += kxy[nh][j]*kxy[nh][j];
            }
        }
        for(let j = 0; j < 2; j++){
            ksum[j] = Math.sqrt(ksum[j]);
        }
        lu = srcobj[SrcPrmsLabel.lu]*0.001; // mm -> m
        if(isf8){
            lu *= 2.0;
        }
        ku = 2*Math.PI/lu;
        let Kxy = srcobj[SrcPrmsLabel.Kxy];
        let bcoef = 1/lu/COEF_K_VALUE;
        let Bpk = [];
        for(let nh = 0; nh < grid.length; nh++){
            for(let j = 0; j < 2; j++){
                kxy[nh][j] /= ksum[j];
                kxy[nh][j] *= Kxy[j];
            }
            Bpk.push([kxy[nh][0]*bcoef*(nh+1), kxy[nh][1]*bcoef*(nh+1)]);
        }
        npoints = (maxnh+1)*POINTSPERPER+1;
        let dlu = lu/(npoints-1);
        z = new Array(npoints);
        Bxy = [new Array(npoints), new Array(npoints)];
        for(let n = 0; n < npoints; n++){
            z[n] = dlu*n-lu*0.5;
            Bxy[0][n] = Bxy[1][n] = 0;
            for(let nh = 0; nh <= maxnh; nh++){
                let kunh = ku*(nh+1);
                for(let j = 0; j < 2; j++){
                    Bxy[j][n] += Bpk[nh][j]*Math.sin(kunh*z[n]+phase[nh][j]);
                }
            }
        }
    }
    else{
        let data;
        if(srctype == SrcTypeLabel.CUSTOM){
            data = srcobj[SrcPrmsLabel.fvsz].data;
        }
        else{
            data = srcobj[SrcPrmsLabel.fvsz1per].data;
        }
        if(!Array.isArray(data)){
            return {power:0, flux:0, K:[0,0]};
        }
        if(data.length == 0){
            return {power:0, flux:0, K:[0,0]};
        }
        z = CopyJSON(data[0]);
        Bxy = [CopyJSON(data[1]), CopyJSON(data[2])];
        npoints = z.length;
        lu = z[npoints-1]-z[0];
        ku = 2*Math.PI/lu;
        if(npoints == 0 || npoints == 1){
            return {power:0, flux:0, K:[0,0]};
        }
        else if(npoints == 2){
            z.insert((z[0]+z[1])*0.5, 1);
            for(let j = 0; j < 2; j++){
                Bxy[j].insert((Bxy[j][0]+Bxy[j][1])*0.5, 1);
            }
            npoints++;
        }
    }

    let Tm2gt = COEF_K_VALUE*Math.PI*2.0;
    let Ixy = Integrate(z, Bxy, Tm2gt, true);
    let Isq = [new Array(npoints),new Array(npoints)];
    let Bsq = new Array(npoints);
    for(let n = 0; n < npoints; n++){
        for(let j = 0; j < 2; j++){
            Isq[j][n] = Ixy[j][n]**2;
        }
        Bsq[n] = Math.hypot(Bxy[0][n], Bxy[1][n])**2;
    }
    let Bsqa = Integrate(z, [Bsq], 1, false)[0];
    if(srctype == SrcTypeLabel.CUSTOM){
        return {power:Bsqa[npoints-1], flux:0, Kxy:0};
    }

    let rz = Integrate(z, Isq, 1, false);
    let FxyR = [new Array(npoints), new Array(npoints)];
    let FxyI = [new Array(npoints), new Array(npoints)];

    let Kxy = [0, 0], K2 = 0;
    for(let j = 0; j < 2; j++){
        Kxy[j] = rz[j][npoints-1]/(z[npoints-1]-z[0]);
        K2 += Kxy[j];
        Kxy[j] = Math.sqrt(2*Kxy[j]);
    }
    for(let n = 0; n < npoints; n++){
        let psi = (z[n]+rz[0][n]+rz[1][n])/(1.0+K2)*ku;
        if(isf8){
            psi *= 2;
            // 2nd harmonic
        }
        for(let j = 0; j < 2; j++){
            FxyR[j][n] = Ixy[j][n]*Math.cos(psi);
            FxyI[j][n] = Ixy[j][n]*Math.sin(psi);
        }
    }
    let fxycoef = 2.0/lu/(1+K2);
    if(isf8){
        fxycoef *= 2;
        // 2nd harmonic
    }
    let FxyRa = Integrate(z, FxyR, fxycoef, false);
    let FxyIa = Integrate(z, FxyI, fxycoef, false);
    let flux = Math.hypot(
                FxyRa[0][npoints-1],
                FxyRa[1][npoints-1],
                FxyIa[0][npoints-1],
                FxyIa[1][npoints-1]
                )**2;
    
    // dump orbit for debugging
    /*
    let debugstr = "z\tbx\tby\tIx\tIy\trzx\trzy\n"
    for(let n = 0; n < npoints; n++){
        let debugline = z[n].toString();
        debugline += "\t"+Bxy[0][n].toString();
        debugline += "\t"+Bxy[1][n].toString();
        debugline += "\t"+Ixy[0][n].toString();
        debugline += "\t"+Ixy[1][n].toString();
        debugline += "\t"+rz[0][n].toString();
        debugline += "\t"+rz[1][n].toString();
        debugstr += debugline+"\n";
    }
    window.__TAURI__.tauri.invoke("write_file", { path: "debug.dat", data: debugstr});
    */

    return {power:Bsqa[npoints-1], flux:flux, Kxy:Kxy};
}
